/*
 * Copyright (C) 2013- Qiming Sun <osirpt.sun@gmail.com>
 *
 * Cartisen GTO to spheric or spinor GTO transformation
 * (and reorder the integrals from block(i,j,k,l)_{contr_idx(I,K,L,J)}
 * to (iI,jJ,kK,lL))
 *
 * Cartesian to real-spheric coefficients
 * ref: H. B. Schlegel and M. J.  Frisch, Int. J. Quant. Chem., 54(1995), 83-87.
 * See also the python implementation xyz2sph_real in testsuit/cart2sph.py
 */

#include <stdlib.h>
#include <complex.h>
#include "config.h"
#include "cint_bas.h"
#include "cart2sph.h"
#include "misc.h"


static double g_trans_cart2sph[] = {
        1, /* factors of s and p are moved to CINTcommon_fac_sp */
        // px
#ifdef PYPZPX
        // py
        0,
        1,
        0,
        // pz
        0,
        0,
        1,
        // px
        1,
        0,
        0,
#else
// by default, p orbitals are ordered px, py, pz
        // px
        1,
        0,
        0,
        // py
        0,
        1,
        0,
        // pz
        0,
        0,
        1,
#endif
        // dxy
        0,
        1.092548430592079070,
        0,
        0,
        0,
        0,
        // dyz
        0,
        0,
        0,
        0,
        1.092548430592079070,
        0,
        // dz2
        -0.315391565252520002,
        0,
        0,
        -0.315391565252520002,
        0,
        0.630783130505040012,
        // dxz
        0,
        0,
        1.092548430592079070,
        0,
        0,
        0,
        // dy2
        0.546274215296039535,
        0,
        0,
        -0.546274215296039535,
        0,
        0,
        // f-3 ~ fyx2
        0,
        1.770130769779930531,
        0,
        0,
        0,
        0,
        -0.590043589926643510,
        0,
        0,
        0,
        // f-2 ~ fxyz
        0,
        0,
        0,
        0,
        2.890611442640554055,
        0,
        0,
        0,
        0,
        0,
        // f-1 ~ fyz2
        0,
        -0.457045799464465739,
        0,
        0,
        0,
        0,
        -0.457045799464465739,
        0,
        1.828183197857862944,
        0,
        // f0 ~ fz3
        0,
        0,
        -1.119528997770346170,
        0,
        0,
        0,
        0,
        -1.119528997770346170,
        0,
        0.746352665180230782,
        // f1 ~ fxz2
        -0.457045799464465739,
        0,
        0,
        -0.457045799464465739,
        0,
        1.828183197857862944,
        0,
        0,
        0,
        0,
        // f2 ~ fzx2
        0,
        0,
        1.445305721320277020,
        0,
        0,
        0,
        0,
        -1.445305721320277020,
        0,
        0,
        // f3 ~ fx3
        0.590043589926643510,
        0,
        0,
        -1.770130769779930530,
        0,
        0,
        0,
        0,
        0,
        0,
        // g-4 ~ gyx3
        0,
        2.503342941796704538,
        0,
        0,
        0,
        0,
        -2.503342941796704530,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        // g-3 ~ gx2yz
        0,
        0,
        0,
        0,
        5.310392309339791593,
        0,
        0,
        0,
        0,
        0,
        0,
        -1.770130769779930530,
        0,
        0,
        0,
        // g-2 ~ gxyz2
        0,
        -0.946174695757560014,
        0,
        0,
        0,
        0,
        -0.946174695757560014,
        0,
        5.677048174545360108,
        0,
        0,
        0,
        0,
        0,
        0,
        // g-1 ~ gyz3
        0,
        0,
        0,
        0,
        -2.007139630671867500,
        0,
        0,
        0,
        0,
        0,
        0,
        -2.007139630671867500,
        0,
        2.676186174229156671,
        0,
        // g0 ~ gz4
        0.317356640745612911,
        0,
        0,
        0.634713281491225822,
        0,
        -2.538853125964903290,
        0,
        0,
        0,
        0,
        0.317356640745612911,
        0,
        -2.538853125964903290,
        0,
        0.846284375321634430,
        // g1 ~ gxz3
        0,
        0,
        -2.007139630671867500,
        0,
        0,
        0,
        0,
        -2.007139630671867500,
        0,
        2.676186174229156671,
        0,
        0,
        0,
        0,
        0,
        // g2 ~ gx2z2
        -0.473087347878780002,
        0,
        0,
        0,
        0,
        2.838524087272680054,
        0,
        0,
        0,
        0,
        0.473087347878780009,
        0,
        -2.838524087272680050,
        0,
        0,
        // g3 ~ gzx3
        0,
        0,
        1.770130769779930531,
        0,
        0,
        0,
        0,
        -5.310392309339791590,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        // g4 ~ gy4
        0.625835735449176134,
        0,
        0,
        -3.755014412695056800,
        0,
        0,
        0,
        0,
        0,
        0,
        0.625835735449176134,
        0,
        0,
        0,
        0,
        // h-5 ~ hyx4
        0,
        3.281910284200850514,
        0,
        0,
        0,
        0,
        -6.563820568401701020,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0.656382056840170102,
        0,
        0,
        0,
        0,
        0,
        // h-4 ~ hx3yz
        0,
        0,
        0,
        0,
        8.302649259524165115,
        0,
        0,
        0,
        0,
        0,
        0,
        -8.302649259524165110,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        // h-3 ~ hyx2z2
        0,
        -1.467714898305751160,
        0,
        0,
        0,
        0,
        -0.978476598870500779,
        0,
        11.741719186446009300,
        0,
        0,
        0,
        0,
        0,
        0,
        0.489238299435250387,
        0,
        -3.913906395482003100,
        0,
        0,
        0,
        // h-2 ~ hxyz3
        0,
        0,
        0,
        0,
        -4.793536784973323750,
        0,
        0,
        0,
        0,
        0,
        0,
        -4.793536784973323750,
        0,
        9.587073569946647510,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        // h-1 ~ hyz4
        0,
        0.452946651195696921,
        0,
        0,
        0,
        0,
        0.905893302391393842,
        0,
        -5.435359814348363050,
        0,
        0,
        0,
        0,
        0,
        0,
        0.452946651195696921,
        0,
        -5.435359814348363050,
        0,
        3.623573209565575370,
        0,
        // h0 ~ hx2y2z
        0,
        0,
        1.754254836801353946,
        0,
        0,
        0,
        0,
        3.508509673602707893,
        0,
        -4.678012898136943850,
        0,
        0,
        0,
        0,
        0,
        0,
        1.754254836801353946,
        0,
        -4.678012898136943850,
        0,
        0.935602579627388771,
        // h1 ~ xz4
        0.452946651195696921,
        0,
        0,
        0.905893302391393842,
        0,
        -5.435359814348363050,
        0,
        0,
        0,
        0,
        0.452946651195696921,
        0,
        -5.435359814348363050,
        0,
        3.623573209565575370,
        0,
        0,
        0,
        0,
        0,
        0,
        // h2 ~ hx2z3
        0,
        0,
        -2.396768392486661870,
        0,
        0,
        0,
        0,
        0,
        0,
        4.793536784973323755,
        0,
        0,
        0,
        0,
        0,
        0,
        2.396768392486661877,
        0,
        -4.793536784973323750,
        0,
        0,
        // h3 ~ hx3z2
        -0.489238299435250389,
        0,
        0,
        0.978476598870500775,
        0,
        3.913906395482003101,
        0,
        0,
        0,
        0,
        1.467714898305751163,
        0,
        -11.741719186446009300,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        // h4 ~ hzy4
        0,
        0,
        2.075662314881041278,
        0,
        0,
        0,
        0,
        -12.453973889286247600,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        2.075662314881041278,
        0,
        0,
        0,
        0,
        // h5 ~ hxy4
        0.656382056840170102,
        0,
        0,
        -6.563820568401701020,
        0,
        0,
        0,
        0,
        0,
        0,
        3.281910284200850514,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        //i-6
        0, 4.0991046311514863, 0, 0, 0, 0, -13.6636821038382887, 0, 0, 0, 0, 0, 0, 0, 0, 4.0991046311514863, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //i-5
        0, 0, 0, 0, 11.8330958111587634, 0, 0, 0, 0, 0, 0, -23.6661916223175268, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.3666191622317525, 0, 0, 0, 0, 0,
        //i-4
        0, -2.0182596029148963, 0, 0, 0, 0, 0, 0, 20.1825960291489679, 0, 0, 0, 0, 0, 0, 2.0182596029148963, 0, -20.1825960291489679, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //i-3
        0, 0, 0, 0, -8.2908473356343109, 0, 0, 0, 0, 0, 0, -5.5272315570895412, 0, 22.1089262283581647, 0, 0, 0, 0, 0, 0, 0, 0, 2.7636157785447706, 0, -7.3696420761193888, 0, 0, 0,
        //i-2
        0, 0.9212052595149236, 0, 0, 0, 0, 1.8424105190298472, 0, -14.7392841522387776, 0, 0, 0, 0, 0, 0, 0.9212052595149236, 0, -14.7392841522387776, 0, 14.7392841522387776, 0, 0, 0, 0, 0, 0, 0, 0,
        //i-1
        0, 0, 0, 0, 2.9131068125936568, 0, 0, 0, 0, 0, 0, 5.8262136251873136, 0, -11.6524272503746271, 0, 0, 0, 0, 0, 0, 0, 0, 2.9131068125936568, 0, -11.6524272503746271, 0, 4.6609709001498505, 0,
        //i0
        -0.3178460113381421, 0, 0, -0.9535380340144264, 0, 5.7212282040865583, 0, 0, 0, 0, -0.9535380340144264, 0, 11.4424564081731166, 0, -7.6283042721154111, 0, 0, 0, 0, 0, 0, -0.3178460113381421, 0, 5.7212282040865583, 0, -7.6283042721154111, 0, 1.0171072362820548,
        //i1
        0, 0, 2.9131068125936568, 0, 0, 0, 0, 5.8262136251873136, 0, -11.6524272503746271, 0, 0, 0, 0, 0, 0, 2.9131068125936568, 0, -11.6524272503746271, 0, 4.6609709001498505, 0, 0, 0, 0, 0, 0, 0,
        //i2
        0.4606026297574618, 0, 0, 0.4606026297574618, 0, -7.3696420761193888, 0, 0, 0, 0, -0.4606026297574618, 0, 0, 0, 7.3696420761193888, 0, 0, 0, 0, 0, 0, -0.4606026297574618, 0, 7.3696420761193888, 0, -7.3696420761193888, 0, 0,
        //i3
        0, 0, -2.7636157785447706, 0, 0, 0, 0, 5.5272315570895412, 0, 7.3696420761193888, 0, 0, 0, 0, 0, 0, 8.2908473356343109, 0, -22.1089262283581647, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //i4
        -0.5045649007287241, 0, 0, 2.5228245036436201, 0, 5.0456490072872420, 0, 0, 0, 0, 2.5228245036436201, 0, -30.2738940437234518, 0, 0, 0, 0, 0, 0, 0, 0, -0.5045649007287241, 0, 5.0456490072872420, 0, 0, 0, 0,
        //i5
        0, 0, 2.3666191622317525, 0, 0, 0, 0, -23.6661916223175268, 0, 0, 0, 0, 0, 0, 0, 0, 11.8330958111587634, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //i6
        0.6831841051919144, 0, 0, -10.2477615778787161, 0, 0, 0, 0, 0, 0, 10.2477615778787161, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.6831841051919144, 0, 0, 0, 0, 0, 0,
        //j-7
        0, 4.9501391276721742, 0, 0, 0, 0, -24.7506956383608703, 0, 0, 0, 0, 0, 0, 0, 0, 14.8504173830165218, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.7071627325245963, 0, 0, 0, 0, 0, 0, 0,
        //j-6
        0, 0, 0, 0, 15.8757639708114002, 0, 0, 0, 0, 0, 0, -52.9192132360380043, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 15.8757639708114002, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //j-5
        0, -2.5945778936013020, 0, 0, 0, 0, 2.5945778936013020, 0, 31.1349347232156219, 0, 0, 0, 0, 0, 0, 4.6702402084823440, 0, -62.2698694464312439, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.5189155787202604, 0, 6.2269869446431247, 0, 0, 0, 0, 0,
        //j-4
        0, 0, 0, 0, -12.4539738892862495, 0, 0, 0, 0, 0, 0, 0, 0, 41.5132462976208316, 0, 0, 0, 0, 0, 0, 0, 0, 12.4539738892862495, 0, -41.5132462976208316, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //j-3
        0, 1.4081304047606462, 0, 0, 0, 0, 2.3468840079344107, 0, -28.1626080952129243, 0, 0, 0, 0, 0, 0, 0.4693768015868821, 0, -18.7750720634752817, 0, 37.5501441269505705, 0, 0, 0, 0, 0, 0, 0, 0, -0.4693768015868821, 0, 9.3875360317376408, 0, -12.5167147089835229, 0, 0, 0,
        //j-2
        0, 0, 0, 0, 6.6379903866747414, 0, 0, 0, 0, 0, 0, 13.2759807733494828, 0, -35.4026153955986160, 0, 0, 0, 0, 0, 0, 0, 0, 6.6379903866747414, 0, -35.4026153955986160, 0, 21.2415692373591725, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //j-1
        0, -0.4516580379125866, 0, 0, 0, 0, -1.3549741137377600, 0, 10.8397929099020782, 0, 0, 0, 0, 0, 0, -1.3549741137377600, 0, 21.6795858198041564, 0, -21.6795858198041564, 0, 0, 0, 0, 0, 0, 0, 0, -0.4516580379125866, 0, 10.8397929099020782, 0, -21.6795858198041564, 0, 5.7812228852811094, 0,
        //j0
        0, 0, -2.3899496919201728, 0, 0, 0, 0, -7.1698490757605189, 0, 14.3396981515210360, 0, 0, 0, 0, 0, 0, -7.1698490757605189, 0, 28.6793963030420720, 0, -11.4717585212168292, 0, 0, 0, 0, 0, 0, 0, 0, -2.3899496919201728, 0, 14.3396981515210360, 0, -11.4717585212168292, 0, 1.0925484305920790,
        //j1
        -0.4516580379125866, 0, 0, -1.3549741137377600, 0, 10.8397929099020782, 0, 0, 0, 0, -1.3549741137377600, 0, 21.6795858198041564, 0, -21.6795858198041564, 0, 0, 0, 0, 0, 0, -0.4516580379125866, 0, 10.8397929099020782, 0, -21.6795858198041564, 0, 5.7812228852811094, 0, 0, 0, 0, 0, 0, 0, 0,
        //j2
        0, 0, 3.3189951933373707, 0, 0, 0, 0, 3.3189951933373707, 0, -17.7013076977993080, 0, 0, 0, 0, 0, 0, -3.3189951933373707, 0, 0, 0, 10.6207846186795862, 0, 0, 0, 0, 0, 0, 0, 0, -3.3189951933373707, 0, 17.7013076977993080, 0, -10.6207846186795862, 0, 0,
        //j3
        0.4693768015868821, 0, 0, -0.4693768015868821, 0, -9.3875360317376408, 0, 0, 0, 0, -2.3468840079344107, 0, 18.7750720634752817, 0, 12.5167147089835229, 0, 0, 0, 0, 0, 0, -1.4081304047606462, 0, 28.1626080952129243, 0, -37.5501441269505705, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //j4
        0, 0, -3.1134934723215624, 0, 0, 0, 0, 15.5674673616078110, 0, 10.3783115744052079, 0, 0, 0, 0, 0, 0, 15.5674673616078110, 0, -62.2698694464312439, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3.1134934723215624, 0, 10.3783115744052079, 0, 0, 0, 0,
        //j5
        -0.5189155787202604, 0, 0, 4.6702402084823440, 0, 6.2269869446431247, 0, 0, 0, 0, 2.5945778936013020, 0, -62.2698694464312439, 0, 0, 0, 0, 0, 0, 0, 0, -2.5945778936013020, 0, 31.1349347232156219, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //j6
        0, 0, 2.6459606618019000, 0, 0, 0, 0, -39.6894099270284997, 0, 0, 0, 0, 0, 0, 0, 0, 39.6894099270284997, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.6459606618019000, 0, 0, 0, 0, 0, 0,
        //j7
        0.7071627325245963, 0, 0, -14.8504173830165218, 0, 0, 0, 0, 0, 0, 24.7506956383608703, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4.9501391276721742, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // l = 8
        0, 5.83141328139864, 0, 0, 0, 0, -40.81989296979048, 0, 0, 0, 0, 0, 0, 0, 0, 40.81989296979048, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -5.83141328139864, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 20.40994648489524, 0, 0, 0, 0, 0, 0, -102.0497324244762, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 61.22983945468572, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.91570664069932, 0, 0, 0, 0, 0, 0, 0,
        0, -3.193996596357255, 0, 0, 0, 0, 7.452658724833595, 0, 44.71595234900157, 0, 0, 0, 0, 0, 0, 7.452658724833595, 0, -149.0531744966719, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3.193996596357255, 0, 44.71595234900157, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -17.24955311049054, 0, 0, 0, 0, 0, 0, 17.24955311049054, 0, 68.99821244196217, 0, 0, 0, 0, 0, 0, 0, 0, 31.04919559888297, 0, -137.9964248839243, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3.449910622098108, 0, 13.79964248839243, 0, 0, 0, 0, 0,
        0, 1.913666099037323, 0, 0, 0, 0, 1.913666099037323, 0, -45.92798637689575, 0, 0, 0, 0, 0, 0, -1.913666099037323, 0, 0, 0, 76.54664396149292, 0, 0, 0, 0, 0, 0, 0, 0, -1.913666099037323, 0, 45.92798637689575, 0, -76.54664396149292, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 11.1173953976599, 0, 0, 0, 0, 0, 0, 18.52899232943316, 0, -74.11596931773265, 0, 0, 0, 0, 0, 0, 0, 0, 3.705798465886632, 0, -49.41064621182176, 0, 59.29277545418611, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3.705798465886632, 0, 24.70532310591088, 0, -19.7642584847287, 0, 0, 0,
        0, -0.9123045168698189, 0, 0, 0, 0, -2.736913550609457, 0, 27.36913550609457, 0, 0, 0, 0, 0, 0, -2.736913550609457, 0, 54.73827101218914, 0, -72.98436134958553, 0, 0, 0, 0, 0, 0, 0, 0, -0.9123045168698189, 0, 27.36913550609457, 0, -72.98436134958553, 0, 29.19374453983421, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -3.8164436064573, 0, 0, 0, 0, 0, 0, -11.4493308193719, 0, 30.5315488516584, 0, 0, 0, 0, 0, 0, 0, 0, -11.4493308193719, 0, 61.06309770331679, 0, -36.63785862199007, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3.8164436064573, 0, 30.5315488516584, 0, -36.63785862199007, 0, 6.978639737521918, 0,
        0.3180369672047749, 0, 0, 1.272147868819099, 0, -10.1771829505528, 0, 0, 0, 0, 1.908221803228649, 0, -30.53154885165839, 0, 30.53154885165839, 0, 0, 0, 0, 0, 0, 1.272147868819099, 0, -30.53154885165839, 0, 61.06309770331677, 0, -16.28349272088447, 0, 0, 0, 0, 0, 0, 0, 0, 0.3180369672047749, 0, -10.1771829505528, 0, 30.53154885165839, 0, -16.28349272088447, 0, 1.16310662292032,
        0, 0, -3.8164436064573, 0, 0, 0, 0, -11.4493308193719, 0, 30.5315488516584, 0, 0, 0, 0, 0, 0, -11.4493308193719, 0, 61.06309770331679, 0, -36.63785862199007, 0, 0, 0, 0, 0, 0, 0, 0, -3.8164436064573, 0, 30.5315488516584, 0, -36.63785862199007, 0, 6.978639737521918, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.4561522584349095, 0, 0, -0.9123045168698189, 0, 13.68456775304729, 0, 0, 0, 0, 0, 0, 13.68456775304729, 0, -36.49218067479276, 0, 0, 0, 0, 0, 0, 0.9123045168698189, 0, -13.68456775304729, 0, 0, 0, 14.5968722699171, 0, 0, 0, 0, 0, 0, 0, 0, 0.4561522584349095, 0, -13.68456775304729, 0, 36.49218067479276, 0, -14.5968722699171, 0, 0,
        0, 0, 3.705798465886632, 0, 0, 0, 0, -3.705798465886632, 0, -24.70532310591088, 0, 0, 0, 0, 0, 0, -18.52899232943316, 0, 49.41064621182176, 0, 19.7642584847287, 0, 0, 0, 0, 0, 0, 0, 0, -11.1173953976599, 0, 74.11596931773265, 0, -59.29277545418611, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.4784165247593308, 0, 0, -1.913666099037323, 0, -11.48199659422394, 0, 0, 0, 0, -4.784165247593307, 0, 57.40998297111968, 0, 19.13666099037323, 0, 0, 0, 0, 0, 0, -1.913666099037323, 0, 57.40998297111968, 0, -114.8199659422394, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.4784165247593308, 0, -11.48199659422394, 0, 19.13666099037323, 0, 0, 0, 0,
        0, 0, -3.449910622098108, 0, 0, 0, 0, 31.04919559888297, 0, 13.79964248839243, 0, 0, 0, 0, 0, 0, 17.24955311049054, 0, -137.9964248839243, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -17.24955311049054, 0, 68.99821244196217, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.5323327660595425, 0, 0, 7.452658724833595, 0, 7.452658724833595, 0, 0, 0, 0, 0, 0, -111.7898808725039, 0, 0, 0, 0, 0, 0, 0, 0, -7.452658724833595, 0, 111.7898808725039, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.5323327660595425, 0, -7.452658724833595, 0, 0, 0, 0, 0, 0,
        0, 0, 2.91570664069932, 0, 0, 0, 0, -61.22983945468572, 0, 0, 0, 0, 0, 0, 0, 0, 102.0497324244762, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -20.40994648489524, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.72892666017483, 0, 0, -20.40994648489524, 0, 0, 0, 0, 0, 0, 51.0248662122381, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -20.40994648489524, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.72892666017483, 0, 0, 0, 0, 0, 0, 0, 0,
        // l = 9
        0, 6.740108566678694, 0, 0, 0, 0, -62.9076799556678, 0, 0, 0, 0, 0, 0, 0, 0, 94.36151993350171, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -26.96043426671477, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.7489009518531882, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 25.41854119163758, 0, 0, 0, 0, 0, 0, -177.9297883414631, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 177.9297883414631, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -25.41854119163758, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -3.814338369408373, 0, 0, 0, 0, 15.25735347763349, 0, 61.02941391053396, 0, 0, 0, 0, 0, 0, 7.628676738816745, 0, -305.1470695526698, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -10.89810962688107, 0, 183.0882417316019, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.5449054813440533, 0, -8.718487701504852, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -22.65129549625621, 0, 0, 0, 0, 0, 0, 52.85302282459782, 0, 105.7060456491956, 0, 0, 0, 0, 0, 0, 0, 0, 52.85302282459782, 0, -352.3534854973187, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -22.65129549625621, 0, 105.7060456491956, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 2.436891395195093, 0, 0, 0, 0, 0, 0, -68.23295906546261, 0, 0, 0, 0, 0, 0, -6.82329590654626, 0, 68.23295906546261, 0, 136.4659181309252, 0, 0, 0, 0, 0, 0, 0, 0, -3.899026232312149, 0, 122.8193263178327, 0, -272.9318362618504, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.4873782790390186, 0, -13.64659181309252, 0, 27.29318362618504, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 16.31079695491669, 0, 0, 0, 0, 0, 0, 16.31079695491669, 0, -130.4863756393335, 0, 0, 0, 0, 0, 0, 0, 0, -16.31079695491669, 0, 0, 0, 130.4863756393335, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -16.31079695491669, 0, 130.4863756393335, 0, -130.4863756393335, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -1.385125560048583, 0, 0, 0, 0, -3.693668160129556, 0, 49.864520161749, 0, 0, 0, 0, 0, 0, -2.770251120097167, 0, 83.107533602915, 0, -166.21506720583, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16.621506720583, 0, -110.8100448038867, 0, 88.64803584310934, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.4617085200161945, 0, -16.621506720583, 0, 55.40502240194333, 0, -29.54934528103645, 0, 0, 0,
        0, 0, 0, 0, -8.46325696792098, 0, 0, 0, 0, 0, 0, -25.38977090376294, 0, 84.63256967920979, 0, 0, 0, 0, 0, 0, 0, 0, -25.38977090376294, 0, 169.2651393584196, 0, -135.4121114867357, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -8.46325696792098, 0, 84.63256967920979, 0, -135.4121114867357, 0, 38.68917471049591, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0.451093112065591, 0, 0, 0, 0, 1.804372448262364, 0, -18.04372448262364, 0, 0, 0, 0, 0, 0, 2.706558672393546, 0, -54.13117344787092, 0, 72.17489793049457, 0, 0, 0, 0, 0, 0, 0, 0, 1.804372448262364, 0, -54.13117344787092, 0, 144.3497958609891, 0, -57.73991834439565, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.451093112065591, 0, -18.04372448262364, 0, 72.17489793049457, 0, -57.73991834439565, 0, 8.248559763485094, 0,
        0, 0, 3.026024588281776, 0, 0, 0, 0, 12.1040983531271, 0, -32.27759560833895, 0, 0, 0, 0, 0, 0, 18.15614752969066, 0, -96.83278682501685, 0, 58.0996720950101, 0, 0, 0, 0, 0, 0, 0, 0, 12.1040983531271, 0, -96.83278682501685, 0, 116.1993441900202, 0, -22.1332084171467, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.026024588281776, 0, -32.27759560833895, 0, 58.0996720950101, 0, -22.1332084171467, 0, 1.229622689841484,
        0.451093112065591, 0, 0, 1.804372448262364, 0, -18.04372448262364, 0, 0, 0, 0, 2.706558672393546, 0, -54.13117344787092, 0, 72.17489793049457, 0, 0, 0, 0, 0, 0, 1.804372448262364, 0, -54.13117344787092, 0, 144.3497958609891, 0, -57.73991834439565, 0, 0, 0, 0, 0, 0, 0, 0, 0.451093112065591, 0, -18.04372448262364, 0, 72.17489793049457, 0, -57.73991834439565, 0, 8.248559763485094, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -4.23162848396049, 0, 0, 0, 0, -8.46325696792098, 0, 42.3162848396049, 0, 0, 0, 0, 0, 0, 0, 0, 42.3162848396049, 0, -67.70605574336784, 0, 0, 0, 0, 0, 0, 0, 0, 8.46325696792098, 0, -42.3162848396049, 0, 0, 0, 19.34458735524795, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4.23162848396049, 0, -42.3162848396049, 0, 67.70605574336784, 0, -19.34458735524795, 0, 0,
        -0.4617085200161945, 0, 0, 0, 0, 16.621506720583, 0, 0, 0, 0, 2.770251120097167, 0, -16.621506720583, 0, -55.40502240194333, 0, 0, 0, 0, 0, 0, 3.693668160129556, 0, -83.107533602915, 0, 110.8100448038867, 0, 29.54934528103645, 0, 0, 0, 0, 0, 0, 0, 0, 1.385125560048583, 0, -49.864520161749, 0, 166.21506720583, 0, -88.64803584310934, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 4.077699238729173, 0, 0, 0, 0, -16.31079695491669, 0, -32.62159390983339, 0, 0, 0, 0, 0, 0, -40.77699238729173, 0, 163.1079695491669, 0, 32.62159390983339, 0, 0, 0, 0, 0, 0, 0, 0, -16.31079695491669, 0, 163.1079695491669, 0, -195.7295634590003, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4.077699238729173, 0, -32.62159390983339, 0, 32.62159390983339, 0, 0, 0, 0,
        0.4873782790390186, 0, 0, -3.899026232312149, 0, -13.64659181309252, 0, 0, 0, 0, -6.82329590654626, 0, 122.8193263178327, 0, 27.29318362618504, 0, 0, 0, 0, 0, 0, 0, 0, 68.23295906546261, 0, -272.9318362618504, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.436891395195093, 0, -68.23295906546261, 0, 136.4659181309252, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -3.775215916042701, 0, 0, 0, 0, 52.85302282459782, 0, 17.61767427486594, 0, 0, 0, 0, 0, 0, 0, 0, -264.2651141229891, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -52.85302282459782, 0, 264.2651141229891, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.775215916042701, 0, -17.61767427486594, 0, 0, 0, 0, 0, 0,
        -0.5449054813440533, 0, 0, 10.89810962688107, 0, 8.718487701504852, 0, 0, 0, 0, -7.628676738816745, 0, -183.0882417316019, 0, 0, 0, 0, 0, 0, 0, 0, -15.25735347763349, 0, 305.1470695526698, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.814338369408373, 0, -61.02941391053396, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 3.177317648954698, 0, 0, 0, 0, -88.96489417073154, 0, 0, 0, 0, 0, 0, 0, 0, 222.4122354268289, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -88.96489417073154, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.177317648954698, 0, 0, 0, 0, 0, 0, 0, 0,
        0.7489009518531882, 0, 0, -26.96043426671477, 0, 0, 0, 0, 0, 0, 94.36151993350171, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -62.9076799556678, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6.740108566678694, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // l = 10
        0, 7.673951182219901, 0, 0, 0, 0, -92.08741418663881, 0, 0, 0, 0, 0, 0, 0, 0, 193.3835697919415, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -92.08741418663881, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 7.673951182219901, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 30.88705769902543, 0, 0, 0, 0, 0, 0, -288.2792051909041, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 432.4188077863561, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -123.5482307961017, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.431895299891715, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -4.453815461763347, 0, 0, 0, 0, 26.72289277058008, 0, 80.16867831174027, 0, 0, 0, 0, 0, 0, 0, 0, -561.1807481821819, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -26.72289277058008, 0, 561.1807481821819, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4.453815461763347, 0, -80.16867831174027, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -28.63763513582592, 0, 0, 0, 0, 0, 0, 114.5505405433037, 0, 152.7340540577382, 0, 0, 0, 0, 0, 0, 0, 0, 57.27527027165184, 0, -763.6702702886912, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -81.82181467378834, 0, 458.2021621732147, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4.091090733689417, 0, -21.81915057967689, 0, 0, 0, 0, 0, 0, 0,
        0, 2.976705744527138, 0, 0, 0, 0, -3.968940992702851, 0, -95.25458382486842, 0, 0, 0, 0, 0, 0, -13.89129347445998, 0, 222.2606955913596, 0, 222.2606955913597, 0, 0, 0, 0, 0, 0, 0, 0, -3.968940992702851, 0, 222.2606955913596, 0, -740.8689853045323, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.976705744527138, 0, -95.25458382486842, 0, 222.2606955913597, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 22.18705464592268, 0, 0, 0, 0, 0, 0, 0, 0, -207.0791766952783, 0, 0, 0, 0, 0, 0, 0, 0, -62.12375300858349, 0, 207.0791766952783, 0, 248.495012034334, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -35.49928743347628, 0, 372.742518051501, 0, -496.990024068668, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4.437410929184535, 0, -41.41583533905566, 0, 49.6990024068668, 0, 0, 0, 0, 0,
        0, -1.870976726712969, 0, 0, 0, 0, -3.741953453425937, 0, 78.58102252194469, 0, 0, 0, 0, 0, 0, 0, 0, 78.58102252194469, 0, -314.3240900877788, 0, 0, 0, 0, 0, 0, 0, 0, 3.741953453425937, 0, -78.58102252194469, 0, 0, 0, 209.5493933918525, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.870976726712969, 0, -78.58102252194469, 0, 314.3240900877788, 0, -209.5493933918525, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -13.89129347445998, 0, 0, 0, 0, 0, 0, -37.04344926522661, 0, 166.6955216935197, 0, 0, 0, 0, 0, 0, 0, 0, -27.78258694891996, 0, 277.8258694891996, 0, -333.3910433870395, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 55.56517389783991, 0, -222.2606955913596, 0, 127.0061117664912, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4.630431158153326, 0, -55.56517389783991, 0, 111.1303477956798, 0, -42.33537058883041, 0, 0, 0,
        0, 0.9081022627604556, 0, 0, 0, 0, 3.632409051041822, 0, -43.58890861250187, 0, 0, 0, 0, 0, 0, 5.448613576562733, 0, -130.7667258375056, 0, 217.9445430625093, 0, 0, 0, 0, 0, 0, 0, 0, 3.632409051041822, 0, -130.7667258375056, 0, 435.8890861250187, 0, -232.4741792666766, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.9081022627604556, 0, -43.58890861250187, 0, 217.9445430625093, 0, -232.4741792666766, 0, 49.815895557145, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 4.718637772708116, 0, 0, 0, 0, 0, 0, 18.87455109083247, 0, -62.91517030277488, 0, 0, 0, 0, 0, 0, 0, 0, 28.3118266362487, 0, -188.7455109083247, 0, 150.9964087266597, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 18.87455109083247, 0, -188.7455109083247, 0, 301.9928174533194, 0, -86.28366212951984, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4.718637772708116, 0, -62.91517030277488, 0, 150.9964087266597, 0, -86.28366212951984, 0, 9.587073569946648, 0,
        -0.3181304937373671, 0, 0, -1.590652468686835, 0, 15.90652468686835, 0, 0, 0, 0, -3.181304937373671, 0, 63.62609874747341, 0, -84.83479832996456, 0, 0, 0, 0, 0, 0, -3.181304937373671, 0, 95.43914812121012, 0, -254.5043949898937, 0, 101.8017579959575, 0, 0, 0, 0, 0, 0, 0, 0, -1.590652468686835, 0, 63.62609874747341, 0, -254.5043949898937, 0, 203.6035159919149, 0, -29.08621657027356, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.3181304937373671, 0, 15.90652468686835, 0, -84.83479832996456, 0, 101.8017579959575, 0, -29.08621657027356, 0, 1.292720736456603,
        0, 0, 4.718637772708116, 0, 0, 0, 0, 18.87455109083247, 0, -62.91517030277488, 0, 0, 0, 0, 0, 0, 28.3118266362487, 0, -188.7455109083247, 0, 150.9964087266597, 0, 0, 0, 0, 0, 0, 0, 0, 18.87455109083247, 0, -188.7455109083247, 0, 301.9928174533194, 0, -86.28366212951984, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4.718637772708116, 0, -62.91517030277488, 0, 150.9964087266597, 0, -86.28366212951984, 0, 9.587073569946648, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.4540511313802278, 0, 0, 1.362153394140683, 0, -21.79445430625093, 0, 0, 0, 0, 0.9081022627604556, 0, -43.58890861250187, 0, 108.9722715312547, 0, 0, 0, 0, 0, 0, -0.9081022627604556, 0, 0, 0, 108.9722715312547, 0, -116.2370896333383, 0, 0, 0, 0, 0, 0, 0, 0, -1.362153394140683, 0, 43.58890861250187, 0, -108.9722715312547, 0, 0, 0, 24.9079477785725, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.4540511313802278, 0, 21.79445430625093, 0, -108.9722715312547, 0, 116.2370896333383, 0, -24.9079477785725, 0, 0,
        0, 0, -4.630431158153326, 0, 0, 0, 0, 0, 0, 55.56517389783991, 0, 0, 0, 0, 0, 0, 27.78258694891996, 0, -55.56517389783991, 0, -111.1303477956798, 0, 0, 0, 0, 0, 0, 0, 0, 37.04344926522661, 0, -277.8258694891996, 0, 222.2606955913596, 0, 42.33537058883041, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 13.89129347445998, 0, -166.6955216935197, 0, 333.3910433870395, 0, -127.0061117664912, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.4677441816782422, 0, 0, 1.403232545034726, 0, 19.64525563048617, 0, 0, 0, 0, 6.548418543495391, 0, -78.58102252194469, 0, -78.58102252194469, 0, 0, 0, 0, 0, 0, 6.548418543495391, 0, -196.4525563048617, 0, 392.9051126097235, 0, 52.38734834796313, 0, 0, 0, 0, 0, 0, 0, 0, 1.403232545034726, 0, -78.58102252194469, 0, 392.9051126097235, 0, -314.3240900877788, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.4677441816782422, 0, 19.64525563048617, 0, -78.58102252194469, 0, 52.38734834796313, 0, 0, 0, 0,
        0, 0, 4.437410929184535, 0, 0, 0, 0, -35.49928743347628, 0, -41.41583533905566, 0, 0, 0, 0, 0, 0, -62.12375300858349, 0, 372.742518051501, 0, 49.6990024068668, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 207.0791766952783, 0, -496.990024068668, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 22.18705464592268, 0, -207.0791766952783, 0, 248.495012034334, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.4961176240878564, 0, 0, -6.449529113142133, 0, -15.8757639708114, 0, 0, 0, 0, -6.945646737229989, 0, 222.2606955913596, 0, 37.04344926522661, 0, 0, 0, 0, 0, 0, 6.945646737229989, 0, 0, 0, -555.6517389783992, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6.449529113142133, 0, -222.2606955913596, 0, 555.6517389783992, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.4961176240878564, 0, 15.8757639708114, 0, -37.04344926522661, 0, 0, 0, 0, 0, 0,
        0, 0, -4.091090733689417, 0, 0, 0, 0, 81.82181467378834, 0, 21.81915057967689, 0, 0, 0, 0, 0, 0, -57.27527027165184, 0, -458.2021621732147, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -114.5505405433037, 0, 763.6702702886912, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 28.63763513582592, 0, -152.7340540577382, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.5567269327204184, 0, 0, 15.0316271834513, 0, 10.02108478896753, 0, 0, 0, 0, -23.38253117425757, 0, -280.590374091091, 0, 0, 0, 0, 0, 0, 0, 0, -23.38253117425757, 0, 701.4759352277273, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 15.0316271834513, 0, -280.590374091091, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.5567269327204184, 0, 10.02108478896753, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 3.431895299891715, 0, 0, 0, 0, -123.5482307961017, 0, 0, 0, 0, 0, 0, 0, 0, 432.4188077863561, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -288.2792051909041, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 30.88705769902543, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.7673951182219901, 0, 0, -34.53278031998956, 0, 0, 0, 0, 0, 0, 161.1529748266179, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -161.1529748266179, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 34.53278031998956, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.7673951182219901, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // l = 11
        0, 8.631063163659167, 0, 0, 0, 0, -129.4659474548875, 0, 0, 0, 0, 0, 0, 0, 0, 362.504652873685, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -258.9318949097751, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 43.15531581829584, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.7846421057871971, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 36.80297698805311, 0, 0, 0, 0, 0, 0, -441.6357238566373, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 927.4350200989384, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -441.6357238566373, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 36.80297698805311, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -5.110940374050938, 0, 0, 0, 0, 42.59116978375781, 0, 102.2188074810188, 0, 0, 0, 0, 0, 0, -23.85105507890438, 0, -954.0422031561751, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -51.10940374050938, 0, 1431.063304734263, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 19.87587923242031, 0, -408.875229924075, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.5678822637834375, 0, 11.35764527566875, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -35.19037680383713, 0, 0, 0, 0, 0, 0, 211.1422608230228, 0, 211.1422608230228, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1477.995825761159, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -211.1422608230228, 0, 1477.995825761159, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 35.19037680383713, 0, -211.1422608230228, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 3.532036427339827, 0, 0, 0, 0, -10.59610928201948, 0, -127.1533113842337, 0, 0, 0, 0, 0, 0, -21.19221856403896, 0, 508.613245536935, 0, 339.0754970246234, 0, 0, 0, 0, 0, 0, 0, 0, 3.027459794862709, 0, 254.3066227684675, 0, -1695.377485123117, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 9.586956017065244, 0, -363.295175383525, 0, 1017.22649107387, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.5045766324771181, 0, 18.16475876917625, 0, -48.43935671780334, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 28.72100542905686, 0, 0, 0, 0, 0, 0, -38.29467390540915, 0, -306.3573912432732, 0, 0, 0, 0, 0, 0, 0, 0, -134.031358668932, 0, 714.8339129009709, 0, 428.9003477405824, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -38.29467390540915, 0, 714.8339129009709, 0, -1429.667825801941, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 28.72100542905686, 0, -306.3573912432732, 0, 428.9003477405824, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -2.369836079783365, 0, 0, 0, 0, -2.369836079783365, 0, 113.7521318296015, 0, 0, 0, 0, 0, 0, 6.63554102339342, 0, 0, 0, -530.8432818714737, 0, 0, 0, 0, 0, 0, 0, 0, 10.4272787510468, 0, -318.5059691228842, 0, 530.8432818714737, 0, 424.674625497179, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.31777051169671, 0, -182.0034109273624, 0, 955.5179073686526, 0, -849.349250994358, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.4739672159566729, 0, 22.7504263659203, 0, -106.1686563742947, 0, 84.9349250994358, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -20.06399012830402, 0, 0, 0, 0, 0, 0, -40.12798025660804, 0, 280.8958617962563, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 280.8958617962563, 0, -674.150068311015, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40.12798025660804, 0, -280.8958617962563, 0, 0, 0, 321.0238420528643, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 20.06399012830402, 0, -280.8958617962563, 0, 674.150068311015, 0, -321.0238420528643, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 1.373687498354136, 0, 0, 0, 0, 5.036854160631831, 0, -76.92649990783158, 0, 0, 0, 0, 0, 0, 6.410541658985967, 0, -205.1373330875509, 0, 461.5589994469895, 0, 0, 0, 0, 0, 0, 0, 0, 2.747374996708271, 0, -153.8529998156632, 0, 769.2649990783159, 0, -615.4119992626527, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.4578958327847119, 0, 0, 0, 153.8529998156632, 0, -410.2746661751018, 0, 175.8319997893294, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.4578958327847119, 0, 25.64216663594386, 0, -153.8529998156632, 0, 205.1373330875509, 0, -58.61066659644312, 0, 0, 0,
        0, 0, 0, 0, 10.27973595067153, 0, 0, 0, 0, 0, 0, 41.11894380268614, 0, -164.4757752107446, 0, 0, 0, 0, 0, 0, 0, 0, 61.67841570402921, 0, -493.4273256322336, 0, 493.4273256322337, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 41.11894380268614, 0, -493.4273256322336, 0, 986.8546512644674, 0, -375.9446290531304, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 10.27973595067153, 0, -164.4757752107446, 0, 493.4273256322337, 0, -375.9446290531304, 0, 62.65743817552173, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -0.4507962425947618, 0, 0, 0, 0, -2.253981212973809, 0, 27.04777455568571, 0, 0, 0, 0, 0, 0, -4.507962425947618, 0, 108.1910982227429, 0, -180.3184970379047, 0, 0, 0, 0, 0, 0, 0, 0, -4.507962425947618, 0, 162.2866473341143, 0, -540.9554911137142, 0, 288.5095952606476, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.253981212973809, 0, 108.1910982227429, 0, -540.9554911137142, 0, 577.0191905212952, 0, -123.6469693974204, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.4507962425947618, 0, 27.04777455568571, 0, -180.3184970379047, 0, 288.5095952606476, 0, -123.6469693974204, 0, 10.99084172421515, 0,
        0, 0, -3.662285987505434, 0, 0, 0, 0, -18.31142993752717, 0, 61.03809979175723, 0, 0, 0, 0, 0, 0, -36.62285987505434, 0, 244.1523991670289, 0, -195.3219193336232, 0, 0, 0, 0, 0, 0, 0, 0, -36.62285987505434, 0, 366.2285987505434, 0, -585.9657580008695, 0, 167.4187880002484, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -18.31142993752717, 0, 244.1523991670289, 0, -585.9657580008695, 0, 334.8375760004968, 0, -37.20417511116631, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3.662285987505434, 0, 61.03809979175723, 0, -195.3219193336232, 0, 167.4187880002484, 0, -37.20417511116631, 0, 1.352879094951502,
        -0.4507962425947618, 0, 0, -2.253981212973809, 0, 27.04777455568571, 0, 0, 0, 0, -4.507962425947618, 0, 108.1910982227429, 0, -180.3184970379047, 0, 0, 0, 0, 0, 0, -4.507962425947618, 0, 162.2866473341143, 0, -540.9554911137142, 0, 288.5095952606476, 0, 0, 0, 0, 0, 0, 0, 0, -2.253981212973809, 0, 108.1910982227429, 0, -540.9554911137142, 0, 577.0191905212952, 0, -123.6469693974204, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.4507962425947618, 0, 27.04777455568571, 0, -180.3184970379047, 0, 288.5095952606476, 0, -123.6469693974204, 0, 10.99084172421515, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 5.139867975335767, 0, 0, 0, 0, 15.4196039260073, 0, -82.23788760537228, 0, 0, 0, 0, 0, 0, 10.27973595067153, 0, -164.4757752107446, 0, 246.7136628161169, 0, 0, 0, 0, 0, 0, 0, 0, -10.27973595067153, 0, 0, 0, 246.7136628161169, 0, -187.9723145265652, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -15.4196039260073, 0, 164.4757752107446, 0, -246.7136628161169, 0, 0, 0, 31.32871908776087, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -5.139867975335767, 0, 82.23788760537228, 0, -246.7136628161169, 0, 187.9723145265652, 0, -31.32871908776087, 0, 0,
        0.4578958327847119, 0, 0, 0.4578958327847119, 0, -25.64216663594386, 0, 0, 0, 0, -2.747374996708271, 0, 0, 0, 153.8529998156632, 0, 0, 0, 0, 0, 0, -6.410541658985967, 0, 153.8529998156632, 0, -153.8529998156632, 0, -205.1373330875509, 0, 0, 0, 0, 0, 0, 0, 0, -5.036854160631831, 0, 205.1373330875509, 0, -769.2649990783159, 0, 410.2746661751018, 0, 58.61066659644312, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.373687498354136, 0, 76.92649990783158, 0, -461.5589994469895, 0, 615.4119992626527, 0, -175.8319997893294, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -5.015997532076005, 0, 0, 0, 0, 15.04799259622802, 0, 70.22396544906408, 0, 0, 0, 0, 0, 0, 70.22396544906407, 0, -280.8958617962563, 0, -168.5375170777538, 0, 0, 0, 0, 0, 0, 0, 0, 70.22396544906407, 0, -702.2396544906408, 0, 842.6875853887689, 0, 80.25596051321608, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 15.04799259622802, 0, -280.8958617962563, 0, 842.6875853887689, 0, -481.5357630792965, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -5.015997532076005, 0, 70.22396544906408, 0, -168.5375170777538, 0, 80.25596051321608, 0, 0, 0, 0,
        -0.4739672159566729, 0, 0, 3.31777051169671, 0, 22.7504263659203, 0, 0, 0, 0, 10.4272787510468, 0, -182.0034109273624, 0, -106.1686563742947, 0, 0, 0, 0, 0, 0, 6.63554102339342, 0, -318.5059691228842, 0, 955.5179073686526, 0, 84.9349250994358, 0, 0, 0, 0, 0, 0, 0, 0, -2.369836079783365, 0, 0, 0, 530.8432818714737, 0, -849.349250994358, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.369836079783365, 0, 113.7521318296015, 0, -530.8432818714737, 0, 424.674625497179, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 4.786834238176144, 0, 0, 0, 0, -62.22884509628987, 0, -51.0595652072122, 0, 0, 0, 0, 0, 0, -67.01567933446601, 0, 714.8339129009709, 0, 71.48339129009707, 0, 0, 0, 0, 0, 0, 0, 0, 67.01567933446601, 0, 0, 0, -1072.250869351456, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 62.22884509628987, 0, -714.8339129009709, 0, 1072.250869351456, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4.786834238176144, 0, 51.0595652072122, 0, -71.48339129009707, 0, 0, 0, 0, 0, 0,
        0.5045766324771181, 0, 0, -9.586956017065244, 0, -18.16475876917625, 0, 0, 0, 0, -3.027459794862709, 0, 363.295175383525, 0, 48.43935671780334, 0, 0, 0, 0, 0, 0, 21.19221856403896, 0, -254.3066227684675, 0, -1017.22649107387, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 10.59610928201948, 0, -508.613245536935, 0, 1695.377485123117, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3.532036427339827, 0, 127.1533113842337, 0, -339.0754970246234, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -4.398797100479641, 0, 0, 0, 0, 118.7675217129503, 0, 26.39278260287784, 0, 0, 0, 0, 0, 0, -184.7494782201449, 0, -738.9979128805796, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -184.7494782201449, 0, 1847.494782201449, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 118.7675217129503, 0, -738.9979128805796, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4.398797100479641, 0, 26.39278260287784, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.5678822637834375, 0, 0, 19.87587923242031, 0, 11.35764527566875, 0, 0, 0, 0, -51.10940374050938, 0, -408.875229924075, 0, 0, 0, 0, 0, 0, 0, 0, -23.85105507890438, 0, 1431.063304734263, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 42.59116978375781, 0, -954.0422031561751, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -5.110940374050938, 0, 102.2188074810188, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 3.680297698805311, 0, 0, 0, 0, -165.613396446239, 0, 0, 0, 0, 0, 0, 0, 0, 772.8625167491152, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -772.8625167491152, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 165.613396446239, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3.680297698805311, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.7846421057871971, 0, 0, -43.15531581829584, 0, 0, 0, 0, 0, 0, 258.9318949097751, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -362.504652873685, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 129.4659474548875, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -8.631063163659167, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // l = 12
        0, 9.609863949407661, 0, 0, 0, 0, -176.1808390724738, 0, 0, 0, 0, 0, 0, 0, 0, 634.2510206609056, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -634.2510206609056, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 176.1808390724738, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -9.609863949407661, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 43.15531581829583, 0, 0, 0, 0, 0, 0, -647.3297372744373, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1812.523264368425, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1294.659474548875, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 215.7765790914791, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3.923210528935984, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -5.784458347938102, 0, 0, 0, 0, 63.62904182731912, 0, 127.2580836546383, 0, 0, 0, 0, 0, 0, -76.35485019278295, 0, -1527.097003855659, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -76.35485019278295, 0, 3206.903708096884, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 63.62904182731912, 0, -1527.097003855659, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -5.784458347938102, 0, 127.2580836546383, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -42.2938455917996, 0, 0, 0, 0, 0, 0, 352.4487132649967, 0, 281.9589706119974, 0, 0, 0, 0, 0, 0, 0, 0, -197.3712794283981, 0, -2631.617059045309, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -422.938455917996, 0, 3947.425588567963, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 164.4760661903318, 0, -1127.835882447989, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4.699316176866622, 0, 31.32877451244415, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 4.101899446670816, 0, 0, 0, 0, -20.50949723335408, 0, -164.0759778668327, 0, 0, 0, 0, 0, 0, -24.6113966800249, 0, 984.455867200996, 0, 492.2279336004979, 0, 0, 0, 0, 0, 0, 0, 0, 24.6113966800249, 0, 0, 0, -3445.595535203485, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 20.50949723335408, 0, -984.455867200996, 0, 3445.595535203485, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4.101899446670816, 0, 164.0759778668327, 0, -492.2279336004979, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 35.89162015836965, 0, 0, 0, 0, 0, 0, -107.6748604751089, 0, -430.6994419004357, 0, 0, 0, 0, 0, 0, 0, 0, -215.3497209502179, 0, 1722.797767601743, 0, 689.1191070406971, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 30.76424585003112, 0, 861.3988838008713, 0, -3445.595535203485, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 97.42011185843189, 0, -1230.569834001245, 0, 2067.357321122091, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -5.127374308338521, 0, 61.52849170006224, 0, -98.44558672009958, 0, 0, 0, 0, 0, 0, 0,
        0, -2.881335616715016, 0, 0, 0, 0, 0.9604452055716719, 0, 155.5921233026108, 0, 0, 0, 0, 0, 0, 17.28801370029009, 0, -207.4561644034811, 0, -829.8246576139245, 0, 0, 0, 0, 0, 0, 0, 0, 17.28801370029009, 0, -726.0965754121839, 0, 1936.257534432491, 0, 774.5030137729962, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.9604452055716719, 0, -207.4561644034811, 0, 1936.257534432491, 0, -2581.676712576654, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.881335616715016, 0, 155.5921233026108, 0, -829.8246576139245, 0, 774.5030137729962, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -26.95242673514147, 0, 0, 0, 0, 0, 0, -26.95242673514147, 0, 431.2388277622634, 0, 0, 0, 0, 0, 0, 0, 0, 75.4667948583961, 0, 0, 0, -1207.468717734338, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 118.5906776346225, 0, -1207.468717734338, 0, 1207.468717734338, 0, 689.9821244196215, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 37.73339742919805, 0, -689.9821244196215, 0, 2173.443691921808, 0, -1379.964248839243, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -5.390485347028293, 0, 86.24776555245269, 0, -241.4937435468676, 0, 137.9964248839243, 0, 0, 0, 0, 0,
        0, 1.848921220493557, 0, 0, 0, 0, 5.54676366148067, 0, -118.3309581115876, 0, 0, 0, 0, 0, 0, 3.697842440987113, 0, -236.6619162231752, 0, 828.3167067811135, 0, 0, 0, 0, 0, 0, 0, 0, -3.697842440987113, 0, 0, 0, 828.3167067811135, 0, -1325.306730849781, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -5.54676366148067, 0, 236.6619162231752, 0, -828.3167067811135, 0, 0, 0, 473.3238324463505, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.848921220493557, 0, 118.3309581115876, 0, -828.3167067811135, 0, 1325.306730849781, 0, -473.3238324463505, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 16.64029098444201, 0, 0, 0, 0, 0, 0, 61.01440027628737, 0, -310.6187650429175, 0, 0, 0, 0, 0, 0, 0, 0, 77.65469126072938, 0, -828.3167067811133, 0, 1118.227554154503, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 33.28058196888402, 0, -621.237530085835, 0, 1863.712590257505, 0, -1064.978623004289, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -5.54676366148067, 0, 0, 0, 372.742518051501, 0, -709.9857486695257, 0, 236.6619162231752, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -5.54676366148067, 0, 103.5395883476392, 0, -372.742518051501, 0, 354.9928743347629, 0, -78.88730540772508, 0, 0, 0,
        0, -0.9057827129626244, 0, 0, 0, 0, -4.528913564813122, 0, 63.4047899073837, 0, 0, 0, 0, 0, 0, -9.057827129626244, 0, 253.6191596295348, 0, -507.2383192590696, 0, 0, 0, 0, 0, 0, 0, 0, -9.057827129626244, 0, 380.4287394443022, 0, -1521.714957777209, 0, 1014.476638518139, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4.528913564813122, 0, 253.6191596295348, 0, -1521.714957777209, 0, 2028.953277036278, 0, -579.7009362960796, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.9057827129626244, 0, 63.4047899073837, 0, -507.2383192590696, 0, 1014.476638518139, 0, -579.7009362960796, 0, 77.2934581728106, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -5.620233931023189, 0, 0, 0, 0, 0, 0, -28.10116965511595, 0, 112.4046786204638, 0, 0, 0, 0, 0, 0, 0, 0, -56.20233931023189, 0, 449.6187144818551, 0, -449.6187144818551, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -56.20233931023189, 0, 674.4280717227828, 0, -1348.856143445566, 0, 513.8499594078344, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -28.10116965511595, 0, 449.6187144818551, 0, -1348.856143445566, 0, 1027.699918815669, 0, -171.2833198026115, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -5.620233931023189, 0, 112.4046786204638, 0, -449.6187144818551, 0, 513.8499594078344, 0, -171.2833198026115, 0, 12.4569687129172, 0,
        0.318183090330888, 0, 0, 1.909098541985328, 0, -22.90918250382393, 0, 0, 0, 0, 4.77274635496332, 0, -114.5459125191197, 0, 190.9098541985328, 0, 0, 0, 0, 0, 0, 6.36366180661776, 0, -229.0918250382393, 0, 763.6394167941311, 0, -407.2743556235366, 0, 0, 0, 0, 0, 0, 0, 0, 4.77274635496332, 0, -229.0918250382393, 0, 1145.459125191197, 0, -1221.82306687061, 0, 261.8192286151307, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.909098541985328, 0, -114.5459125191197, 0, 763.6394167941311, 0, -1221.82306687061, 0, 523.6384572302613, 0, -46.5456406426899, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.318183090330888, 0, -22.90918250382393, 0, 190.9098541985328, 0, -407.2743556235366, 0, 261.8192286151307, 0, -46.5456406426899, 0, 1.410473958869391,
        0, 0, -5.620233931023189, 0, 0, 0, 0, -28.10116965511595, 0, 112.4046786204638, 0, 0, 0, 0, 0, 0, -56.20233931023189, 0, 449.6187144818551, 0, -449.6187144818551, 0, 0, 0, 0, 0, 0, 0, 0, -56.20233931023189, 0, 674.4280717227828, 0, -1348.856143445566, 0, 513.8499594078344, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -28.10116965511595, 0, 449.6187144818551, 0, -1348.856143445566, 0, 1027.699918815669, 0, -171.2833198026115, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -5.620233931023189, 0, 112.4046786204638, 0, -449.6187144818551, 0, 513.8499594078344, 0, -171.2833198026115, 0, 12.4569687129172, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.4528913564813122, 0, 0, -1.811565425925249, 0, 31.70239495369185, 0, 0, 0, 0, -2.264456782406561, 0, 95.10718486107555, 0, -253.6191596295348, 0, 0, 0, 0, 0, 0, 0, 0, 63.4047899073837, 0, -507.2383192590696, 0, 507.2383192590696, 0, 0, 0, 0, 0, 0, 0, 0, 2.264456782406561, 0, -63.4047899073837, 0, 0, 0, 507.2383192590696, 0, -289.8504681480398, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.811565425925249, 0, -95.10718486107555, 0, 507.2383192590696, 0, -507.2383192590696, 0, 0, 0, 38.6467290864053, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.4528913564813122, 0, -31.70239495369185, 0, 253.6191596295348, 0, -507.2383192590696, 0, 289.8504681480398, 0, -38.6467290864053, 0, 0,
        0, 0, 5.54676366148067, 0, 0, 0, 0, 5.54676366148067, 0, -103.5395883476392, 0, 0, 0, 0, 0, 0, -33.28058196888402, 0, 0, 0, 372.742518051501, 0, 0, 0, 0, 0, 0, 0, 0, -77.65469126072938, 0, 621.237530085835, 0, -372.742518051501, 0, -354.9928743347629, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -61.01440027628737, 0, 828.3167067811133, 0, -1863.712590257505, 0, 709.9857486695257, 0, 78.88730540772508, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -16.64029098444201, 0, 310.6187650429175, 0, -1118.227554154503, 0, 1064.978623004289, 0, -236.6619162231752, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.4622303051233891, 0, 0, -0.9244606102467783, 0, -29.5827395278969, 0, 0, 0, 0, -7.857915187097616, 0, 88.74821858369071, 0, 207.0791766952784, 0, 0, 0, 0, 0, 0, -12.9424485434549, 0, 414.1583533905567, 0, -828.3167067811135, 0, -331.3266827124453, 0, 0, 0, 0, 0, 0, 0, 0, -7.857915187097616, 0, 414.1583533905567, 0, -2070.791766952784, 0, 1656.633413562227, 0, 118.3309581115876, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.9244606102467783, 0, 88.74821858369071, 0, -828.3167067811135, 0, 1656.633413562227, 0, -709.9857486695257, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.4622303051233891, 0, -29.5827395278969, 0, 207.0791766952784, 0, -331.3266827124453, 0, 118.3309581115876, 0, 0, 0, 0,
        0, 0, -5.390485347028293, 0, 0, 0, 0, 37.73339742919805, 0, 86.24776555245269, 0, 0, 0, 0, 0, 0, 118.5906776346225, 0, -689.9821244196215, 0, -241.4937435468676, 0, 0, 0, 0, 0, 0, 0, 0, 75.4667948583961, 0, -1207.468717734338, 0, 2173.443691921808, 0, 137.9964248839243, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -26.95242673514147, 0, 0, 0, 1207.468717734338, 0, -1379.964248839243, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -26.95242673514147, 0, 431.2388277622634, 0, -1207.468717734338, 0, 689.9821244196215, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.480222602785836, 0, 0, 5.762671233430032, 0, 25.93202055043514, 0, 0, 0, 0, 12.96601027521757, 0, -337.1162671556568, 0, -138.3041096023208, 0, 0, 0, 0, 0, 0, 0, 0, -363.048287706092, 0, 1936.257534432491, 0, 129.0838356288327, 0, 0, 0, 0, 0, 0, 0, 0, -12.96601027521757, 0, 363.048287706092, 0, 0, 0, -1936.257534432491, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -5.762671233430032, 0, 337.1162671556568, 0, -1936.257534432491, 0, 1936.257534432491, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.480222602785836, 0, -25.93202055043514, 0, 138.3041096023208, 0, -129.0838356288327, 0, 0, 0, 0, 0, 0,
        0, 0, 5.127374308338521, 0, 0, 0, 0, -97.42011185843189, 0, -61.52849170006224, 0, 0, 0, 0, 0, 0, -30.76424585003112, 0, 1230.569834001245, 0, 98.44558672009958, 0, 0, 0, 0, 0, 0, 0, 0, 215.3497209502179, 0, -861.3988838008713, 0, -2067.357321122091, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 107.6748604751089, 0, -1722.797767601743, 0, 3445.595535203485, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -35.89162015836965, 0, 430.6994419004357, 0, -689.1191070406971, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.512737430833852, 0, 0, -13.33117320168015, 0, -20.50949723335408, 0, 0, 0, 0, 7.691061462507781, 0, 553.7564253005602, 0, 61.52849170006224, 0, 0, 0, 0, 0, 0, 43.06994419004357, 0, -861.3988838008714, 0, -1722.797767601743, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 7.691061462507781, 0, -861.3988838008714, 0, 4306.994419004357, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -13.33117320168015, 0, 553.7564253005602, 0, -1722.797767601743, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.512737430833852, 0, -20.50949723335408, 0, 61.52849170006224, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -4.699316176866622, 0, 0, 0, 0, 164.4760661903318, 0, 31.32877451244415, 0, 0, 0, 0, 0, 0, -422.938455917996, 0, -1127.835882447989, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -197.3712794283981, 0, 3947.425588567963, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 352.4487132649967, 0, -2631.617059045309, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -42.2938455917996, 0, 281.9589706119974, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.5784458347938102, 0, 0, 25.45161673092765, 0, 12.72580836546383, 0, 0, 0, 0, -95.44356274097868, 0, -572.6613764458722, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2672.419756747403, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 95.44356274097868, 0, -2672.419756747403, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -25.45161673092765, 0, 572.6613764458722, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.5784458347938102, 0, -12.72580836546383, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 3.923210528935984, 0, 0, 0, 0, -215.7765790914791, 0, 0, 0, 0, 0, 0, 0, 0, 1294.659474548875, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1812.523264368425, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 647.3297372744373, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -43.15531581829583, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.8008219957839717, 0, 0, -52.85425172174213, 0, 0, 0, 0, 0, 0, 396.406887913066, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -739.9595241043899, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 396.406887913066, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -52.85425172174213, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.8008219957839717, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // l = 13
        0, 10.60900254488917, 0, 0, 0, 0, -233.3980559875617, 0, 0, 0, 0, 0, 0, 0, 0, 1050.291251944028, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1400.38833592537, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 583.4951399689042, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -63.65401526933501, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.8160771188376283, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 49.93431784259574, 0, 0, 0, 0, 0, 0, -915.4624937809218, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3295.664977611319, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3295.664977611319, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 915.4624937809218, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -49.93431784259574, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -6.473297372744374, 0, 0, 0, 0, 90.62616321842124, 0, 155.359136945865, 0, 0, 0, 0, 0, 0, -174.7790290640981, 0, -2330.387054187975, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -77.67956847293249, 0, 6525.083751726329, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 161.8324343186094, 0, -4660.774108375949, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -31.77800528438147, 0, 776.7956847293249, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.5884815793403977, 0, -14.12355790416954, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -49.93431784259574, 0, 0, 0, 0, 0, 0, 549.2774962685531, 0, 366.1849975123687, 0, 0, 0, 0, 0, 0, 0, 0, -659.1329955222637, 0, -4394.219970148425, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -659.1329955222637, 0, 9227.861937311692, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 549.2774962685531, 0, -4394.219970148425, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -49.93431784259574, 0, 366.1849975123687, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 4.685411261829863, 0, 0, 0, 0, -34.35968258675233, 0, -206.158095520514, 0, 0, 0, 0, 0, 0, -17.17984129337616, 0, 1717.984129337616, 0, 687.1936517350465, 0, 0, 0, 0, 0, 0, 0, 0, 68.71936517350465, 0, -962.0711124290651, 0, -6413.807416193768, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 28.63306882229361, 0, -2061.58095520514, 0, 9620.711124290651, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -17.70044254469059, 0, 801.7259270242209, 0, -2748.774606940186, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.5206012513144292, 0, -22.90645505783488, 0, 76.35485019278295, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 43.68089589976209, 0, 0, 0, 0, 0, 0, -218.4044794988104, 0, -582.4119453301611, 0, 0, 0, 0, 0, 0, 0, 0, -262.0853753985725, 0, 3494.471671980967, 0, 1048.34150159429, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 262.0853753985725, 0, 0, 0, -7338.39051116003, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 218.4044794988104, 0, -3494.471671980967, 0, 7338.39051116003, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -43.68089589976209, 0, 582.4119453301611, 0, -1048.34150159429, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -3.404978058421841, 0, 0, 0, 0, 6.809956116843682, 0, 204.2986835053105, 0, 0, 0, 0, 0, 0, 30.64480252579657, 0, -612.8960505159314, 0, -1225.792101031863, 0, 0, 0, 0, 0, 0, 0, 0, 17.51131572902661, 0, -1225.792101031863, 0, 4903.168404127451, 0, 1307.511574433987, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -12.16063592293515, 0, 175.1131572902661, 0, 2451.584202063726, 0, -6537.557872169935, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -8.755657864513306, 0, 554.5249980858427, 0, -3502.263145805322, 0, 3922.534723301961, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.4864254369174059, 0, -29.18552621504435, 0, 175.1131572902661, 0, -186.7873677762839, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -34.53278031998955, 0, 0, 0, 0, 0, 0, 11.51092677332985, 0, 621.5900457598119, 0, 0, 0, 0, 0, 0, 0, 0, 207.1966819199373, 0, -828.7867276797492, 0, -1989.088146431398, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 207.1966819199373, 0, -2900.753546879122, 0, 4641.205675006596, 0, 1326.058764287599, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 11.51092677332985, 0, -828.7867276797492, 0, 4641.205675006596, 0, -4420.195880958662, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -34.53278031998955, 0, 621.5900457598119, 0, -1989.088146431398, 0, 1326.058764287599, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 2.334148624627139, 0, 0, 0, 0, 4.668297249254278, 0, -168.058700973154, 0, 0, 0, 0, 0, 0, -4.20146752432885, 0, -168.058700973154, 0, 1344.469607785232, 0, 0, 0, 0, 0, 0, 0, 0, -16.8058700973154, 0, 470.5643627248312, 0, 0, 0, -2509.6766011991, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -13.53806202283741, 0, 739.4582842818776, 0, -3764.51490179865, 0, 2509.6766011991, 0, 1075.575686228186, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.800978349552567, 0, 235.2821813624156, 0, -2151.151372456371, 0, 4517.41788215838, 0, -2151.151372456371, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.4668297249254278, 0, -33.6117401946308, 0, 268.8939215570464, 0, -501.93532023982, 0, 215.1151372456371, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 23.76708941910389, 0, 0, 0, 0, 0, 0, 71.30126825731166, 0, -507.0312409408829, 0, 0, 0, 0, 0, 0, 0, 0, 47.53417883820777, 0, -1014.062481881766, 0, 2129.531211951708, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -47.53417883820777, 0, 0, 0, 2129.531211951708, 0, -2433.749956516238, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -71.30126825731166, 0, 1014.062481881766, 0, -2129.531211951708, 0, 0, 0, 676.0416545878439, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -23.76708941910389, 0, 507.0312409408829, 0, -2129.531211951708, 0, 2433.749956516238, 0, -676.0416545878439, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -1.36713941034431, 0, 0, 0, 0, -6.379983914940114, 0, 109.3711528275448, 0, 0, 0, 0, 0, 0, -11.39282841953592, 0, 401.0275603676643, 0, -1020.797426390418, 0, 0, 0, 0, 0, 0, 0, 0, -9.114262735628734, 0, 510.3987131952091, 0, -2722.126470374449, 0, 2449.913823337004, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.278565683907184, 0, 218.7423056550896, 0, -2041.594852780836, 0, 4083.189705561673, 0, -1749.938445240717, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.9114262735628734, 0, -36.45705094251494, 0, 0, 0, 816.6379411123346, 0, -1166.625630160478, 0, 311.1001680427941, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.4557131367814367, 0, -36.45705094251494, 0, 340.2658087968061, 0, -816.6379411123346, 0, 583.312815080239, 0, -103.7000560142647, 0, 0, 0,
        0, 0, 0, 0, -12.09143589391947, 0, 0, 0, 0, 0, 0, -60.45717946959737, 0, 282.1335041914544, 0, 0, 0, 0, 0, 0, 0, 0, -120.9143589391947, 0, 1128.534016765818, 0, -1354.240820118981, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -120.9143589391947, 0, 1692.801025148726, 0, -4062.722460356943, 0, 1934.629743027116, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -60.45717946959737, 0, 1128.534016765818, 0, -4062.722460356943, 0, 3869.259486054231, 0, -859.8354413453848, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -12.09143589391947, 0, 282.1335041914544, 0, -1354.240820118981, 0, 1934.629743027116, 0, -859.8354413453848, 0, 93.80022996495107, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0.4506212100730813, 0, 0, 0, 0, 2.703727260438488, 0, -37.85218164613883, 0, 0, 0, 0, 0, 0, 6.75931815109622, 0, -189.2609082306941, 0, 378.5218164613883, 0, 0, 0, 0, 0, 0, 0, 0, 9.012424201461626, 0, -378.5218164613883, 0, 1514.087265845553, 0, -1009.391510563702, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6.75931815109622, 0, -378.5218164613883, 0, 2271.13089876833, 0, -3028.174531691106, 0, 865.1927233403161, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.703727260438488, 0, -189.2609082306941, 0, 1514.087265845553, 0, -3028.174531691106, 0, 1730.385446680632, 0, -230.7180595574176, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.4506212100730813, 0, -37.85218164613883, 0, 378.5218164613883, 0, -1009.391510563702, 0, 865.1927233403161, 0, -230.7180595574176, 0, 13.98291270044955, 0,
        0, 0, 4.298652372786529, 0, 0, 0, 0, 25.79191423671917, 0, -103.1676569468767, 0, 0, 0, 0, 0, 0, 64.47978559179793, 0, -515.8382847343835, 0, 515.8382847343835, 0, 0, 0, 0, 0, 0, 0, 0, 85.97304745573058, 0, -1031.676569468767, 0, 2063.353138937534, 0, -786.0392910238224, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64.47978559179793, 0, -1031.676569468767, 0, 3095.029708406301, 0, -2358.117873071467, 0, 393.0196455119112, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 25.79191423671917, 0, -515.8382847343835, 0, 2063.353138937534, 0, -2358.117873071467, 0, 786.0392910238224, 0, -57.16649389264163, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4.298652372786529, 0, -103.1676569468767, 0, 515.8382847343835, 0, -786.0392910238224, 0, 393.0196455119112, 0, -57.16649389264163, 0, 1.46580753570876,
        0.4506212100730813, 0, 0, 2.703727260438488, 0, -37.85218164613883, 0, 0, 0, 0, 6.75931815109622, 0, -189.2609082306941, 0, 378.5218164613883, 0, 0, 0, 0, 0, 0, 9.012424201461626, 0, -378.5218164613883, 0, 1514.087265845553, 0, -1009.391510563702, 0, 0, 0, 0, 0, 0, 0, 0, 6.75931815109622, 0, -378.5218164613883, 0, 2271.13089876833, 0, -3028.174531691106, 0, 865.1927233403161, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.703727260438488, 0, -189.2609082306941, 0, 1514.087265845553, 0, -3028.174531691106, 0, 1730.385446680632, 0, -230.7180595574176, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.4506212100730813, 0, -37.85218164613883, 0, 378.5218164613883, 0, -1009.391510563702, 0, 865.1927233403161, 0, -230.7180595574176, 0, 13.98291270044955, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -6.045717946959737, 0, 0, 0, 0, -24.18287178783895, 0, 141.0667520957272, 0, 0, 0, 0, 0, 0, -30.22858973479868, 0, 423.2002562871816, 0, -677.1204100594905, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 282.1335041914544, 0, -1354.240820118981, 0, 967.3148715135579, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 30.22858973479868, 0, -282.1335041914544, 0, 0, 0, 967.3148715135579, 0, -429.9177206726924, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 24.18287178783895, 0, -423.2002562871816, 0, 1354.240820118981, 0, -967.3148715135579, 0, 0, 0, 46.90011498247553, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6.045717946959737, 0, -141.0667520957272, 0, 677.1204100594905, 0, -967.3148715135579, 0, 429.9177206726924, 0, -46.90011498247553, 0, 0,
        -0.4557131367814367, 0, 0, -0.9114262735628734, 0, 36.45705094251494, 0, 0, 0, 0, 2.278565683907184, 0, 36.45705094251494, 0, -340.2658087968061, 0, 0, 0, 0, 0, 0, 9.114262735628734, 0, -218.7423056550896, 0, 0, 0, 816.6379411123346, 0, 0, 0, 0, 0, 0, 0, 0, 11.39282841953592, 0, -510.3987131952091, 0, 2041.594852780836, 0, -816.6379411123346, 0, -583.312815080239, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6.379983914940114, 0, -401.0275603676643, 0, 2722.126470374449, 0, -4083.189705561673, 0, 1166.625630160478, 0, 103.7000560142647, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.36713941034431, 0, -109.3711528275448, 0, 1020.797426390418, 0, -2449.913823337004, 0, 1749.938445240717, 0, -311.1001680427941, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 5.941772354775972, 0, 0, 0, 0, -11.88354470955194, 0, -126.7578102352207, 0, 0, 0, 0, 0, 0, -101.0101300311915, 0, 380.2734307056622, 0, 532.3828029879271, 0, 0, 0, 0, 0, 0, 0, 0, -166.3696259337272, 0, 1774.60934329309, 0, -2129.531211951708, 0, -608.4374891290595, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -101.0101300311915, 0, 1774.60934329309, 0, -5323.828029879271, 0, 3042.187445645297, 0, 169.010413646961, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -11.88354470955194, 0, 380.2734307056622, 0, -2129.531211951708, 0, 3042.187445645297, 0, -1014.062481881766, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5.941772354775972, 0, -126.7578102352207, 0, 532.3828029879271, 0, -608.4374891290595, 0, 169.010413646961, 0, 0, 0, 0,
        0.4668297249254278, 0, 0, -2.800978349552567, 0, -33.6117401946308, 0, 0, 0, 0, -13.53806202283741, 0, 235.2821813624156, 0, 268.8939215570464, 0, 0, 0, 0, 0, 0, -16.8058700973154, 0, 739.4582842818776, 0, -2151.151372456371, 0, -501.93532023982, 0, 0, 0, 0, 0, 0, 0, 0, -4.20146752432885, 0, 470.5643627248312, 0, -3764.51490179865, 0, 4517.41788215838, 0, 215.1151372456371, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4.668297249254278, 0, -168.058700973154, 0, 0, 0, 2509.6766011991, 0, -2151.151372456371, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.334148624627139, 0, -168.058700973154, 0, 1344.469607785232, 0, -2509.6766011991, 0, 1075.575686228186, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -5.755463386664925, 0, 0, 0, 0, 69.0655606399791, 0, 103.5983409599687, 0, 0, 0, 0, 0, 0, 155.397511439953, 0, -1346.778432479592, 0, -331.5146910718997, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1450.376773439561, 0, 4641.205675006596, 0, 221.0097940479331, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -155.397511439953, 0, 1450.376773439561, 0, 0, 0, -3315.146910718997, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -69.0655606399791, 0, 1346.778432479592, 0, -4641.205675006596, 0, 3315.146910718997, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5.755463386664925, 0, -103.5983409599687, 0, 331.5146910718997, 0, -221.0097940479331, 0, 0, 0, 0, 0, 0,
        -0.4864254369174059, 0, 0, 8.755657864513306, 0, 29.18552621504435, 0, 0, 0, 0, 12.16063592293515, 0, -554.5249980858427, 0, -175.1131572902661, 0, 0, 0, 0, 0, 0, -17.51131572902661, 0, -175.1131572902661, 0, 3502.263145805322, 0, 186.7873677762839, 0, 0, 0, 0, 0, 0, 0, 0, -30.64480252579657, 0, 1225.792101031863, 0, -2451.584202063726, 0, -3922.534723301961, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -6.809956116843682, 0, 612.8960505159314, 0, -4903.168404127451, 0, 6537.557872169935, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.404978058421841, 0, -204.2986835053105, 0, 1225.792101031863, 0, -1307.511574433987, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 5.460111987470261, 0, 0, 0, 0, -141.9629116742268, 0, -72.80149316627014, 0, 0, 0, 0, 0, 0, 81.90167981205391, 0, 1965.640315489294, 0, 131.0426876992863, 0, 0, 0, 0, 0, 0, 0, 0, 458.6494069475019, 0, -3057.662712983346, 0, -3669.195255580015, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 81.90167981205391, 0, -3057.662712983346, 0, 9172.988138950038, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -141.9629116742268, 0, 1965.640315489294, 0, -3669.195255580015, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5.460111987470261, 0, -72.80149316627014, 0, 131.0426876992863, 0, 0, 0, 0, 0, 0, 0, 0,
        0.5206012513144292, 0, 0, -17.70044254469059, 0, -22.90645505783488, 0, 0, 0, 0, 28.63306882229361, 0, 801.7259270242209, 0, 76.35485019278295, 0, 0, 0, 0, 0, 0, 68.71936517350465, 0, -2061.58095520514, 0, -2748.774606940186, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -17.17984129337616, 0, -962.0711124290651, 0, 9620.711124290651, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -34.35968258675233, 0, 1717.984129337616, 0, -6413.807416193768, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4.685411261829863, 0, -206.158095520514, 0, 687.1936517350465, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -4.993431784259574, 0, 0, 0, 0, 219.7109985074212, 0, 36.61849975123687, 0, 0, 0, 0, 0, 0, -823.9162444028297, 0, -1647.832488805659, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 7689.884947759744, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 823.9162444028297, 0, -7689.884947759744, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -219.7109985074212, 0, 1647.832488805659, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4.993431784259574, 0, -36.61849975123687, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.5884815793403977, 0, 0, 31.77800528438147, 0, 14.12355790416954, 0, 0, 0, 0, -161.8324343186094, 0, -776.7956847293249, 0, 0, 0, 0, 0, 0, 0, 0, 77.67956847293249, 0, 4660.774108375949, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 174.7790290640981, 0, -6525.083751726329, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -90.62616321842124, 0, 2330.387054187975, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6.473297372744374, 0, -155.359136945865, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 4.161193153549645, 0, 0, 0, 0, -274.6387481342766, 0, 0, 0, 0, 0, 0, 0, 0, 2059.790611007074, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3844.942473879872, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2059.790611007074, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -274.6387481342766, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4.161193153549645, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.8160771188376283, 0, 0, -63.65401526933501, 0, 0, 0, 0, 0, 0, 583.4951399689042, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1400.38833592537, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1050.291251944028, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -233.3980559875617, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 10.60900254488917, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // l = 14
        0, 11.62730916290334, 0, 0, 0, 0, -302.3100382354867, 0, 0, 0, 0, 0, 0, 0, 0, 1662.705210295177, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2850.351789077446, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1662.705210295177, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -302.3100382354867, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 11.62730916290334, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 57.13122714353754, 0, 0, 0, 0, 0, 0, -1256.886997157826, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5655.991487210216, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -7541.321982946955, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3142.217492894565, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -342.7873628612252, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4.394709780272118, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -7.176531019523646, 0, 0, 0, 0, 124.3932043384099, 0, 186.5898065076148, 0, 0, 0, 0, 0, 0, -342.0813119306271, 0, -3420.813119306271, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12314.92722950258, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 342.0813119306271, 0, -12314.92722950258, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -124.3932043384099, 0, 3420.813119306271, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 7.176531019523646, 0, -186.5898065076148, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -58.09962199636542, 0, 0, 0, 0, 0, 0, 813.3947079491158, 0, 464.7969759709233, 0, 0, 0, 0, 0, 0, 0, 0, -1568.689793901866, 0, -6971.95463956385, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -697.195463956385, 0, 19521.47299077878, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1452.490549909135, 0, -13943.9092791277, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -285.2163261639757, 0, 2323.984879854617, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5.281783817851402, 0, -42.25427054281121, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 5.281783817851402, 0, 0, 0, 0, -52.81783817851402, 0, -253.5256232568673, 0, 0, 0, 0, 0, 0, 11.61992439927308, 0, 2788.78185582554, 0, 929.5939519418467, 0, 0, 0, 0, 0, 0, 0, 0, 139.439092791277, 0, -3346.538226990648, 0, -11155.12742330216, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 11.61992439927308, 0, -3346.538226990648, 0, 23425.76758893454, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -52.81783817851402, 0, 2788.78185582554, 0, -11155.12742330216, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5.281783817851402, 0, -253.5256232568673, 0, 929.5939519418467, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 52.07313853625346, 0, 0, 0, 0, 0, 0, -381.8696825991921, 0, -763.7393651983841, 0, 0, 0, 0, 0, 0, 0, 0, -190.934841299596, 0, 6364.494709986534, 0, 1527.478730396768, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 763.7393651983841, 0, -3564.117037592459, 0, -14256.46815036984, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 318.2247354993267, 0, -7637.393651983841, 0, 21384.70222555475, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -196.720745581402, 0, 2970.097531327049, 0, -6109.914921587073, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5.78590428180594, 0, -84.85992946648712, 0, 169.7198589329742, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -3.940231044985851, 0, 0, 0, 0, 15.7609241799434, 0, 260.0552489690662, 0, 0, 0, 0, 0, 0, 43.34254149484436, 0, -1300.276244845331, 0, -1733.701659793774, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1560.331493814397, 0, 10402.20995876265, 0, 2080.441991752529, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -43.34254149484436, 0, 1560.331493814397, 0, 0, 0, -14563.0939422677, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -15.7609241799434, 0, 1300.276244845331, 0, -10402.20995876265, 0, 14563.0939422677, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.940231044985851, 0, -260.0552489690662, 0, 1733.701659793774, 0, -2080.441991752529, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -42.78485868831644, 0, 0, 0, 0, 0, 0, 85.56971737663287, 0, 855.6971737663287, 0, 0, 0, 0, 0, 0, 0, 0, 385.0637281948479, 0, -2567.091521298986, 0, -3080.509825558783, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 220.0364161113417, 0, -5134.183042597972, 0, 12322.03930223513, 0, 2347.055105187644, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -152.8030667439873, 0, 733.4547203711389, 0, 6161.019651117567, 0, -11735.27552593822, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -110.0182080556708, 0, 2322.606614508606, 0, -8801.456644453667, 0, 7041.165315562933, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6.112122669759491, 0, -122.2424533951898, 0, 440.0728322226833, 0, -335.2935864553778, 0, 0, 0, 0, 0, 0, 0,
        0, 2.829363009969403, 0, 0, 0, 0, 1.886242006646268, 0, -226.3490407975522, 0, 0, 0, 0, 0, 0, -17.91929906313955, 0, 75.44968026585074, 0, 2037.14136717797, 0, 0, 0, 0, 0, 0, 0, 0, -33.95235611963283, 0, 1358.094244785313, 0, -2716.188489570627, 0, -4345.901583313003, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -17.91929906313955, 0, 1358.094244785313, 0, -9506.659713497193, 0, 10140.43702773034, 0, 2172.950791656501, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.886242006646268, 0, 75.44968026585074, 0, -2716.188489570627, 0, 10140.43702773034, 0, -7243.169305521671, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.829363009969403, 0, -226.3490407975522, 0, 2037.14136717797, 0, -4345.901583313003, 0, 2172.950791656501, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 31.633240116575, 0, 0, 0, 0, 0, 0, 63.26648023315, 0, -759.1977627977999, 0, 0, 0, 0, 0, 0, 0, 0, -56.939832209835, 0, -759.1977627977999, 0, 3644.14926142944, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -227.75932883934, 0, 2125.75373583384, 0, 0, 0, -4858.86568190592, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -183.472792676135, 0, 3340.47015631032, 0, -10203.61793200243, 0, 4858.86568190592, 0, 1619.62189396864, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -37.95988813989, 0, 1062.87686791692, 0, -5830.638818287104, 0, 8745.958227430655, 0, -3239.24378793728, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6.326648023315, 0, -151.83955255956, 0, 728.8298522858879, 0, -971.7731363811839, 0, 323.924378793728, 0, 0, 0, 0, 0,
        0, -1.835933153488193, 0, 0, 0, 0, -7.343732613952774, 0, 165.2339838139374, 0, 0, 0, 0, 0, 0, -9.179665767440967, 0, 495.7019514418122, 0, -1762.495827348666, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 330.4679676278748, 0, -3524.991654697331, 0, 4934.988316576264, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 9.179665767440967, 0, -330.4679676278748, 0, 0, 0, 4934.988316576264, 0, -4229.989985636798, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 7.343732613952774, 0, -495.7019514418122, 0, 3524.991654697331, 0, -4934.988316576264, 0, 0, 0, 939.997774585955, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.835933153488193, 0, -165.2339838139374, 0, 1762.495827348666, 0, -4934.988316576264, 0, 4229.989985636798, 0, -939.997774585955, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -19.37540204348254, 0, 0, 0, 0, 0, 0, -90.41854286958517, 0, 516.677387826201, 0, 0, 0, 0, 0, 0, 0, 0, -161.4616836956878, 0, 1894.483755362737, 0, -2893.393371826726, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -129.1693469565502, 0, 2411.161143188938, 0, -7715.715658204601, 0, 4960.10292313153, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -32.29233673913756, 0, 1033.354775652402, 0, -5786.786743653451, 0, 8266.838205219216, 0, -2755.612735073072, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12.91693469565502, 0, -172.225795942067, 0, 0, 0, 1653.367641043843, 0, -1837.075156715381, 0, 400.8163978288105, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6.458467347827512, 0, -172.225795942067, 0, 964.4644572755752, 0, -1653.367641043843, 0, 918.5375783576907, 0, -133.6054659429368, 0, 0, 0,
        0, 0.9043663200508067, 0, 0, 0, 0, 5.42619792030484, 0, -86.81916672487744, 0, 0, 0, 0, 0, 0, 13.5654948007621, 0, -434.0958336243872, 0, 1012.890278456903, 0, 0, 0, 0, 0, 0, 0, 0, 18.08732640101613, 0, -868.1916672487744, 0, 4051.561113827614, 0, -3241.248891062091, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 13.5654948007621, 0, -868.1916672487744, 0, 6077.341670741421, 0, -9723.746673186273, 0, 3472.766668995098, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5.42619792030484, 0, -434.0958336243872, 0, 4051.561113827614, 0, -9723.746673186273, 0, 6945.533337990195, 0, -1234.761482309368, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.9043663200508067, 0, -86.81916672487744, 0, 1012.890278456903, 0, -3241.248891062091, 0, 3472.766668995098, 0, -1234.761482309368, 0, 112.2510438463062, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 6.521478277491721, 0, 0, 0, 0, 0, 0, 39.12886966495032, 0, -182.6013917697682, 0, 0, 0, 0, 0, 0, 0, 0, 97.82217416237581, 0, -913.0069588488409, 0, 1095.608350618609, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 130.4295655498344, 0, -1826.013917697682, 0, 4382.433402474436, 0, -2086.873048797351, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 97.82217416237581, 0, -1826.013917697682, 0, 6573.650103711654, 0, -6260.619146392052, 0, 1391.248699198234, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 39.12886966495032, 0, -913.0069588488409, 0, 4382.433402474436, 0, -6260.619146392052, 0, 2782.497398396467, 0, -303.5451707341601, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6.521478277491721, 0, -182.6013917697682, 0, 1095.608350618609, 0, -2086.873048797351, 0, 1391.248699198234, 0, -303.5451707341601, 0, 15.56641901200821, 0,
        -0.3182155563368222, 0, 0, -2.227508894357756, 0, 31.18512452100858, 0, 0, 0, 0, -6.682526683073267, 0, 187.1107471260515, 0, -374.2214942521029, 0, 0, 0, 0, 0, 0, -11.13754447178878, 0, 467.7768678151287, 0, -1871.107471260515, 0, 1247.404980840343, 0, 0, 0, 0, 0, 0, 0, 0, -11.13754447178878, 0, 623.7024904201716, 0, -3742.214942521029, 0, 4989.619923361373, 0, -1425.605692388964, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -6.682526683073267, 0, 467.7768678151287, 0, -3742.214942521029, 0, 7484.429885042059, 0, -4276.817077166891, 0, 570.2422769555854, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.227508894357756, 0, 187.1107471260515, 0, -1871.107471260515, 0, 4989.619923361373, 0, -4276.817077166891, 0, 1140.484553911171, 0, -69.12027599461642, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.3182155563368222, 0, 31.18512452100858, 0, -374.2214942521029, 0, 1247.404980840343, 0, -1425.605692388964, 0, 570.2422769555854, 0, -69.12027599461642, 0, 1.519126944936625,
        0, 0, 6.521478277491721, 0, 0, 0, 0, 39.12886966495032, 0, -182.6013917697682, 0, 0, 0, 0, 0, 0, 97.82217416237581, 0, -913.0069588488409, 0, 1095.608350618609, 0, 0, 0, 0, 0, 0, 0, 0, 130.4295655498344, 0, -1826.013917697682, 0, 4382.433402474436, 0, -2086.873048797351, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 97.82217416237581, 0, -1826.013917697682, 0, 6573.650103711654, 0, -6260.619146392052, 0, 1391.248699198234, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 39.12886966495032, 0, -913.0069588488409, 0, 4382.433402474436, 0, -6260.619146392052, 0, 2782.497398396467, 0, -303.5451707341601, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6.521478277491721, 0, -182.6013917697682, 0, 1095.608350618609, 0, -2086.873048797351, 0, 1391.248699198234, 0, -303.5451707341601, 0, 15.56641901200821, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.4521831600254033, 0, 0, 2.260915800127017, 0, -43.40958336243872, 0, 0, 0, 0, 4.06964844022863, 0, -173.6383334497549, 0, 506.4451392284517, 0, 0, 0, 0, 0, 0, 2.260915800127017, 0, -217.0479168121936, 0, 1519.335417685355, 0, -1620.624445531046, 0, 0, 0, 0, 0, 0, 0, 0, -2.260915800127017, 0, 0, 0, 1012.890278456903, 0, -3241.248891062091, 0, 1736.383334497549, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4.06964844022863, 0, 217.0479168121936, 0, -1012.890278456903, 0, 0, 0, 1736.383334497549, 0, -617.380741154684, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.260915800127017, 0, 173.6383334497549, 0, -1519.335417685355, 0, 3241.248891062091, 0, -1736.383334497549, 0, 0, 0, 56.12552192315309, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.4521831600254033, 0, 43.40958336243872, 0, -506.4451392284517, 0, 1620.624445531046, 0, -1736.383334497549, 0, 617.380741154684, 0, -56.12552192315309, 0, 0,
        0, 0, -6.458467347827512, 0, 0, 0, 0, -12.91693469565502, 0, 172.225795942067, 0, 0, 0, 0, 0, 0, 32.29233673913756, 0, 172.225795942067, 0, -964.4644572755752, 0, 0, 0, 0, 0, 0, 0, 0, 129.1693469565502, 0, -1033.354775652402, 0, 0, 0, 1653.367641043843, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 161.4616836956878, 0, -2411.161143188938, 0, 5786.786743653451, 0, -1653.367641043843, 0, -918.5375783576907, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 90.41854286958517, 0, -1894.483755362737, 0, 7715.715658204601, 0, -8266.838205219216, 0, 1837.075156715381, 0, 133.6054659429368, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 19.37540204348254, 0, -516.677387826201, 0, 2893.393371826726, 0, -4960.10292313153, 0, 2755.612735073072, 0, -400.8163978288105, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.4589832883720484, 0, 0, 0.4589832883720484, 0, 41.30849595348435, 0, 0, 0, 0, 8.720682479068919, 0, -82.6169919069687, 0, -440.6239568371664, 0, 0, 0, 0, 0, 0, 20.65424797674218, 0, -702.244431209234, 0, 1321.871870511499, 0, 1233.747079144066, 0, 0, 0, 0, 0, 0, 0, 0, 20.65424797674218, 0, -1156.637886697562, 0, 6168.73539572033, 0, -4934.988316576264, 0, -1057.497496409199, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 8.720682479068919, 0, -702.244431209234, 0, 6168.73539572033, 0, -12337.47079144066, 0, 5287.487482045997, 0, 234.9994436464888, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.4589832883720484, 0, -82.6169919069687, 0, 1321.871870511499, 0, -4934.988316576264, 0, 5287.487482045997, 0, -1409.996661878933, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.4589832883720484, 0, 41.30849595348435, 0, -440.6239568371664, 0, 1233.747079144066, 0, -1057.497496409199, 0, 234.9994436464888, 0, 0, 0, 0,
        0, 0, 6.326648023315, 0, 0, 0, 0, -37.95988813989, 0, -151.83955255956, 0, 0, 0, 0, 0, 0, -183.472792676135, 0, 1062.87686791692, 0, 728.8298522858879, 0, 0, 0, 0, 0, 0, 0, 0, -227.75932883934, 0, 3340.47015631032, 0, -5830.638818287104, 0, -971.7731363811839, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -56.939832209835, 0, 2125.75373583384, 0, -10203.61793200243, 0, 8745.958227430655, 0, 323.924378793728, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 63.26648023315, 0, -759.1977627977999, 0, 0, 0, 4858.86568190592, 0, -3239.24378793728, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 31.633240116575, 0, -759.1977627977999, 0, 3644.14926142944, 0, -4858.86568190592, 0, 1619.62189396864, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.4715605016615671, 0, 0, -5.187165518277238, 0, -37.72484013292537, 0, 0, 0, 0, -18.39085956480112, 0, 452.6980815951044, 0, 339.5235611963283, 0, 0, 0, 0, 0, 0, -12.73213354486231, 0, 1018.570683588985, 0, -4413.806295552268, 0, -724.3169305521671, 0, 0, 0, 0, 0, 0, 0, 0, 12.73213354486231, 0, 0, 0, -4753.329856748596, 0, 10140.43702773034, 0, 362.1584652760835, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 18.39085956480112, 0, -1018.570683588985, 0, 4753.329856748596, 0, 0, 0, -5432.376979141253, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5.187165518277238, 0, -452.6980815951044, 0, 4413.806295552268, 0, -10140.43702773034, 0, 5432.376979141253, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.4715605016615671, 0, 37.72484013292537, 0, -339.5235611963283, 0, 724.3169305521671, 0, -362.1584652760835, 0, 0, 0, 0, 0, 0,
        0, 0, -6.112122669759491, 0, 0, 0, 0, 110.0182080556708, 0, 122.2424533951898, 0, 0, 0, 0, 0, 0, 152.8030667439873, 0, -2322.606614508606, 0, -440.0728322226833, 0, 0, 0, 0, 0, 0, 0, 0, -220.0364161113417, 0, -733.4547203711389, 0, 8801.456644453667, 0, 335.2935864553778, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -385.0637281948479, 0, 5134.183042597972, 0, -6161.019651117567, 0, -7041.165315562933, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -85.56971737663287, 0, 2567.091521298986, 0, -12322.03930223513, 0, 11735.27552593822, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 42.78485868831644, 0, -855.6971737663287, 0, 3080.509825558783, 0, -2347.055105187644, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.4925288806232314, 0, 0, 12.31322201558078, 0, 32.50690612113327, 0, 0, 0, 0, 5.417817686855545, 0, -845.179559149465, 0, -216.7127074742218, 0, 0, 0, 0, 0, 0, -48.7603591816999, 0, 487.603591816999, 0, 5851.243101803988, 0, 260.0552489690662, 0, 0, 0, 0, 0, 0, 0, 0, -48.7603591816999, 0, 2730.580114175195, 0, -9101.933713917315, 0, -7281.546971133852, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5.417817686855545, 0, 487.603591816999, 0, -9101.933713917315, 0, 18203.86742783463, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12.31322201558078, 0, -845.179559149465, 0, 5851.243101803988, 0, -7281.546971133852, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.4925288806232314, 0, 32.50690612113327, 0, -216.7127074742218, 0, 260.0552489690662, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 5.78590428180594, 0, 0, 0, 0, -196.720745581402, 0, -84.85992946648712, 0, 0, 0, 0, 0, 0, 318.2247354993267, 0, 2970.097531327049, 0, 169.7198589329742, 0, 0, 0, 0, 0, 0, 0, 0, 763.7393651983841, 0, -7637.393651983841, 0, -6109.914921587073, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -190.934841299596, 0, -3564.117037592459, 0, 21384.70222555475, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -381.8696825991921, 0, 6364.494709986534, 0, -14256.46815036984, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 52.07313853625346, 0, -763.7393651983841, 0, 1527.478730396768, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.5281783817851402, 0, 0, -22.71167041676103, 0, -25.35256232568673, 0, 0, 0, 0, 63.90958419600196, 0, 1115.512742330216, 0, 92.95939519418467, 0, 0, 0, 0, 0, 0, 87.14943299454813, 0, -4183.17278373831, 0, -4183.17278373831, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -87.14943299454813, 0, 0, 0, 19521.47299077878, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -63.90958419600196, 0, 4183.17278373831, 0, -19521.47299077878, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 22.71167041676103, 0, -1115.512742330216, 0, 4183.17278373831, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.5281783817851402, 0, 25.35256232568673, 0, -92.95939519418467, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -5.281783817851402, 0, 0, 0, 0, 285.2163261639757, 0, 42.25427054281121, 0, 0, 0, 0, 0, 0, -1452.490549909135, 0, -2323.984879854617, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 697.195463956385, 0, 13943.9092791277, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1568.689793901866, 0, -19521.47299077878, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -813.3947079491158, 0, 6971.95463956385, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 58.09962199636542, 0, -464.7969759709233, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.5980442516269705, 0, 0, 38.87287635575308, 0, 15.54915054230123, 0, 0, 0, 0, -256.5609839479703, 0, -1026.243935791881, 0, 0, 0, 0, 0, 0, 0, 0, 256.5609839479703, 0, 7696.82951843911, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 256.5609839479703, 0, -14367.41510108634, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -256.5609839479703, 0, 7696.82951843911, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 38.87287635575308, 0, -1026.243935791881, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.5980442516269705, 0, 15.54915054230123, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 4.394709780272118, 0, 0, 0, 0, -342.7873628612252, 0, 0, 0, 0, 0, 0, 0, 0, 3142.217492894565, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -7541.321982946955, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5655.991487210216, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1256.886997157826, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 57.13122714353754, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.830522083064524, 0, 0, -75.57750955887168, 0, 0, 0, 0, 0, 0, 831.3526051475885, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2494.057815442766, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2494.057815442766, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -831.3526051475885, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 75.57750955887168, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.830522083064524, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // l = 15
        0, 12.66375976286059, 0, 0, 0, 0, -384.1340461401045, 0, 0, 0, 0, 0, 0, 0, 0, 2535.28470452469, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -5432.752938267193, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4225.47450754115, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1152.402138420314, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 88.64631834002413, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.8442506508573726, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 64.73811759282017, 0, 0, 0, 0, 0, 0, -1683.191057413324, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 9257.550815773284, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -15870.0871127542, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 9257.550815773284, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1683.191057413324, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64.73811759282017, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -7.893350484654575, 0, 0, 0, 0, 165.7603601777461, 0, 221.0138135703281, 0, 0, 0, 0, 0, 0, -607.7879873184023, 0, -4862.303898547218, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 260.480565993601, 0, 21880.36754346248, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 607.7879873184023, 0, -29173.82339128331, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -386.7741737480742, 0, 12155.75974636805, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 46.75292210141556, 0, -1326.082881421969, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.6071808065118904, 0, 17.00106258233293, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -66.77884815276176, 0, 0, 0, 0, 0, 0, 1157.500034647871, 0, 578.7500173239353, 0, 0, 0, 0, 0, 0, 0, 0, -3183.125095281644, 0, -10610.41698427215, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 38197.50114337973, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3183.125095281644, 0, -38197.50114337973, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1157.500034647871, 0, 10610.41698427215, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66.77884815276176, 0, -578.7500173239353, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 5.890314521234328, 0, 0, 0, 0, -76.57408877604626, 0, -306.2963551041851, 0, 0, 0, 0, 0, 0, 76.57408877604626, 0, 4288.148971458591, 0, 1225.18542041674, 0, 0, 0, 0, 0, 0, 0, 0, 229.7222663281388, 0, -8270.001587812996, 0, -18377.7813062511, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -76.57408877604626, 0, -3675.556261250221, 0, 51457.78765750309, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -118.3417735629806, 0, 7657.408877604626, 0, -36755.56261250221, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 28.38060632958358, 0, -1503.636652329636, 0, 6125.927102083701, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.5354831382940298, 0, 27.84512319128955, 0, -111.3804927651582, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 61.05447148378159, 0, 0, 0, 0, 0, 0, -610.5447148378159, 0, -976.8715437405055, 0, 0, 0, 0, 0, 0, 0, 0, 134.3198372643195, 0, 10745.58698114556, 0, 2149.117396229112, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1611.838047171834, 0, -12894.70437737467, 0, -25789.40875474934, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 134.3198372643195, 0, -12894.70437737467, 0, 54157.75838497362, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -610.5447148378159, 0, 10745.58698114556, 0, -25789.40875474934, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 61.05447148378159, 0, -976.8715437405055, 0, 2149.117396229112, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -4.486569049517132, 0, 0, 0, 0, 28.4149373136085, 0, 323.0329715652335, 0, 0, 0, 0, 0, 0, 49.35225954468845, 0, -2368.908458145046, 0, -2368.908458145046, 0, 0, 0, 0, 0, 0, 0, 0, -49.35225954468845, 0, -1184.454229072523, 0, 19740.90381787538, 0, 3158.544610860061, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -93.22093469552263, 0, 4737.816916290091, 0, -11054.90613801021, 0, -29479.74970136057, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -10.46866111553997, 0, 1974.090381787538, 0, -23689.08458145046, 0, 44219.62455204085, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16.45075318156282, 0, -1220.34678146866, 0, 9212.421781675177, 0, -12634.17844344024, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.4985076721685702, 0, 35.89255239613705, 0, -263.2120509050051, 0, 350.9494012066734, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -51.69118335186266, 0, 0, 0, 0, 0, 0, 206.7647334074506, 0, 1137.206033740979, 0, 0, 0, 0, 0, 0, 0, 0, 568.6030168704893, 0, -5686.030168704893, 0, -4548.824134963914, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -6823.236202445871, 0, 27292.94480978348, 0, 3898.992115683355, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -568.6030168704893, 0, 6823.236202445871, 0, 0, 0, -27292.94480978348, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -206.7647334074506, 0, 5686.030168704893, 0, -27292.94480978348, 0, 27292.94480978348, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 51.69118335186266, 0, -1137.206033740979, 0, 4548.824134963914, 0, -3898.992115683355, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 3.334384020345036, 0, 0, 0, 0, -3.334384020345036, 0, -293.4257937903632, 0, 0, 0, 0, 0, 0, -36.6782242237954, 0, 586.8515875807264, 0, 2934.257937903632, 0, 0, 0, 0, 0, 0, 0, 0, -47.15771685916551, 0, 2640.832144113269, 0, -8802.773813710896, 0, -7042.219050968717, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -5.239746317685057, 0, 1509.046939493296, 0, -17605.54762742179, 0, 28168.87620387487, 0, 4024.125171982124, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 20.48264469640522, 0, -1047.949263537011, 0, 2515.078232488827, 0, 14084.43810193743, 0, -20120.62585991062, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 8.097789763695088, 0, -754.5234697466482, 0, 7964.414402881287, 0, -20120.62585991062, 0, 12072.37551594637, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.4763405743350052, 0, 41.91797054148046, 0, -419.1797054148046, 0, 1006.031292995531, 0, -574.8750245688748, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 40.21623606427654, 0, 0, 0, 0, 0, 0, 26.81082404285103, 0, -1072.432961714041, 0, 0, 0, 0, 0, 0, 0, 0, -254.7028284070847, 0, 357.4776539046803, 0, 5791.137993255822, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -482.5948327713185, 0, 6434.597770284246, 0, -7721.517324341095, 0, -8824.591227818395, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -254.7028284070847, 0, 6434.597770284246, 0, -27025.31063519383, 0, 20590.71286490959, 0, 3431.785477484931, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 26.81082404285103, 0, 357.4776539046803, 0, -7721.517324341095, 0, 20590.71286490959, 0, -11439.28492494977, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40.21623606427654, 0, -1072.432961714041, 0, 5791.137993255822, 0, -8824.591227818395, 0, 3431.785477484931, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -2.312653286194929, 0, 0, 0, 0, -6.937959858584787, 0, 231.2653286194929, 0, 0, 0, 0, 0, 0, -0.4625306572389858, 0, 462.5306572389858, 0, -2775.183943433915, 0, 0, 0, 0, 0, 0, 0, 0, 20.81387957575436, 0, -416.2775915150872, 0, -2775.183943433915, 0, 8880.588618988527, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 30.06449272053408, 0, -1665.110366060349, 0, 7770.515041614961, 0, 0, 0, -8880.588618988527, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16.1885730033645, 0, -1341.338905993059, 0, 12210.80935110922, 0, -24865.64813316788, 0, 8880.588618988527, 0, 2368.156965063607, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.312653286194929, 0, -277.5183943433915, 0, 3885.257520807481, 0, -14208.94179038164, 0, 15985.05951417935, 0, -4736.313930127214, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.4625306572389858, 0, 46.25306572389858, 0, -555.0367886867829, 0, 1776.117723797705, 0, -1776.117723797705, 0, 473.6313930127214, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -27.44175328360903, 0, 0, 0, 0, 0, 0, -109.7670131344361, 0, 823.252598508271, 0, 0, 0, 0, 0, 0, 0, 0, -137.2087664180452, 0, 2469.757795524813, 0, -5268.816630452934, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1646.505197016542, 0, -10537.63326090587, 0, 10537.63326090587, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 137.2087664180452, 0, -1646.505197016542, 0, 0, 0, 10537.63326090587, 0, -7025.088840603912, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 109.7670131344361, 0, -2469.757795524813, 0, 10537.63326090587, 0, -10537.63326090587, 0, 0, 0, 1277.288880109802, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 27.44175328360903, 0, -823.252598508271, 0, 5268.816630452934, 0, -10537.63326090587, 0, 7025.088840603912, 0, -1277.288880109802, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 1.363030880952603, 0, 0, 0, 0, 7.723841658731416, 0, -147.2073351428811, 0, 0, 0, 0, 0, 0, 17.71940145238384, 0, -686.9675640001119, 0, 1962.764468571748, 0, 0, 0, 0, 0, 0, 0, 0, 20.44546321428904, 0, -1226.727792857343, 0, 7196.803051429743, 0, -7327.654016001193, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 11.35859067460502, 0, -981.3822342858741, 0, 9159.567520001492, 0, -19540.41070933652, 0, 9421.269449144391, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.363030880952603, 0, -245.3455585714685, 0, 3925.528937143496, 0, -14655.30803200239, 0, 15702.11574857399, 0, -4187.230866286396, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.363030880952603, 0, 98.13822342858741, 0, -654.2548228572494, 0, 0, 0, 3140.423149714797, 0, -2791.487244190931, 0, 507.543135307442, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.454343626984201, 0, 49.0691117142937, 0, -654.2548228572494, 0, 2442.551338667064, 0, -3140.423149714797, 0, 1395.743622095465, 0, -169.1810451024807, 0, 0, 0,
        0, 0, 0, 0, 13.90024211921377, 0, 0, 0, 0, 0, 0, 83.40145271528264, 0, -444.8077478148407, 0, 0, 0, 0, 0, 0, 0, 0, 208.5036317882066, 0, -2224.038739074204, 0, 3113.654234703885, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 278.0048423842755, 0, -4448.077478148407, 0, 12454.61693881554, 0, -7116.923965037452, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 208.5036317882066, 0, -4448.077478148407, 0, 18681.92540822331, 0, -21350.77189511235, 0, 5930.769970864543, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 83.40145271528264, 0, -2224.038739074204, 0, 12454.61693881554, 0, -21350.77189511235, 0, 11861.53994172909, 0, -1725.31490061514, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 13.90024211921377, 0, -444.8077478148407, 0, 3113.654234703885, 0, -7116.923965037452, 0, 5930.769970864543, 0, -1725.31490061514, 0, 132.7165308165492, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -0.4505094349975498, 0, 0, 0, 0, -3.153566044982849, 0, 50.45705671972558, 0, 0, 0, 0, 0, 0, -9.460698134948546, 0, 302.7423403183535, 0, -706.3987940761581, 0, 0, 0, 0, 0, 0, 0, 0, -15.76783022491424, 0, 756.8558507958837, 0, -3531.99397038079, 0, 2825.595176304632, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -15.76783022491424, 0, 1009.141134394512, 0, -7063.987940761581, 0, 11302.38070521853, 0, -4036.564537578046, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -9.460698134948546, 0, 756.8558507958837, 0, -7063.987940761581, 0, 16953.57105782779, 0, -12109.69361273414, 0, 2152.834420041625, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3.153566044982849, 0, 302.7423403183535, 0, -3531.99397038079, 0, 11302.38070521853, 0, -12109.69361273414, 0, 4305.668840083249, 0, -391.4244400075681, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.4505094349975498, 0, 50.45705671972558, 0, -706.3987940761581, 0, 2825.595176304632, 0, -4036.564537578046, 0, 2152.834420041625, 0, -391.4244400075681, 0, 17.20546989044255, 0,
        0, 0, -4.935083598341307, 0, 0, 0, 0, -34.54558518838915, 0, 161.2127308791494, 0, 0, 0, 0, 0, 0, -103.6367555651675, 0, 967.2763852748962, 0, -1160.731662329875, 0, 0, 0, 0, 0, 0, 0, 0, -172.7279259419458, 0, 2418.190963187241, 0, -5803.658311649377, 0, 2763.646815071132, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -172.7279259419458, 0, 3224.254617582987, 0, -11607.31662329875, 0, 11054.58726028453, 0, -2456.574946729895, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -103.6367555651675, 0, 2418.190963187241, 0, -11607.31662329875, 0, 16581.88089042679, 0, -7369.724840189685, 0, 803.9699825661475, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -34.54558518838915, 0, 967.2763852748962, 0, -5803.658311649377, 0, 11054.58726028453, 0, -7369.724840189685, 0, 1607.939965132295, 0, -82.4584597503741, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4.935083598341307, 0, 161.2127308791494, 0, -1160.731662329875, 0, 2763.646815071132, 0, -2456.574946729895, 0, 803.9699825661475, 0, -82.4584597503741, 0, 1.570637328578554,
        -0.4505094349975498, 0, 0, -3.153566044982849, 0, 50.45705671972558, 0, 0, 0, 0, -9.460698134948546, 0, 302.7423403183535, 0, -706.3987940761581, 0, 0, 0, 0, 0, 0, -15.76783022491424, 0, 756.8558507958837, 0, -3531.99397038079, 0, 2825.595176304632, 0, 0, 0, 0, 0, 0, 0, 0, -15.76783022491424, 0, 1009.141134394512, 0, -7063.987940761581, 0, 11302.38070521853, 0, -4036.564537578046, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -9.460698134948546, 0, 756.8558507958837, 0, -7063.987940761581, 0, 16953.57105782779, 0, -12109.69361273414, 0, 2152.834420041625, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3.153566044982849, 0, 302.7423403183535, 0, -3531.99397038079, 0, 11302.38070521853, 0, -12109.69361273414, 0, 4305.668840083249, 0, -391.4244400075681, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.4505094349975498, 0, 50.45705671972558, 0, -706.3987940761581, 0, 2825.595176304632, 0, -4036.564537578046, 0, 2152.834420041625, 0, -391.4244400075681, 0, 17.20546989044255, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 6.950121059606886, 0, 0, 0, 0, 34.75060529803443, 0, -222.4038739074204, 0, 0, 0, 0, 0, 0, 62.55108953646198, 0, -889.6154956296814, 0, 1556.827117351943, 0, 0, 0, 0, 0, 0, 0, 0, 34.75060529803443, 0, -1112.019369537102, 0, 4670.481352055828, 0, -3558.461982518726, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -34.75060529803443, 0, 0, 0, 3113.654234703885, 0, -7116.923965037452, 0, 2965.384985432271, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -62.55108953646198, 0, 1112.019369537102, 0, -3113.654234703885, 0, 0, 0, 2965.384985432271, 0, -862.6574503075699, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -34.75060529803443, 0, 889.6154956296814, 0, -4670.481352055828, 0, 7116.923965037452, 0, -2965.384985432271, 0, 0, 0, 66.35826540827461, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -6.950121059606886, 0, 222.4038739074204, 0, -1556.827117351943, 0, 3558.461982518726, 0, -2965.384985432271, 0, 862.6574503075699, 0, -66.35826540827461, 0, 0,
        0.454343626984201, 0, 0, 1.363030880952603, 0, -49.0691117142937, 0, 0, 0, 0, -1.363030880952603, 0, -98.13822342858741, 0, 654.2548228572494, 0, 0, 0, 0, 0, 0, -11.35859067460502, 0, 245.3455585714685, 0, 654.2548228572494, 0, -2442.551338667064, 0, 0, 0, 0, 0, 0, 0, 0, -20.44546321428904, 0, 981.3822342858741, 0, -3925.528937143496, 0, 0, 0, 3140.423149714797, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -17.71940145238384, 0, 1226.727792857343, 0, -9159.567520001492, 0, 14655.30803200239, 0, -3140.423149714797, 0, -1395.743622095465, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -7.723841658731416, 0, 686.9675640001119, 0, -7196.803051429743, 0, 19540.41070933652, 0, -15702.11574857399, 0, 2791.487244190931, 0, 169.1810451024807, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.363030880952603, 0, 147.2073351428811, 0, -1962.764468571748, 0, 7327.654016001193, 0, -9421.269449144391, 0, 4187.230866286396, 0, -507.543135307442, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -6.860438320902258, 0, 0, 0, 0, 6.860438320902258, 0, 205.8131496270677, 0, 0, 0, 0, 0, 0, 130.3483280971429, 0, -411.6262992541355, 0, -1317.204157613234, 0, 0, 0, 0, 0, 0, 0, 0, 308.7197244406016, 0, -3498.823543660152, 0, 3951.612472839701, 0, 2634.408315226467, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 308.7197244406016, 0, -5762.768189557897, 0, 18440.85820658527, 0, -10537.63326090587, 0, -1756.272210150978, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 130.3483280971429, 0, -3498.823543660152, 0, 18440.85820658527, 0, -26344.08315226467, 0, 8781.36105075489, 0, 319.3222200274506, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6.860438320902258, 0, -411.6262992541355, 0, 3951.612472839701, 0, -10537.63326090587, 0, 8781.36105075489, 0, -1915.933320164703, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -6.860438320902258, 0, 205.8131496270677, 0, -1317.204157613234, 0, 2634.408315226467, 0, -1756.272210150978, 0, 319.3222200274506, 0, 0, 0, 0,
        -0.4625306572389858, 0, 0, 2.312653286194929, 0, 46.25306572389858, 0, 0, 0, 0, 16.1885730033645, 0, -277.5183943433915, 0, -555.0367886867829, 0, 0, 0, 0, 0, 0, 30.06449272053408, 0, -1341.338905993059, 0, 3885.257520807481, 0, 1776.117723797705, 0, 0, 0, 0, 0, 0, 0, 0, 20.81387957575436, 0, -1665.110366060349, 0, 12210.80935110922, 0, -14208.94179038164, 0, -1776.117723797705, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.4625306572389858, 0, -416.2775915150872, 0, 7770.515041614961, 0, -24865.64813316788, 0, 15985.05951417935, 0, 473.6313930127214, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -6.937959858584787, 0, 462.5306572389858, 0, -2775.183943433915, 0, 0, 0, 8880.588618988527, 0, -4736.313930127214, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.312653286194929, 0, 231.2653286194929, 0, -2775.183943433915, 0, 8880.588618988527, 0, -8880.588618988527, 0, 2368.156965063607, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 6.702706010712756, 0, 0, 0, 0, -73.72976611784032, 0, -178.7388269523402, 0, 0, 0, 0, 0, 0, -261.4055344177975, 0, 2144.865923428082, 0, 965.1896655426369, 0, 0, 0, 0, 0, 0, 0, 0, -180.9730622892444, 0, 4825.948327713185, 0, -12547.46565205428, 0, -1470.765204636399, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 180.9730622892444, 0, 0, 0, -13512.65531759692, 0, 20590.71286490959, 0, 571.9642462474885, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 261.4055344177975, 0, -4825.948327713185, 0, 13512.65531759692, 0, 0, 0, -8579.463693712328, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 73.72976611784032, 0, -2144.865923428082, 0, 12547.46565205428, 0, -20590.71286490959, 0, 8579.463693712328, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -6.702706010712756, 0, 178.7388269523402, 0, -965.1896655426369, 0, 1470.765204636399, 0, -571.9642462474885, 0, 0, 0, 0, 0, 0,
        0.4763405743350052, 0, 0, -8.097789763695088, 0, -41.91797054148046, 0, 0, 0, 0, -20.48264469640522, 0, 754.5234697466482, 0, 419.1797054148046, 0, 0, 0, 0, 0, 0, 5.239746317685057, 0, 1047.949263537011, 0, -7964.414402881287, 0, -1006.031292995531, 0, 0, 0, 0, 0, 0, 0, 0, 47.15771685916551, 0, -1509.046939493296, 0, -2515.078232488827, 0, 20120.62585991062, 0, 574.8750245688748, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 36.6782242237954, 0, -2640.832144113269, 0, 17605.54762742179, 0, -14084.43810193743, 0, -12072.37551594637, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.334384020345036, 0, -586.8515875807264, 0, 8802.773813710896, 0, -28168.87620387487, 0, 20120.62585991062, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3.334384020345036, 0, 293.4257937903632, 0, -2934.257937903632, 0, 7042.219050968717, 0, -4024.125171982124, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -6.461397918982832, 0, 0, 0, 0, 161.5349479745708, 0, 142.1507542176223, 0, 0, 0, 0, 0, 0, 71.07537710881116, 0, -3695.91960965818, 0, -568.6030168704893, 0, 0, 0, 0, 0, 0, 0, 0, -639.6783939793004, 0, 2132.261313264335, 0, 15352.28145550321, 0, 487.3740144604194, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -639.6783939793004, 0, 11940.66335428027, 0, -23881.32670856055, 0, -13646.47240489174, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 71.07537710881116, 0, 2132.261313264335, 0, -23881.32670856055, 0, 34116.18101222936, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 161.5349479745708, 0, -3695.91960965818, 0, 15352.28145550321, 0, -13646.47240489174, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -6.461397918982832, 0, 142.1507542176223, 0, -568.6030168704893, 0, 487.3740144604194, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.4985076721685702, 0, 0, 16.45075318156282, 0, 35.89255239613705, 0, 0, 0, 0, -10.46866111553997, 0, -1220.34678146866, 0, -263.2120509050051, 0, 0, 0, 0, 0, 0, -93.22093469552263, 0, 1974.090381787538, 0, 9212.421781675177, 0, 350.9494012066734, 0, 0, 0, 0, 0, 0, 0, 0, -49.35225954468845, 0, 4737.816916290091, 0, -23689.08458145046, 0, -12634.17844344024, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 49.35225954468845, 0, -1184.454229072523, 0, -11054.90613801021, 0, 44219.62455204085, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 28.4149373136085, 0, -2368.908458145046, 0, 19740.90381787538, 0, -29479.74970136057, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4.486569049517132, 0, 323.0329715652335, 0, -2368.908458145046, 0, 3158.544610860061, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 6.105447148378159, 0, 0, 0, 0, -262.5342273802609, 0, -97.68715437405055, 0, 0, 0, 0, 0, 0, 738.7591049537573, 0, 4298.234792458224, 0, 214.9117396229112, 0, 0, 0, 0, 0, 0, 0, 0, 1007.398779482396, 0, -16118.38047171834, 0, -9671.028283031004, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1007.398779482396, 0, 0, 0, 45131.46532081135, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -738.7591049537573, 0, 16118.38047171834, 0, -45131.46532081135, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 262.5342273802609, 0, -4298.234792458224, 0, 9671.028283031004, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -6.105447148378159, 0, 97.68715437405055, 0, -214.9117396229112, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.5354831382940298, 0, 0, -28.38060632958358, 0, -27.84512319128955, 0, 0, 0, 0, 118.3417735629806, 0, 1503.636652329636, 0, 111.3804927651582, 0, 0, 0, 0, 0, 0, 76.57408877604626, 0, -7657.408877604626, 0, -6125.927102083701, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -229.7222663281388, 0, 3675.556261250221, 0, 36755.56261250221, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -76.57408877604626, 0, 8270.001587812996, 0, -51457.78765750309, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 76.57408877604626, 0, -4288.148971458591, 0, 18377.7813062511, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -5.890314521234328, 0, 306.2963551041851, 0, -1225.18542041674, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -5.564904012730147, 0, 0, 0, 0, 361.7187608274595, 0, 48.22916811032794, 0, 0, 0, 0, 0, 0, -2387.343821461233, 0, -3183.125095281644, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2387.343821461233, 0, 23873.43821461233, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2387.343821461233, 0, -44563.75133394302, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2387.343821461233, 0, 23873.43821461233, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 361.7187608274595, 0, -3183.125095281644, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -5.564904012730147, 0, 48.22916811032794, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.6071808065118904, 0, 0, 46.75292210141556, 0, 17.00106258233293, 0, 0, 0, 0, -386.7741737480742, 0, -1326.082881421969, 0, 0, 0, 0, 0, 0, 0, 0, 607.7879873184023, 0, 12155.75974636805, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 260.480565993601, 0, -29173.82339128331, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -607.7879873184023, 0, 21880.36754346248, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 165.7603601777461, 0, -4862.303898547218, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -7.893350484654575, 0, 221.0138135703281, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 4.624151256630012, 0, 0, 0, 0, -420.7977643533311, 0, 0, 0, 0, 0, 0, 0, 0, 4628.775407886642, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -13886.32622365993, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 13886.32622365993, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4628.775407886642, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 420.7977643533311, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4.624151256630012, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.8442506508573726, 0, 0, -88.64631834002413, 0, 0, 0, 0, 0, 0, 1152.402138420314, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4225.47450754115, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5432.752938267193, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2535.28470452469, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 384.1340461401045, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -12.66375976286059, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
};

/*
 * Real part of cartesian to spinor transformation
 * / xyz_alpha \
 * \ xyz_beta  /
 */
static double g_trans_cart2jR[] = {
        0,
        1,
        1,
        0,
       -0.577350269189625764,  // p1/2
        0,
        0,
        0,
        0,
        0.577350269189625764,
        0,
        0,
       -0.577350269189625764,
       -0.577350269189625764,
        0,
        0,
        0,  // p3/2
        0,
        0,
        0.707106781186547524,
        0,
        0,
        0.408248290463863016,
        0,
        0,
        0,
        0,
        0.816496580927726033,
        0,
        0,
        0.816496580927726033,
       -0.408248290463863016,
        0,
        0,
       -0.707106781186547524,
        0,
        0,
        0,
        0,
        0,
       -0.345494149471335479,  // d3/2
        0,
        0,
        0.345494149471335479,
        0,
        0,
        0,
        0,
        0.345494149471335479,
        0,
        0,
        0,
        0,
        0,
       -0.598413420602149016,
        0,
        0,
        0,
       -0.199471140200716338,
        0,
        0,
       -0.199471140200716338,
        0,
        0.398942280401432677,
        0.199471140200716338,
        0,
        0,
        0.199471140200716338,
        0,
       -0.398942280401432677,
        0,
        0,
       -0.598413420602149016,
        0,
        0,
        0,
        0,
        0,
        0.345494149471335479,
        0,
        0,
        0,
        0.345494149471335479,
        0,
        0,
       -0.345494149471335479,
        0,
        0,
        0,  // d5/2
        0,
        0,
        0,
        0,
        0,
        0.386274202023189580,
        0,
        0,
       -0.386274202023189580,
        0,
        0,
        0.172747074735667739,
        0,
        0,
       -0.172747074735667739,
        0,
        0,
        0,
        0,
        0.690988298942670958,
        0,
        0,
        0,
        0,
        0,
        0.488602511902919921,
        0,
        0,
        0,
       -0.244301255951459960,
        0,
        0,
       -0.244301255951459960,
        0,
        0.488602511902919921,
       -0.244301255951459960,
        0,
        0,
       -0.244301255951459960,
        0,
        0.488602511902919921,
        0,
        0,
       -0.488602511902919921,
        0,
        0,
        0,
        0,
        0,
       -0.690988298942670958,
        0,
        0,
        0,
        0.172747074735667739,
        0,
        0,
       -0.172747074735667739,
        0,
        0,
        0.386274202023189580,
        0,
        0,
       -0.386274202023189580,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
       -0.386274202023189580,  // f5/2
        0,
        0,
        1.158822606069568741,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0.386274202023189580,
        0,
        0,
        0,
        0,
       -0.386274202023189580,
        0,
        0,
        0,
        0,
       -0.863735373678338698,
        0,
        0,
        0,
        0,
        0.863735373678338698,
        0,
        0,
       -0.172747074735667739,
        0,
        0,
       -0.172747074735667739,
        0,
        0.690988298942670958,
        0,
        0,
        0,
        0,
        0.244301255951459960,
        0,
        0,
        0.244301255951459960,
        0,
       -0.977205023805839843,
        0,
        0,
        0,
        0,
        0,
        0,
       -0.732903767854379882,
        0,
        0,
        0,
        0,
       -0.732903767854379882,
        0,
        0.488602511902919921,
        0,
        0,
        0.732903767854379882,
        0,
        0,
        0,
        0,
        0.732903767854379882,
        0,
       -0.488602511902919921,
        0.244301255951459960,
        0,
        0,
        0.244301255951459960,
        0,
       -0.977205023805839843,
        0,
        0,
        0,
        0,
       -0.172747074735667739,
        0,
        0,
       -0.172747074735667739,
        0,
        0.690988298942670958,
        0,
        0,
        0,
        0,
        0,
        0,
        0.863735373678338698,
        0,
        0,
        0,
        0,
       -0.863735373678338698,
        0,
        0,
        0,
        0,
       -0.386274202023189580,
        0,
        0,
        0,
        0,
        0.386274202023189580,
        0,
        0,
       -0.386274202023189580,
        0,
        0,
        1.158822606069568741,
        0,
        0,
        0,
        0,
        0,
        0,
        0,  // f7/2
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0.417223823632784089,
        0,
        0,
       -1.251671470898352269,
        0,
        0,
        0,
        0,
        0,
        0,
        0.157695782626260003,
        0,
        0,
       -0.473087347878780009,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0.946174695757560018,
        0,
        0,
        0,
        0,
       -0.946174695757560018,
        0,
        0,
        0,
        0,
        0.546274215296039535,
        0,
        0,
        0,
        0,
       -0.546274215296039535,
        0,
        0,
       -0.273137107648019767,
        0,
        0,
       -0.273137107648019767,
        0,
        1.092548430592079070,
        0,
        0,
        0,
        0,
       -0.211571093830408607,
        0,
        0,
       -0.211571093830408607,
        0,
        0.846284375321634430,
        0,
        0,
        0,
        0,
        0,
        0,
       -0.846284375321634430,
        0,
        0,
        0,
        0,
       -0.846284375321634430,
        0,
        0.564189583547756286,
        0,
        0,
       -0.846284375321634430,
        0,
        0,
        0,
        0,
       -0.846284375321634430,
        0,
        0.564189583547756286,
        0.211571093830408607,
        0,
        0,
        0.211571093830408607,
        0,
       -0.846284375321634430,
        0,
        0,
        0,
        0,
        0.273137107648019767,
        0,
        0,
        0.273137107648019767,
        0,
       -1.092548430592079070,
        0,
        0,
        0,
        0,
        0,
        0,
        0.546274215296039535,
        0,
        0,
        0,
        0,
       -0.546274215296039535,
        0,
        0,
        0,
        0,
        0.946174695757560018,
        0,
        0,
        0,
        0,
       -0.946174695757560018,
        0,
        0,
       -0.157695782626260003,
        0,
        0,
        0.473087347878780009,
        0,
        0,
        0,
        0,
        0,
        0,
       -0.417223823632784089,
        0,
        0,
        1.251671470898352269,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        // g7/2, -7/2
        -0.417223823632784089, 0, 0, 2.503342941796704538, 0, 0, 0, 0, 0, 0, -0.417223823632784089, 0, 0, 0, 0,
        0, 0, 0.417223823632784089, 0, 0, 0, 0, -1.251671470898352269, 0, 0, 0, 0, 0, 0, 0,
        // g7/2, -5/2
        0, 0, -1.103870478383820021, 0, 0, 0, 0, 3.311611435151460063, 0, 0, 0, 0, 0, 0, 0,
        -0.157695782626260003, 0, 0, 0, 0, 0.946174695757560018, 0, 0, 0, 0, 0.157695782626260003, 0, -0.946174695757560018, 0, 0,
        // g7/2, -3/2
        0.273137107648019767, 0, 0, 0, 0, -1.638822645888118605, 0, 0, 0, 0, -0.273137107648019767, 0, 1.638822645888118605, 0, 0,
        0, 0, -0.819411322944059302, 0, 0, 0, 0, -0.819411322944059302, 0, 1.092548430592079070, 0, 0, 0, 0, 0,
        // g7/2, -1/2
        0, 0, 1.057855469152043038, 0, 0, 0, 0, 1.057855469152043038, 0, -1.410473958869390717, 0, 0, 0, 0, 0,
        0.211571093830408607, 0, 0, 0.423142187660817215, 0, -1.692568750643268860, 0, 0, 0, 0, 0.211571093830408607, 0, -1.692568750643268860, 0, 0.564189583547756286,
        // g7/2, 1/2
        -0.211571093830408607, 0, 0, -0.423142187660817215, 0, 1.692568750643268860, 0, 0, 0, 0, -0.211571093830408607, 0, 1.692568750643268860, 0, -0.564189583547756286,
        0, 0, 1.057855469152043038, 0, 0, 0, 0, 1.057855469152043038, 0, -1.410473958869390717, 0, 0, 0, 0, 0,
        // g7/2, 3/2
        0, 0, -0.819411322944059302, 0, 0, 0, 0, -0.819411322944059302, 0, 1.092548430592079070, 0, 0, 0, 0, 0,
        -0.273137107648019767, 0, 0, 0, 0, 1.638822645888118605, 0, 0, 0, 0, 0.273137107648019767, 0, -1.638822645888118605, 0, 0,
        // g7/2, 5/2
        0.157695782626260003, 0, 0, 0, 0, -0.946174695757560018, 0, 0, 0, 0, -0.157695782626260003, 0, 0.946174695757560018, 0, 0,
        0, 0, -1.103870478383820021, 0, 0, 0, 0, 3.311611435151460063, 0, 0, 0, 0, 0, 0, 0,
        // g7/2, 7/2
        0, 0, 0.417223823632784089, 0, 0, 0, 0, -1.251671470898352269, 0, 0, 0, 0, 0, 0, 0,
        0.417223823632784089, 0, 0, -2.503342941796704538, 0, 0, 0, 0, 0, 0, 0.417223823632784089, 0, 0, 0, 0,
        // g9/2, -9/2
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.442532692444982632, 0, 0, -2.655196154669895796, 0, 0, 0, 0, 0, 0, 0.442532692444982632, 0, 0, 0, 0,
        // g9/2, -7/2
        0.147510897481660877, 0, 0, -0.885065384889965265, 0, 0, 0, 0, 0, 0, 0.147510897481660877, 0, 0, 0, 0,
        0, 0, 1.180087179853287020, 0, 0, 0, 0, -3.540261539559861062, 0, 0, 0, 0, 0, 0, 0,
        // g9/2, -5/2
        0, 0, 0.590043589926643510, 0, 0, 0, 0, -1.770130769779930531, 0, 0, 0, 0, 0, 0, 0,
        -0.295021794963321755, 0, 0, 0, 0, 1.770130769779930531, 0, 0, 0, 0, 0.295021794963321755, 0, -1.770130769779930531, 0, 0,
        // g9/2, -3/2
        -0.193137101011594790, 0, 0, 0, 0, 1.158822606069568741, 0, 0, 0, 0, 0.193137101011594790, 0, -1.158822606069568741, 0, 0,
        0, 0, -1.158822606069568741, 0, 0, 0, 0, -1.158822606069568741, 0, 1.545096808092758321, 0, 0, 0, 0, 0,
        // g9/2, -1/2
        0, 0, -0.946174695757560018, 0, 0, 0, 0, -0.946174695757560018, 0, 1.261566261010080024, 0, 0, 0, 0, 0,
        0.236543673939390004, 0, 0, 0.473087347878780009, 0, -1.892349391515120036, 0, 0, 0, 0, 0.236543673939390004, 0, -1.892349391515120036, 0, 0.630783130505040012,
        // g9/2, 1/2
        0.236543673939390004, 0, 0, 0.473087347878780009, 0, -1.892349391515120036, 0, 0, 0, 0, 0.236543673939390004, 0, -1.892349391515120036, 0, 0.630783130505040012,
        0, 0, 0.946174695757560018, 0, 0, 0, 0, 0.946174695757560018, 0, -1.261566261010080024, 0, 0, 0, 0, 0,
        // g9/2, 3/2
        0, 0, 1.158822606069568741, 0, 0, 0, 0, 1.158822606069568741, 0, -1.545096808092758321, 0, 0, 0, 0, 0,
        -0.193137101011594790, 0, 0, 0, 0, 1.158822606069568741, 0, 0, 0, 0, 0.193137101011594790, 0, -1.158822606069568741, 0, 0,
        // g9/2, 5/2
        -0.295021794963321755, 0, 0, 0, 0, 1.770130769779930531, 0, 0, 0, 0, 0.295021794963321755, 0, -1.770130769779930531, 0, 0,
        0, 0, -0.590043589926643510, 0, 0, 0, 0, 1.770130769779930531, 0, 0, 0, 0, 0, 0, 0,
        // g9/2, 7/2
        0, 0, -1.180087179853287020, 0, 0, 0, 0, 3.540261539559861062, 0, 0, 0, 0, 0, 0, 0,
        0.147510897481660877, 0, 0, -0.885065384889965265, 0, 0, 0, 0, 0, 0, 0.147510897481660877, 0, 0, 0, 0,
        // g9/2, 9/2
        0.442532692444982632, 0, 0, -2.655196154669895796, 0, 0, 0, 0, 0, 0, 0.442532692444982632, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // h9/2, -9/2
        -0.442532692444982632, 0, 0, 4.425326924449826327, 0, 0, 0, 0, 0, 0, -2.212663462224913163, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0.442532692444982632, 0, 0, 0, 0, -2.655196154669895796, 0, 0, 0, 0, 0, 0, 0, 0, 0.442532692444982632, 0, 0, 0, 0,
        // h9/2, -7/2
        0, 0, -1.327598077334947898, 0, 0, 0, 0, 7.965588464009687389, 0, 0, 0, 0, 0, 0, 0, 0, -1.327598077334947898, 0, 0, 0, 0,
        -0.147510897481660877, 0, 0, 0.295021794963321755, 0, 1.180087179853287020, 0, 0, 0, 0, 0.442532692444982632, 0, -3.540261539559861062, 0, 0, 0, 0, 0, 0, 0, 0,
        // h9/2, -5/2
        0.295021794963321755, 0, 0, -0.590043589926643510, 0, -2.360174359706574041, 0, 0, 0, 0, -0.885065384889965265, 0, 7.080523079119722124, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -0.885065384889965265, 0, 0, 0, 0, 0, 0, 1.770130769779930531, 0, 0, 0, 0, 0, 0, 0.885065384889965265, 0, -1.770130769779930531, 0, 0,
        // h9/2, -3/2
        0, 0, 1.351959707081163531, 0, 0, 0, 0, 0, 0, -2.703919414162327062, 0, 0, 0, 0, 0, 0, -1.351959707081163531, 0, 2.703919414162327062, 0, 0,
        0.193137101011594790, 0, 0, 0.386274202023189580, 0, -2.317645212139137482, 0, 0, 0, 0, 0.193137101011594790, 0, -2.317645212139137482, 0, 1.545096808092758321, 0, 0, 0, 0, 0, 0,
        // h9/2, -1/2
        -0.236543673939390004, 0, 0, -0.473087347878780009, 0, 2.838524087272680054, 0, 0, 0, 0, -0.236543673939390004, 0, 2.838524087272680054, 0, -1.892349391515120036, 0, 0, 0, 0, 0, 0,
        0, 0, 1.182718369696950022, 0, 0, 0, 0, 2.365436739393900045, 0, -3.153915652525200060, 0, 0, 0, 0, 0, 0, 1.182718369696950022, 0, -3.153915652525200060, 0, 0.630783130505040012,
        // h9/2, 1/2
        0, 0, -1.182718369696950022, 0, 0, 0, 0, -2.365436739393900045, 0, 3.153915652525200060, 0, 0, 0, 0, 0, 0, -1.182718369696950022, 0, 3.153915652525200060, 0, -0.630783130505040012,
        -0.236543673939390004, 0, 0, -0.473087347878780009, 0, 2.838524087272680054, 0, 0, 0, 0, -0.236543673939390004, 0, 2.838524087272680054, 0, -1.892349391515120036, 0, 0, 0, 0, 0, 0,
        // h9/2, 3/2
        0.193137101011594790, 0, 0, 0.386274202023189580, 0, -2.317645212139137482, 0, 0, 0, 0, 0.193137101011594790, 0, -2.317645212139137482, 0, 1.545096808092758321, 0, 0, 0, 0, 0, 0,
        0, 0, -1.351959707081163531, 0, 0, 0, 0, 0, 0, 2.703919414162327062, 0, 0, 0, 0, 0, 0, 1.351959707081163531, 0, -2.703919414162327062, 0, 0,
        // h9/2, 5/2
        0, 0, 0.885065384889965265, 0, 0, 0, 0, 0, 0, -1.770130769779930531, 0, 0, 0, 0, 0, 0, -0.885065384889965265, 0, 1.770130769779930531, 0, 0,
        0.295021794963321755, 0, 0, -0.590043589926643510, 0, -2.360174359706574041, 0, 0, 0, 0, -0.885065384889965265, 0, 7.080523079119722124, 0, 0, 0, 0, 0, 0, 0, 0,
        // h9/2, 7/2
        -0.147510897481660877, 0, 0, 0.295021794963321755, 0, 1.180087179853287020, 0, 0, 0, 0, 0.442532692444982632, 0, -3.540261539559861062, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 1.327598077334947898, 0, 0, 0, 0, -7.965588464009687389, 0, 0, 0, 0, 0, 0, 0, 0, 1.327598077334947898, 0, 0, 0, 0,
        // h9/2, 9/2
        0, 0, -0.442532692444982632, 0, 0, 0, 0, 2.655196154669895796, 0, 0, 0, 0, 0, 0, 0, 0, -0.442532692444982632, 0, 0, 0, 0,
        -0.442532692444982632, 0, 0, 4.425326924449826327, 0, 0, 0, 0, 0, 0, -2.212663462224913163, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // h11/2, -11/2
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.464132203440858160, 0, 0, -4.641322034408581606, 0, 0, 0, 0, 0, 0, 2.320661017204290803, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // h11/2, -9/2
        0.139941124721293271, 0, 0, -1.399411247212932717, 0, 0, 0, 0, 0, 0, 0.699705623606466358, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 1.399411247212932717, 0, 0, 0, 0, -8.396467483277596306, 0, 0, 0, 0, 0, 0, 0, 0, 1.399411247212932717, 0, 0, 0, 0,
        // h11/2, -7/2
        0, 0, 0.625835735449176134, 0, 0, 0, 0, -3.755014412695056807, 0, 0, 0, 0, 0, 0, 0, 0, 0.625835735449176134, 0, 0, 0, 0,
        -0.312917867724588067, 0, 0, 0.625835735449176134, 0, 2.503342941796704538, 0, 0, 0, 0, 0.938753603173764201, 0, -7.510028825390113615, 0, 0, 0, 0, 0, 0, 0, 0,
        // h11/2, -5/2
        -0.180663215165034628, 0, 0, 0.361326430330069256, 0, 1.445305721320277027, 0, 0, 0, 0, 0.541989645495103885, 0, -4.335917163960831083, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -1.445305721320277027, 0, 0, 0, 0, 0, 0, 2.890611442640554055, 0, 0, 0, 0, 0, 0, 1.445305721320277027, 0, -2.890611442640554055, 0, 0,
        // h11/2, -3/2
        0, 0, -1.021985476433282363, 0, 0, 0, 0, 0, 0, 2.043970952866564726, 0, 0, 0, 0, 0, 0, 1.021985476433282363, 0, -2.043970952866564726, 0, 0,
        0.255496369108320590, 0, 0, 0.510992738216641181, 0, -3.065956429299847090, 0, 0, 0, 0, 0.255496369108320590, 0, -3.065956429299847090, 0, 2.043970952866564726, 0, 0, 0, 0, 0, 0,
        // h11/2, -1/2
        0.215933843419584674, 0, 0, 0.431867686839169349, 0, -2.591206121035016094, 0, 0, 0, 0, 0.215933843419584674, 0, -2.591206121035016094, 0, 1.727470747356677396, 0, 0, 0, 0, 0, 0,
        0, 0, 1.295603060517508047, 0, 0, 0, 0, 2.591206121035016094, 0, -3.454941494713354792, 0, 0, 0, 0, 0, 0, 1.295603060517508047, 0, -3.454941494713354792, 0, 0.690988298942670958,
        // h11/2, 1/2
        0, 0, 1.295603060517508047, 0, 0, 0, 0, 2.591206121035016094, 0, -3.454941494713354792, 0, 0, 0, 0, 0, 0, 1.295603060517508047, 0, -3.454941494713354792, 0, 0.690988298942670958,
        -0.215933843419584674, 0, 0, -0.431867686839169349, 0, 2.591206121035016094, 0, 0, 0, 0, -0.215933843419584674, 0, 2.591206121035016094, 0, -1.727470747356677396, 0, 0, 0, 0, 0, 0,
        // h11/2, 3/2
        -0.255496369108320590, 0, 0, -0.510992738216641181, 0, 3.065956429299847090, 0, 0, 0, 0, -0.255496369108320590, 0, 3.065956429299847090, 0, -2.043970952866564726, 0, 0, 0, 0, 0, 0,
        0, 0, -1.021985476433282363, 0, 0, 0, 0, 0, 0, 2.043970952866564726, 0, 0, 0, 0, 0, 0, 1.021985476433282363, 0, -2.043970952866564726, 0, 0,
        // h11/2, 5/2
        0, 0, -1.445305721320277027, 0, 0, 0, 0, 0, 0, 2.890611442640554055, 0, 0, 0, 0, 0, 0, 1.445305721320277027, 0, -2.890611442640554055, 0, 0,
        0.180663215165034628, 0, 0, -0.361326430330069256, 0, -1.445305721320277027, 0, 0, 0, 0, -0.541989645495103885, 0, 4.335917163960831083, 0, 0, 0, 0, 0, 0, 0, 0,
        // h11/2, 7/2
        0.312917867724588067, 0, 0, -0.625835735449176134, 0, -2.503342941796704538, 0, 0, 0, 0, -0.938753603173764201, 0, 7.510028825390113615, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0.625835735449176134, 0, 0, 0, 0, -3.755014412695056807, 0, 0, 0, 0, 0, 0, 0, 0, 0.625835735449176134, 0, 0, 0, 0,
        // h11/2, 9/2
        0, 0, 1.399411247212932717, 0, 0, 0, 0, -8.396467483277596306, 0, 0, 0, 0, 0, 0, 0, 0, 1.399411247212932717, 0, 0, 0, 0,
        -0.139941124721293271, 0, 0, 1.399411247212932717, 0, 0, 0, 0, 0, 0, -0.699705623606466358, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // h11/2, 11/2
        -0.464132203440858160, 0, 0, 4.641322034408581606, 0, 0, 0, 0, 0, 0, -2.320661017204290803, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //i11/2,-11/2
        -0.4641322034408583, 0, 0, 6.9619830516128740, 0, 0, 0, 0, 0, 0, -6.9619830516128740, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.4641322034408583, 0, 0, 0, 0, 0, 0,
        0, 0, 0.4641322034408583, 0, 0, 0, 0, -4.6413220344085833, 0, 0, 0, 0, 0, 0, 0, 0, 2.3206610172042916, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //i11/2,-9/2
        0, 0, -1.5393523719342264, 0, 0, 0, 0, 15.3935237193422640, 0, 0, 0, 0, 0, 0, 0, 0, -7.6967618596711320, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.1399411247212932, 0, 0, 0.6997056236064662, 0, 1.3994112472129330, 0, 0, 0, 0, 0.6997056236064662, 0, -8.3964674832775970, 0, 0, 0, 0, 0, 0, 0, 0, -0.1399411247212932, 0, 1.3994112472129330, 0, 0, 0, 0,
        //i11/2,-7/2
        0.3129178677245880, 0, 0, -1.5645893386229399, 0, -3.1291786772458812, 0, 0, 0, 0, -1.5645893386229399, 0, 18.7750720634752852, 0, 0, 0, 0, 0, 0, 0, 0, 0.3129178677245880, 0, -3.1291786772458812, 0, 0, 0, 0,
        0, 0, -0.9387536031737643, 0, 0, 0, 0, 1.8775072063475287, 0, 2.5033429417967050, 0, 0, 0, 0, 0, 0, 2.8162608095212929, 0, -7.5100288253901146, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //i11/2,-5/2
        0, 0, 1.6259689364853118, 0, 0, 0, 0, -3.2519378729706236, 0, -4.3359171639608318, 0, 0, 0, 0, 0, 0, -4.8779068094559346, 0, 13.0077514918824946, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.1806632151650347, 0, 0, 0.1806632151650347, 0, -2.8906114426405547, 0, 0, 0, 0, -0.1806632151650347, 0, 0, 0, 2.8906114426405547, 0, 0, 0, 0, 0, 0, -0.1806632151650347, 0, 2.8906114426405547, 0, -2.8906114426405547, 0, 0,
        //i11/2,-3/2
        -0.2554963691083206, 0, 0, -0.2554963691083206, 0, 4.0879419057331301, 0, 0, 0, 0, 0.2554963691083206, 0, 0, 0, -4.0879419057331301, 0, 0, 0, 0, 0, 0, 0.2554963691083206, 0, -4.0879419057331301, 0, 4.0879419057331301, 0, 0,
        0, 0, 1.2774818455416030, 0, 0, 0, 0, 2.5549636910832061, 0, -5.1099273821664122, 0, 0, 0, 0, 0, 0, 1.2774818455416030, 0, -5.1099273821664122, 0, 2.0439709528665646, 0, 0, 0, 0, 0, 0, 0,
        //i11/2,-1/2
        0, 0, -1.5115369039370925, 0, 0, 0, 0, -3.0230738078741850, 0, 6.0461476157483700, 0, 0, 0, 0, 0, 0, -1.5115369039370925, 0, 6.0461476157483700, 0, -2.4184590462993478, 0, 0, 0, 0, 0, 0, 0,
        -0.2159338434195847, 0, 0, -0.6478015302587540, 0, 3.8868091815525241, 0, 0, 0, 0, -0.6478015302587540, 0, 7.7736183631050482, 0, -5.1824122420700318, 0, 0, 0, 0, 0, 0, -0.2159338434195847, 0, 3.8868091815525241, 0, -5.1824122420700318, 0, 0.6909882989426709,
        //i11/2,1/2
        0.2159338434195847, 0, 0, 0.6478015302587540, 0, -3.8868091815525241, 0, 0, 0, 0, 0.6478015302587540, 0, -7.7736183631050482, 0, 5.1824122420700318, 0, 0, 0, 0, 0, 0, 0.2159338434195847, 0, -3.8868091815525241, 0, 5.1824122420700318, 0, -0.6909882989426709,
        0, 0, -1.5115369039370925, 0, 0, 0, 0, -3.0230738078741850, 0, 6.0461476157483700, 0, 0, 0, 0, 0, 0, -1.5115369039370925, 0, 6.0461476157483700, 0, -2.4184590462993478, 0, 0, 0, 0, 0, 0, 0,
        //i11/2,3/2
        0, 0, 1.2774818455416030, 0, 0, 0, 0, 2.5549636910832061, 0, -5.1099273821664122, 0, 0, 0, 0, 0, 0, 1.2774818455416030, 0, -5.1099273821664122, 0, 2.0439709528665646, 0, 0, 0, 0, 0, 0, 0,
        0.2554963691083206, 0, 0, 0.2554963691083206, 0, -4.0879419057331301, 0, 0, 0, 0, -0.2554963691083206, 0, 0, 0, 4.0879419057331301, 0, 0, 0, 0, 0, 0, -0.2554963691083206, 0, 4.0879419057331301, 0, -4.0879419057331301, 0, 0,
        //i11/2,5/2
        -0.1806632151650347, 0, 0, -0.1806632151650347, 0, 2.8906114426405547, 0, 0, 0, 0, 0.1806632151650347, 0, 0, 0, -2.8906114426405547, 0, 0, 0, 0, 0, 0, 0.1806632151650347, 0, -2.8906114426405547, 0, 2.8906114426405547, 0, 0,
        0, 0, 1.6259689364853118, 0, 0, 0, 0, -3.2519378729706236, 0, -4.3359171639608318, 0, 0, 0, 0, 0, 0, -4.8779068094559346, 0, 13.0077514918824946, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //i11/2,7/2
        0, 0, -0.9387536031737643, 0, 0, 0, 0, 1.8775072063475287, 0, 2.5033429417967050, 0, 0, 0, 0, 0, 0, 2.8162608095212929, 0, -7.5100288253901146, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.3129178677245880, 0, 0, 1.5645893386229399, 0, 3.1291786772458812, 0, 0, 0, 0, 1.5645893386229399, 0, -18.7750720634752852, 0, 0, 0, 0, 0, 0, 0, 0, -0.3129178677245880, 0, 3.1291786772458812, 0, 0, 0, 0,
        //i11/2,9/2
        0.1399411247212932, 0, 0, -0.6997056236064662, 0, -1.3994112472129330, 0, 0, 0, 0, -0.6997056236064662, 0, 8.3964674832775970, 0, 0, 0, 0, 0, 0, 0, 0, 0.1399411247212932, 0, -1.3994112472129330, 0, 0, 0, 0,
        0, 0, -1.5393523719342264, 0, 0, 0, 0, 15.3935237193422640, 0, 0, 0, 0, 0, 0, 0, 0, -7.6967618596711320, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //i11/2,11/2
        0, 0, 0.4641322034408583, 0, 0, 0, 0, -4.6413220344085833, 0, 0, 0, 0, 0, 0, 0, 0, 2.3206610172042916, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.4641322034408583, 0, 0, -6.9619830516128740, 0, 0, 0, 0, 0, 0, 6.9619830516128740, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.4641322034408583, 0, 0, 0, 0, 0, 0,
        //i13/2,-13/2
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.4830841135800663, 0, 0, -7.2462617037009949, 0, 0, 0, 0, 0, 0, 7.2462617037009949, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.4830841135800663, 0, 0, 0, 0, 0, 0,
        //i13/2,-11/2
        0.1339834262980768, 0, 0, -2.0097513944711523, 0, 0, 0, 0, 0, 0, 2.0097513944711523, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.1339834262980768, 0, 0, 0, 0, 0, 0,
        0, 0, 1.6078011155769223, 0, 0, 0, 0, -16.0780111557692216, 0, 0, 0, 0, 0, 0, 0, 0, 8.0390055778846108, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //i13/2,-9/2
        0, 0, 0.6563820568401703, 0, 0, 0, 0, -6.5638205684017032, 0, 0, 0, 0, 0, 0, 0, 0, 3.2819102842008516, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.3281910284200850, 0, 0, 1.6409551421004249, 0, 3.2819102842008507, 0, 0, 0, 0, 1.6409551421004249, 0, -19.6914617052051035, 0, 0, 0, 0, 0, 0, 0, 0, -0.3281910284200850, 0, 3.2819102842008507, 0, 0, 0, 0,
        //i13/2,-7/2
        -0.1713921747991747, 0, 0, 0.8569608739958732, 0, 1.7139217479917468, 0, 0, 0, 0, 0.8569608739958732, 0, -10.2835304879504807, 0, 0, 0, 0, 0, 0, 0, 0, -0.1713921747991747, 0, 1.7139217479917468, 0, 0, 0, 0,
        0, 0, -1.7139217479917466, 0, 0, 0, 0, 3.4278434959834931, 0, 4.5704579946446584, 0, 0, 0, 0, 0, 0, 5.1417652439752395, 0, -13.7113739839339726, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //i13/2,-5/2
        0, 0, -1.0839792909902080, 0, 0, 0, 0, 2.1679585819804159, 0, 2.8906114426405547, 0, 0, 0, 0, 0, 0, 3.2519378729706232, 0, -8.6718343279216636, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.2709948227475520, 0, 0, 0.2709948227475520, 0, -4.3359171639608318, 0, 0, 0, 0, -0.2709948227475520, 0, 0, 0, 4.3359171639608318, 0, 0, 0, 0, 0, 0, -0.2709948227475520, 0, 4.3359171639608318, 0, -4.3359171639608318, 0, 0,
        //i13/2,-3/2
        0.2019876150713442, 0, 0, 0.2019876150713442, 0, -3.2318018411415070, 0, 0, 0, 0, -0.2019876150713442, 0, 0, 0, 3.2318018411415070, 0, 0, 0, 0, 0, 0, -0.2019876150713442, 0, 3.2318018411415070, 0, -3.2318018411415070, 0, 0,
        0, 0, 1.6159009205707533, 0, 0, 0, 0, 3.2318018411415066, 0, -6.4636036822830132, 0, 0, 0, 0, 0, 0, 1.6159009205707533, 0, -6.4636036822830132, 0, 2.5854414729132049, 0, 0, 0, 0, 0, 0, 0,
        //i13/2,-1/2
        0, 0, 1.3994112472129328, 0, 0, 0, 0, 2.7988224944258655, 0, -5.5976449888517310, 0, 0, 0, 0, 0, 0, 1.3994112472129328, 0, -5.5976449888517310, 0, 2.2390579955406920, 0, 0, 0, 0, 0, 0, 0,
        -0.2332352078688221, 0, 0, -0.6997056236064663, 0, 4.1982337416387976, 0, 0, 0, 0, -0.6997056236064663, 0, 8.3964674832775952, 0, -5.5976449888517301, 0, 0, 0, 0, 0, 0, -0.2332352078688221, 0, 4.1982337416387976, 0, -5.5976449888517301, 0, 0.7463526651802307,
        //i13/2,1/2
        -0.2332352078688221, 0, 0, -0.6997056236064663, 0, 4.1982337416387976, 0, 0, 0, 0, -0.6997056236064663, 0, 8.3964674832775952, 0, -5.5976449888517301, 0, 0, 0, 0, 0, 0, -0.2332352078688221, 0, 4.1982337416387976, 0, -5.5976449888517301, 0, 0.7463526651802307,
        0, 0, -1.3994112472129328, 0, 0, 0, 0, -2.7988224944258655, 0, 5.5976449888517310, 0, 0, 0, 0, 0, 0, -1.3994112472129328, 0, 5.5976449888517310, 0, -2.2390579955406920, 0, 0, 0, 0, 0, 0, 0,
        //i13/2,3/2
        0, 0, -1.6159009205707533, 0, 0, 0, 0, -3.2318018411415066, 0, 6.4636036822830132, 0, 0, 0, 0, 0, 0, -1.6159009205707533, 0, 6.4636036822830132, 0, -2.5854414729132049, 0, 0, 0, 0, 0, 0, 0,
        0.2019876150713442, 0, 0, 0.2019876150713442, 0, -3.2318018411415070, 0, 0, 0, 0, -0.2019876150713442, 0, 0, 0, 3.2318018411415070, 0, 0, 0, 0, 0, 0, -0.2019876150713442, 0, 3.2318018411415070, 0, -3.2318018411415070, 0, 0,
        //i13/2,5/2
        0.2709948227475520, 0, 0, 0.2709948227475520, 0, -4.3359171639608318, 0, 0, 0, 0, -0.2709948227475520, 0, 0, 0, 4.3359171639608318, 0, 0, 0, 0, 0, 0, -0.2709948227475520, 0, 4.3359171639608318, 0, -4.3359171639608318, 0, 0,
        0, 0, 1.0839792909902080, 0, 0, 0, 0, -2.1679585819804159, 0, -2.8906114426405547, 0, 0, 0, 0, 0, 0, -3.2519378729706232, 0, 8.6718343279216636, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //i13/2,7/2
        0, 0, 1.7139217479917466, 0, 0, 0, 0, -3.4278434959834931, 0, -4.5704579946446584, 0, 0, 0, 0, 0, 0, -5.1417652439752395, 0, 13.7113739839339726, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.1713921747991747, 0, 0, 0.8569608739958732, 0, 1.7139217479917468, 0, 0, 0, 0, 0.8569608739958732, 0, -10.2835304879504807, 0, 0, 0, 0, 0, 0, 0, 0, -0.1713921747991747, 0, 1.7139217479917468, 0, 0, 0, 0,
        //i13/2,9/2
        -0.3281910284200850, 0, 0, 1.6409551421004249, 0, 3.2819102842008507, 0, 0, 0, 0, 1.6409551421004249, 0, -19.6914617052051035, 0, 0, 0, 0, 0, 0, 0, 0, -0.3281910284200850, 0, 3.2819102842008507, 0, 0, 0, 0,
        0, 0, -0.6563820568401703, 0, 0, 0, 0, 6.5638205684017032, 0, 0, 0, 0, 0, 0, 0, 0, -3.2819102842008516, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //i13/2,11/2
        0, 0, -1.6078011155769223, 0, 0, 0, 0, 16.0780111557692216, 0, 0, 0, 0, 0, 0, 0, 0, -8.0390055778846108, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.1339834262980768, 0, 0, -2.0097513944711523, 0, 0, 0, 0, 0, 0, 2.0097513944711523, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.1339834262980768, 0, 0, 0, 0, 0, 0,
        //i13/2,13/2
        0.4830841135800663, 0, 0, -7.2462617037009949, 0, 0, 0, 0, 0, 0, 7.2462617037009949, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.4830841135800663, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //j13/2,-13/2
        -0.4830841135800663, 0, 0, 10.1447663851813932, 0, 0, 0, 0, 0, 0, -16.9079439753023237, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.3815887950604644, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0.4830841135800661, 0, 0, 0, 0, -7.2462617037009922, 0, 0, 0, 0, 0, 0, 0, 0, 7.2462617037009922, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.4830841135800661, 0, 0, 0, 0, 0, 0,
        //j13/2,-11/2
        0, 0, -1.7417845418749984, 0, 0, 0, 0, 26.1267681281249757, 0, 0, 0, 0, 0, 0, 0, 0, -26.1267681281249757, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.7417845418749984, 0, 0, 0, 0, 0, 0,
        -0.1339834262980768, 0, 0, 1.2058508366826914, 0, 1.6078011155769216, 0, 0, 0, 0, 0.6699171314903840, 0, -16.0780111557692145, 0, 0, 0, 0, 0, 0, 0, 0, -0.6699171314903840, 0, 8.0390055778846072, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //j13/2,-9/2
        0.3281910284200851, 0, 0, -2.9537192557807663, 0, -3.9382923410410213, 0, 0, 0, 0, -1.6409551421004256, 0, 39.3829234104102142, 0, 0, 0, 0, 0, 0, 0, 0, 1.6409551421004256, 0, -19.6914617052051071, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -0.9845730852602553, 0, 0, 0, 0, 4.9228654263012768, 0, 3.2819102842008512, 0, 0, 0, 0, 0, 0, 4.9228654263012768, 0, -19.6914617052051071, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.9845730852602553, 0, 3.2819102842008512, 0, 0, 0, 0,
        //j13/2,-7/2
        0, 0, 1.8853139227909212, 0, 0, 0, 0, -9.4265696139546051, 0, -6.2843797426364043, 0, 0, 0, 0, 0, 0, -9.4265696139546051, 0, 37.7062784558184205, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.8853139227909212, 0, -6.2843797426364043, 0, 0, 0, 0,
        0.1713921747991747, 0, 0, -0.1713921747991747, 0, -3.4278434959834923, 0, 0, 0, 0, -0.8569608739958733, 0, 6.8556869919669845, 0, 4.5704579946446575, 0, 0, 0, 0, 0, 0, -0.5141765243975239, 0, 10.2835304879504772, 0, -13.7113739839339726, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //j13/2,-5/2
        -0.2709948227475520, 0, 0, 0.2709948227475520, 0, 5.4198964549510391, 0, 0, 0, 0, 1.3549741137377600, 0, -10.8397929099020782, 0, -7.2265286066013861, 0, 0, 0, 0, 0, 0, 0.8129844682426559, 0, -16.2596893648531164, 0, 21.6795858198041600, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 1.3549741137377600, 0, 0, 0, 0, 1.3549741137377600, 0, -7.2265286066013861, 0, 0, 0, 0, 0, 0, -1.3549741137377600, 0, 0, 0, 4.3359171639608327, 0, 0, 0, 0, 0, 0, 0, 0, -1.3549741137377600, 0, 7.2265286066013861, 0, -4.3359171639608327, 0, 0,
        //j13/2,-3/2
        0, 0, -1.8178885356420982, 0, 0, 0, 0, -1.8178885356420982, 0, 9.6954055234245224, 0, 0, 0, 0, 0, 0, 1.8178885356420982, 0, 0, 0, -5.8172433140547142, 0, 0, 0, 0, 0, 0, 0, 0, 1.8178885356420982, 0, -9.6954055234245224, 0, 5.8172433140547142, 0, 0,
        -0.2019876150713442, 0, 0, -0.6059628452140327, 0, 4.8477027617122603, 0, 0, 0, 0, -0.6059628452140327, 0, 9.6954055234245207, 0, -9.6954055234245207, 0, 0, 0, 0, 0, 0, -0.2019876150713442, 0, 4.8477027617122603, 0, -9.6954055234245207, 0, 2.5854414729132063, 0, 0, 0, 0, 0, 0, 0, 0,
        //j13/2,-1/2
        0.2332352078688221, 0, 0, 0.6997056236064664, 0, -5.5976449888517301, 0, 0, 0, 0, 0.6997056236064664, 0, -11.1952899777034602, 0, 11.1952899777034602, 0, 0, 0, 0, 0, 0, 0.2332352078688221, 0, -5.5976449888517301, 0, 11.1952899777034602, 0, -2.9854106607209236, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -1.6326464550817545, 0, 0, 0, 0, -4.8979393652452643, 0, 9.7958787304905268, 0, 0, 0, 0, 0, 0, -4.8979393652452643, 0, 19.5917574609810536, 0, -7.8367029843924216, 0, 0, 0, 0, 0, 0, 0, 0, -1.6326464550817545, 0, 9.7958787304905268, 0, -7.8367029843924216, 0, 0.7463526651802307,
        //j13/2,1/2
        0, 0, 1.6326464550817545, 0, 0, 0, 0, 4.8979393652452643, 0, -9.7958787304905268, 0, 0, 0, 0, 0, 0, 4.8979393652452643, 0, -19.5917574609810536, 0, 7.8367029843924216, 0, 0, 0, 0, 0, 0, 0, 0, 1.6326464550817545, 0, -9.7958787304905268, 0, 7.8367029843924216, 0, -0.7463526651802307,
        0.2332352078688221, 0, 0, 0.6997056236064664, 0, -5.5976449888517301, 0, 0, 0, 0, 0.6997056236064664, 0, -11.1952899777034602, 0, 11.1952899777034602, 0, 0, 0, 0, 0, 0, 0.2332352078688221, 0, -5.5976449888517301, 0, 11.1952899777034602, 0, -2.9854106607209236, 0, 0, 0, 0, 0, 0, 0, 0,
        //j13/2,3/2
        -0.2019876150713442, 0, 0, -0.6059628452140327, 0, 4.8477027617122603, 0, 0, 0, 0, -0.6059628452140327, 0, 9.6954055234245207, 0, -9.6954055234245207, 0, 0, 0, 0, 0, 0, -0.2019876150713442, 0, 4.8477027617122603, 0, -9.6954055234245207, 0, 2.5854414729132063, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 1.8178885356420982, 0, 0, 0, 0, 1.8178885356420982, 0, -9.6954055234245224, 0, 0, 0, 0, 0, 0, -1.8178885356420982, 0, 0, 0, 5.8172433140547142, 0, 0, 0, 0, 0, 0, 0, 0, -1.8178885356420982, 0, 9.6954055234245224, 0, -5.8172433140547142, 0, 0,
        //j13/2,5/2
        0, 0, -1.3549741137377600, 0, 0, 0, 0, -1.3549741137377600, 0, 7.2265286066013861, 0, 0, 0, 0, 0, 0, 1.3549741137377600, 0, 0, 0, -4.3359171639608327, 0, 0, 0, 0, 0, 0, 0, 0, 1.3549741137377600, 0, -7.2265286066013861, 0, 4.3359171639608327, 0, 0,
        -0.2709948227475520, 0, 0, 0.2709948227475520, 0, 5.4198964549510391, 0, 0, 0, 0, 1.3549741137377600, 0, -10.8397929099020782, 0, -7.2265286066013861, 0, 0, 0, 0, 0, 0, 0.8129844682426559, 0, -16.2596893648531164, 0, 21.6795858198041600, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //j13/2,7/2
        0.1713921747991747, 0, 0, -0.1713921747991747, 0, -3.4278434959834923, 0, 0, 0, 0, -0.8569608739958733, 0, 6.8556869919669845, 0, 4.5704579946446575, 0, 0, 0, 0, 0, 0, -0.5141765243975239, 0, 10.2835304879504772, 0, -13.7113739839339726, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -1.8853139227909212, 0, 0, 0, 0, 9.4265696139546051, 0, 6.2843797426364043, 0, 0, 0, 0, 0, 0, 9.4265696139546051, 0, -37.7062784558184205, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.8853139227909212, 0, 6.2843797426364043, 0, 0, 0, 0,
        //j13/2,9/2
        0, 0, 0.9845730852602553, 0, 0, 0, 0, -4.9228654263012768, 0, -3.2819102842008512, 0, 0, 0, 0, 0, 0, -4.9228654263012768, 0, 19.6914617052051071, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.9845730852602553, 0, -3.2819102842008512, 0, 0, 0, 0,
        0.3281910284200851, 0, 0, -2.9537192557807663, 0, -3.9382923410410213, 0, 0, 0, 0, -1.6409551421004256, 0, 39.3829234104102142, 0, 0, 0, 0, 0, 0, 0, 0, 1.6409551421004256, 0, -19.6914617052051071, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //j13/2,11/2
        -0.1339834262980768, 0, 0, 1.2058508366826914, 0, 1.6078011155769216, 0, 0, 0, 0, 0.6699171314903840, 0, -16.0780111557692145, 0, 0, 0, 0, 0, 0, 0, 0, -0.6699171314903840, 0, 8.0390055778846072, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 1.7417845418749984, 0, 0, 0, 0, -26.1267681281249757, 0, 0, 0, 0, 0, 0, 0, 0, 26.1267681281249757, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.7417845418749984, 0, 0, 0, 0, 0, 0,
        //j13/2,13/2
        0, 0, -0.4830841135800661, 0, 0, 0, 0, 7.2462617037009922, 0, 0, 0, 0, 0, 0, 0, 0, -7.2462617037009922, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.4830841135800661, 0, 0, 0, 0, 0, 0,
        -0.4830841135800663, 0, 0, 10.1447663851813932, 0, 0, 0, 0, 0, 0, -16.9079439753023237, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.3815887950604644, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //j15/2,-15/2
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.5000395635705508, 0, 0, -10.5008308349815653, 0, 0, 0, 0, 0, 0, 17.5013847249692773, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3.5002769449938556, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //j15/2,-13/2
        0.1291096601435712, 0, 0, -2.7113028630149949, 0, 0, 0, 0, 0, 0, 4.5188381050249919, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.9037676210049984, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 1.8075352420099966, 0, 0, 0, 0, -27.1130286301499481, 0, 0, 0, 0, 0, 0, 0, 0, 27.1130286301499481, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.8075352420099966, 0, 0, 0, 0, 0, 0,
        //j15/2,-11/2
        0, 0, 0.6831841051919142, 0, 0, 0, 0, -10.2477615778787126, 0, 0, 0, 0, 0, 0, 0, 0, 10.2477615778787126, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.6831841051919142, 0, 0, 0, 0, 0, 0,
        -0.3415920525959572, 0, 0, 3.0743284733636154, 0, 4.0991046311514863, 0, 0, 0, 0, 1.7079602629797861, 0, -40.9910463115148644, 0, 0, 0, 0, 0, 0, 0, 0, -1.7079602629797861, 0, 20.4955231557574322, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //j15/2,-9/2
        -0.1640955142100426, 0, 0, 1.4768596278903832, 0, 1.9691461705205107, 0, 0, 0, 0, 0.8204775710502128, 0, -19.6914617052051071, 0, 0, 0, 0, 0, 0, 0, 0, -0.8204775710502128, 0, 9.8457308526025535, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -1.9691461705205107, 0, 0, 0, 0, 9.8457308526025535, 0, 6.5638205684017024, 0, 0, 0, 0, 0, 0, 9.8457308526025535, 0, -39.3829234104102142, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.9691461705205107, 0, 6.5638205684017024, 0, 0, 0, 0,
        //j15/2,-7/2
        0, 0, -1.1368870716237374, 0, 0, 0, 0, 5.6844353581186864, 0, 3.7896235720791247, 0, 0, 0, 0, 0, 0, 5.6844353581186864, 0, -22.7377414324747456, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.1368870716237374, 0, 3.7896235720791247, 0, 0, 0, 0,
        0.2842217679059343, 0, 0, -0.2842217679059343, 0, -5.6844353581186855, 0, 0, 0, 0, -1.4211088395296716, 0, 11.3688707162373710, 0, 7.5792471441582485, 0, 0, 0, 0, 0, 0, -0.8526653037178029, 0, 17.0533060743560583, 0, -22.7377414324747456, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //j15/2,-5/2
        0.1916222768312404, 0, 0, -0.1916222768312404, 0, -3.8324455366248085, 0, 0, 0, 0, -0.9581113841562022, 0, 7.6648910732496169, 0, 5.1099273821664122, 0, 0, 0, 0, 0, 0, -0.5748668304937213, 0, 11.4973366098744254, 0, -15.3297821464992357, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 1.9162227683124053, 0, 0, 0, 0, 1.9162227683124053, 0, -10.2198547643328261, 0, 0, 0, 0, 0, 0, -1.9162227683124053, 0, 0, 0, 6.1319128585996969, 0, 0, 0, 0, 0, 0, 0, 0, -1.9162227683124053, 0, 10.2198547643328261, 0, -6.1319128585996969, 0, 0,
        //j15/2,-3/2
        0, 0, 1.4842997738594836, 0, 0, 0, 0, 1.4842997738594836, 0, -7.9162654605839124, 0, 0, 0, 0, 0, 0, -1.4842997738594836, 0, 0, 0, 4.7497592763503480, 0, 0, 0, 0, 0, 0, 0, 0, -1.4842997738594836, 0, 7.9162654605839124, 0, -4.7497592763503480, 0, 0,
        -0.2473832956432473, 0, 0, -0.7421498869297418, 0, 5.9371990954379337, 0, 0, 0, 0, -0.7421498869297418, 0, 11.8743981908758673, 0, -11.8743981908758673, 0, 0, 0, 0, 0, 0, -0.2473832956432473, 0, 5.9371990954379337, 0, -11.8743981908758673, 0, 3.1665061842335653, 0, 0, 0, 0, 0, 0, 0, 0,
        //j15/2,-1/2
        -0.2181715595945335, 0, 0, -0.6545146787836006, 0, 5.2361174302688038, 0, 0, 0, 0, -0.6545146787836006, 0, 10.4722348605376077, 0, -10.4722348605376077, 0, 0, 0, 0, 0, 0, -0.2181715595945335, 0, 5.2361174302688038, 0, -10.4722348605376077, 0, 2.7925959628100294, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -1.7453724767562677, 0, 0, 0, 0, -5.2361174302688038, 0, 10.4722348605376059, 0, 0, 0, 0, 0, 0, -5.2361174302688038, 0, 20.9444697210752118, 0, -8.3777878884300847, 0, 0, 0, 0, 0, 0, 0, 0, -1.7453724767562677, 0, 10.4722348605376059, 0, -8.3777878884300847, 0, 0.7978845608028652,
        //j15/2,1/2
        0, 0, -1.7453724767562677, 0, 0, 0, 0, -5.2361174302688038, 0, 10.4722348605376059, 0, 0, 0, 0, 0, 0, -5.2361174302688038, 0, 20.9444697210752118, 0, -8.3777878884300847, 0, 0, 0, 0, 0, 0, 0, 0, -1.7453724767562677, 0, 10.4722348605376059, 0, -8.3777878884300847, 0, 0.7978845608028652,
        0.2181715595945335, 0, 0, 0.6545146787836006, 0, -5.2361174302688038, 0, 0, 0, 0, 0.6545146787836006, 0, -10.4722348605376077, 0, 10.4722348605376077, 0, 0, 0, 0, 0, 0, 0.2181715595945335, 0, -5.2361174302688038, 0, 10.4722348605376077, 0, -2.7925959628100294, 0, 0, 0, 0, 0, 0, 0, 0,
        //j15/2,3/2
        0.2473832956432473, 0, 0, 0.7421498869297418, 0, -5.9371990954379337, 0, 0, 0, 0, 0.7421498869297418, 0, -11.8743981908758673, 0, 11.8743981908758673, 0, 0, 0, 0, 0, 0, 0.2473832956432473, 0, -5.9371990954379337, 0, 11.8743981908758673, 0, -3.1665061842335653, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 1.4842997738594836, 0, 0, 0, 0, 1.4842997738594836, 0, -7.9162654605839124, 0, 0, 0, 0, 0, 0, -1.4842997738594836, 0, 0, 0, 4.7497592763503480, 0, 0, 0, 0, 0, 0, 0, 0, -1.4842997738594836, 0, 7.9162654605839124, 0, -4.7497592763503480, 0, 0,
        //j15/2,5/2
        0, 0, 1.9162227683124053, 0, 0, 0, 0, 1.9162227683124053, 0, -10.2198547643328261, 0, 0, 0, 0, 0, 0, -1.9162227683124053, 0, 0, 0, 6.1319128585996969, 0, 0, 0, 0, 0, 0, 0, 0, -1.9162227683124053, 0, 10.2198547643328261, 0, -6.1319128585996969, 0, 0,
        -0.1916222768312404, 0, 0, 0.1916222768312404, 0, 3.8324455366248085, 0, 0, 0, 0, 0.9581113841562022, 0, -7.6648910732496169, 0, -5.1099273821664122, 0, 0, 0, 0, 0, 0, 0.5748668304937213, 0, -11.4973366098744254, 0, 15.3297821464992357, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //j15/2,7/2
        -0.2842217679059343, 0, 0, 0.2842217679059343, 0, 5.6844353581186855, 0, 0, 0, 0, 1.4211088395296716, 0, -11.3688707162373710, 0, -7.5792471441582485, 0, 0, 0, 0, 0, 0, 0.8526653037178029, 0, -17.0533060743560583, 0, 22.7377414324747456, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -1.1368870716237374, 0, 0, 0, 0, 5.6844353581186864, 0, 3.7896235720791247, 0, 0, 0, 0, 0, 0, 5.6844353581186864, 0, -22.7377414324747456, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.1368870716237374, 0, 3.7896235720791247, 0, 0, 0, 0,
        //j15/2,9/2
        0, 0, -1.9691461705205107, 0, 0, 0, 0, 9.8457308526025535, 0, 6.5638205684017024, 0, 0, 0, 0, 0, 0, 9.8457308526025535, 0, -39.3829234104102142, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.9691461705205107, 0, 6.5638205684017024, 0, 0, 0, 0,
        0.1640955142100426, 0, 0, -1.4768596278903832, 0, -1.9691461705205107, 0, 0, 0, 0, -0.8204775710502128, 0, 19.6914617052051071, 0, 0, 0, 0, 0, 0, 0, 0, 0.8204775710502128, 0, -9.8457308526025535, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //j15/2,11/2
        0.3415920525959572, 0, 0, -3.0743284733636154, 0, -4.0991046311514863, 0, 0, 0, 0, -1.7079602629797861, 0, 40.9910463115148644, 0, 0, 0, 0, 0, 0, 0, 0, 1.7079602629797861, 0, -20.4955231557574322, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0.6831841051919142, 0, 0, 0, 0, -10.2477615778787126, 0, 0, 0, 0, 0, 0, 0, 0, 10.2477615778787126, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.6831841051919142, 0, 0, 0, 0, 0, 0,
        //j15/2,13/2
        0, 0, 1.8075352420099966, 0, 0, 0, 0, -27.1130286301499481, 0, 0, 0, 0, 0, 0, 0, 0, 27.1130286301499481, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.8075352420099966, 0, 0, 0, 0, 0, 0,
        -0.1291096601435712, 0, 0, 2.7113028630149949, 0, 0, 0, 0, 0, 0, -4.5188381050249919, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.9037676210049984, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //j15/2,15/2
        -0.5000395635705508, 0, 0, 10.5008308349815653, 0, 0, 0, 0, 0, 0, -17.5013847249692773, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.5002769449938556, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 15/2, mj = -15/2
        -0.500039563570550664, 0, 0, 14.0011077799754186, 0, 0, 0, 0, 0, 0, -35.0027694499385465, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 14.0011077799754186, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.500039563570550664, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0.500039563570550664, 0, 0, 0, 0, -10.5008308349815639, 0, 0, 0, 0, 0, 0, 0, 0, 17.5013847249692732, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3.50027694499385465, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 15/2, mj = -13/2
        0, 0, -1.93664490215356767, 0, 0, 0, 0, 40.669542945224921, 0, 0, 0, 0, 0, 0, 0, 0, -67.7825715753748683, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 13.5565143150749737, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.129109660143571178, 0, 0, 1.80753524200999649, 0, 1.80753524200999649, 0, 0, 0, 0, 0, 0, -27.1130286301499473, 0, 0, 0, 0, 0, 0, 0, 0, -1.80753524200999649, 0, 27.1130286301499473, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.129109660143571178, 0, -1.80753524200999649, 0, 0, 0, 0, 0, 0,
        // j = 15/2, mj = -11/2
        0.341592052595957161, 0, 0, -4.78228873634340025, 0, -4.78228873634340025, 0, 0, 0, 0, 0, 0, 71.7343310451510038, 0, 0, 0, 0, 0, 0, 0, 0, 4.78228873634340025, 0, -71.7343310451510038, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.341592052595957161, 0, 4.78228873634340025, 0, 0, 0, 0, 0, 0,
        0, 0, -1.02477615778787148, 0, 0, 0, 0, 9.22298542009084335, 0, 4.09910463115148593, 0, 0, 0, 0, 0, 0, 5.12388078893935741, 0, -40.9910463115148593, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -5.12388078893935741, 0, 20.4955231557574297, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 15/2, mj = -9/2
        0, 0, 2.13324168473055283, 0, 0, 0, 0, -19.1991751625749755, 0, -8.53296673892221134, 0, 0, 0, 0, 0, 0, -10.6662084236527642, 0, 85.3296673892221134, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 10.6662084236527642, 0, -42.6648336946110567, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.164095514210042526, 0, 0, -0.656382056840170103, 0, -3.93829234104102062, 0, 0, 0, 0, -1.64095514210042526, 0, 19.6914617052051031, 0, 6.56382056840170103, 0, 0, 0, 0, 0, 0, -0.656382056840170103, 0, 19.6914617052051031, 0, -39.3829234104102062, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.164095514210042526, 0, -3.93829234104102062, 0, 6.56382056840170103, 0, 0, 0, 0,
        // j = 15/2, mj = -7/2
        -0.284221767905934336, 0, 0, 1.13688707162373734, 0, 6.82132242974242407, 0, 0, 0, 0, 2.84221767905934336, 0, -34.1066121487121203, 0, -11.3688707162373734, 0, 0, 0, 0, 0, 0, 1.13688707162373734, 0, -34.1066121487121203, 0, 68.2132242974242407, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.284221767905934336, 0, 6.82132242974242407, 0, -11.3688707162373734, 0, 0, 0, 0,
        0, 0, 1.42110883952967168, 0, 0, 0, 0, -1.42110883952967168, 0, -9.47405893019781121, 0, 0, 0, 0, 0, 0, -7.1055441976483584, 0, 18.9481178603956224, 0, 7.57924714415824896, 0, 0, 0, 0, 0, 0, 0, 0, -4.26332651858901504, 0, 28.4221767905934336, 0, -22.7377414324747469, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 15/2, mj = -5/2
        0, 0, -2.10784504514364487, 0, 0, 0, 0, 2.10784504514364487, 0, 14.0523003009576325, 0, 0, 0, 0, 0, 0, 10.5392252257182244, 0, -28.104600601915265, 0, -11.241840240766106, 0, 0, 0, 0, 0, 0, 0, 0, 6.32353513543093462, 0, -42.1569009028728975, 0, 33.725520722298318, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.191622276831240443, 0, 0, -0.383244553662480886, 0, 5.74866830493721329, 0, 0, 0, 0, 0, 0, 5.74866830493721329, 0, -15.3297821464992355, 0, 0, 0, 0, 0, 0, 0.383244553662480886, 0, -5.74866830493721329, 0, 0, 0, 6.13191285859969418, 0, 0, 0, 0, 0, 0, 0, 0, 0.191622276831240443, 0, -5.74866830493721329, 0, 15.3297821464992355, 0, -6.13191285859969418, 0, 0,
        // j = 15/2, mj = -3/2
        0.247383295643247195, 0, 0, 0.49476659128649439, 0, -7.42149886929741585, 0, 0, 0, 0, 0, 0, -7.42149886929741585, 0, 19.7906636514597756, 0, 0, 0, 0, 0, 0, -0.49476659128649439, 0, 7.42149886929741585, 0, 0, 0, -7.91626546058391024, 0, 0, 0, 0, 0, 0, 0, 0, -0.247383295643247195, 0, 7.42149886929741585, 0, -19.7906636514597756, 0, 7.91626546058391024, 0, 0,
        0, 0, -1.73168306950273036, 0, 0, 0, 0, -5.19504920850819109, 0, 13.8534645560218429, 0, 0, 0, 0, 0, 0, -5.19504920850819109, 0, 27.7069291120436858, 0, -16.6241574672262115, 0, 0, 0, 0, 0, 0, 0, 0, -1.73168306950273036, 0, 13.8534645560218429, 0, -16.6241574672262115, 0, 3.16650618423356409, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 15/2, mj = -1/2
        0, 0, 1.96354403635080146, 0, 0, 0, 0, 5.89063210905240439, 0, -15.7083522908064117, 0, 0, 0, 0, 0, 0, 5.89063210905240439, 0, -31.4167045816128234, 0, 18.850022748967694, 0, 0, 0, 0, 0, 0, 0, 0, 1.96354403635080146, 0, -15.7083522908064117, 0, 18.850022748967694, 0, -3.5904805236128941, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.218171559594533496, 0, 0, 0.872686238378133983, 0, -6.98148990702507186, 0, 0, 0, 0, 1.30902935756720097, 0, -20.9444697210752156, 0, 20.9444697210752156, 0, 0, 0, 0, 0, 0, 0.872686238378133983, 0, -20.9444697210752156, 0, 41.8889394421504312, 0, -11.170383851240115, 0, 0, 0, 0, 0, 0, 0, 0, 0.218171559594533496, 0, -6.98148990702507186, 0, 20.9444697210752156, 0, -11.170383851240115, 0, 0.797884560802865356,
        // j = 15/2, mj = 1/2
        -0.218171559594533496, 0, 0, -0.872686238378133983, 0, 6.98148990702507186, 0, 0, 0, 0, -1.30902935756720097, 0, 20.9444697210752156, 0, -20.9444697210752156, 0, 0, 0, 0, 0, 0, -0.872686238378133983, 0, 20.9444697210752156, 0, -41.8889394421504312, 0, 11.170383851240115, 0, 0, 0, 0, 0, 0, 0, 0, -0.218171559594533496, 0, 6.98148990702507186, 0, -20.9444697210752156, 0, 11.170383851240115, 0, -0.797884560802865356,
        0, 0, 1.96354403635080146, 0, 0, 0, 0, 5.89063210905240439, 0, -15.7083522908064117, 0, 0, 0, 0, 0, 0, 5.89063210905240439, 0, -31.4167045816128234, 0, 18.850022748967694, 0, 0, 0, 0, 0, 0, 0, 0, 1.96354403635080146, 0, -15.7083522908064117, 0, 18.850022748967694, 0, -3.5904805236128941, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 15/2, mj = 3/2
        0, 0, -1.73168306950273036, 0, 0, 0, 0, -5.19504920850819109, 0, 13.8534645560218429, 0, 0, 0, 0, 0, 0, -5.19504920850819109, 0, 27.7069291120436858, 0, -16.6241574672262115, 0, 0, 0, 0, 0, 0, 0, 0, -1.73168306950273036, 0, 13.8534645560218429, 0, -16.6241574672262115, 0, 3.16650618423356409, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.247383295643247195, 0, 0, -0.49476659128649439, 0, 7.42149886929741585, 0, 0, 0, 0, 0, 0, 7.42149886929741585, 0, -19.7906636514597756, 0, 0, 0, 0, 0, 0, 0.49476659128649439, 0, -7.42149886929741585, 0, 0, 0, 7.91626546058391024, 0, 0, 0, 0, 0, 0, 0, 0, 0.247383295643247195, 0, -7.42149886929741585, 0, 19.7906636514597756, 0, -7.91626546058391024, 0, 0,
        // j = 15/2, mj = 5/2
        0.191622276831240443, 0, 0, 0.383244553662480886, 0, -5.74866830493721329, 0, 0, 0, 0, 0, 0, -5.74866830493721329, 0, 15.3297821464992355, 0, 0, 0, 0, 0, 0, -0.383244553662480886, 0, 5.74866830493721329, 0, 0, 0, -6.13191285859969418, 0, 0, 0, 0, 0, 0, 0, 0, -0.191622276831240443, 0, 5.74866830493721329, 0, -15.3297821464992355, 0, 6.13191285859969418, 0, 0,
        0, 0, -2.10784504514364487, 0, 0, 0, 0, 2.10784504514364487, 0, 14.0523003009576325, 0, 0, 0, 0, 0, 0, 10.5392252257182244, 0, -28.104600601915265, 0, -11.241840240766106, 0, 0, 0, 0, 0, 0, 0, 0, 6.32353513543093462, 0, -42.1569009028728975, 0, 33.725520722298318, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 15/2, mj = 7/2
        0, 0, 1.42110883952967168, 0, 0, 0, 0, -1.42110883952967168, 0, -9.47405893019781121, 0, 0, 0, 0, 0, 0, -7.1055441976483584, 0, 18.9481178603956224, 0, 7.57924714415824896, 0, 0, 0, 0, 0, 0, 0, 0, -4.26332651858901504, 0, 28.4221767905934336, 0, -22.7377414324747469, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.284221767905934336, 0, 0, -1.13688707162373734, 0, -6.82132242974242407, 0, 0, 0, 0, -2.84221767905934336, 0, 34.1066121487121203, 0, 11.3688707162373734, 0, 0, 0, 0, 0, 0, -1.13688707162373734, 0, 34.1066121487121203, 0, -68.2132242974242407, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.284221767905934336, 0, -6.82132242974242407, 0, 11.3688707162373734, 0, 0, 0, 0,
        // j = 15/2, mj = 9/2
        -0.164095514210042526, 0, 0, 0.656382056840170103, 0, 3.93829234104102062, 0, 0, 0, 0, 1.64095514210042526, 0, -19.6914617052051031, 0, -6.56382056840170103, 0, 0, 0, 0, 0, 0, 0.656382056840170103, 0, -19.6914617052051031, 0, 39.3829234104102062, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.164095514210042526, 0, 3.93829234104102062, 0, -6.56382056840170103, 0, 0, 0, 0,
        0, 0, 2.13324168473055283, 0, 0, 0, 0, -19.1991751625749755, 0, -8.53296673892221134, 0, 0, 0, 0, 0, 0, -10.6662084236527642, 0, 85.3296673892221134, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 10.6662084236527642, 0, -42.6648336946110567, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 15/2, mj = 11/2
        0, 0, -1.02477615778787148, 0, 0, 0, 0, 9.22298542009084335, 0, 4.09910463115148593, 0, 0, 0, 0, 0, 0, 5.12388078893935741, 0, -40.9910463115148593, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -5.12388078893935741, 0, 20.4955231557574297, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.341592052595957161, 0, 0, 4.78228873634340025, 0, 4.78228873634340025, 0, 0, 0, 0, 0, 0, -71.7343310451510038, 0, 0, 0, 0, 0, 0, 0, 0, -4.78228873634340025, 0, 71.7343310451510038, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.341592052595957161, 0, -4.78228873634340025, 0, 0, 0, 0, 0, 0,
        // j = 15/2, mj = 13/2
        0.129109660143571178, 0, 0, -1.80753524200999649, 0, -1.80753524200999649, 0, 0, 0, 0, 0, 0, 27.1130286301499473, 0, 0, 0, 0, 0, 0, 0, 0, 1.80753524200999649, 0, -27.1130286301499473, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.129109660143571178, 0, 1.80753524200999649, 0, 0, 0, 0, 0, 0,
        0, 0, -1.93664490215356767, 0, 0, 0, 0, 40.669542945224921, 0, 0, 0, 0, 0, 0, 0, 0, -67.7825715753748683, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 13.5565143150749737, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 15/2, mj = 15/2
        0, 0, 0.500039563570550664, 0, 0, 0, 0, -10.5008308349815639, 0, 0, 0, 0, 0, 0, 0, 0, 17.5013847249692732, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3.50027694499385465, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.500039563570550664, 0, 0, -14.0011077799754186, 0, 0, 0, 0, 0, 0, 35.0027694499385465, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -14.0011077799754186, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.500039563570550664, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = -17/2
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.51542898439728431, 0, 0, -14.4320115631239607, 0, 0, 0, 0, 0, 0, 36.0800289078099017, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -14.4320115631239607, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.51542898439728431, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = -15/2
        0.125009890892637666, 0, 0, -3.50027694499385465, 0, 0, 0, 0, 0, 0, 8.75069236248463662, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3.50027694499385465, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.125009890892637666, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 2.00015825428220266, 0, 0, 0, 0, -42.0033233399262558, 0, 0, 0, 0, 0, 0, 0, 0, 70.005538899877093, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -14.0011077799754186, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = -13/2
        0, 0, 0.707162732524596178, 0, 0, 0, 0, -14.8504173830165197, 0, 0, 0, 0, 0, 0, 0, 0, 24.7506956383608662, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4.95013912767217325, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.353581366262298089, 0, 0, 4.95013912767217325, 0, 4.95013912767217325, 0, 0, 0, 0, 0, 0, -74.2520869150825987, 0, 0, 0, 0, 0, 0, 0, 0, -4.95013912767217325, 0, 74.2520869150825987, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.353581366262298089, 0, -4.95013912767217325, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = -11/2
        -0.158126394107949853, 0, 0, 2.21376951751129794, 0, 2.21376951751129794, 0, 0, 0, 0, 0, 0, -33.2065427626694691, 0, 0, 0, 0, 0, 0, 0, 0, -2.21376951751129794, 0, 33.2065427626694691, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.158126394107949853, 0, -2.21376951751129794, 0, 0, 0, 0, 0, 0,
        0, 0, -2.21376951751129794, 0, 0, 0, 0, 19.9239256576016814, 0, 8.85507807004519175, 0, 0, 0, 0, 0, 0, 11.0688475875564897, 0, -88.5507807004519175, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -11.0688475875564897, 0, 44.2753903502259587, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = -9/2
        0, 0, -1.18330958111587602, 0, 0, 0, 0, 10.6497862300428841, 0, 4.73323832446350406, 0, 0, 0, 0, 0, 0, 5.91654790557938008, 0, -47.3323832446350406, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -5.91654790557938008, 0, 23.6661916223175203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.295827395278969004, 0, 0, -1.18330958111587602, 0, -7.0998574866952561, 0, 0, 0, 0, -2.95827395278969004, 0, 35.4992874334762805, 0, 11.8330958111587602, 0, 0, 0, 0, 0, 0, -1.18330958111587602, 0, 35.4992874334762805, 0, -70.998574866952561, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.295827395278969004, 0, -7.0998574866952561, 0, 11.8330958111587602, 0, 0, 0, 0,
        // j = 17/2, mj = -7/2
        0.183464362288218895, 0, 0, -0.733857449152875582, 0, -4.40314469491725349, 0, 0, 0, 0, -1.83464362288218895, 0, 22.0157234745862674, 0, 7.33857449152875582, 0, 0, 0, 0, 0, 0, -0.733857449152875582, 0, 22.0157234745862674, 0, -44.0314469491725349, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.183464362288218895, 0, -4.40314469491725349, 0, 7.33857449152875582, 0, 0, 0, 0,
        0, 0, 2.20157234745862674, 0, 0, 0, 0, -2.20157234745862674, 0, -14.6771489830575116, 0, 0, 0, 0, 0, 0, -11.0078617372931337, 0, 29.3542979661150233, 0, 11.7417191864460093, 0, 0, 0, 0, 0, 0, 0, 0, -6.60471704237588023, 0, 44.0314469491725349, 0, -35.2251575593380279, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = -5/2
        0, 0, 1.55674673616078096, 0, 0, 0, 0, -1.55674673616078096, 0, -10.3783115744052064, 0, 0, 0, 0, 0, 0, -7.7837336808039048, 0, 20.7566231488104128, 0, 8.30264925952416512, 0, 0, 0, 0, 0, 0, 0, 0, -4.67024020848234288, 0, 31.1349347232156192, 0, -24.9079477785724953, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.25945778936013016, 0, 0, -0.51891557872026032, 0, 7.7837336808039048, 0, 0, 0, 0, 0, 0, 7.7837336808039048, 0, -20.7566231488104128, 0, 0, 0, 0, 0, 0, 0.51891557872026032, 0, -7.7837336808039048, 0, 0, 0, 8.30264925952416512, 0, 0, 0, 0, 0, 0, 0, 0, 0.25945778936013016, 0, -7.7837336808039048, 0, 20.7566231488104128, 0, -8.30264925952416512, 0, 0,
        // j = 17/2, mj = -3/2
        -0.206975714696966254, 0, 0, -0.413951429393932508, 0, 6.20927144090898762, 0, 0, 0, 0, 0, 0, 6.20927144090898762, 0, -16.5580571757573003, 0, 0, 0, 0, 0, 0, 0.413951429393932508, 0, -6.20927144090898762, 0, 0, 0, 6.62322287030292013, 0, 0, 0, 0, 0, 0, 0, 0, 0.206975714696966254, 0, -6.20927144090898762, 0, 16.5580571757573003, 0, -6.62322287030292013, 0, 0,
        0, 0, -2.06975714696966254, 0, 0, 0, 0, -6.20927144090898762, 0, 16.5580571757573003, 0, 0, 0, 0, 0, 0, -6.20927144090898762, 0, 33.1161143515146006, 0, -19.8696686109087604, 0, 0, 0, 0, 0, 0, 0, 0, -2.06975714696966254, 0, 16.5580571757573003, 0, -19.8696686109087604, 0, 3.78469878303024007, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = -1/2
        0, 0, -1.85124707101607532, 0, 0, 0, 0, -5.55374121304822595, 0, 14.8099765681286025, 0, 0, 0, 0, 0, 0, -5.55374121304822595, 0, 29.6199531362572051, 0, -17.771971881754323, 0, 0, 0, 0, 0, 0, 0, 0, -1.85124707101607532, 0, 14.8099765681286025, 0, -17.771971881754323, 0, 3.38513750128653772, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.231405883877009415, 0, 0, 0.925623535508037658, 0, -7.40498828406430127, 0, 0, 0, 0, 1.38843530326205649, 0, -22.2149648521929038, 0, 22.2149648521929038, 0, 0, 0, 0, 0, 0, 0.925623535508037658, 0, -22.2149648521929038, 0, 44.4299297043858076, 0, -11.847981254502882, 0, 0, 0, 0, 0, 0, 0, 0, 0.231405883877009415, 0, -7.40498828406430127, 0, 22.2149648521929038, 0, -11.847981254502882, 0, 0.84628437532163443,
        // j = 17/2, mj = 1/2
        0.231405883877009415, 0, 0, 0.925623535508037658, 0, -7.40498828406430127, 0, 0, 0, 0, 1.38843530326205649, 0, -22.2149648521929038, 0, 22.2149648521929038, 0, 0, 0, 0, 0, 0, 0.925623535508037658, 0, -22.2149648521929038, 0, 44.4299297043858076, 0, -11.847981254502882, 0, 0, 0, 0, 0, 0, 0, 0, 0.231405883877009415, 0, -7.40498828406430127, 0, 22.2149648521929038, 0, -11.847981254502882, 0, 0.84628437532163443,
        0, 0, 1.85124707101607532, 0, 0, 0, 0, 5.55374121304822595, 0, -14.8099765681286025, 0, 0, 0, 0, 0, 0, 5.55374121304822595, 0, -29.6199531362572051, 0, 17.771971881754323, 0, 0, 0, 0, 0, 0, 0, 0, 1.85124707101607532, 0, -14.8099765681286025, 0, 17.771971881754323, 0, -3.38513750128653772, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = 3/2
        0, 0, 2.06975714696966254, 0, 0, 0, 0, 6.20927144090898762, 0, -16.5580571757573003, 0, 0, 0, 0, 0, 0, 6.20927144090898762, 0, -33.1161143515146006, 0, 19.8696686109087604, 0, 0, 0, 0, 0, 0, 0, 0, 2.06975714696966254, 0, -16.5580571757573003, 0, 19.8696686109087604, 0, -3.78469878303024007, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.206975714696966254, 0, 0, -0.413951429393932508, 0, 6.20927144090898762, 0, 0, 0, 0, 0, 0, 6.20927144090898762, 0, -16.5580571757573003, 0, 0, 0, 0, 0, 0, 0.413951429393932508, 0, -6.20927144090898762, 0, 0, 0, 6.62322287030292013, 0, 0, 0, 0, 0, 0, 0, 0, 0.206975714696966254, 0, -6.20927144090898762, 0, 16.5580571757573003, 0, -6.62322287030292013, 0, 0,
        // j = 17/2, mj = 5/2
        -0.25945778936013016, 0, 0, -0.51891557872026032, 0, 7.7837336808039048, 0, 0, 0, 0, 0, 0, 7.7837336808039048, 0, -20.7566231488104128, 0, 0, 0, 0, 0, 0, 0.51891557872026032, 0, -7.7837336808039048, 0, 0, 0, 8.30264925952416512, 0, 0, 0, 0, 0, 0, 0, 0, 0.25945778936013016, 0, -7.7837336808039048, 0, 20.7566231488104128, 0, -8.30264925952416512, 0, 0,
        0, 0, -1.55674673616078096, 0, 0, 0, 0, 1.55674673616078096, 0, 10.3783115744052064, 0, 0, 0, 0, 0, 0, 7.7837336808039048, 0, -20.7566231488104128, 0, -8.30264925952416512, 0, 0, 0, 0, 0, 0, 0, 0, 4.67024020848234288, 0, -31.1349347232156192, 0, 24.9079477785724953, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = 7/2
        0, 0, -2.20157234745862674, 0, 0, 0, 0, 2.20157234745862674, 0, 14.6771489830575116, 0, 0, 0, 0, 0, 0, 11.0078617372931337, 0, -29.3542979661150233, 0, -11.7417191864460093, 0, 0, 0, 0, 0, 0, 0, 0, 6.60471704237588023, 0, -44.0314469491725349, 0, 35.2251575593380279, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.183464362288218895, 0, 0, -0.733857449152875582, 0, -4.40314469491725349, 0, 0, 0, 0, -1.83464362288218895, 0, 22.0157234745862674, 0, 7.33857449152875582, 0, 0, 0, 0, 0, 0, -0.733857449152875582, 0, 22.0157234745862674, 0, -44.0314469491725349, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.183464362288218895, 0, -4.40314469491725349, 0, 7.33857449152875582, 0, 0, 0, 0,
        // j = 17/2, mj = 9/2
        0.295827395278969004, 0, 0, -1.18330958111587602, 0, -7.0998574866952561, 0, 0, 0, 0, -2.95827395278969004, 0, 35.4992874334762805, 0, 11.8330958111587602, 0, 0, 0, 0, 0, 0, -1.18330958111587602, 0, 35.4992874334762805, 0, -70.998574866952561, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.295827395278969004, 0, -7.0998574866952561, 0, 11.8330958111587602, 0, 0, 0, 0,
        0, 0, 1.18330958111587602, 0, 0, 0, 0, -10.6497862300428841, 0, -4.73323832446350406, 0, 0, 0, 0, 0, 0, -5.91654790557938008, 0, 47.3323832446350406, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5.91654790557938008, 0, -23.6661916223175203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = 11/2
        0, 0, 2.21376951751129794, 0, 0, 0, 0, -19.9239256576016814, 0, -8.85507807004519175, 0, 0, 0, 0, 0, 0, -11.0688475875564897, 0, 88.5507807004519175, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 11.0688475875564897, 0, -44.2753903502259587, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.158126394107949853, 0, 0, 2.21376951751129794, 0, 2.21376951751129794, 0, 0, 0, 0, 0, 0, -33.2065427626694691, 0, 0, 0, 0, 0, 0, 0, 0, -2.21376951751129794, 0, 33.2065427626694691, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.158126394107949853, 0, -2.21376951751129794, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = 13/2
        -0.353581366262298089, 0, 0, 4.95013912767217325, 0, 4.95013912767217325, 0, 0, 0, 0, 0, 0, -74.2520869150825987, 0, 0, 0, 0, 0, 0, 0, 0, -4.95013912767217325, 0, 74.2520869150825987, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.353581366262298089, 0, -4.95013912767217325, 0, 0, 0, 0, 0, 0,
        0, 0, -0.707162732524596178, 0, 0, 0, 0, 14.8504173830165197, 0, 0, 0, 0, 0, 0, 0, 0, -24.7506956383608662, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4.95013912767217325, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = 15/2
        0, 0, -2.00015825428220266, 0, 0, 0, 0, 42.0033233399262558, 0, 0, 0, 0, 0, 0, 0, 0, -70.005538899877093, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 14.0011077799754186, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.125009890892637666, 0, 0, -3.50027694499385465, 0, 0, 0, 0, 0, 0, 8.75069236248463662, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3.50027694499385465, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.125009890892637666, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = 17/2
        0.51542898439728431, 0, 0, -14.4320115631239607, 0, 0, 0, 0, 0, 0, 36.0800289078099017, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -14.4320115631239607, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.51542898439728431, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = -17/2
        -0.51542898439728431, 0, 0, 18.5554434383022352, 0, 0, 0, 0, 0, 0, -64.944052034057823, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 43.296034689371882, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4.63886085957555879, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0.51542898439728431, 0, 0, 0, 0, -14.4320115631239607, 0, 0, 0, 0, 0, 0, 0, 0, 36.0800289078099017, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -14.4320115631239607, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.51542898439728431, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = -15/2
        0, 0, -2.12516814517484032, 0, 0, 0, 0, 59.504708064895529, 0, 0, 0, 0, 0, 0, 0, 0, -148.761770162238823, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 59.504708064895529, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.12516814517484032, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.125009890892637666, 0, 0, 2.50019781785275332, 0, 2.00015825428220266, 0, 0, 0, 0, -1.75013847249692732, 0, -42.0033233399262558, 0, 0, 0, 0, 0, 0, 0, 0, -3.50027694499385465, 0, 70.005538899877093, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.875069236248463662, 0, -14.0011077799754186, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = -13/2
        0.353581366262298089, 0, 0, -7.07162732524596178, 0, -5.65730186019676943, 0, 0, 0, 0, 4.95013912767217325, 0, 118.803339064132158, 0, 0, 0, 0, 0, 0, 0, 0, 9.9002782553443465, 0, -198.00556510688693, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.47506956383608662, 0, 39.601113021377386, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -1.06074409878689427, 0, 0, 0, 0, 14.8504173830165197, 0, 4.95013912767217325, 0, 0, 0, 0, 0, 0, 0, 0, -74.2520869150825987, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -14.8504173830165197, 0, 74.2520869150825987, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.06074409878689427, 0, -4.95013912767217325, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = -11/2
        0, 0, 2.37189591161924779, 0, 0, 0, 0, -33.2065427626694691, 0, -11.0688475875564897, 0, 0, 0, 0, 0, 0, 0, 0, 166.032713813347345, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 33.2065427626694691, 0, -166.032713813347345, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.37189591161924779, 0, 11.0688475875564897, 0, 0, 0, 0, 0, 0,
        0.158126394107949853, 0, 0, -1.26501115286359882, 0, -4.42753903502259587, 0, 0, 0, 0, -2.21376951751129794, 0, 39.8478513152033629, 0, 8.85507807004519175, 0, 0, 0, 0, 0, 0, 0, 0, 22.1376951751129794, 0, -88.5507807004519175, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.790631970539749263, 0, -22.1376951751129794, 0, 44.2753903502259587, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = -9/2
        -0.295827395278969004, 0, 0, 2.36661916223175203, 0, 8.28316706781113211, 0, 0, 0, 0, 4.14158353390556606, 0, -74.548503610300189, 0, -16.5663341356222642, 0, 0, 0, 0, 0, 0, 0, 0, -41.4158353390556606, 0, 165.663341356222642, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.47913697639484502, 0, 41.4158353390556606, 0, -82.8316706781113211, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 1.47913697639484502, 0, 0, 0, 0, -5.91654790557938008, 0, -11.8330958111587602, 0, 0, 0, 0, 0, 0, -14.7913697639484502, 0, 59.1654790557938008, 0, 11.8330958111587602, 0, 0, 0, 0, 0, 0, 0, 0, -5.91654790557938008, 0, 59.1654790557938008, 0, -70.998574866952561, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.47913697639484502, 0, -11.8330958111587602, 0, 11.8330958111587602, 0, 0, 0, 0,
        // j = 17/2, mj = -7/2
        0, 0, -2.38503670974684564, 0, 0, 0, 0, 9.54014683898738256, 0, 19.0802936779747651, 0, 0, 0, 0, 0, 0, 23.8503670974684564, 0, -95.4014683898738256, 0, -19.0802936779747651, 0, 0, 0, 0, 0, 0, 0, 0, 9.54014683898738256, 0, -95.4014683898738256, 0, 114.481762067848591, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.38503670974684564, 0, 19.0802936779747651, 0, -19.0802936779747651, 0, 0, 0, 0,
        -0.183464362288218895, 0, 0, 0, 0, 6.60471704237588023, 0, 0, 0, 0, 1.10078617372931337, 0, -6.60471704237588023, 0, -22.0157234745862674, 0, 0, 0, 0, 0, 0, 1.46771489830575116, 0, -33.0235852118794012, 0, 44.0314469491725349, 0, 11.7417191864460093, 0, 0, 0, 0, 0, 0, 0, 0, 0.550393086864656686, 0, -19.8141511271276407, 0, 66.0471704237588023, 0, -35.2251575593380279, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = -5/2
        0.25945778936013016, 0, 0, 0, 0, -9.34048041696468576, 0, 0, 0, 0, -1.55674673616078096, 0, 9.34048041696468576, 0, 31.1349347232156192, 0, 0, 0, 0, 0, 0, -2.07566231488104128, 0, 46.7024020848234288, 0, -62.2698694464312384, 0, -16.6052985190483302, 0, 0, 0, 0, 0, 0, 0, 0, -0.77837336808039048, 0, 28.0214412508940573, 0, -93.4048041696468576, 0, 49.8158955571449907, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -1.81620452552091112, 0, 0, 0, 0, -3.63240905104182224, 0, 18.1620452552091112, 0, 0, 0, 0, 0, 0, 0, 0, 18.1620452552091112, 0, -29.0592724083345779, 0, 0, 0, 0, 0, 0, 0, 0, 3.63240905104182224, 0, -18.1620452552091112, 0, 0, 0, 8.30264925952416512, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.81620452552091112, 0, -18.1620452552091112, 0, 29.0592724083345779, 0, -8.30264925952416512, 0, 0,
        // j = 17/2, mj = -3/2
        0, 0, 2.27673286166662879, 0, 0, 0, 0, 4.55346572333325759, 0, -22.7673286166662879, 0, 0, 0, 0, 0, 0, 0, 0, -22.7673286166662879, 0, 36.4277257866660607, 0, 0, 0, 0, 0, 0, 0, 0, -4.55346572333325759, 0, 22.7673286166662879, 0, 0, 0, -10.4079216533331602, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.27673286166662879, 0, 22.7673286166662879, 0, -36.4277257866660607, 0, 10.4079216533331602, 0, 0,
        0.206975714696966254, 0, 0, 0.827902858787865016, 0, -8.27902858787865016, 0, 0, 0, 0, 1.24185428818179752, 0, -24.8370857636359505, 0, 33.1161143515146006, 0, 0, 0, 0, 0, 0, 0.827902858787865016, 0, -24.8370857636359505, 0, 66.2322287030292013, 0, -26.4928914812116805, 0, 0, 0, 0, 0, 0, 0, 0, 0.206975714696966254, 0, -8.27902858787865016, 0, 33.1161143515146006, 0, -26.4928914812116805, 0, 3.78469878303024007, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = -1/2
        -0.231405883877009415, 0, 0, -0.925623535508037658, 0, 9.25623535508037658, 0, 0, 0, 0, -1.38843530326205649, 0, 27.7687060652411297, 0, -37.0249414203215063, 0, 0, 0, 0, 0, 0, -0.925623535508037658, 0, 27.7687060652411297, 0, -74.0498828406430127, 0, 29.6199531362572051, 0, 0, 0, 0, 0, 0, 0, 0, -0.231405883877009415, 0, 9.25623535508037658, 0, -37.0249414203215063, 0, 29.6199531362572051, 0, -4.23142187660817215, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 2.08265295489308473, 0, 0, 0, 0, 8.33061181957233892, 0, -22.2149648521929038, 0, 0, 0, 0, 0, 0, 12.4959177293585084, 0, -66.6448945565787114, 0, 39.9869367339472268, 0, 0, 0, 0, 0, 0, 0, 0, 8.33061181957233892, 0, -66.6448945565787114, 0, 79.9738734678944537, 0, -15.2331187557894197, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.08265295489308473, 0, -22.2149648521929038, 0, 39.9869367339472268, 0, -15.2331187557894197, 0, 0.84628437532163443,
        // j = 17/2, mj = 1/2
        0, 0, -2.08265295489308473, 0, 0, 0, 0, -8.33061181957233892, 0, 22.2149648521929038, 0, 0, 0, 0, 0, 0, -12.4959177293585084, 0, 66.6448945565787114, 0, -39.9869367339472268, 0, 0, 0, 0, 0, 0, 0, 0, -8.33061181957233892, 0, 66.6448945565787114, 0, -79.9738734678944537, 0, 15.2331187557894197, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.08265295489308473, 0, 22.2149648521929038, 0, -39.9869367339472268, 0, 15.2331187557894197, 0, -0.84628437532163443,
        -0.231405883877009415, 0, 0, -0.925623535508037658, 0, 9.25623535508037658, 0, 0, 0, 0, -1.38843530326205649, 0, 27.7687060652411297, 0, -37.0249414203215063, 0, 0, 0, 0, 0, 0, -0.925623535508037658, 0, 27.7687060652411297, 0, -74.0498828406430127, 0, 29.6199531362572051, 0, 0, 0, 0, 0, 0, 0, 0, -0.231405883877009415, 0, 9.25623535508037658, 0, -37.0249414203215063, 0, 29.6199531362572051, 0, -4.23142187660817215, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = 3/2
        0.206975714696966254, 0, 0, 0.827902858787865016, 0, -8.27902858787865016, 0, 0, 0, 0, 1.24185428818179752, 0, -24.8370857636359505, 0, 33.1161143515146006, 0, 0, 0, 0, 0, 0, 0.827902858787865016, 0, -24.8370857636359505, 0, 66.2322287030292013, 0, -26.4928914812116805, 0, 0, 0, 0, 0, 0, 0, 0, 0.206975714696966254, 0, -8.27902858787865016, 0, 33.1161143515146006, 0, -26.4928914812116805, 0, 3.78469878303024007, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -2.27673286166662879, 0, 0, 0, 0, -4.55346572333325759, 0, 22.7673286166662879, 0, 0, 0, 0, 0, 0, 0, 0, 22.7673286166662879, 0, -36.4277257866660607, 0, 0, 0, 0, 0, 0, 0, 0, 4.55346572333325759, 0, -22.7673286166662879, 0, 0, 0, 10.4079216533331602, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.27673286166662879, 0, -22.7673286166662879, 0, 36.4277257866660607, 0, -10.4079216533331602, 0, 0,
        // j = 17/2, mj = 5/2
        0, 0, 1.81620452552091112, 0, 0, 0, 0, 3.63240905104182224, 0, -18.1620452552091112, 0, 0, 0, 0, 0, 0, 0, 0, -18.1620452552091112, 0, 29.0592724083345779, 0, 0, 0, 0, 0, 0, 0, 0, -3.63240905104182224, 0, 18.1620452552091112, 0, 0, 0, -8.30264925952416512, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.81620452552091112, 0, 18.1620452552091112, 0, -29.0592724083345779, 0, 8.30264925952416512, 0, 0,
        0.25945778936013016, 0, 0, 0, 0, -9.34048041696468576, 0, 0, 0, 0, -1.55674673616078096, 0, 9.34048041696468576, 0, 31.1349347232156192, 0, 0, 0, 0, 0, 0, -2.07566231488104128, 0, 46.7024020848234288, 0, -62.2698694464312384, 0, -16.6052985190483302, 0, 0, 0, 0, 0, 0, 0, 0, -0.77837336808039048, 0, 28.0214412508940573, 0, -93.4048041696468576, 0, 49.8158955571449907, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = 7/2
        -0.183464362288218895, 0, 0, 0, 0, 6.60471704237588023, 0, 0, 0, 0, 1.10078617372931337, 0, -6.60471704237588023, 0, -22.0157234745862674, 0, 0, 0, 0, 0, 0, 1.46771489830575116, 0, -33.0235852118794012, 0, 44.0314469491725349, 0, 11.7417191864460093, 0, 0, 0, 0, 0, 0, 0, 0, 0.550393086864656686, 0, -19.8141511271276407, 0, 66.0471704237588023, 0, -35.2251575593380279, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 2.38503670974684564, 0, 0, 0, 0, -9.54014683898738256, 0, -19.0802936779747651, 0, 0, 0, 0, 0, 0, -23.8503670974684564, 0, 95.4014683898738256, 0, 19.0802936779747651, 0, 0, 0, 0, 0, 0, 0, 0, -9.54014683898738256, 0, 95.4014683898738256, 0, -114.481762067848591, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.38503670974684564, 0, -19.0802936779747651, 0, 19.0802936779747651, 0, 0, 0, 0,
        // j = 17/2, mj = 9/2
        0, 0, -1.47913697639484502, 0, 0, 0, 0, 5.91654790557938008, 0, 11.8330958111587602, 0, 0, 0, 0, 0, 0, 14.7913697639484502, 0, -59.1654790557938008, 0, -11.8330958111587602, 0, 0, 0, 0, 0, 0, 0, 0, 5.91654790557938008, 0, -59.1654790557938008, 0, 70.998574866952561, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.47913697639484502, 0, 11.8330958111587602, 0, -11.8330958111587602, 0, 0, 0, 0,
        -0.295827395278969004, 0, 0, 2.36661916223175203, 0, 8.28316706781113211, 0, 0, 0, 0, 4.14158353390556606, 0, -74.548503610300189, 0, -16.5663341356222642, 0, 0, 0, 0, 0, 0, 0, 0, -41.4158353390556606, 0, 165.663341356222642, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.47913697639484502, 0, 41.4158353390556606, 0, -82.8316706781113211, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = 11/2
        0.158126394107949853, 0, 0, -1.26501115286359882, 0, -4.42753903502259587, 0, 0, 0, 0, -2.21376951751129794, 0, 39.8478513152033629, 0, 8.85507807004519175, 0, 0, 0, 0, 0, 0, 0, 0, 22.1376951751129794, 0, -88.5507807004519175, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.790631970539749263, 0, -22.1376951751129794, 0, 44.2753903502259587, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -2.37189591161924779, 0, 0, 0, 0, 33.2065427626694691, 0, 11.0688475875564897, 0, 0, 0, 0, 0, 0, 0, 0, -166.032713813347345, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -33.2065427626694691, 0, 166.032713813347345, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.37189591161924779, 0, -11.0688475875564897, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = 13/2
        0, 0, 1.06074409878689427, 0, 0, 0, 0, -14.8504173830165197, 0, -4.95013912767217325, 0, 0, 0, 0, 0, 0, 0, 0, 74.2520869150825987, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 14.8504173830165197, 0, -74.2520869150825987, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.06074409878689427, 0, 4.95013912767217325, 0, 0, 0, 0, 0, 0,
        0.353581366262298089, 0, 0, -7.07162732524596178, 0, -5.65730186019676943, 0, 0, 0, 0, 4.95013912767217325, 0, 118.803339064132158, 0, 0, 0, 0, 0, 0, 0, 0, 9.9002782553443465, 0, -198.00556510688693, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.47506956383608662, 0, 39.601113021377386, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = 15/2
        -0.125009890892637666, 0, 0, 2.50019781785275332, 0, 2.00015825428220266, 0, 0, 0, 0, -1.75013847249692732, 0, -42.0033233399262558, 0, 0, 0, 0, 0, 0, 0, 0, -3.50027694499385465, 0, 70.005538899877093, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.875069236248463662, 0, -14.0011077799754186, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 2.12516814517484032, 0, 0, 0, 0, -59.504708064895529, 0, 0, 0, 0, 0, 0, 0, 0, 148.761770162238823, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -59.504708064895529, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.12516814517484032, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = 17/2
        0, 0, -0.51542898439728431, 0, 0, 0, 0, 14.4320115631239607, 0, 0, 0, 0, 0, 0, 0, 0, -36.0800289078099017, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 14.4320115631239607, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.51542898439728431, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.51542898439728431, 0, 0, 18.5554434383022352, 0, 0, 0, 0, 0, 0, -64.944052034057823, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 43.296034689371882, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4.63886085957555879, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = -19/2
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.52955294149244958, 0, 0, -19.0639058937281849, 0, 0, 0, 0, 0, 0, 66.723670628048647, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -44.4824470853657647, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4.76597647343204622, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = -17/2
        0.121487776695804978, 0, 0, -4.37355996104897921, 0, 0, 0, 0, 0, 0, 15.3074598636714272, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -10.2049732424476182, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.0933899902622448, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 2.18677998052448961, 0, 0, 0, 0, -61.229839454685709, 0, 0, 0, 0, 0, 0, 0, 0, 153.074598636714272, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -61.229839454685709, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.18677998052448961, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = -15/2
        0, 0, 0.728926660174829869, 0, 0, 0, 0, -20.4099464848952363, 0, 0, 0, 0, 0, 0, 0, 0, 51.0248662122380908, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -20.4099464848952363, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.728926660174829869, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.364463330087414934, 0, 0, 7.28926660174829869, 0, 5.83141328139863895, 0, 0, 0, 0, -5.10248662122380908, 0, -122.459678909371418, 0, 0, 0, 0, 0, 0, 0, 0, -10.2049732424476182, 0, 204.099464848952363, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.55124331061190454, 0, -40.8198929697904727, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = -13/2
        -0.153105222743980097, 0, 0, 3.06210445487960195, 0, 2.44968356390368156, 0, 0, 0, 0, -2.14347311841572136, 0, -51.4433548419773127, 0, 0, 0, 0, 0, 0, 0, 0, -4.28694623683144273, 0, 85.7389247366288545, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.07173655920786068, 0, -17.1477849473257709, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -2.44968356390368156, 0, 0, 0, 0, 34.2955698946515418, 0, 11.4318566315505139, 0, 0, 0, 0, 0, 0, 0, 0, -171.477849473257709, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -34.2955698946515418, 0, 171.477849473257709, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.44968356390368156, 0, -11.4318566315505139, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = -11/2
        0, 0, -1.22484178195184078, 0, 0, 0, 0, 17.1477849473257709, 0, 5.71592831577525697, 0, 0, 0, 0, 0, 0, 0, 0, -85.7389247366288545, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -17.1477849473257709, 0, 85.7389247366288545, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.22484178195184078, 0, -5.71592831577525697, 0, 0, 0, 0, 0, 0,
        0.306210445487960195, 0, 0, -2.44968356390368156, 0, -8.57389247366288545, 0, 0, 0, 0, -4.28694623683144273, 0, 77.1650322629659691, 0, 17.1477849473257709, 0, 0, 0, 0, 0, 0, 0, 0, 42.8694623683144273, 0, -171.477849473257709, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.53105222743980097, 0, -42.8694623683144273, 0, 85.7389247366288545, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = -9/2
        0.176790683131149045, 0, 0, -1.41432546504919236, 0, -4.95013912767217325, 0, 0, 0, 0, -2.47506956383608662, 0, 44.5512521490495592, 0, 9.9002782553443465, 0, 0, 0, 0, 0, 0, 0, 0, 24.7506956383608662, 0, -99.002782553443465, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.883953415655745223, 0, -24.7506956383608662, 0, 49.5013912767217325, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 2.47506956383608662, 0, 0, 0, 0, -9.9002782553443465, 0, -19.800556510688693, 0, 0, 0, 0, 0, 0, -24.7506956383608662, 0, 99.002782553443465, 0, 19.800556510688693, 0, 0, 0, 0, 0, 0, 0, 0, -9.9002782553443465, 0, 99.002782553443465, 0, -118.803339064132158, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.47506956383608662, 0, -19.800556510688693, 0, 19.800556510688693, 0, 0, 0, 0,
        // j = 19/2, mj = -7/2
        0, 0, 1.62031337522288607, 0, 0, 0, 0, -6.48125350089154427, 0, -12.9625070017830885, 0, 0, 0, 0, 0, 0, -16.2031337522288607, 0, 64.8125350089154427, 0, 12.9625070017830885, 0, 0, 0, 0, 0, 0, 0, 0, -6.48125350089154427, 0, 64.8125350089154427, 0, -77.7750420106985312, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.62031337522288607, 0, -12.9625070017830885, 0, 12.9625070017830885, 0, 0, 0, 0,
        -0.270052229203814345, 0, 0, 0, 0, 9.7218802513373164, 0, 0, 0, 0, 1.62031337522288607, 0, -9.7218802513373164, 0, -32.4062675044577213, 0, 0, 0, 0, 0, 0, 2.16041783363051476, 0, -48.609401256686582, 0, 64.8125350089154427, 0, 17.2833426690441181, 0, 0, 0, 0, 0, 0, 0, 0, 0.810156687611443034, 0, -29.1656407540119492, 0, 97.218802513373164, 0, -51.8500280071323542, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = -5/2
        -0.198164159898117403, 0, 0, 0, 0, 7.13390975633222651, 0, 0, 0, 0, 1.18898495938870442, 0, -7.13390975633222651, 0, -23.7796991877740884, 0, 0, 0, 0, 0, 0, 1.58531327918493922, 0, -35.6695487816611326, 0, 47.5593983755481767, 0, 12.6825062334795138, 0, 0, 0, 0, 0, 0, 0, 0, 0.594492479694352209, 0, -21.4017292689966795, 0, 71.3390975633222651, 0, -38.0475187004385414, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -2.37796991877740884, 0, 0, 0, 0, -4.75593983755481767, 0, 23.7796991877740884, 0, 0, 0, 0, 0, 0, 0, 0, 23.7796991877740884, 0, -38.0475187004385414, 0, 0, 0, 0, 0, 0, 0, 0, 4.75593983755481767, 0, -23.7796991877740884, 0, 0, 0, 10.8707196286967261, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.37796991877740884, 0, -23.7796991877740884, 0, 38.0475187004385414, 0, -10.8707196286967261, 0, 0,
        // j = 19/2, mj = -3/2
        0, 0, -1.9416043082307367, 0, 0, 0, 0, -3.88320861646147339, 0, 19.416043082307367, 0, 0, 0, 0, 0, 0, 0, 0, 19.416043082307367, 0, -31.0656689316917871, 0, 0, 0, 0, 0, 0, 0, 0, 3.88320861646147339, 0, -19.416043082307367, 0, 0, 0, 8.87590540905479632, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.9416043082307367, 0, -19.416043082307367, 0, 31.0656689316917871, 0, -8.87590540905479632, 0, 0,
        0.242700538528842087, 0, 0, 0.970802154115368348, 0, -9.70802154115368348, 0, 0, 0, 0, 1.45620323117305252, 0, -29.1240646234610504, 0, 38.8320861646147339, 0, 0, 0, 0, 0, 0, 0.970802154115368348, 0, -29.1240646234610504, 0, 77.6641723292294678, 0, -31.0656689316917871, 0, 0, 0, 0, 0, 0, 0, 0, 0.242700538528842087, 0, -9.70802154115368348, 0, 38.8320861646147339, 0, -31.0656689316917871, 0, 4.43795270452739816, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = -1/2
        0.219530897104735508, 0, 0, 0.878123588418942033, 0, -8.78123588418942033, 0, 0, 0, 0, 1.31718538262841305, 0, -26.343707652568261, 0, 35.1249435367576813, 0, 0, 0, 0, 0, 0, 0.878123588418942033, 0, -26.343707652568261, 0, 70.2498870735153626, 0, -28.0999548294061451, 0, 0, 0, 0, 0, 0, 0, 0, 0.219530897104735508, 0, -8.78123588418942033, 0, 35.1249435367576813, 0, -28.0999548294061451, 0, 4.01427926134373501, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 2.19530897104735508, 0, 0, 0, 0, 8.78123588418942033, 0, -23.4166290245051209, 0, 0, 0, 0, 0, 0, 13.1718538262841305, 0, -70.2498870735153626, 0, 42.1499322441092176, 0, 0, 0, 0, 0, 0, 0, 0, 8.78123588418942033, 0, -70.2498870735153626, 0, 84.2998644882184352, 0, -16.05711704537494, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.19530897104735508, 0, -23.4166290245051209, 0, 42.1499322441092176, 0, -16.05711704537494, 0, 0.892062058076385557,
        // j = 19/2, mj = 1/2
        0, 0, 2.19530897104735508, 0, 0, 0, 0, 8.78123588418942033, 0, -23.4166290245051209, 0, 0, 0, 0, 0, 0, 13.1718538262841305, 0, -70.2498870735153626, 0, 42.1499322441092176, 0, 0, 0, 0, 0, 0, 0, 0, 8.78123588418942033, 0, -70.2498870735153626, 0, 84.2998644882184352, 0, -16.05711704537494, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.19530897104735508, 0, -23.4166290245051209, 0, 42.1499322441092176, 0, -16.05711704537494, 0, 0.892062058076385557,
        -0.219530897104735508, 0, 0, -0.878123588418942033, 0, 8.78123588418942033, 0, 0, 0, 0, -1.31718538262841305, 0, 26.343707652568261, 0, -35.1249435367576813, 0, 0, 0, 0, 0, 0, -0.878123588418942033, 0, 26.343707652568261, 0, -70.2498870735153626, 0, 28.0999548294061451, 0, 0, 0, 0, 0, 0, 0, 0, -0.219530897104735508, 0, 8.78123588418942033, 0, -35.1249435367576813, 0, 28.0999548294061451, 0, -4.01427926134373501, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = 3/2
        -0.242700538528842087, 0, 0, -0.970802154115368348, 0, 9.70802154115368348, 0, 0, 0, 0, -1.45620323117305252, 0, 29.1240646234610504, 0, -38.8320861646147339, 0, 0, 0, 0, 0, 0, -0.970802154115368348, 0, 29.1240646234610504, 0, -77.6641723292294678, 0, 31.0656689316917871, 0, 0, 0, 0, 0, 0, 0, 0, -0.242700538528842087, 0, 9.70802154115368348, 0, -38.8320861646147339, 0, 31.0656689316917871, 0, -4.43795270452739816, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -1.9416043082307367, 0, 0, 0, 0, -3.88320861646147339, 0, 19.416043082307367, 0, 0, 0, 0, 0, 0, 0, 0, 19.416043082307367, 0, -31.0656689316917871, 0, 0, 0, 0, 0, 0, 0, 0, 3.88320861646147339, 0, -19.416043082307367, 0, 0, 0, 8.87590540905479632, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.9416043082307367, 0, -19.416043082307367, 0, 31.0656689316917871, 0, -8.87590540905479632, 0, 0,
        // j = 19/2, mj = 5/2
        0, 0, -2.37796991877740884, 0, 0, 0, 0, -4.75593983755481767, 0, 23.7796991877740884, 0, 0, 0, 0, 0, 0, 0, 0, 23.7796991877740884, 0, -38.0475187004385414, 0, 0, 0, 0, 0, 0, 0, 0, 4.75593983755481767, 0, -23.7796991877740884, 0, 0, 0, 10.8707196286967261, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.37796991877740884, 0, -23.7796991877740884, 0, 38.0475187004385414, 0, -10.8707196286967261, 0, 0,
        0.198164159898117403, 0, 0, 0, 0, -7.13390975633222651, 0, 0, 0, 0, -1.18898495938870442, 0, 7.13390975633222651, 0, 23.7796991877740884, 0, 0, 0, 0, 0, 0, -1.58531327918493922, 0, 35.6695487816611326, 0, -47.5593983755481767, 0, -12.6825062334795138, 0, 0, 0, 0, 0, 0, 0, 0, -0.594492479694352209, 0, 21.4017292689966795, 0, -71.3390975633222651, 0, 38.0475187004385414, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = 7/2
        0.270052229203814345, 0, 0, 0, 0, -9.7218802513373164, 0, 0, 0, 0, -1.62031337522288607, 0, 9.7218802513373164, 0, 32.4062675044577213, 0, 0, 0, 0, 0, 0, -2.16041783363051476, 0, 48.609401256686582, 0, -64.8125350089154427, 0, -17.2833426690441181, 0, 0, 0, 0, 0, 0, 0, 0, -0.810156687611443034, 0, 29.1656407540119492, 0, -97.218802513373164, 0, 51.8500280071323542, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 1.62031337522288607, 0, 0, 0, 0, -6.48125350089154427, 0, -12.9625070017830885, 0, 0, 0, 0, 0, 0, -16.2031337522288607, 0, 64.8125350089154427, 0, 12.9625070017830885, 0, 0, 0, 0, 0, 0, 0, 0, -6.48125350089154427, 0, 64.8125350089154427, 0, -77.7750420106985312, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.62031337522288607, 0, -12.9625070017830885, 0, 12.9625070017830885, 0, 0, 0, 0,
        // j = 19/2, mj = 9/2
        0, 0, 2.47506956383608662, 0, 0, 0, 0, -9.9002782553443465, 0, -19.800556510688693, 0, 0, 0, 0, 0, 0, -24.7506956383608662, 0, 99.002782553443465, 0, 19.800556510688693, 0, 0, 0, 0, 0, 0, 0, 0, -9.9002782553443465, 0, 99.002782553443465, 0, -118.803339064132158, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.47506956383608662, 0, -19.800556510688693, 0, 19.800556510688693, 0, 0, 0, 0,
        -0.176790683131149045, 0, 0, 1.41432546504919236, 0, 4.95013912767217325, 0, 0, 0, 0, 2.47506956383608662, 0, -44.5512521490495592, 0, -9.9002782553443465, 0, 0, 0, 0, 0, 0, 0, 0, -24.7506956383608662, 0, 99.002782553443465, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.883953415655745223, 0, 24.7506956383608662, 0, -49.5013912767217325, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = 11/2
        -0.306210445487960195, 0, 0, 2.44968356390368156, 0, 8.57389247366288545, 0, 0, 0, 0, 4.28694623683144273, 0, -77.1650322629659691, 0, -17.1477849473257709, 0, 0, 0, 0, 0, 0, 0, 0, -42.8694623683144273, 0, 171.477849473257709, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.53105222743980097, 0, 42.8694623683144273, 0, -85.7389247366288545, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -1.22484178195184078, 0, 0, 0, 0, 17.1477849473257709, 0, 5.71592831577525697, 0, 0, 0, 0, 0, 0, 0, 0, -85.7389247366288545, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -17.1477849473257709, 0, 85.7389247366288545, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.22484178195184078, 0, -5.71592831577525697, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = 13/2
        0, 0, -2.44968356390368156, 0, 0, 0, 0, 34.2955698946515418, 0, 11.4318566315505139, 0, 0, 0, 0, 0, 0, 0, 0, -171.477849473257709, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -34.2955698946515418, 0, 171.477849473257709, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.44968356390368156, 0, -11.4318566315505139, 0, 0, 0, 0, 0, 0,
        0.153105222743980097, 0, 0, -3.06210445487960195, 0, -2.44968356390368156, 0, 0, 0, 0, 2.14347311841572136, 0, 51.4433548419773127, 0, 0, 0, 0, 0, 0, 0, 0, 4.28694623683144273, 0, -85.7389247366288545, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.07173655920786068, 0, 17.1477849473257709, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = 15/2
        0.364463330087414934, 0, 0, -7.28926660174829869, 0, -5.83141328139863895, 0, 0, 0, 0, 5.10248662122380908, 0, 122.459678909371418, 0, 0, 0, 0, 0, 0, 0, 0, 10.2049732424476182, 0, -204.099464848952363, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.55124331061190454, 0, 40.8198929697904727, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0.728926660174829869, 0, 0, 0, 0, -20.4099464848952363, 0, 0, 0, 0, 0, 0, 0, 0, 51.0248662122380908, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -20.4099464848952363, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.728926660174829869, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = 17/2
        0, 0, 2.18677998052448961, 0, 0, 0, 0, -61.229839454685709, 0, 0, 0, 0, 0, 0, 0, 0, 153.074598636714272, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -61.229839454685709, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.18677998052448961, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.121487776695804978, 0, 0, 4.37355996104897921, 0, 0, 0, 0, 0, 0, -15.3074598636714272, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 10.2049732424476182, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.0933899902622448, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = 19/2
        -0.52955294149244958, 0, 0, 19.0639058937281849, 0, 0, 0, 0, 0, 0, -66.723670628048647, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 44.4824470853657647, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4.76597647343204622, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = -19/2
        -0.52955294149244958, 0, 0, 23.8298823671602311, 0, 0, 0, 0, 0, 0, -111.206117713414412, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 111.206117713414412, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -23.8298823671602311, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.52955294149244958, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0.52955294149244958, 0, 0, 0, 0, -19.0639058937281849, 0, 0, 0, 0, 0, 0, 0, 0, 66.723670628048647, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -44.4824470853657647, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4.76597647343204622, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = -17/2
        0, 0, -2.30826775722029458, 0, 0, 0, 0, 83.0976392599306051, 0, 0, 0, 0, 0, 0, 0, 0, -290.841737409757118, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 193.894491606504745, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -20.7744098149826513, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.121487776695804978, 0, 0, 3.28016997078673441, 0, 2.18677998052448961, 0, 0, 0, 0, -5.10248662122380908, 0, -61.229839454685709, 0, 0, 0, 0, 0, 0, 0, 0, -5.10248662122380908, 0, 153.074598636714272, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.28016997078673441, 0, -61.229839454685709, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.121487776695804978, 0, 2.18677998052448961, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = -15/2
        0.364463330087414934, 0, 0, -9.84050991236020323, 0, -6.56033994157346882, 0, 0, 0, 0, 15.3074598636714272, 0, 183.689518364057127, 0, 0, 0, 0, 0, 0, 0, 0, 15.3074598636714272, 0, -459.223795910142817, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -9.84050991236020323, 0, 183.689518364057127, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.364463330087414934, 0, -6.56033994157346882, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -1.0933899902622448, 0, 0, 0, 0, 21.8677998052448961, 0, 5.83141328139863895, 0, 0, 0, 0, 0, 0, -15.3074598636714272, 0, -122.459678909371418, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -30.6149197273428545, 0, 204.099464848952363, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 7.65372993183571362, 0, -40.8198929697904727, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = -13/2
        0, 0, 2.60278878664766166, 0, 0, 0, 0, -52.0557757329532331, 0, -13.8815401954541955, 0, 0, 0, 0, 0, 0, 36.4390430130672632, 0, 291.512344104538105, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 72.8780860261345264, 0, -485.853906840896842, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -18.2195215065336316, 0, 97.1707813681793685, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.153105222743980097, 0, 0, -1.99036789567174127, 0, -4.89936712780736312, 0, 0, 0, 0, -2.14347311841572136, 0, 68.5911397893030836, 0, 11.4318566315505139, 0, 0, 0, 0, 0, 0, 2.14347311841572136, 0, 0, 0, -171.477849473257709, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.99036789567174127, 0, -68.5911397893030836, 0, 171.477849473257709, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.153105222743980097, 0, 4.89936712780736312, 0, -11.4318566315505139, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = -11/2
        -0.306210445487960195, 0, 0, 3.98073579134348253, 0, 9.79873425561472623, 0, 0, 0, 0, 4.28694623683144273, 0, -137.182279578606167, 0, -22.8637132631010279, 0, 0, 0, 0, 0, 0, -4.28694623683144273, 0, 0, 0, 342.955698946515418, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3.98073579134348253, 0, 137.182279578606167, 0, -342.955698946515418, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.306210445487960195, 0, -9.79873425561472623, 0, 22.8637132631010279, 0, 0, 0, 0, 0, 0,
        0, 0, 1.53105222743980097, 0, 0, 0, 0, -12.2484178195184078, 0, -14.2898207894381424, 0, 0, 0, 0, 0, 0, -21.4347311841572136, 0, 128.608387104943282, 0, 17.1477849473257709, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 71.4491039471907121, 0, -171.477849473257709, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 7.65526113719900487, 0, -71.4491039471907121, 0, 85.7389247366288545, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = -9/2
        0, 0, -2.65186024696723567, 0, 0, 0, 0, 21.2148819757378853, 0, 24.7506956383608662, 0, 0, 0, 0, 0, 0, 37.1260434575412994, 0, -222.756260745247796, 0, -29.7008347660330395, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -123.753478191804331, 0, 297.008347660330395, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -13.2593012348361783, 0, 123.753478191804331, 0, -148.504173830165197, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.176790683131149045, 0, 0, 0.530372049393447134, 0, 7.42520869150825987, 0, 0, 0, 0, 2.47506956383608662, 0, -29.7008347660330395, 0, -29.7008347660330395, 0, 0, 0, 0, 0, 0, 2.47506956383608662, 0, -74.2520869150825987, 0, 148.504173830165197, 0, 19.800556510688693, 0, 0, 0, 0, 0, 0, 0, 0, 0.530372049393447134, 0, -29.7008347660330395, 0, 148.504173830165197, 0, -118.803339064132158, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.176790683131149045, 0, 7.42520869150825987, 0, -29.7008347660330395, 0, 19.800556510688693, 0, 0, 0, 0,
        // j = 19/2, mj = -7/2
        0.270052229203814345, 0, 0, -0.810156687611443034, 0, -11.3421936265602025, 0, 0, 0, 0, -3.78073120885340082, 0, 45.3687745062408099, 0, 45.3687745062408099, 0, 0, 0, 0, 0, 0, -3.78073120885340082, 0, 113.421936265602025, 0, -226.843872531204049, 0, -30.2458496708272066, 0, 0, 0, 0, 0, 0, 0, 0, -0.810156687611443034, 0, 45.3687745062408099, 0, -226.843872531204049, 0, 181.47509802496324, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.270052229203814345, 0, -11.3421936265602025, 0, 45.3687745062408099, 0, -30.2458496708272066, 0, 0, 0, 0,
        0, 0, -1.89036560442670041, 0, 0, 0, 0, 0, 0, 22.6843872531204049, 0, 0, 0, 0, 0, 0, 11.3421936265602025, 0, -22.6843872531204049, 0, -45.3687745062408099, 0, 0, 0, 0, 0, 0, 0, 0, 15.1229248354136033, 0, -113.421936265602025, 0, 90.7375490124816198, 0, 17.2833426690441181, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5.67109681328010124, 0, -68.0531617593612148, 0, 136.10632351872243, 0, -51.8500280071323542, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = -5/2
        0, 0, 2.57613407867552624, 0, 0, 0, 0, 0, 0, -30.9136089441063149, 0, 0, 0, 0, 0, 0, -15.4568044720531574, 0, 30.9136089441063149, 0, 61.8272178882126298, 0, 0, 0, 0, 0, 0, 0, 0, -20.6090726294042099, 0, 154.568044720531574, 0, -123.65443577642526, 0, -23.5532258621762399, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -7.72840223602657872, 0, 92.7408268323189446, 0, -185.481653664637889, 0, 70.6596775865287197, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.198164159898117403, 0, 0, 0.594492479694352209, 0, -9.51187967510963535, 0, 0, 0, 0, 0.396328319796234806, 0, -19.0237593502192707, 0, 47.5593983755481767, 0, 0, 0, 0, 0, 0, -0.396328319796234806, 0, 0, 0, 47.5593983755481767, 0, -50.7300249339180552, 0, 0, 0, 0, 0, 0, 0, 0, -0.594492479694352209, 0, 19.0237593502192707, 0, -47.5593983755481767, 0, 0, 0, 10.8707196286967261, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.198164159898117403, 0, 9.51187967510963535, 0, -47.5593983755481767, 0, 50.7300249339180552, 0, -10.8707196286967261, 0, 0,
        // j = 19/2, mj = -3/2
        -0.242700538528842087, 0, 0, -0.728101615586526261, 0, 11.6496258493844202, 0, 0, 0, 0, -0.485401077057684174, 0, 23.2992516987688403, 0, -58.2481292469221009, 0, 0, 0, 0, 0, 0, 0.485401077057684174, 0, 0, 0, -58.2481292469221009, 0, 62.1313378633835742, 0, 0, 0, 0, 0, 0, 0, 0, 0.728101615586526261, 0, -23.2992516987688403, 0, 58.2481292469221009, 0, 0, 0, -13.3138581135821945, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.242700538528842087, 0, -11.6496258493844202, 0, 58.2481292469221009, 0, -62.1313378633835742, 0, 13.3138581135821945, 0, 0,
        0, 0, 2.18430484675957878, 0, 0, 0, 0, 8.73721938703831513, 0, -29.1240646234610504, 0, 0, 0, 0, 0, 0, 13.1058290805574727, 0, -87.3721938703831513, 0, 69.897755096306521, 0, 0, 0, 0, 0, 0, 0, 0, 8.73721938703831513, 0, -87.3721938703831513, 0, 139.795510192613042, 0, -39.9415743407465834, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.18430484675957878, 0, -29.1240646234610504, 0, 69.897755096306521, 0, -39.9415743407465834, 0, 4.43795270452739816, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = -1/2
        0, 0, -2.41483986815209059, 0, 0, 0, 0, -9.65935947260836236, 0, 32.1978649086945412, 0, 0, 0, 0, 0, 0, -14.4890392089125435, 0, 96.5935947260836236, 0, -77.2748757808668989, 0, 0, 0, 0, 0, 0, 0, 0, -9.65935947260836236, 0, 96.5935947260836236, 0, -154.549751561733798, 0, 44.1570718747810851, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.41483986815209059, 0, 32.1978649086945412, 0, -77.2748757808668989, 0, 44.1570718747810851, 0, -4.90634131942012056, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.219530897104735508, 0, 0, -1.09765448552367754, 0, 10.9765448552367754, 0, 0, 0, 0, -2.19530897104735508, 0, 43.9061794209471016, 0, -58.5415725612628022, 0, 0, 0, 0, 0, 0, -2.19530897104735508, 0, 65.8592691314206525, 0, -175.624717683788407, 0, 70.2498870735153626, 0, 0, 0, 0, 0, 0, 0, 0, -1.09765448552367754, 0, 43.9061794209471016, 0, -175.624717683788407, 0, 140.499774147030725, 0, -20.071396306718675, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.219530897104735508, 0, 10.9765448552367754, 0, -58.5415725612628022, 0, 70.2498870735153626, 0, -20.071396306718675, 0, 0.892062058076385557,
        // j = 19/2, mj = 1/2
        0.219530897104735508, 0, 0, 1.09765448552367754, 0, -10.9765448552367754, 0, 0, 0, 0, 2.19530897104735508, 0, -43.9061794209471016, 0, 58.5415725612628022, 0, 0, 0, 0, 0, 0, 2.19530897104735508, 0, -65.8592691314206525, 0, 175.624717683788407, 0, -70.2498870735153626, 0, 0, 0, 0, 0, 0, 0, 0, 1.09765448552367754, 0, -43.9061794209471016, 0, 175.624717683788407, 0, -140.499774147030725, 0, 20.071396306718675, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.219530897104735508, 0, -10.9765448552367754, 0, 58.5415725612628022, 0, -70.2498870735153626, 0, 20.071396306718675, 0, -0.892062058076385557,
        0, 0, -2.41483986815209059, 0, 0, 0, 0, -9.65935947260836236, 0, 32.1978649086945412, 0, 0, 0, 0, 0, 0, -14.4890392089125435, 0, 96.5935947260836236, 0, -77.2748757808668989, 0, 0, 0, 0, 0, 0, 0, 0, -9.65935947260836236, 0, 96.5935947260836236, 0, -154.549751561733798, 0, 44.1570718747810851, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.41483986815209059, 0, 32.1978649086945412, 0, -77.2748757808668989, 0, 44.1570718747810851, 0, -4.90634131942012056, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = 3/2
        0, 0, 2.18430484675957878, 0, 0, 0, 0, 8.73721938703831513, 0, -29.1240646234610504, 0, 0, 0, 0, 0, 0, 13.1058290805574727, 0, -87.3721938703831513, 0, 69.897755096306521, 0, 0, 0, 0, 0, 0, 0, 0, 8.73721938703831513, 0, -87.3721938703831513, 0, 139.795510192613042, 0, -39.9415743407465834, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.18430484675957878, 0, -29.1240646234610504, 0, 69.897755096306521, 0, -39.9415743407465834, 0, 4.43795270452739816, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.242700538528842087, 0, 0, 0.728101615586526261, 0, -11.6496258493844202, 0, 0, 0, 0, 0.485401077057684174, 0, -23.2992516987688403, 0, 58.2481292469221009, 0, 0, 0, 0, 0, 0, -0.485401077057684174, 0, 0, 0, 58.2481292469221009, 0, -62.1313378633835742, 0, 0, 0, 0, 0, 0, 0, 0, -0.728101615586526261, 0, 23.2992516987688403, 0, -58.2481292469221009, 0, 0, 0, 13.3138581135821945, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.242700538528842087, 0, 11.6496258493844202, 0, -58.2481292469221009, 0, 62.1313378633835742, 0, -13.3138581135821945, 0, 0,
        // j = 19/2, mj = 5/2
        -0.198164159898117403, 0, 0, -0.594492479694352209, 0, 9.51187967510963535, 0, 0, 0, 0, -0.396328319796234806, 0, 19.0237593502192707, 0, -47.5593983755481767, 0, 0, 0, 0, 0, 0, 0.396328319796234806, 0, 0, 0, -47.5593983755481767, 0, 50.7300249339180552, 0, 0, 0, 0, 0, 0, 0, 0, 0.594492479694352209, 0, -19.0237593502192707, 0, 47.5593983755481767, 0, 0, 0, -10.8707196286967261, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.198164159898117403, 0, -9.51187967510963535, 0, 47.5593983755481767, 0, -50.7300249339180552, 0, 10.8707196286967261, 0, 0,
        0, 0, 2.57613407867552624, 0, 0, 0, 0, 0, 0, -30.9136089441063149, 0, 0, 0, 0, 0, 0, -15.4568044720531574, 0, 30.9136089441063149, 0, 61.8272178882126298, 0, 0, 0, 0, 0, 0, 0, 0, -20.6090726294042099, 0, 154.568044720531574, 0, -123.65443577642526, 0, -23.5532258621762399, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -7.72840223602657872, 0, 92.7408268323189446, 0, -185.481653664637889, 0, 70.6596775865287197, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = 7/2
        0, 0, -1.89036560442670041, 0, 0, 0, 0, 0, 0, 22.6843872531204049, 0, 0, 0, 0, 0, 0, 11.3421936265602025, 0, -22.6843872531204049, 0, -45.3687745062408099, 0, 0, 0, 0, 0, 0, 0, 0, 15.1229248354136033, 0, -113.421936265602025, 0, 90.7375490124816198, 0, 17.2833426690441181, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5.67109681328010124, 0, -68.0531617593612148, 0, 136.10632351872243, 0, -51.8500280071323542, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.270052229203814345, 0, 0, 0.810156687611443034, 0, 11.3421936265602025, 0, 0, 0, 0, 3.78073120885340082, 0, -45.3687745062408099, 0, -45.3687745062408099, 0, 0, 0, 0, 0, 0, 3.78073120885340082, 0, -113.421936265602025, 0, 226.843872531204049, 0, 30.2458496708272066, 0, 0, 0, 0, 0, 0, 0, 0, 0.810156687611443034, 0, -45.3687745062408099, 0, 226.843872531204049, 0, -181.47509802496324, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.270052229203814345, 0, 11.3421936265602025, 0, -45.3687745062408099, 0, 30.2458496708272066, 0, 0, 0, 0,
        // j = 19/2, mj = 9/2
        0.176790683131149045, 0, 0, -0.530372049393447134, 0, -7.42520869150825987, 0, 0, 0, 0, -2.47506956383608662, 0, 29.7008347660330395, 0, 29.7008347660330395, 0, 0, 0, 0, 0, 0, -2.47506956383608662, 0, 74.2520869150825987, 0, -148.504173830165197, 0, -19.800556510688693, 0, 0, 0, 0, 0, 0, 0, 0, -0.530372049393447134, 0, 29.7008347660330395, 0, -148.504173830165197, 0, 118.803339064132158, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.176790683131149045, 0, -7.42520869150825987, 0, 29.7008347660330395, 0, -19.800556510688693, 0, 0, 0, 0,
        0, 0, -2.65186024696723567, 0, 0, 0, 0, 21.2148819757378853, 0, 24.7506956383608662, 0, 0, 0, 0, 0, 0, 37.1260434575412994, 0, -222.756260745247796, 0, -29.7008347660330395, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -123.753478191804331, 0, 297.008347660330395, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -13.2593012348361783, 0, 123.753478191804331, 0, -148.504173830165197, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = 11/2
        0, 0, 1.53105222743980097, 0, 0, 0, 0, -12.2484178195184078, 0, -14.2898207894381424, 0, 0, 0, 0, 0, 0, -21.4347311841572136, 0, 128.608387104943282, 0, 17.1477849473257709, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 71.4491039471907121, 0, -171.477849473257709, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 7.65526113719900487, 0, -71.4491039471907121, 0, 85.7389247366288545, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.306210445487960195, 0, 0, -3.98073579134348253, 0, -9.79873425561472623, 0, 0, 0, 0, -4.28694623683144273, 0, 137.182279578606167, 0, 22.8637132631010279, 0, 0, 0, 0, 0, 0, 4.28694623683144273, 0, 0, 0, -342.955698946515418, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.98073579134348253, 0, -137.182279578606167, 0, 342.955698946515418, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.306210445487960195, 0, 9.79873425561472623, 0, -22.8637132631010279, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = 13/2
        -0.153105222743980097, 0, 0, 1.99036789567174127, 0, 4.89936712780736312, 0, 0, 0, 0, 2.14347311841572136, 0, -68.5911397893030836, 0, -11.4318566315505139, 0, 0, 0, 0, 0, 0, -2.14347311841572136, 0, 0, 0, 171.477849473257709, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.99036789567174127, 0, 68.5911397893030836, 0, -171.477849473257709, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.153105222743980097, 0, -4.89936712780736312, 0, 11.4318566315505139, 0, 0, 0, 0, 0, 0,
        0, 0, 2.60278878664766166, 0, 0, 0, 0, -52.0557757329532331, 0, -13.8815401954541955, 0, 0, 0, 0, 0, 0, 36.4390430130672632, 0, 291.512344104538105, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 72.8780860261345264, 0, -485.853906840896842, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -18.2195215065336316, 0, 97.1707813681793685, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = 15/2
        0, 0, -1.0933899902622448, 0, 0, 0, 0, 21.8677998052448961, 0, 5.83141328139863895, 0, 0, 0, 0, 0, 0, -15.3074598636714272, 0, -122.459678909371418, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -30.6149197273428545, 0, 204.099464848952363, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 7.65372993183571362, 0, -40.8198929697904727, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.364463330087414934, 0, 0, 9.84050991236020323, 0, 6.56033994157346882, 0, 0, 0, 0, -15.3074598636714272, 0, -183.689518364057127, 0, 0, 0, 0, 0, 0, 0, 0, -15.3074598636714272, 0, 459.223795910142817, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 9.84050991236020323, 0, -183.689518364057127, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.364463330087414934, 0, 6.56033994157346882, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = 17/2
        0.121487776695804978, 0, 0, -3.28016997078673441, 0, -2.18677998052448961, 0, 0, 0, 0, 5.10248662122380908, 0, 61.229839454685709, 0, 0, 0, 0, 0, 0, 0, 0, 5.10248662122380908, 0, -153.074598636714272, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3.28016997078673441, 0, 61.229839454685709, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.121487776695804978, 0, -2.18677998052448961, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -2.30826775722029458, 0, 0, 0, 0, 83.0976392599306051, 0, 0, 0, 0, 0, 0, 0, 0, -290.841737409757118, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 193.894491606504745, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -20.7744098149826513, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = 19/2
        0, 0, 0.52955294149244958, 0, 0, 0, 0, -19.0639058937281849, 0, 0, 0, 0, 0, 0, 0, 0, 66.723670628048647, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -44.4824470853657647, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4.76597647343204622, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.52955294149244958, 0, 0, -23.8298823671602311, 0, 0, 0, 0, 0, 0, 111.206117713414412, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -111.206117713414412, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 23.8298823671602311, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.52955294149244958, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = -21/2
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.542630291944221461, 0, 0, -24.4183631374899657, 0, 0, 0, 0, 0, 0, 113.952361308286507, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -113.952361308286507, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 24.4183631374899657, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.542630291944221461, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = -19/2
        0.11841163748620862, 0, 0, -5.32852368687938788, 0, 0, 0, 0, 0, 0, 24.8664438721038101, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -24.8664438721038101, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5.32852368687938788, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.11841163748620862, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 2.36823274972417239, 0, 0, 0, 0, -85.2563789900702061, 0, 0, 0, 0, 0, 0, 0, 0, 298.397326465245721, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -198.931550976830481, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 21.3140947475175515, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = -17/2
        0, 0, 0.748900951853188297, 0, 0, 0, 0, -26.9604342667147787, 0, 0, 0, 0, 0, 0, 0, 0, 94.3615199335017254, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -62.9076799556678169, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6.74010856667869467, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.374450475926594148, 0, 0, 10.110162850018042, 0, 6.74010856667869467, 0, 0, 0, 0, -15.7269199889169542, 0, -188.723039867003451, 0, 0, 0, 0, 0, 0, 0, 0, -15.7269199889169542, 0, 471.807599667508627, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 10.110162850018042, 0, -188.723039867003451, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.374450475926594148, 0, 6.74010856667869467, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = -15/2
        -0.148791531444953757, 0, 0, 4.01737134901375144, 0, 2.67824756600916763, 0, 0, 0, 0, -6.2492443206880578, 0, -74.9909318482566936, 0, 0, 0, 0, 0, 0, 0, 0, -6.2492443206880578, 0, 187.477329620641734, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4.01737134901375144, 0, -74.9909318482566936, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.148791531444953757, 0, 2.67824756600916763, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -2.67824756600916763, 0, 0, 0, 0, 53.5649513201833526, 0, 14.2839870187155607, 0, 0, 0, 0, 0, 0, -37.4954659241283468, 0, -299.963727393026774, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -74.9909318482566936, 0, 499.939545655044624, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 18.7477329620641734, 0, -99.9879091310089248, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = -13/2
        0, 0, -1.26253801041429866, 0, 0, 0, 0, 25.2507602082859732, 0, 6.73353605554292619, 0, 0, 0, 0, 0, 0, -17.6755321458001813, 0, -141.40425716640145, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -35.3510642916003625, 0, 235.673761944002417, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 8.83776607290009063, 0, -47.1347523888004833, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.315634502603574665, 0, 0, -4.10324853384647065, 0, -10.1003040833143893, 0, 0, 0, 0, -4.41888303645004531, 0, 141.40425716640145, 0, 23.5673761944002417, 0, 0, 0, 0, 0, 0, 4.41888303645004531, 0, 0, 0, -353.510642916003625, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4.10324853384647065, 0, -141.40425716640145, 0, 353.510642916003625, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.315634502603574665, 0, 10.1003040833143893, 0, -23.5673761944002417, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = -11/2
        0.171176842882893189, 0, 0, -2.22529895747761146, 0, -5.47765897225258205, 0, 0, 0, 0, -2.39647580036050465, 0, 76.6872256115361487, 0, 12.7812042685893581, 0, 0, 0, 0, 0, 0, 2.39647580036050465, 0, 0, 0, -191.718064028840372, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.22529895747761146, 0, -76.6872256115361487, 0, 191.718064028840372, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.171176842882893189, 0, 5.47765897225258205, 0, -12.7812042685893581, 0, 0, 0, 0, 0, 0,
        0, 0, 2.73882948612629102, 0, 0, 0, 0, -21.9106358890103282, 0, -25.5624085371787162, 0, 0, 0, 0, 0, 0, -38.3436128057680743, 0, 230.061676834608446, 0, 30.6748902446144595, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 127.812042685893581, 0, -306.748902446144595, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 13.6941474306314551, 0, -127.812042685893581, 0, 153.374451223072297, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = -9/2
        0, 0, 1.67718368337461811, 0, 0, 0, 0, -13.4174694669969449, 0, -15.6537143781631024, 0, 0, 0, 0, 0, 0, -23.4805715672446536, 0, 140.883429403467921, 0, 18.7844572537957229, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 78.2685718908155119, 0, -187.844572537957229, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 8.38591841687309056, 0, -78.2685718908155119, 0, 93.9222862689786143, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.279530613895769685, 0, 0, 0.838591841687309056, 0, 11.7402857836223268, 0, 0, 0, 0, 3.9134285945407756, 0, -46.9611431344893072, 0, -46.9611431344893072, 0, 0, 0, 0, 0, 0, 3.9134285945407756, 0, -117.402857836223268, 0, 234.805715672446536, 0, 31.3074287563262048, 0, 0, 0, 0, 0, 0, 0, 0, 0.838591841687309056, 0, -46.9611431344893072, 0, 234.805715672446536, 0, -187.844572537957229, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.279530613895769685, 0, 11.7402857836223268, 0, -46.9611431344893072, 0, 31.3074287563262048, 0, 0, 0, 0,
        // j = 21/2, mj = -7/2
        -0.190955762544560929, 0, 0, 0.572867287633682787, 0, 8.02014202687155901, 0, 0, 0, 0, 2.673380675623853, 0, -32.0805681074862361, 0, -32.0805681074862361, 0, 0, 0, 0, 0, 0, 2.673380675623853, 0, -80.2014202687155901, 0, 160.40284053743118, 0, 21.387045404990824, 0, 0, 0, 0, 0, 0, 0, 0, 0.572867287633682787, 0, -32.0805681074862361, 0, 160.40284053743118, 0, -128.322272429944944, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.190955762544560929, 0, 8.02014202687155901, 0, -32.0805681074862361, 0, 21.387045404990824, 0, 0, 0, 0,
        0, 0, -2.673380675623853, 0, 0, 0, 0, 0, 0, 32.0805681074862361, 0, 0, 0, 0, 0, 0, 16.040284053743118, 0, -32.0805681074862361, 0, -64.1611362149724721, 0, 0, 0, 0, 0, 0, 0, 0, 21.387045404990824, 0, -160.40284053743118, 0, 128.322272429944944, 0, 24.4423376057037989, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 8.02014202687155901, 0, -96.2417043224587082, 0, 192.483408644917416, 0, -73.3270128171113967, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = -5/2
        0, 0, -2.02088583643044288, 0, 0, 0, 0, 0, 0, 24.2506300371653145, 0, 0, 0, 0, 0, 0, 12.1253150185826573, 0, -24.2506300371653145, 0, -48.501260074330629, 0, 0, 0, 0, 0, 0, 0, 0, 16.167086691443543, 0, -121.253150185826573, 0, 97.0025201486612581, 0, 18.4766705045069063, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6.06265750929132863, 0, -72.7518901114959435, 0, 145.503780222991887, 0, -55.4300115135207189, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.25261072955380536, 0, 0, 0.757832188661416079, 0, -12.1253150185826573, 0, 0, 0, 0, 0.505221459107610719, 0, -24.2506300371653145, 0, 60.6265750929132863, 0, 0, 0, 0, 0, 0, -0.505221459107610719, 0, 0, 0, 60.6265750929132863, 0, -64.668346765774172, 0, 0, 0, 0, 0, 0, 0, 0, -0.757832188661416079, 0, 24.2506300371653145, 0, -60.6265750929132863, 0, 0, 0, 13.8575028783801797, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.25261072955380536, 0, 12.1253150185826573, 0, -60.6265750929132863, 0, 64.668346765774172, 0, -13.8575028783801797, 0, 0,
        // j = 21/2, mj = -3/2
        0.210184831878141177, 0, 0, 0.630554495634423532, 0, -10.0888719301507765, 0, 0, 0, 0, 0.420369663756282355, 0, -20.177743860301553, 0, 50.4443596507538826, 0, 0, 0, 0, 0, 0, -0.420369663756282355, 0, 0, 0, 50.4443596507538826, 0, -53.8073169608041414, 0, 0, 0, 0, 0, 0, 0, 0, -0.630554495634423532, 0, 20.177743860301553, 0, -50.4443596507538826, 0, 0, 0, 11.5301393487437446, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.210184831878141177, 0, 10.0888719301507765, 0, -50.4443596507538826, 0, 53.8073169608041414, 0, -11.5301393487437446, 0, 0,
        0, 0, 2.52221798253769413, 0, 0, 0, 0, 10.0888719301507765, 0, -33.6295731005025884, 0, 0, 0, 0, 0, 0, 15.1333078952261648, 0, -100.888719301507765, 0, 80.7109754412062121, 0, 0, 0, 0, 0, 0, 0, 0, 10.0888719301507765, 0, -100.888719301507765, 0, 161.421950882412424, 0, -46.1205573949749784, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.52221798253769413, 0, -33.6295731005025884, 0, 80.7109754412062121, 0, -46.1205573949749784, 0, 5.12450637721944204, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = -1/2
        0, 0, 2.30245947330177705, 0, 0, 0, 0, 9.20983789320710822, 0, -30.6994596440236941, 0, 0, 0, 0, 0, 0, 13.8147568398106623, 0, -92.0983789320710822, 0, 73.6787031456568658, 0, 0, 0, 0, 0, 0, 0, 0, 9.20983789320710822, 0, -92.0983789320710822, 0, 147.357406291313732, 0, -42.1021160832324947, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.30245947330177705, 0, -30.6994596440236941, 0, 73.6787031456568658, 0, -42.1021160832324947, 0, 4.67801289813694386, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.230245947330177705, 0, 0, -1.15122973665088853, 0, 11.5122973665088853, 0, 0, 0, 0, -2.30245947330177705, 0, 46.0491894660355411, 0, -61.3989192880473881, 0, 0, 0, 0, 0, 0, -2.30245947330177705, 0, 69.0737841990533116, 0, -184.196757864142164, 0, 73.6787031456568658, 0, 0, 0, 0, 0, 0, 0, 0, -1.15122973665088853, 0, 46.0491894660355411, 0, -184.196757864142164, 0, 147.357406291313732, 0, -21.0510580416162474, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.230245947330177705, 0, 11.5122973665088853, 0, -61.3989192880473881, 0, 73.6787031456568658, 0, -21.0510580416162474, 0, 0.935602579627388772,
        // j = 21/2, mj = 1/2
        -0.230245947330177705, 0, 0, -1.15122973665088853, 0, 11.5122973665088853, 0, 0, 0, 0, -2.30245947330177705, 0, 46.0491894660355411, 0, -61.3989192880473881, 0, 0, 0, 0, 0, 0, -2.30245947330177705, 0, 69.0737841990533116, 0, -184.196757864142164, 0, 73.6787031456568658, 0, 0, 0, 0, 0, 0, 0, 0, -1.15122973665088853, 0, 46.0491894660355411, 0, -184.196757864142164, 0, 147.357406291313732, 0, -21.0510580416162474, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.230245947330177705, 0, 11.5122973665088853, 0, -61.3989192880473881, 0, 73.6787031456568658, 0, -21.0510580416162474, 0, 0.935602579627388772,
        0, 0, -2.30245947330177705, 0, 0, 0, 0, -9.20983789320710822, 0, 30.6994596440236941, 0, 0, 0, 0, 0, 0, -13.8147568398106623, 0, 92.0983789320710822, 0, -73.6787031456568658, 0, 0, 0, 0, 0, 0, 0, 0, -9.20983789320710822, 0, 92.0983789320710822, 0, -147.357406291313732, 0, 42.1021160832324947, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.30245947330177705, 0, 30.6994596440236941, 0, -73.6787031456568658, 0, 42.1021160832324947, 0, -4.67801289813694386, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = 3/2
        0, 0, -2.52221798253769413, 0, 0, 0, 0, -10.0888719301507765, 0, 33.6295731005025884, 0, 0, 0, 0, 0, 0, -15.1333078952261648, 0, 100.888719301507765, 0, -80.7109754412062121, 0, 0, 0, 0, 0, 0, 0, 0, -10.0888719301507765, 0, 100.888719301507765, 0, -161.421950882412424, 0, 46.1205573949749784, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.52221798253769413, 0, 33.6295731005025884, 0, -80.7109754412062121, 0, 46.1205573949749784, 0, -5.12450637721944204, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.210184831878141177, 0, 0, 0.630554495634423532, 0, -10.0888719301507765, 0, 0, 0, 0, 0.420369663756282355, 0, -20.177743860301553, 0, 50.4443596507538826, 0, 0, 0, 0, 0, 0, -0.420369663756282355, 0, 0, 0, 50.4443596507538826, 0, -53.8073169608041414, 0, 0, 0, 0, 0, 0, 0, 0, -0.630554495634423532, 0, 20.177743860301553, 0, -50.4443596507538826, 0, 0, 0, 11.5301393487437446, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.210184831878141177, 0, 10.0888719301507765, 0, -50.4443596507538826, 0, 53.8073169608041414, 0, -11.5301393487437446, 0, 0,
        // j = 21/2, mj = 5/2
        0.25261072955380536, 0, 0, 0.757832188661416079, 0, -12.1253150185826573, 0, 0, 0, 0, 0.505221459107610719, 0, -24.2506300371653145, 0, 60.6265750929132863, 0, 0, 0, 0, 0, 0, -0.505221459107610719, 0, 0, 0, 60.6265750929132863, 0, -64.668346765774172, 0, 0, 0, 0, 0, 0, 0, 0, -0.757832188661416079, 0, 24.2506300371653145, 0, -60.6265750929132863, 0, 0, 0, 13.8575028783801797, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.25261072955380536, 0, 12.1253150185826573, 0, -60.6265750929132863, 0, 64.668346765774172, 0, -13.8575028783801797, 0, 0,
        0, 0, 2.02088583643044288, 0, 0, 0, 0, 0, 0, -24.2506300371653145, 0, 0, 0, 0, 0, 0, -12.1253150185826573, 0, 24.2506300371653145, 0, 48.501260074330629, 0, 0, 0, 0, 0, 0, 0, 0, -16.167086691443543, 0, 121.253150185826573, 0, -97.0025201486612581, 0, -18.4766705045069063, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -6.06265750929132863, 0, 72.7518901114959435, 0, -145.503780222991887, 0, 55.4300115135207189, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = 7/2
        0, 0, 2.673380675623853, 0, 0, 0, 0, 0, 0, -32.0805681074862361, 0, 0, 0, 0, 0, 0, -16.040284053743118, 0, 32.0805681074862361, 0, 64.1611362149724721, 0, 0, 0, 0, 0, 0, 0, 0, -21.387045404990824, 0, 160.40284053743118, 0, -128.322272429944944, 0, -24.4423376057037989, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -8.02014202687155901, 0, 96.2417043224587082, 0, -192.483408644917416, 0, 73.3270128171113967, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.190955762544560929, 0, 0, 0.572867287633682787, 0, 8.02014202687155901, 0, 0, 0, 0, 2.673380675623853, 0, -32.0805681074862361, 0, -32.0805681074862361, 0, 0, 0, 0, 0, 0, 2.673380675623853, 0, -80.2014202687155901, 0, 160.40284053743118, 0, 21.387045404990824, 0, 0, 0, 0, 0, 0, 0, 0, 0.572867287633682787, 0, -32.0805681074862361, 0, 160.40284053743118, 0, -128.322272429944944, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.190955762544560929, 0, 8.02014202687155901, 0, -32.0805681074862361, 0, 21.387045404990824, 0, 0, 0, 0,
        // j = 21/2, mj = 9/2
        -0.279530613895769685, 0, 0, 0.838591841687309056, 0, 11.7402857836223268, 0, 0, 0, 0, 3.9134285945407756, 0, -46.9611431344893072, 0, -46.9611431344893072, 0, 0, 0, 0, 0, 0, 3.9134285945407756, 0, -117.402857836223268, 0, 234.805715672446536, 0, 31.3074287563262048, 0, 0, 0, 0, 0, 0, 0, 0, 0.838591841687309056, 0, -46.9611431344893072, 0, 234.805715672446536, 0, -187.844572537957229, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.279530613895769685, 0, 11.7402857836223268, 0, -46.9611431344893072, 0, 31.3074287563262048, 0, 0, 0, 0,
        0, 0, -1.67718368337461811, 0, 0, 0, 0, 13.4174694669969449, 0, 15.6537143781631024, 0, 0, 0, 0, 0, 0, 23.4805715672446536, 0, -140.883429403467921, 0, -18.7844572537957229, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -78.2685718908155119, 0, 187.844572537957229, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -8.38591841687309056, 0, 78.2685718908155119, 0, -93.9222862689786143, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = 11/2
        0, 0, -2.73882948612629102, 0, 0, 0, 0, 21.9106358890103282, 0, 25.5624085371787162, 0, 0, 0, 0, 0, 0, 38.3436128057680743, 0, -230.061676834608446, 0, -30.6748902446144595, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -127.812042685893581, 0, 306.748902446144595, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -13.6941474306314551, 0, 127.812042685893581, 0, -153.374451223072297, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.171176842882893189, 0, 0, -2.22529895747761146, 0, -5.47765897225258205, 0, 0, 0, 0, -2.39647580036050465, 0, 76.6872256115361487, 0, 12.7812042685893581, 0, 0, 0, 0, 0, 0, 2.39647580036050465, 0, 0, 0, -191.718064028840372, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.22529895747761146, 0, -76.6872256115361487, 0, 191.718064028840372, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.171176842882893189, 0, 5.47765897225258205, 0, -12.7812042685893581, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = 13/2
        0.315634502603574665, 0, 0, -4.10324853384647065, 0, -10.1003040833143893, 0, 0, 0, 0, -4.41888303645004531, 0, 141.40425716640145, 0, 23.5673761944002417, 0, 0, 0, 0, 0, 0, 4.41888303645004531, 0, 0, 0, -353.510642916003625, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4.10324853384647065, 0, -141.40425716640145, 0, 353.510642916003625, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.315634502603574665, 0, 10.1003040833143893, 0, -23.5673761944002417, 0, 0, 0, 0, 0, 0,
        0, 0, 1.26253801041429866, 0, 0, 0, 0, -25.2507602082859732, 0, -6.73353605554292619, 0, 0, 0, 0, 0, 0, 17.6755321458001813, 0, 141.40425716640145, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 35.3510642916003625, 0, -235.673761944002417, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -8.83776607290009063, 0, 47.1347523888004833, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = 15/2
        0, 0, 2.67824756600916763, 0, 0, 0, 0, -53.5649513201833526, 0, -14.2839870187155607, 0, 0, 0, 0, 0, 0, 37.4954659241283468, 0, 299.963727393026774, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 74.9909318482566936, 0, -499.939545655044624, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -18.7477329620641734, 0, 99.9879091310089248, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.148791531444953757, 0, 0, 4.01737134901375144, 0, 2.67824756600916763, 0, 0, 0, 0, -6.2492443206880578, 0, -74.9909318482566936, 0, 0, 0, 0, 0, 0, 0, 0, -6.2492443206880578, 0, 187.477329620641734, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4.01737134901375144, 0, -74.9909318482566936, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.148791531444953757, 0, 2.67824756600916763, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = 17/2
        -0.374450475926594148, 0, 0, 10.110162850018042, 0, 6.74010856667869467, 0, 0, 0, 0, -15.7269199889169542, 0, -188.723039867003451, 0, 0, 0, 0, 0, 0, 0, 0, -15.7269199889169542, 0, 471.807599667508627, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 10.110162850018042, 0, -188.723039867003451, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.374450475926594148, 0, 6.74010856667869467, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -0.748900951853188297, 0, 0, 0, 0, 26.9604342667147787, 0, 0, 0, 0, 0, 0, 0, 0, -94.3615199335017254, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 62.9076799556678169, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -6.74010856667869467, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = 19/2
        0, 0, -2.36823274972417239, 0, 0, 0, 0, 85.2563789900702061, 0, 0, 0, 0, 0, 0, 0, 0, -298.397326465245721, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 198.931550976830481, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -21.3140947475175515, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.11841163748620862, 0, 0, -5.32852368687938788, 0, 0, 0, 0, 0, 0, 24.8664438721038101, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -24.8664438721038101, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5.32852368687938788, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.11841163748620862, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = 21/2
        0.542630291944221461, 0, 0, -24.4183631374899657, 0, 0, 0, 0, 0, 0, 113.952361308286507, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -113.952361308286507, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 24.4183631374899657, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.542630291944221461, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = -21/2
        -0.542630291944221461, 0, 0, 29.8446660569321803, 0, 0, 0, 0, 0, 0, -179.067996341593082, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 250.695194878230315, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -89.533998170796541, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5.96893321138643607, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0.542630291944221461, 0, 0, 0, 0, -24.4183631374899657, 0, 0, 0, 0, 0, 0, 0, 0, 113.952361308286507, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -113.952361308286507, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 24.4183631374899657, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.542630291944221461, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = -19/2
        0, 0, -2.48664438721038101, 0, 0, 0, 0, 111.898997424467146, 0, 0, 0, 0, 0, 0, 0, 0, -522.195321314180012, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 522.195321314180012, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -111.898997424467146, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.48664438721038101, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.11841163748620862, 0, 0, 4.14440731201730169, 0, 2.36823274972417239, 0, 0, 0, 0, -10.6570473737587758, 0, -85.2563789900702061, 0, 0, 0, 0, 0, 0, 0, 0, -4.97328877442076202, 0, 298.397326465245721, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 8.88087281146564647, 0, -198.931550976830481, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.06570473737587758, 0, 21.3140947475175515, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = -17/2
        0.374450475926594148, 0, 0, -13.1057666574307952, 0, -7.48900951853188297, 0, 0, 0, 0, 33.7005428333934733, 0, 269.604342667147787, 0, 0, 0, 0, 0, 0, 0, 0, 15.7269199889169542, 0, -943.615199335017254, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -28.0837856944945611, 0, 629.076799556678169, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.37005428333934733, 0, -67.4010856667869467, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -1.12335142777978244, 0, 0, 0, 0, 30.330488550054126, 0, 6.74010856667869467, 0, 0, 0, 0, 0, 0, -47.1807599667508627, 0, -188.723039867003451, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -47.1807599667508627, 0, 471.807599667508627, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 30.330488550054126, 0, -188.723039867003451, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.12335142777978244, 0, 6.74010856667869467, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = -15/2
        0, 0, 2.82703909745412139, 0, 0, 0, 0, -76.3300556312612774, 0, -16.9622345847247283, 0, 0, 0, 0, 0, 0, 118.735642093073098, 0, 474.942568372292393, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 118.735642093073098, 0, -1187.35642093073098, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -76.3300556312612774, 0, 474.942568372292393, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.82703909745412139, 0, -16.9622345847247283, 0, 0, 0, 0, 0, 0, 0, 0,
        0.148791531444953757, 0, 0, -2.82703909745412139, 0, -5.35649513201833526, 0, 0, 0, 0, -0.892749188669722543, 0, 107.129902640366705, 0, 14.2839870187155607, 0, 0, 0, 0, 0, 0, 6.2492443206880578, 0, -74.9909318482566936, 0, -299.963727393026774, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.1246221603440289, 0, -149.981863696513387, 0, 499.939545655044624, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.0415407201146763, 0, 37.4954659241283468, 0, -99.9879091310089248, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = -13/2
        -0.315634502603574665, 0, 0, 5.99705554946791864, 0, 11.3628420937286879, 0, 0, 0, 0, 1.89380701562144799, 0, -227.256841874573759, 0, -30.3009122499431679, 0, 0, 0, 0, 0, 0, -13.2566491093501359, 0, 159.079789312201631, 0, 636.319157248806525, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -6.62832455467506797, 0, 318.159578624403263, 0, -1060.53192874801088, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.20944151822502266, 0, -79.5398946561008156, 0, 212.106385749602175, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 1.57817251301787333, 0, 0, 0, 0, -20.5162426692323532, 0, -16.8338401388573155, 0, 0, 0, 0, 0, 0, -22.0944151822502266, 0, 235.673761944002417, 0, 23.5673761944002417, 0, 0, 0, 0, 0, 0, 0, 0, 22.0944151822502266, 0, 0, 0, -353.510642916003625, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 20.5162426692323532, 0, -235.673761944002417, 0, 353.510642916003625, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.57817251301787333, 0, 16.8338401388573155, 0, -23.5673761944002417, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = -11/2
        0, 0, -2.91000632900918421, 0, 0, 0, 0, 37.8300822771193948, 0, 31.0400675094312983, 0, 0, 0, 0, 0, 0, 40.740088606128579, 0, -434.560945132038176, 0, -43.4560945132038176, 0, 0, 0, 0, 0, 0, 0, 0, -40.740088606128579, 0, 0, 0, 651.841417698057264, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -37.8300822771193948, 0, 434.560945132038176, 0, -651.841417698057264, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.91000632900918421, 0, -31.0400675094312983, 0, 43.4560945132038176, 0, 0, 0, 0, 0, 0,
        -0.171176842882893189, 0, 0, 1.19823790018025232, 0, 8.21648845837887307, 0, 0, 0, 0, 3.76589054342365016, 0, -65.7319076670309846, 0, -38.3436128057680743, 0, 0, 0, 0, 0, 0, 2.39647580036050465, 0, -115.030838417304223, 0, 345.092515251912669, 0, 30.6748902446144595, 0, 0, 0, 0, 0, 0, 0, 0, -0.855884214414465945, 0, 0, 0, 191.718064028840372, 0, -306.748902446144595, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.855884214414465945, 0, 41.0824422918943654, 0, -191.718064028840372, 0, 153.374451223072297, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = -9/2
        0.279530613895769685, 0, 0, -1.9567142972703878, 0, -13.4174694669969449, 0, 0, 0, 0, -6.14967350570693308, 0, 107.339755735975559, 0, 62.6148575126524095, 0, 0, 0, 0, 0, 0, -3.9134285945407756, 0, 187.844572537957229, 0, -563.533717613871686, 0, -50.0918860101219276, 0, 0, 0, 0, 0, 0, 0, 0, 1.39765306947884843, 0, 0, 0, -313.074287563262048, 0, 500.918860101219276, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.39765306947884843, 0, -67.0873473349847245, 0, 313.074287563262048, 0, -250.459430050609638, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -1.9567142972703878, 0, 0, 0, 0, 5.87014289181116339, 0, 27.3940001617854292, 0, 0, 0, 0, 0, 0, 27.3940001617854292, 0, -109.576000647141717, 0, -65.74560038828503, 0, 0, 0, 0, 0, 0, 0, 0, 27.3940001617854292, 0, -273.940001617854292, 0, 328.72800194142515, 0, 31.3074287563262048, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5.87014289181116339, 0, -109.576000647141717, 0, 328.72800194142515, 0, -187.844572537957229, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.9567142972703878, 0, 27.3940001617854292, 0, -65.74560038828503, 0, 31.3074287563262048, 0, 0, 0, 0,
        // j = 21/2, mj = -7/2
        0, 0, 2.86433643816841393, 0, 0, 0, 0, -8.5930093145052418, 0, -40.1007101343577951, 0, 0, 0, 0, 0, 0, -40.1007101343577951, 0, 160.40284053743118, 0, 96.2417043224587082, 0, 0, 0, 0, 0, 0, 0, 0, -40.1007101343577951, 0, 401.007101343577951, 0, -481.208521612293541, 0, -45.8293830106946229, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -8.5930093145052418, 0, 160.40284053743118, 0, -481.208521612293541, 0, 274.976298064167738, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.86433643816841393, 0, -40.1007101343577951, 0, 96.2417043224587082, 0, -45.8293830106946229, 0, 0, 0, 0,
        0.190955762544560929, 0, 0, 0.190955762544560929, 0, -10.693522702495412, 0, 0, 0, 0, -1.14573457526736557, 0, 0, 0, 64.1611362149724721, 0, 0, 0, 0, 0, 0, -2.673380675623853, 0, 64.1611362149724721, 0, -64.1611362149724721, 0, -85.5481816199632961, 0, 0, 0, 0, 0, 0, 0, 0, -2.10051338799017022, 0, 85.5481816199632961, 0, -320.805681074862361, 0, 171.096363239926592, 0, 24.4423376057037989, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.572867287633682787, 0, 32.0805681074862361, 0, -192.483408644917416, 0, 256.644544859889888, 0, -73.3270128171113967, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = -5/2
        -0.25261072955380536, 0, 0, -0.25261072955380536, 0, 14.1462008550131001, 0, 0, 0, 0, 1.51566437732283216, 0, 0, 0, -84.8772051300786008, 0, 0, 0, 0, 0, 0, 3.53655021375327503, 0, -84.8772051300786008, 0, 84.8772051300786008, 0, 113.169606840104801, 0, 0, 0, 0, 0, 0, 0, 0, 2.77871802509185895, 0, -113.169606840104801, 0, 424.386025650393004, 0, -226.339213680209602, 0, -32.334173382887086, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.757832188661416079, 0, -42.4386025650393004, 0, 254.631615390235802, 0, -339.508820520314403, 0, 97.0025201486612581, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 2.27349656598424824, 0, 0, 0, 0, 6.82048969795274471, 0, -36.3759450557479718, 0, 0, 0, 0, 0, 0, 4.54699313196849647, 0, -72.7518901114959435, 0, 109.127835167243915, 0, 0, 0, 0, 0, 0, 0, 0, -4.54699313196849647, 0, 0, 0, 109.127835167243915, 0, -83.1450172702810783, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -6.82048969795274471, 0, 72.7518901114959435, 0, -109.127835167243915, 0, 0, 0, 13.8575028783801797, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.27349656598424824, 0, 36.3759450557479718, 0, -109.127835167243915, 0, 83.1450172702810783, 0, -13.8575028783801797, 0, 0,
        // j = 21/2, mj = -3/2
        0, 0, -2.73240281441583531, 0, 0, 0, 0, -8.19720844324750592, 0, 43.7184450306533649, 0, 0, 0, 0, 0, 0, -5.46480562883167061, 0, 87.4368900613067298, 0, -131.155335091960095, 0, 0, 0, 0, 0, 0, 0, 0, 5.46480562883167061, 0, 0, 0, -131.155335091960095, 0, 99.9278743557791198, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 8.19720844324750592, 0, -87.4368900613067298, 0, 131.155335091960095, 0, 0, 0, -16.6546457259631866, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.73240281441583531, 0, -43.7184450306533649, 0, 131.155335091960095, 0, -99.9278743557791198, 0, 16.6546457259631866, 0, 0,
        -0.210184831878141177, 0, 0, -1.05092415939070589, 0, 12.6110899126884706, 0, 0, 0, 0, -2.10184831878141177, 0, 50.4443596507538826, 0, -84.073932751256471, 0, 0, 0, 0, 0, 0, -2.10184831878141177, 0, 75.6665394761308239, 0, -252.221798253769413, 0, 134.518292402010354, 0, 0, 0, 0, 0, 0, 0, 0, -1.05092415939070589, 0, 50.4443596507538826, 0, -252.221798253769413, 0, 269.036584804020707, 0, -57.650696743718723, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.210184831878141177, 0, 12.6110899126884706, 0, -84.073932751256471, 0, 134.518292402010354, 0, -57.650696743718723, 0, 5.12450637721944204, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = -1/2
        0.230245947330177705, 0, 0, 1.15122973665088853, 0, -13.8147568398106623, 0, 0, 0, 0, 2.30245947330177705, 0, -55.2590273592426493, 0, 92.0983789320710822, 0, 0, 0, 0, 0, 0, 2.30245947330177705, 0, -82.888541038863974, 0, 276.295136796213247, 0, -147.357406291313732, 0, 0, 0, 0, 0, 0, 0, 0, 1.15122973665088853, 0, -55.2590273592426493, 0, 276.295136796213247, 0, -294.714812582627463, 0, 63.1531741248487421, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.230245947330177705, 0, -13.8147568398106623, 0, 92.0983789320710822, 0, -147.357406291313732, 0, 63.1531741248487421, 0, -5.61361547776433263, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -2.53270542063195476, 0, 0, 0, 0, -12.6635271031597738, 0, 42.2117570105325793, 0, 0, 0, 0, 0, 0, -25.3270542063195476, 0, 168.847028042130317, 0, -135.077622433704254, 0, 0, 0, 0, 0, 0, 0, 0, -25.3270542063195476, 0, 253.270542063195476, 0, -405.232867301112762, 0, 115.78081922888936, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -12.6635271031597738, 0, 168.847028042130317, 0, -405.232867301112762, 0, 231.561638457778721, 0, -25.7290709397531912, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.53270542063195476, 0, 42.2117570105325793, 0, -135.077622433704254, 0, 115.78081922888936, 0, -25.7290709397531912, 0, 0.935602579627388772,
        // j = 21/2, mj = 1/2
        0, 0, 2.53270542063195476, 0, 0, 0, 0, 12.6635271031597738, 0, -42.2117570105325793, 0, 0, 0, 0, 0, 0, 25.3270542063195476, 0, -168.847028042130317, 0, 135.077622433704254, 0, 0, 0, 0, 0, 0, 0, 0, 25.3270542063195476, 0, -253.270542063195476, 0, 405.232867301112762, 0, -115.78081922888936, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12.6635271031597738, 0, -168.847028042130317, 0, 405.232867301112762, 0, -231.561638457778721, 0, 25.7290709397531912, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.53270542063195476, 0, -42.2117570105325793, 0, 135.077622433704254, 0, -115.78081922888936, 0, 25.7290709397531912, 0, -0.935602579627388772,
        0.230245947330177705, 0, 0, 1.15122973665088853, 0, -13.8147568398106623, 0, 0, 0, 0, 2.30245947330177705, 0, -55.2590273592426493, 0, 92.0983789320710822, 0, 0, 0, 0, 0, 0, 2.30245947330177705, 0, -82.888541038863974, 0, 276.295136796213247, 0, -147.357406291313732, 0, 0, 0, 0, 0, 0, 0, 0, 1.15122973665088853, 0, -55.2590273592426493, 0, 276.295136796213247, 0, -294.714812582627463, 0, 63.1531741248487421, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.230245947330177705, 0, -13.8147568398106623, 0, 92.0983789320710822, 0, -147.357406291313732, 0, 63.1531741248487421, 0, -5.61361547776433263, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = 3/2
        -0.210184831878141177, 0, 0, -1.05092415939070589, 0, 12.6110899126884706, 0, 0, 0, 0, -2.10184831878141177, 0, 50.4443596507538826, 0, -84.073932751256471, 0, 0, 0, 0, 0, 0, -2.10184831878141177, 0, 75.6665394761308239, 0, -252.221798253769413, 0, 134.518292402010354, 0, 0, 0, 0, 0, 0, 0, 0, -1.05092415939070589, 0, 50.4443596507538826, 0, -252.221798253769413, 0, 269.036584804020707, 0, -57.650696743718723, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.210184831878141177, 0, 12.6110899126884706, 0, -84.073932751256471, 0, 134.518292402010354, 0, -57.650696743718723, 0, 5.12450637721944204, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 2.73240281441583531, 0, 0, 0, 0, 8.19720844324750592, 0, -43.7184450306533649, 0, 0, 0, 0, 0, 0, 5.46480562883167061, 0, -87.4368900613067298, 0, 131.155335091960095, 0, 0, 0, 0, 0, 0, 0, 0, -5.46480562883167061, 0, 0, 0, 131.155335091960095, 0, -99.9278743557791198, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -8.19720844324750592, 0, 87.4368900613067298, 0, -131.155335091960095, 0, 0, 0, 16.6546457259631866, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.73240281441583531, 0, 43.7184450306533649, 0, -131.155335091960095, 0, 99.9278743557791198, 0, -16.6546457259631866, 0, 0,
        // j = 21/2, mj = 5/2
        0, 0, -2.27349656598424824, 0, 0, 0, 0, -6.82048969795274471, 0, 36.3759450557479718, 0, 0, 0, 0, 0, 0, -4.54699313196849647, 0, 72.7518901114959435, 0, -109.127835167243915, 0, 0, 0, 0, 0, 0, 0, 0, 4.54699313196849647, 0, 0, 0, -109.127835167243915, 0, 83.1450172702810783, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6.82048969795274471, 0, -72.7518901114959435, 0, 109.127835167243915, 0, 0, 0, -13.8575028783801797, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.27349656598424824, 0, -36.3759450557479718, 0, 109.127835167243915, 0, -83.1450172702810783, 0, 13.8575028783801797, 0, 0,
        -0.25261072955380536, 0, 0, -0.25261072955380536, 0, 14.1462008550131001, 0, 0, 0, 0, 1.51566437732283216, 0, 0, 0, -84.8772051300786008, 0, 0, 0, 0, 0, 0, 3.53655021375327503, 0, -84.8772051300786008, 0, 84.8772051300786008, 0, 113.169606840104801, 0, 0, 0, 0, 0, 0, 0, 0, 2.77871802509185895, 0, -113.169606840104801, 0, 424.386025650393004, 0, -226.339213680209602, 0, -32.334173382887086, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.757832188661416079, 0, -42.4386025650393004, 0, 254.631615390235802, 0, -339.508820520314403, 0, 97.0025201486612581, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = 7/2
        0.190955762544560929, 0, 0, 0.190955762544560929, 0, -10.693522702495412, 0, 0, 0, 0, -1.14573457526736557, 0, 0, 0, 64.1611362149724721, 0, 0, 0, 0, 0, 0, -2.673380675623853, 0, 64.1611362149724721, 0, -64.1611362149724721, 0, -85.5481816199632961, 0, 0, 0, 0, 0, 0, 0, 0, -2.10051338799017022, 0, 85.5481816199632961, 0, -320.805681074862361, 0, 171.096363239926592, 0, 24.4423376057037989, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.572867287633682787, 0, 32.0805681074862361, 0, -192.483408644917416, 0, 256.644544859889888, 0, -73.3270128171113967, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -2.86433643816841393, 0, 0, 0, 0, 8.5930093145052418, 0, 40.1007101343577951, 0, 0, 0, 0, 0, 0, 40.1007101343577951, 0, -160.40284053743118, 0, -96.2417043224587082, 0, 0, 0, 0, 0, 0, 0, 0, 40.1007101343577951, 0, -401.007101343577951, 0, 481.208521612293541, 0, 45.8293830106946229, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 8.5930093145052418, 0, -160.40284053743118, 0, 481.208521612293541, 0, -274.976298064167738, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.86433643816841393, 0, 40.1007101343577951, 0, -96.2417043224587082, 0, 45.8293830106946229, 0, 0, 0, 0,
        // j = 21/2, mj = 9/2
        0, 0, 1.9567142972703878, 0, 0, 0, 0, -5.87014289181116339, 0, -27.3940001617854292, 0, 0, 0, 0, 0, 0, -27.3940001617854292, 0, 109.576000647141717, 0, 65.74560038828503, 0, 0, 0, 0, 0, 0, 0, 0, -27.3940001617854292, 0, 273.940001617854292, 0, -328.72800194142515, 0, -31.3074287563262048, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -5.87014289181116339, 0, 109.576000647141717, 0, -328.72800194142515, 0, 187.844572537957229, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.9567142972703878, 0, -27.3940001617854292, 0, 65.74560038828503, 0, -31.3074287563262048, 0, 0, 0, 0,
        0.279530613895769685, 0, 0, -1.9567142972703878, 0, -13.4174694669969449, 0, 0, 0, 0, -6.14967350570693308, 0, 107.339755735975559, 0, 62.6148575126524095, 0, 0, 0, 0, 0, 0, -3.9134285945407756, 0, 187.844572537957229, 0, -563.533717613871686, 0, -50.0918860101219276, 0, 0, 0, 0, 0, 0, 0, 0, 1.39765306947884843, 0, 0, 0, -313.074287563262048, 0, 500.918860101219276, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.39765306947884843, 0, -67.0873473349847245, 0, 313.074287563262048, 0, -250.459430050609638, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = 11/2
        -0.171176842882893189, 0, 0, 1.19823790018025232, 0, 8.21648845837887307, 0, 0, 0, 0, 3.76589054342365016, 0, -65.7319076670309846, 0, -38.3436128057680743, 0, 0, 0, 0, 0, 0, 2.39647580036050465, 0, -115.030838417304223, 0, 345.092515251912669, 0, 30.6748902446144595, 0, 0, 0, 0, 0, 0, 0, 0, -0.855884214414465945, 0, 0, 0, 191.718064028840372, 0, -306.748902446144595, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.855884214414465945, 0, 41.0824422918943654, 0, -191.718064028840372, 0, 153.374451223072297, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 2.91000632900918421, 0, 0, 0, 0, -37.8300822771193948, 0, -31.0400675094312983, 0, 0, 0, 0, 0, 0, -40.740088606128579, 0, 434.560945132038176, 0, 43.4560945132038176, 0, 0, 0, 0, 0, 0, 0, 0, 40.740088606128579, 0, 0, 0, -651.841417698057264, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 37.8300822771193948, 0, -434.560945132038176, 0, 651.841417698057264, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.91000632900918421, 0, 31.0400675094312983, 0, -43.4560945132038176, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = 13/2
        0, 0, -1.57817251301787333, 0, 0, 0, 0, 20.5162426692323532, 0, 16.8338401388573155, 0, 0, 0, 0, 0, 0, 22.0944151822502266, 0, -235.673761944002417, 0, -23.5673761944002417, 0, 0, 0, 0, 0, 0, 0, 0, -22.0944151822502266, 0, 0, 0, 353.510642916003625, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -20.5162426692323532, 0, 235.673761944002417, 0, -353.510642916003625, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.57817251301787333, 0, -16.8338401388573155, 0, 23.5673761944002417, 0, 0, 0, 0, 0, 0,
        -0.315634502603574665, 0, 0, 5.99705554946791864, 0, 11.3628420937286879, 0, 0, 0, 0, 1.89380701562144799, 0, -227.256841874573759, 0, -30.3009122499431679, 0, 0, 0, 0, 0, 0, -13.2566491093501359, 0, 159.079789312201631, 0, 636.319157248806525, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -6.62832455467506797, 0, 318.159578624403263, 0, -1060.53192874801088, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.20944151822502266, 0, -79.5398946561008156, 0, 212.106385749602175, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = 15/2
        0.148791531444953757, 0, 0, -2.82703909745412139, 0, -5.35649513201833526, 0, 0, 0, 0, -0.892749188669722543, 0, 107.129902640366705, 0, 14.2839870187155607, 0, 0, 0, 0, 0, 0, 6.2492443206880578, 0, -74.9909318482566936, 0, -299.963727393026774, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.1246221603440289, 0, -149.981863696513387, 0, 499.939545655044624, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.0415407201146763, 0, 37.4954659241283468, 0, -99.9879091310089248, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -2.82703909745412139, 0, 0, 0, 0, 76.3300556312612774, 0, 16.9622345847247283, 0, 0, 0, 0, 0, 0, -118.735642093073098, 0, -474.942568372292393, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -118.735642093073098, 0, 1187.35642093073098, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 76.3300556312612774, 0, -474.942568372292393, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.82703909745412139, 0, 16.9622345847247283, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = 17/2
        0, 0, 1.12335142777978244, 0, 0, 0, 0, -30.330488550054126, 0, -6.74010856667869467, 0, 0, 0, 0, 0, 0, 47.1807599667508627, 0, 188.723039867003451, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 47.1807599667508627, 0, -471.807599667508627, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -30.330488550054126, 0, 188.723039867003451, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.12335142777978244, 0, -6.74010856667869467, 0, 0, 0, 0, 0, 0, 0, 0,
        0.374450475926594148, 0, 0, -13.1057666574307952, 0, -7.48900951853188297, 0, 0, 0, 0, 33.7005428333934733, 0, 269.604342667147787, 0, 0, 0, 0, 0, 0, 0, 0, 15.7269199889169542, 0, -943.615199335017254, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -28.0837856944945611, 0, 629.076799556678169, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.37005428333934733, 0, -67.4010856667869467, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = 19/2
        -0.11841163748620862, 0, 0, 4.14440731201730169, 0, 2.36823274972417239, 0, 0, 0, 0, -10.6570473737587758, 0, -85.2563789900702061, 0, 0, 0, 0, 0, 0, 0, 0, -4.97328877442076202, 0, 298.397326465245721, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 8.88087281146564647, 0, -198.931550976830481, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.06570473737587758, 0, 21.3140947475175515, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 2.48664438721038101, 0, 0, 0, 0, -111.898997424467146, 0, 0, 0, 0, 0, 0, 0, 0, 522.195321314180012, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -522.195321314180012, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 111.898997424467146, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.48664438721038101, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = 21/2
        0, 0, -0.542630291944221461, 0, 0, 0, 0, 24.4183631374899657, 0, 0, 0, 0, 0, 0, 0, 0, -113.952361308286507, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 113.952361308286507, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -24.4183631374899657, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.542630291944221461, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.542630291944221461, 0, 0, 29.8446660569321803, 0, 0, 0, 0, 0, 0, -179.067996341593082, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 250.695194878230315, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -89.533998170796541, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5.96893321138643607, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = -23/2
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.554825753806619302, 0, 0, -30.5154164593640616, 0, 0, 0, 0, 0, 0, 183.09249875618437, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -256.329498258658117, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 91.5462493780921848, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -6.10308329187281232, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = -21/2
        0.115689166958762041, 0, 0, -6.36290418273191227, 0, 0, 0, 0, 0, 0, 38.1774250963914736, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -53.4483951349480631, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 19.0887125481957368, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.27258083654638245, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 2.54516167309276491, 0, 0, 0, 0, -114.532275289174421, 0, 0, 0, 0, 0, 0, 0, 0, 534.483951349480631, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -534.483951349480631, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 114.532275289174421, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.54516167309276491, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = -19/2
        0, 0, 0.767395118221990013, 0, 0, 0, 0, -34.5327803199895506, 0, 0, 0, 0, 0, 0, 0, 0, 161.152974826617903, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -161.152974826617903, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 34.5327803199895506, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.767395118221990013, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.383697559110995006, 0, 0, 13.4294145688848252, 0, 7.67395118221990013, 0, 0, 0, 0, -34.5327803199895506, 0, -276.262242559916405, 0, 0, 0, 0, 0, 0, 0, 0, -16.1152974826617903, 0, 966.917848959707416, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 28.7773169333246255, 0, -644.611899306471611, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3.45327803199895506, 0, 69.0655606399791011, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = -17/2
        -0.145024045724314041, 0, 0, 5.07584160035099142, 0, 2.90048091448628081, 0, 0, 0, 0, -13.0521641151882637, 0, -104.417312921506109, 0, 0, 0, 0, 0, 0, 0, 0, -6.09100992042118971, 0, 365.460595225271382, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 10.876803429323553, 0, -243.640396816847588, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.30521641151882637, 0, 26.1043282303765273, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -2.90048091448628081, 0, 0, 0, 0, 78.3129846911295819, 0, 17.4028854869176849, 0, 0, 0, 0, 0, 0, -121.820198408423794, 0, -487.280793633695176, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -121.820198408423794, 0, 1218.20198408423794, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 78.3129846911295819, 0, -487.280793633695176, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.90048091448628081, 0, 17.4028854869176849, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = -15/2
        0, 0, -1.29713449844641568, 0, 0, 0, 0, 35.0226314580532234, 0, 7.78280699067849409, 0, 0, 0, 0, 0, 0, -54.4796489347494586, 0, -217.918595738997834, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -54.4796489347494586, 0, 544.796489347494586, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 35.0226314580532234, 0, -217.918595738997834, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.29713449844641568, 0, 7.78280699067849409, 0, 0, 0, 0, 0, 0, 0, 0,
        0.32428362461160392, 0, 0, -6.16138886762047449, 0, -11.6742104860177411, 0, 0, 0, 0, -1.94570174766962352, 0, 233.484209720354823, 0, 31.1312279627139763, 0, 0, 0, 0, 0, 0, 13.6199122336873647, 0, -163.438946804248376, 0, -653.755787216993503, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6.80995611684368233, 0, -326.877893608496752, 0, 1089.59297869498917, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.26998537228122744, 0, 81.7194734021241879, 0, -217.918595738997834, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = -13/2
        0.166353989393607054, 0, 0, -3.16072579847853403, 0, -5.98874361816985396, 0, 0, 0, 0, -0.998123936361642327, 0, 119.774872363397079, 0, 15.9699829817862772, 0, 0, 0, 0, 0, 0, 6.98686755453149629, 0, -83.8424106543779555, 0, -335.369642617511822, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.49343377726574814, 0, -167.684821308755911, 0, 558.949404362519703, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.16447792575524938, 0, 41.9212053271889777, 0, -111.789880872503941, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 2.99437180908492698, 0, 0, 0, 0, -38.9268335181040507, 0, -31.9399659635725545, 0, 0, 0, 0, 0, 0, -41.9212053271889777, 0, 447.159523490015762, 0, 44.7159523490015762, 0, 0, 0, 0, 0, 0, 0, 0, 41.9212053271889777, 0, 0, 0, -670.739285235023644, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 38.9268335181040507, 0, -447.159523490015762, 0, 670.739285235023644, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.99437180908492698, 0, 31.9399659635725545, 0, -44.7159523490015762, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = -11/2
        0, 0, 1.72880137002900928, 0, 0, 0, 0, -22.4744178103771206, 0, -18.440547946976099, 0, 0, 0, 0, 0, 0, -24.2032191804061299, 0, 258.167671257665386, 0, 25.8167671257665386, 0, 0, 0, 0, 0, 0, 0, 0, 24.2032191804061299, 0, 0, 0, -387.251506886498079, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 22.4744178103771206, 0, -258.167671257665386, 0, 387.251506886498079, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.72880137002900928, 0, 18.440547946976099, 0, -25.8167671257665386, 0, 0, 0, 0, 0, 0,
        -0.288133561671501547, 0, 0, 2.01693493170051083, 0, 13.8304109602320742, 0, 0, 0, 0, 6.33893835677303403, 0, -110.643287681856594, 0, -64.5419178144163464, 0, 0, 0, 0, 0, 0, 4.03386986340102165, 0, -193.625753443249039, 0, 580.877260329747118, 0, 51.6335342515330771, 0, 0, 0, 0, 0, 0, 0, 0, -1.44066780835750773, 0, 0, 0, 322.709589072081732, 0, -516.335342515330771, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.44066780835750773, 0, 69.1520548011603712, 0, -322.709589072081732, 0, 258.167671257665386, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = -9/2
        -0.184892122049355627, 0, 0, 1.29424485434548939, 0, 8.87482185836907012, 0, 0, 0, 0, 4.0676266850858238, 0, -70.998574866952561, 0, -41.4158353390556606, 0, 0, 0, 0, 0, 0, 2.58848970869097878, 0, -124.247506017166982, 0, 372.742518051500945, 0, 33.1326682712445284, 0, 0, 0, 0, 0, 0, 0, 0, -0.924460610246778137, 0, 0, 0, 207.079176695278303, 0, -331.326682712445284, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.924460610246778137, 0, 44.3741092918453506, 0, -207.079176695278303, 0, 165.663341356222642, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -2.95827395278969004, 0, 0, 0, 0, 8.87482185836907012, 0, 41.4158353390556606, 0, 0, 0, 0, 0, 0, 41.4158353390556606, 0, -165.663341356222642, 0, -99.3980048137335853, 0, 0, 0, 0, 0, 0, 0, 0, 41.4158353390556606, 0, -414.158353390556606, 0, 496.990024068667927, 0, 47.3323832446350406, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 8.87482185836907012, 0, -165.663341356222642, 0, 496.990024068667927, 0, -283.994299467810244, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.95827395278969004, 0, 41.4158353390556606, 0, -99.3980048137335853, 0, 47.3323832446350406, 0, 0, 0, 0,
        // j = 23/2, mj = -7/2
        0, 0, -2.09181557262512238, 0, 0, 0, 0, 6.27544671787536713, 0, 29.2854180167517133, 0, 0, 0, 0, 0, 0, 29.2854180167517133, 0, -117.141672067006853, 0, -70.2850032402041119, 0, 0, 0, 0, 0, 0, 0, 0, 29.2854180167517133, 0, -292.854180167517133, 0, 351.425016201020559, 0, 33.469049162001958, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6.27544671787536713, 0, -117.141672067006853, 0, 351.425016201020559, 0, -200.814294972011748, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.09181557262512238, 0, 29.2854180167517133, 0, -70.2850032402041119, 0, 33.469049162001958, 0, 0, 0, 0,
        0.261476946578140297, 0, 0, 0.261476946578140297, 0, -14.6427090083758566, 0, 0, 0, 0, -1.56886167946884178, 0, 0, 0, 87.8562540502551398, 0, 0, 0, 0, 0, 0, -3.66067725209396416, 0, 87.8562540502551398, 0, -87.8562540502551398, 0, -117.141672067006853, 0, 0, 0, 0, 0, 0, 0, 0, -2.87624641235954327, 0, 117.141672067006853, 0, -439.281270251275699, 0, 234.283344134013706, 0, 33.469049162001958, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.784430839734420891, 0, 43.9281270251275699, 0, -263.568762150765419, 0, 351.425016201020559, 0, -100.407147486005874, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = -5/2
        0.202539171902860758, 0, 0, 0.202539171902860758, 0, -11.3421936265602025, 0, 0, 0, 0, -1.21523503141716455, 0, 0, 0, 68.0531617593612148, 0, 0, 0, 0, 0, 0, -2.83554840664005062, 0, 68.0531617593612148, 0, -68.0531617593612148, 0, -90.7375490124816198, 0, 0, 0, 0, 0, 0, 0, 0, -2.22793089093146834, 0, 90.7375490124816198, 0, -340.265808796806074, 0, 181.47509802496324, 0, 25.9250140035661771, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.607617515708582275, 0, 34.0265808796806074, 0, -204.159485278083644, 0, 272.212647037444859, 0, -77.7750420106985312, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 2.83554840664005062, 0, 0, 0, 0, 8.50664521992015185, 0, -45.3687745062408099, 0, 0, 0, 0, 0, 0, 5.67109681328010124, 0, -90.7375490124816198, 0, 136.10632351872243, 0, 0, 0, 0, 0, 0, 0, 0, -5.67109681328010124, 0, 0, 0, 136.10632351872243, 0, -103.700056014264708, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -8.50664521992015185, 0, 90.7375490124816198, 0, -136.10632351872243, 0, 0, 0, 17.2833426690441181, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.83554840664005062, 0, 45.3687745062408099, 0, -136.10632351872243, 0, 103.700056014264708, 0, -17.2833426690441181, 0, 0,
        // j = 23/2, mj = -3/2
        0, 0, 2.39647580036050465, 0, 0, 0, 0, 7.18942740108151394, 0, -38.3436128057680743, 0, 0, 0, 0, 0, 0, 4.79295160072100929, 0, -76.6872256115361487, 0, 115.030838417304223, 0, 0, 0, 0, 0, 0, 0, 0, -4.79295160072100929, 0, 0, 0, 115.030838417304223, 0, -87.6425435560413128, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -7.18942740108151394, 0, 76.6872256115361487, 0, -115.030838417304223, 0, 0, 0, 14.6070905926735521, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.39647580036050465, 0, 38.3436128057680743, 0, -115.030838417304223, 0, 87.6425435560413128, 0, -14.6070905926735521, 0, 0,
        -0.239647580036050465, 0, 0, -1.19823790018025232, 0, 14.3788548021630279, 0, 0, 0, 0, -2.39647580036050465, 0, 57.5154192086521115, 0, -95.8590320144201859, 0, 0, 0, 0, 0, 0, -2.39647580036050465, 0, 86.2731288129781673, 0, -287.577096043260558, 0, 153.374451223072297, 0, 0, 0, 0, 0, 0, 0, 0, -1.19823790018025232, 0, 57.5154192086521115, 0, -287.577096043260558, 0, 306.748902446144595, 0, -65.7319076670309846, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.239647580036050465, 0, 14.3788548021630279, 0, -95.8590320144201859, 0, 153.374451223072297, 0, -65.7319076670309846, 0, 5.84283623706942085, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = -1/2
        -0.220443711424950199, 0, 0, -1.10221855712475099, 0, 13.2266226854970119, 0, 0, 0, 0, -2.20443711424950199, 0, 52.9064907419880478, 0, -88.1774845699800796, 0, 0, 0, 0, 0, 0, -2.20443711424950199, 0, 79.3597361129820716, 0, -264.532453709940239, 0, 141.083975311968127, 0, 0, 0, 0, 0, 0, 0, 0, -1.10221855712475099, 0, 52.9064907419880478, 0, -264.532453709940239, 0, 282.167950623936255, 0, -60.4645608479863403, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.220443711424950199, 0, 13.2266226854970119, 0, -88.1774845699800796, 0, 141.083975311968127, 0, -60.4645608479863403, 0, 5.37462763093211914, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -2.64532453709940239, 0, 0, 0, 0, -13.2266226854970119, 0, 44.0887422849900398, 0, 0, 0, 0, 0, 0, -26.4532453709940239, 0, 176.354969139960159, 0, -141.083975311968127, 0, 0, 0, 0, 0, 0, 0, 0, -26.4532453709940239, 0, 264.532453709940239, 0, -423.251925935904382, 0, 120.929121695972681, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -13.2266226854970119, 0, 176.354969139960159, 0, -423.251925935904382, 0, 241.858243391945361, 0, -26.8731381546605957, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.64532453709940239, 0, 44.0887422849900398, 0, -141.083975311968127, 0, 120.929121695972681, 0, -26.8731381546605957, 0, 0.977205023805839843,
        // j = 23/2, mj = 1/2
        0, 0, -2.64532453709940239, 0, 0, 0, 0, -13.2266226854970119, 0, 44.0887422849900398, 0, 0, 0, 0, 0, 0, -26.4532453709940239, 0, 176.354969139960159, 0, -141.083975311968127, 0, 0, 0, 0, 0, 0, 0, 0, -26.4532453709940239, 0, 264.532453709940239, 0, -423.251925935904382, 0, 120.929121695972681, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -13.2266226854970119, 0, 176.354969139960159, 0, -423.251925935904382, 0, 241.858243391945361, 0, -26.8731381546605957, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.64532453709940239, 0, 44.0887422849900398, 0, -141.083975311968127, 0, 120.929121695972681, 0, -26.8731381546605957, 0, 0.977205023805839843,
        0.220443711424950199, 0, 0, 1.10221855712475099, 0, -13.2266226854970119, 0, 0, 0, 0, 2.20443711424950199, 0, -52.9064907419880478, 0, 88.1774845699800796, 0, 0, 0, 0, 0, 0, 2.20443711424950199, 0, -79.3597361129820716, 0, 264.532453709940239, 0, -141.083975311968127, 0, 0, 0, 0, 0, 0, 0, 0, 1.10221855712475099, 0, -52.9064907419880478, 0, 264.532453709940239, 0, -282.167950623936255, 0, 60.4645608479863403, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.220443711424950199, 0, -13.2266226854970119, 0, 88.1774845699800796, 0, -141.083975311968127, 0, 60.4645608479863403, 0, -5.37462763093211914, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = 3/2
        0.239647580036050465, 0, 0, 1.19823790018025232, 0, -14.3788548021630279, 0, 0, 0, 0, 2.39647580036050465, 0, -57.5154192086521115, 0, 95.8590320144201859, 0, 0, 0, 0, 0, 0, 2.39647580036050465, 0, -86.2731288129781673, 0, 287.577096043260558, 0, -153.374451223072297, 0, 0, 0, 0, 0, 0, 0, 0, 1.19823790018025232, 0, -57.5154192086521115, 0, 287.577096043260558, 0, -306.748902446144595, 0, 65.7319076670309846, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.239647580036050465, 0, -14.3788548021630279, 0, 95.8590320144201859, 0, -153.374451223072297, 0, 65.7319076670309846, 0, -5.84283623706942085, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 2.39647580036050465, 0, 0, 0, 0, 7.18942740108151394, 0, -38.3436128057680743, 0, 0, 0, 0, 0, 0, 4.79295160072100929, 0, -76.6872256115361487, 0, 115.030838417304223, 0, 0, 0, 0, 0, 0, 0, 0, -4.79295160072100929, 0, 0, 0, 115.030838417304223, 0, -87.6425435560413128, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -7.18942740108151394, 0, 76.6872256115361487, 0, -115.030838417304223, 0, 0, 0, 14.6070905926735521, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.39647580036050465, 0, 38.3436128057680743, 0, -115.030838417304223, 0, 87.6425435560413128, 0, -14.6070905926735521, 0, 0,
        // j = 23/2, mj = 5/2
        0, 0, 2.83554840664005062, 0, 0, 0, 0, 8.50664521992015185, 0, -45.3687745062408099, 0, 0, 0, 0, 0, 0, 5.67109681328010124, 0, -90.7375490124816198, 0, 136.10632351872243, 0, 0, 0, 0, 0, 0, 0, 0, -5.67109681328010124, 0, 0, 0, 136.10632351872243, 0, -103.700056014264708, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -8.50664521992015185, 0, 90.7375490124816198, 0, -136.10632351872243, 0, 0, 0, 17.2833426690441181, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.83554840664005062, 0, 45.3687745062408099, 0, -136.10632351872243, 0, 103.700056014264708, 0, -17.2833426690441181, 0, 0,
        -0.202539171902860758, 0, 0, -0.202539171902860758, 0, 11.3421936265602025, 0, 0, 0, 0, 1.21523503141716455, 0, 0, 0, -68.0531617593612148, 0, 0, 0, 0, 0, 0, 2.83554840664005062, 0, -68.0531617593612148, 0, 68.0531617593612148, 0, 90.7375490124816198, 0, 0, 0, 0, 0, 0, 0, 0, 2.22793089093146834, 0, -90.7375490124816198, 0, 340.265808796806074, 0, -181.47509802496324, 0, -25.9250140035661771, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.607617515708582275, 0, -34.0265808796806074, 0, 204.159485278083644, 0, -272.212647037444859, 0, 77.7750420106985312, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = 7/2
        -0.261476946578140297, 0, 0, -0.261476946578140297, 0, 14.6427090083758566, 0, 0, 0, 0, 1.56886167946884178, 0, 0, 0, -87.8562540502551398, 0, 0, 0, 0, 0, 0, 3.66067725209396416, 0, -87.8562540502551398, 0, 87.8562540502551398, 0, 117.141672067006853, 0, 0, 0, 0, 0, 0, 0, 0, 2.87624641235954327, 0, -117.141672067006853, 0, 439.281270251275699, 0, -234.283344134013706, 0, -33.469049162001958, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.784430839734420891, 0, -43.9281270251275699, 0, 263.568762150765419, 0, -351.425016201020559, 0, 100.407147486005874, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -2.09181557262512238, 0, 0, 0, 0, 6.27544671787536713, 0, 29.2854180167517133, 0, 0, 0, 0, 0, 0, 29.2854180167517133, 0, -117.141672067006853, 0, -70.2850032402041119, 0, 0, 0, 0, 0, 0, 0, 0, 29.2854180167517133, 0, -292.854180167517133, 0, 351.425016201020559, 0, 33.469049162001958, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6.27544671787536713, 0, -117.141672067006853, 0, 351.425016201020559, 0, -200.814294972011748, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.09181557262512238, 0, 29.2854180167517133, 0, -70.2850032402041119, 0, 33.469049162001958, 0, 0, 0, 0,
        // j = 23/2, mj = 9/2
        0, 0, -2.95827395278969004, 0, 0, 0, 0, 8.87482185836907012, 0, 41.4158353390556606, 0, 0, 0, 0, 0, 0, 41.4158353390556606, 0, -165.663341356222642, 0, -99.3980048137335853, 0, 0, 0, 0, 0, 0, 0, 0, 41.4158353390556606, 0, -414.158353390556606, 0, 496.990024068667927, 0, 47.3323832446350406, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 8.87482185836907012, 0, -165.663341356222642, 0, 496.990024068667927, 0, -283.994299467810244, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.95827395278969004, 0, 41.4158353390556606, 0, -99.3980048137335853, 0, 47.3323832446350406, 0, 0, 0, 0,
        0.184892122049355627, 0, 0, -1.29424485434548939, 0, -8.87482185836907012, 0, 0, 0, 0, -4.0676266850858238, 0, 70.998574866952561, 0, 41.4158353390556606, 0, 0, 0, 0, 0, 0, -2.58848970869097878, 0, 124.247506017166982, 0, -372.742518051500945, 0, -33.1326682712445284, 0, 0, 0, 0, 0, 0, 0, 0, 0.924460610246778137, 0, 0, 0, -207.079176695278303, 0, 331.326682712445284, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.924460610246778137, 0, -44.3741092918453506, 0, 207.079176695278303, 0, -165.663341356222642, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = 11/2
        0.288133561671501547, 0, 0, -2.01693493170051083, 0, -13.8304109602320742, 0, 0, 0, 0, -6.33893835677303403, 0, 110.643287681856594, 0, 64.5419178144163464, 0, 0, 0, 0, 0, 0, -4.03386986340102165, 0, 193.625753443249039, 0, -580.877260329747118, 0, -51.6335342515330771, 0, 0, 0, 0, 0, 0, 0, 0, 1.44066780835750773, 0, 0, 0, -322.709589072081732, 0, 516.335342515330771, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.44066780835750773, 0, -69.1520548011603712, 0, 322.709589072081732, 0, -258.167671257665386, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 1.72880137002900928, 0, 0, 0, 0, -22.4744178103771206, 0, -18.440547946976099, 0, 0, 0, 0, 0, 0, -24.2032191804061299, 0, 258.167671257665386, 0, 25.8167671257665386, 0, 0, 0, 0, 0, 0, 0, 0, 24.2032191804061299, 0, 0, 0, -387.251506886498079, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 22.4744178103771206, 0, -258.167671257665386, 0, 387.251506886498079, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.72880137002900928, 0, 18.440547946976099, 0, -25.8167671257665386, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = 13/2
        0, 0, 2.99437180908492698, 0, 0, 0, 0, -38.9268335181040507, 0, -31.9399659635725545, 0, 0, 0, 0, 0, 0, -41.9212053271889777, 0, 447.159523490015762, 0, 44.7159523490015762, 0, 0, 0, 0, 0, 0, 0, 0, 41.9212053271889777, 0, 0, 0, -670.739285235023644, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 38.9268335181040507, 0, -447.159523490015762, 0, 670.739285235023644, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.99437180908492698, 0, 31.9399659635725545, 0, -44.7159523490015762, 0, 0, 0, 0, 0, 0,
        -0.166353989393607054, 0, 0, 3.16072579847853403, 0, 5.98874361816985396, 0, 0, 0, 0, 0.998123936361642327, 0, -119.774872363397079, 0, -15.9699829817862772, 0, 0, 0, 0, 0, 0, -6.98686755453149629, 0, 83.8424106543779555, 0, 335.369642617511822, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3.49343377726574814, 0, 167.684821308755911, 0, -558.949404362519703, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.16447792575524938, 0, -41.9212053271889777, 0, 111.789880872503941, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = 15/2
        -0.32428362461160392, 0, 0, 6.16138886762047449, 0, 11.6742104860177411, 0, 0, 0, 0, 1.94570174766962352, 0, -233.484209720354823, 0, -31.1312279627139763, 0, 0, 0, 0, 0, 0, -13.6199122336873647, 0, 163.438946804248376, 0, 653.755787216993503, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -6.80995611684368233, 0, 326.877893608496752, 0, -1089.59297869498917, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.26998537228122744, 0, -81.7194734021241879, 0, 217.918595738997834, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -1.29713449844641568, 0, 0, 0, 0, 35.0226314580532234, 0, 7.78280699067849409, 0, 0, 0, 0, 0, 0, -54.4796489347494586, 0, -217.918595738997834, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -54.4796489347494586, 0, 544.796489347494586, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 35.0226314580532234, 0, -217.918595738997834, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.29713449844641568, 0, 7.78280699067849409, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = 17/2
        0, 0, -2.90048091448628081, 0, 0, 0, 0, 78.3129846911295819, 0, 17.4028854869176849, 0, 0, 0, 0, 0, 0, -121.820198408423794, 0, -487.280793633695176, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -121.820198408423794, 0, 1218.20198408423794, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 78.3129846911295819, 0, -487.280793633695176, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.90048091448628081, 0, 17.4028854869176849, 0, 0, 0, 0, 0, 0, 0, 0,
        0.145024045724314041, 0, 0, -5.07584160035099142, 0, -2.90048091448628081, 0, 0, 0, 0, 13.0521641151882637, 0, 104.417312921506109, 0, 0, 0, 0, 0, 0, 0, 0, 6.09100992042118971, 0, -365.460595225271382, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -10.876803429323553, 0, 243.640396816847588, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.30521641151882637, 0, -26.1043282303765273, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = 19/2
        0.383697559110995006, 0, 0, -13.4294145688848252, 0, -7.67395118221990013, 0, 0, 0, 0, 34.5327803199895506, 0, 276.262242559916405, 0, 0, 0, 0, 0, 0, 0, 0, 16.1152974826617903, 0, -966.917848959707416, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -28.7773169333246255, 0, 644.611899306471611, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.45327803199895506, 0, -69.0655606399791011, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0.767395118221990013, 0, 0, 0, 0, -34.5327803199895506, 0, 0, 0, 0, 0, 0, 0, 0, 161.152974826617903, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -161.152974826617903, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 34.5327803199895506, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.767395118221990013, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = 21/2
        0, 0, 2.54516167309276491, 0, 0, 0, 0, -114.532275289174421, 0, 0, 0, 0, 0, 0, 0, 0, 534.483951349480631, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -534.483951349480631, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 114.532275289174421, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.54516167309276491, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.115689166958762041, 0, 0, 6.36290418273191227, 0, 0, 0, 0, 0, 0, -38.1774250963914736, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 53.4483951349480631, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -19.0887125481957368, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.27258083654638245, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = 23/2
        -0.554825753806619302, 0, 0, 30.5154164593640616, 0, 0, 0, 0, 0, 0, -183.09249875618437, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 256.329498258658117, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -91.5462493780921848, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6.10308329187281232, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = -23/2
        -0.554825753806619302, 0, 0, 36.6184997512368739, 0, 0, 0, 0, 0, 0, -274.638748134276554, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 512.658996517316235, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -274.638748134276554, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 36.6184997512368739, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.554825753806619302, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0.554825753806619302, 0, 0, 0, 0, -30.5154164593640616, 0, 0, 0, 0, 0, 0, 0, 0, 183.09249875618437, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -256.329498258658117, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 91.5462493780921848, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -6.10308329187281232, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = -21/2
        0, 0, -2.66085084005152695, 0, 0, 0, 0, 146.346796202833982, 0, 0, 0, 0, 0, 0, 0, 0, -878.080777217003893, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1229.31308810380545, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -439.040388608501947, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 29.2693592405667964, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.115689166958762041, 0, 0, 5.09032334618552982, 0, 2.54516167309276491, 0, 0, 0, 0, -19.0887125481957368, 0, -114.532275289174421, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 534.483951349480631, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 19.0887125481957368, 0, -534.483951349480631, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -5.09032334618552982, 0, 114.532275289174421, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.115689166958762041, 0, -2.54516167309276491, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = -19/2
        0.383697559110995006, 0, 0, -16.8826926008837803, 0, -8.44134630044189014, 0, 0, 0, 0, 63.310097253314176, 0, 379.860583519885056, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1772.68272309279693, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -63.310097253314176, 0, 1772.68272309279693, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16.8826926008837803, 0, -379.860583519885056, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.383697559110995006, 0, 8.44134630044189014, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -1.15109267733298502, 0, 0, 0, 0, 40.2882437066544757, 0, 7.67395118221990013, 0, 0, 0, 0, 0, 0, -103.598340959968652, 0, -276.262242559916405, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -48.3458924479853708, 0, 966.917848959707416, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 86.3319507999738764, 0, -644.611899306471611, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -10.3598340959968652, 0, 69.0655606399791011, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = -17/2
        0, 0, 3.04550496021059485, 0, 0, 0, 0, -106.59267360737082, 0, -20.3033664014039657, 0, 0, 0, 0, 0, 0, 274.095446418953537, 0, 730.921190450542765, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 127.911208328844984, 0, -2558.22416657689968, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -228.412872015794614, 0, 1705.48277771793312, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 27.4095446418953537, 0, -182.730297612635691, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.145024045724314041, 0, 0, -3.77062518883216506, 0, -5.80096182897256162, 0, 0, 0, 0, 2.17536068586471061, 0, 156.625969382259164, 0, 17.4028854869176849, 0, 0, 0, 0, 0, 0, 12.1820198408423794, 0, -243.640396816847588, 0, -487.280793633695176, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.17536068586471061, 0, -243.640396816847588, 0, 1218.20198408423794, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3.77062518883216506, 0, 156.625969382259164, 0, -487.280793633695176, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.145024045724314041, 0, -5.80096182897256162, 0, 17.4028854869176849, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = -15/2
        -0.32428362461160392, 0, 0, 8.43137423990170193, 0, 12.9713449844641568, 0, 0, 0, 0, -4.8642543691740588, 0, -350.226314580532234, 0, -38.9140349533924704, 0, 0, 0, 0, 0, 0, -27.2398244673747293, 0, 544.796489347494586, 0, 1089.59297869498917, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4.8642543691740588, 0, 544.796489347494586, 0, -2723.98244673747293, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 8.43137423990170193, 0, -350.226314580532234, 0, 1089.59297869498917, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.32428362461160392, 0, 12.9713449844641568, 0, -38.9140349533924704, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 1.6214181230580196, 0, 0, 0, 0, -30.8069443381023724, 0, -19.4570174766962352, 0, 0, 0, 0, 0, 0, -9.72850873834811761, 0, 389.140349533924704, 0, 31.1312279627139763, 0, 0, 0, 0, 0, 0, 0, 0, 68.0995611684368233, 0, -272.398244673747293, 0, -653.755787216993503, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 34.0497805842184116, 0, -544.796489347494586, 0, 1089.59297869498917, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -11.3499268614061372, 0, 136.199122336873647, 0, -217.918595738997834, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = -13/2
        0, 0, -3.16072579847853403, 0, 0, 0, 0, 60.0537901710921467, 0, 37.9287095817424084, 0, 0, 0, 0, 0, 0, 18.9643547908712042, 0, -758.574191634848168, 0, -60.6859353307878535, 0, 0, 0, 0, 0, 0, 0, 0, -132.750483536098429, 0, 531.001934144393718, 0, 1274.40464194654492, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -66.3752417680492147, 0, 1062.00386828878744, 0, -2124.00773657757487, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 22.1250805893497382, 0, -265.500967072196859, 0, 424.801547315514974, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.166353989393607054, 0, 0, 1.99624787272328465, 0, 8.98311542725478094, 0, 0, 0, 0, 4.49155771362739047, 0, -116.780500554312152, 0, -47.9099489453588317, 0, 0, 0, 0, 0, 0, 0, 0, -125.763615981566933, 0, 670.739285235023644, 0, 44.7159523490015762, 0, 0, 0, 0, 0, 0, 0, 0, -4.49155771362739047, 0, 125.763615981566933, 0, 0, 0, -670.739285235023644, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.99624787272328465, 0, 116.780500554312152, 0, -670.739285235023644, 0, 670.739285235023644, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.166353989393607054, 0, -8.98311542725478094, 0, 47.9099489453588317, 0, -44.7159523490015762, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = -11/2
        0.288133561671501547, 0, 0, -3.45760274005801856, 0, -15.5592123302610835, 0, 0, 0, 0, -7.77960616513054176, 0, 202.269760293394086, 0, 82.9824657613924454, 0, 0, 0, 0, 0, 0, 0, 0, 217.828972623655169, 0, -1161.75452065949424, 0, -77.4503013772996157, 0, 0, 0, 0, 0, 0, 0, 0, 7.77960616513054176, 0, -217.828972623655169, 0, 0, 0, 1161.75452065949424, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.45760274005801856, 0, -202.269760293394086, 0, 1161.75452065949424, 0, -1161.75452065949424, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.288133561671501547, 0, 15.5592123302610835, 0, -82.9824657613924454, 0, 77.4503013772996157, 0, 0, 0, 0, 0, 0,
        0, 0, -2.01693493170051083, 0, 0, 0, 0, 14.1185445219035758, 0, 32.2709589072081732, 0, 0, 0, 0, 0, 0, 44.3725684974112382, 0, -258.167671257665386, 0, -90.358684940182885, 0, 0, 0, 0, 0, 0, 0, 0, 28.2370890438071516, 0, -451.793424700914425, 0, 813.228164461645965, 0, 51.6335342515330771, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -10.0846746585025541, 0, 0, 0, 451.793424700914425, 0, -516.335342515330771, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -10.0846746585025541, 0, 161.354794536040866, 0, -451.793424700914425, 0, 258.167671257665386, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = -9/2
        0, 0, 3.14316607483904567, 0, 0, 0, 0, -22.0021625238733197, 0, -50.2906571974247307, 0, 0, 0, 0, 0, 0, -69.1496536464590047, 0, 402.325257579397845, 0, 140.813840152789246, 0, 0, 0, 0, 0, 0, 0, 0, -44.0043250477466393, 0, 704.06920076394623, 0, -1267.32456137510321, 0, -80.4650515158795691, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 15.7158303741952283, 0, 0, 0, -704.06920076394623, 0, 804.650515158795691, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 15.7158303741952283, 0, -251.453285987123653, 0, 704.06920076394623, 0, -402.325257579397845, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.184892122049355627, 0, 0, -0.369784244098711255, 0, -11.8330958111587602, 0, 0, 0, 0, -3.14316607483904567, 0, 35.4992874334762805, 0, 82.8316706781113211, 0, 0, 0, 0, 0, 0, -5.17697941738195757, 0, 165.663341356222642, 0, -331.326682712445284, 0, -132.530673084978114, 0, 0, 0, 0, 0, 0, 0, 0, -3.14316607483904567, 0, 165.663341356222642, 0, -828.316706781113211, 0, 662.653365424890569, 0, 47.3323832446350406, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.369784244098711255, 0, 35.4992874334762805, 0, -331.326682712445284, 0, 662.653365424890569, 0, -283.994299467810244, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.184892122049355627, 0, -11.8330958111587602, 0, 82.8316706781113211, 0, -132.530673084978114, 0, 47.3323832446350406, 0, 0, 0, 0,
        // j = 23/2, mj = -7/2
        -0.261476946578140297, 0, 0, 0.522953893156280594, 0, 16.734524581000979, 0, 0, 0, 0, 4.44510809182838505, 0, -50.203573743002937, 0, -117.141672067006853, 0, 0, 0, 0, 0, 0, 7.32135450418792832, 0, -234.283344134013706, 0, 468.566688268027412, 0, 187.426675307210965, 0, 0, 0, 0, 0, 0, 0, 0, 4.44510809182838505, 0, -234.283344134013706, 0, 1171.41672067006853, 0, -937.133376536054825, 0, -66.938098324003916, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.522953893156280594, 0, -50.203573743002937, 0, 468.566688268027412, 0, -937.133376536054825, 0, 401.628589944023496, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.261476946578140297, 0, 16.734524581000979, 0, -117.141672067006853, 0, 187.426675307210965, 0, -66.938098324003916, 0, 0, 0, 0,
        0, 0, 2.35329251920326267, 0, 0, 0, 0, 2.35329251920326267, 0, -43.9281270251275699, 0, 0, 0, 0, 0, 0, -14.119755115219576, 0, 0, 0, 158.141257290459252, 0, 0, 0, 0, 0, 0, 0, 0, -32.9460952688456774, 0, 263.568762150765419, 0, -158.141257290459252, 0, -150.610721229008811, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -25.8862177112358894, 0, 351.425016201020559, 0, -790.706286452296258, 0, 301.221442458017622, 0, 33.469049162001958, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -7.05987755760978802, 0, 131.78438107538271, 0, -474.423771871377755, 0, 451.832163687026433, 0, -100.407147486005874, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = -5/2
        0, 0, -3.03808757854291138, 0, 0, 0, 0, -3.03808757854291138, 0, 56.7109681328010124, 0, 0, 0, 0, 0, 0, 18.2285254712574683, 0, 0, 0, -204.159485278083644, 0, 0, 0, 0, 0, 0, 0, 0, 42.5332260996007593, 0, -340.265808796806074, 0, 204.159485278083644, 0, 194.437605026746328, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 33.4189633639720251, 0, -453.687745062408099, 0, 1020.79742639041822, 0, -388.875210053492656, 0, -43.2083566726102951, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 9.11426273562873413, 0, -170.132904398403037, 0, 612.478455834250933, 0, -583.312815080238984, 0, 129.625070017830885, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.202539171902860758, 0, 0, -0.810156687611443034, 0, 14.1777420332002531, 0, 0, 0, 0, -1.01269585951430379, 0, 42.5332260996007593, 0, -113.421936265602025, 0, 0, 0, 0, 0, 0, 0, 0, 28.3554840664005062, 0, -226.843872531204049, 0, 226.843872531204049, 0, 0, 0, 0, 0, 0, 0, 0, 1.01269585951430379, 0, -28.3554840664005062, 0, 0, 0, 226.843872531204049, 0, -129.625070017830885, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.810156687611443034, 0, -42.5332260996007593, 0, 226.843872531204049, 0, -226.843872531204049, 0, 0, 0, 17.2833426690441181, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.202539171902860758, 0, -14.1777420332002531, 0, 113.421936265602025, 0, -226.843872531204049, 0, 129.625070017830885, 0, -17.2833426690441181, 0, 0,
        // j = 23/2, mj = -3/2
        0.239647580036050465, 0, 0, 0.958590320144201859, 0, -16.7753306025235325, 0, 0, 0, 0, 1.19823790018025232, 0, -50.3259918075705976, 0, 134.20264482018826, 0, 0, 0, 0, 0, 0, 0, 0, -33.5506612050470651, 0, 268.40528964037652, 0, -268.40528964037652, 0, 0, 0, 0, 0, 0, 0, 0, -1.19823790018025232, 0, 33.5506612050470651, 0, 0, 0, -268.40528964037652, 0, 153.374451223072297, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.958590320144201859, 0, 50.3259918075705976, 0, -268.40528964037652, 0, 268.40528964037652, 0, 0, 0, -20.449926829742973, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.239647580036050465, 0, 16.7753306025235325, 0, -134.20264482018826, 0, 268.40528964037652, 0, -153.374451223072297, 0, 20.449926829742973, 0, 0,
        0, 0, -2.63612338039655511, 0, 0, 0, 0, -13.1806169019827756, 0, 52.7224676079311022, 0, 0, 0, 0, 0, 0, -26.3612338039655511, 0, 210.889870431724409, 0, -210.889870431724409, 0, 0, 0, 0, 0, 0, 0, 0, -26.3612338039655511, 0, 316.334805647586613, 0, -632.669611295173227, 0, 241.01699477911361, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -13.1806169019827756, 0, 210.889870431724409, 0, -632.669611295173227, 0, 482.03398955822722, 0, -80.3389982597045367, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.63612338039655511, 0, 52.7224676079311022, 0, -210.889870431724409, 0, 241.01699477911361, 0, -80.3389982597045367, 0, 5.84283623706942085, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = -1/2
        0, 0, 2.86576824852435259, 0, 0, 0, 0, 14.3288412426217629, 0, -57.3153649704870517, 0, 0, 0, 0, 0, 0, 28.6576824852435259, 0, -229.261459881948207, 0, 229.261459881948207, 0, 0, 0, 0, 0, 0, 0, 0, 28.6576824852435259, 0, -343.89218982292231, 0, 687.784379645844621, 0, -262.013097007940808, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 14.3288412426217629, 0, -229.261459881948207, 0, 687.784379645844621, 0, -524.026194015881616, 0, 87.337699002646936, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.86576824852435259, 0, -57.3153649704870517, 0, 229.261459881948207, 0, -262.013097007940808, 0, 87.337699002646936, 0, -6.35183265473795898, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.220443711424950199, 0, 0, 1.32266226854970119, 0, -15.8719472225964143, 0, 0, 0, 0, 3.30665567137425298, 0, -79.3597361129820716, 0, 132.266226854970119, 0, 0, 0, 0, 0, 0, 4.40887422849900398, 0, -158.719472225964143, 0, 529.064907419880478, 0, -282.167950623936255, 0, 0, 0, 0, 0, 0, 0, 0, 3.30665567137425298, 0, -158.719472225964143, 0, 793.597361129820716, 0, -846.503851871808764, 0, 181.393682543959021, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.32266226854970119, 0, -79.3597361129820716, 0, 529.064907419880478, 0, -846.503851871808764, 0, 362.787365087918042, 0, -32.2477657855927148, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.220443711424950199, 0, -15.8719472225964143, 0, 132.266226854970119, 0, -282.167950623936255, 0, 181.393682543959021, 0, -32.2477657855927148, 0, 0.977205023805839843,
        // j = 23/2, mj = 1/2
        -0.220443711424950199, 0, 0, -1.32266226854970119, 0, 15.8719472225964143, 0, 0, 0, 0, -3.30665567137425298, 0, 79.3597361129820716, 0, -132.266226854970119, 0, 0, 0, 0, 0, 0, -4.40887422849900398, 0, 158.719472225964143, 0, -529.064907419880478, 0, 282.167950623936255, 0, 0, 0, 0, 0, 0, 0, 0, -3.30665567137425298, 0, 158.719472225964143, 0, -793.597361129820716, 0, 846.503851871808764, 0, -181.393682543959021, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.32266226854970119, 0, 79.3597361129820716, 0, -529.064907419880478, 0, 846.503851871808764, 0, -362.787365087918042, 0, 32.2477657855927148, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.220443711424950199, 0, 15.8719472225964143, 0, -132.266226854970119, 0, 282.167950623936255, 0, -181.393682543959021, 0, 32.2477657855927148, 0, -0.977205023805839843,
        0, 0, 2.86576824852435259, 0, 0, 0, 0, 14.3288412426217629, 0, -57.3153649704870517, 0, 0, 0, 0, 0, 0, 28.6576824852435259, 0, -229.261459881948207, 0, 229.261459881948207, 0, 0, 0, 0, 0, 0, 0, 0, 28.6576824852435259, 0, -343.89218982292231, 0, 687.784379645844621, 0, -262.013097007940808, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 14.3288412426217629, 0, -229.261459881948207, 0, 687.784379645844621, 0, -524.026194015881616, 0, 87.337699002646936, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.86576824852435259, 0, -57.3153649704870517, 0, 229.261459881948207, 0, -262.013097007940808, 0, 87.337699002646936, 0, -6.35183265473795898, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = 3/2
        0, 0, -2.63612338039655511, 0, 0, 0, 0, -13.1806169019827756, 0, 52.7224676079311022, 0, 0, 0, 0, 0, 0, -26.3612338039655511, 0, 210.889870431724409, 0, -210.889870431724409, 0, 0, 0, 0, 0, 0, 0, 0, -26.3612338039655511, 0, 316.334805647586613, 0, -632.669611295173227, 0, 241.01699477911361, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -13.1806169019827756, 0, 210.889870431724409, 0, -632.669611295173227, 0, 482.03398955822722, 0, -80.3389982597045367, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.63612338039655511, 0, 52.7224676079311022, 0, -210.889870431724409, 0, 241.01699477911361, 0, -80.3389982597045367, 0, 5.84283623706942085, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.239647580036050465, 0, 0, -0.958590320144201859, 0, 16.7753306025235325, 0, 0, 0, 0, -1.19823790018025232, 0, 50.3259918075705976, 0, -134.20264482018826, 0, 0, 0, 0, 0, 0, 0, 0, 33.5506612050470651, 0, -268.40528964037652, 0, 268.40528964037652, 0, 0, 0, 0, 0, 0, 0, 0, 1.19823790018025232, 0, -33.5506612050470651, 0, 0, 0, 268.40528964037652, 0, -153.374451223072297, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.958590320144201859, 0, -50.3259918075705976, 0, 268.40528964037652, 0, -268.40528964037652, 0, 0, 0, 20.449926829742973, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.239647580036050465, 0, -16.7753306025235325, 0, 134.20264482018826, 0, -268.40528964037652, 0, 153.374451223072297, 0, -20.449926829742973, 0, 0,
        // j = 23/2, mj = 5/2
        0.202539171902860758, 0, 0, 0.810156687611443034, 0, -14.1777420332002531, 0, 0, 0, 0, 1.01269585951430379, 0, -42.5332260996007593, 0, 113.421936265602025, 0, 0, 0, 0, 0, 0, 0, 0, -28.3554840664005062, 0, 226.843872531204049, 0, -226.843872531204049, 0, 0, 0, 0, 0, 0, 0, 0, -1.01269585951430379, 0, 28.3554840664005062, 0, 0, 0, -226.843872531204049, 0, 129.625070017830885, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.810156687611443034, 0, 42.5332260996007593, 0, -226.843872531204049, 0, 226.843872531204049, 0, 0, 0, -17.2833426690441181, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.202539171902860758, 0, 14.1777420332002531, 0, -113.421936265602025, 0, 226.843872531204049, 0, -129.625070017830885, 0, 17.2833426690441181, 0, 0,
        0, 0, -3.03808757854291138, 0, 0, 0, 0, -3.03808757854291138, 0, 56.7109681328010124, 0, 0, 0, 0, 0, 0, 18.2285254712574683, 0, 0, 0, -204.159485278083644, 0, 0, 0, 0, 0, 0, 0, 0, 42.5332260996007593, 0, -340.265808796806074, 0, 204.159485278083644, 0, 194.437605026746328, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 33.4189633639720251, 0, -453.687745062408099, 0, 1020.79742639041822, 0, -388.875210053492656, 0, -43.2083566726102951, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 9.11426273562873413, 0, -170.132904398403037, 0, 612.478455834250933, 0, -583.312815080238984, 0, 129.625070017830885, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = 7/2
        0, 0, 2.35329251920326267, 0, 0, 0, 0, 2.35329251920326267, 0, -43.9281270251275699, 0, 0, 0, 0, 0, 0, -14.119755115219576, 0, 0, 0, 158.141257290459252, 0, 0, 0, 0, 0, 0, 0, 0, -32.9460952688456774, 0, 263.568762150765419, 0, -158.141257290459252, 0, -150.610721229008811, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -25.8862177112358894, 0, 351.425016201020559, 0, -790.706286452296258, 0, 301.221442458017622, 0, 33.469049162001958, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -7.05987755760978802, 0, 131.78438107538271, 0, -474.423771871377755, 0, 451.832163687026433, 0, -100.407147486005874, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.261476946578140297, 0, 0, -0.522953893156280594, 0, -16.734524581000979, 0, 0, 0, 0, -4.44510809182838505, 0, 50.203573743002937, 0, 117.141672067006853, 0, 0, 0, 0, 0, 0, -7.32135450418792832, 0, 234.283344134013706, 0, -468.566688268027412, 0, -187.426675307210965, 0, 0, 0, 0, 0, 0, 0, 0, -4.44510809182838505, 0, 234.283344134013706, 0, -1171.41672067006853, 0, 937.133376536054825, 0, 66.938098324003916, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.522953893156280594, 0, 50.203573743002937, 0, -468.566688268027412, 0, 937.133376536054825, 0, -401.628589944023496, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.261476946578140297, 0, -16.734524581000979, 0, 117.141672067006853, 0, -187.426675307210965, 0, 66.938098324003916, 0, 0, 0, 0,
        // j = 23/2, mj = 9/2
        -0.184892122049355627, 0, 0, 0.369784244098711255, 0, 11.8330958111587602, 0, 0, 0, 0, 3.14316607483904567, 0, -35.4992874334762805, 0, -82.8316706781113211, 0, 0, 0, 0, 0, 0, 5.17697941738195757, 0, -165.663341356222642, 0, 331.326682712445284, 0, 132.530673084978114, 0, 0, 0, 0, 0, 0, 0, 0, 3.14316607483904567, 0, -165.663341356222642, 0, 828.316706781113211, 0, -662.653365424890569, 0, -47.3323832446350406, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.369784244098711255, 0, -35.4992874334762805, 0, 331.326682712445284, 0, -662.653365424890569, 0, 283.994299467810244, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.184892122049355627, 0, 11.8330958111587602, 0, -82.8316706781113211, 0, 132.530673084978114, 0, -47.3323832446350406, 0, 0, 0, 0,
        0, 0, 3.14316607483904567, 0, 0, 0, 0, -22.0021625238733197, 0, -50.2906571974247307, 0, 0, 0, 0, 0, 0, -69.1496536464590047, 0, 402.325257579397845, 0, 140.813840152789246, 0, 0, 0, 0, 0, 0, 0, 0, -44.0043250477466393, 0, 704.06920076394623, 0, -1267.32456137510321, 0, -80.4650515158795691, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 15.7158303741952283, 0, 0, 0, -704.06920076394623, 0, 804.650515158795691, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 15.7158303741952283, 0, -251.453285987123653, 0, 704.06920076394623, 0, -402.325257579397845, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = 11/2
        0, 0, -2.01693493170051083, 0, 0, 0, 0, 14.1185445219035758, 0, 32.2709589072081732, 0, 0, 0, 0, 0, 0, 44.3725684974112382, 0, -258.167671257665386, 0, -90.358684940182885, 0, 0, 0, 0, 0, 0, 0, 0, 28.2370890438071516, 0, -451.793424700914425, 0, 813.228164461645965, 0, 51.6335342515330771, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -10.0846746585025541, 0, 0, 0, 451.793424700914425, 0, -516.335342515330771, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -10.0846746585025541, 0, 161.354794536040866, 0, -451.793424700914425, 0, 258.167671257665386, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.288133561671501547, 0, 0, 3.45760274005801856, 0, 15.5592123302610835, 0, 0, 0, 0, 7.77960616513054176, 0, -202.269760293394086, 0, -82.9824657613924454, 0, 0, 0, 0, 0, 0, 0, 0, -217.828972623655169, 0, 1161.75452065949424, 0, 77.4503013772996157, 0, 0, 0, 0, 0, 0, 0, 0, -7.77960616513054176, 0, 217.828972623655169, 0, 0, 0, -1161.75452065949424, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3.45760274005801856, 0, 202.269760293394086, 0, -1161.75452065949424, 0, 1161.75452065949424, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.288133561671501547, 0, -15.5592123302610835, 0, 82.9824657613924454, 0, -77.4503013772996157, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = 13/2
        0.166353989393607054, 0, 0, -1.99624787272328465, 0, -8.98311542725478094, 0, 0, 0, 0, -4.49155771362739047, 0, 116.780500554312152, 0, 47.9099489453588317, 0, 0, 0, 0, 0, 0, 0, 0, 125.763615981566933, 0, -670.739285235023644, 0, -44.7159523490015762, 0, 0, 0, 0, 0, 0, 0, 0, 4.49155771362739047, 0, -125.763615981566933, 0, 0, 0, 670.739285235023644, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.99624787272328465, 0, -116.780500554312152, 0, 670.739285235023644, 0, -670.739285235023644, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.166353989393607054, 0, 8.98311542725478094, 0, -47.9099489453588317, 0, 44.7159523490015762, 0, 0, 0, 0, 0, 0,
        0, 0, -3.16072579847853403, 0, 0, 0, 0, 60.0537901710921467, 0, 37.9287095817424084, 0, 0, 0, 0, 0, 0, 18.9643547908712042, 0, -758.574191634848168, 0, -60.6859353307878535, 0, 0, 0, 0, 0, 0, 0, 0, -132.750483536098429, 0, 531.001934144393718, 0, 1274.40464194654492, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -66.3752417680492147, 0, 1062.00386828878744, 0, -2124.00773657757487, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 22.1250805893497382, 0, -265.500967072196859, 0, 424.801547315514974, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = 15/2
        0, 0, 1.6214181230580196, 0, 0, 0, 0, -30.8069443381023724, 0, -19.4570174766962352, 0, 0, 0, 0, 0, 0, -9.72850873834811761, 0, 389.140349533924704, 0, 31.1312279627139763, 0, 0, 0, 0, 0, 0, 0, 0, 68.0995611684368233, 0, -272.398244673747293, 0, -653.755787216993503, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 34.0497805842184116, 0, -544.796489347494586, 0, 1089.59297869498917, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -11.3499268614061372, 0, 136.199122336873647, 0, -217.918595738997834, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.32428362461160392, 0, 0, -8.43137423990170193, 0, -12.9713449844641568, 0, 0, 0, 0, 4.8642543691740588, 0, 350.226314580532234, 0, 38.9140349533924704, 0, 0, 0, 0, 0, 0, 27.2398244673747293, 0, -544.796489347494586, 0, -1089.59297869498917, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4.8642543691740588, 0, -544.796489347494586, 0, 2723.98244673747293, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -8.43137423990170193, 0, 350.226314580532234, 0, -1089.59297869498917, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.32428362461160392, 0, -12.9713449844641568, 0, 38.9140349533924704, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = 17/2
        -0.145024045724314041, 0, 0, 3.77062518883216506, 0, 5.80096182897256162, 0, 0, 0, 0, -2.17536068586471061, 0, -156.625969382259164, 0, -17.4028854869176849, 0, 0, 0, 0, 0, 0, -12.1820198408423794, 0, 243.640396816847588, 0, 487.280793633695176, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.17536068586471061, 0, 243.640396816847588, 0, -1218.20198408423794, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.77062518883216506, 0, -156.625969382259164, 0, 487.280793633695176, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.145024045724314041, 0, 5.80096182897256162, 0, -17.4028854869176849, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 3.04550496021059485, 0, 0, 0, 0, -106.59267360737082, 0, -20.3033664014039657, 0, 0, 0, 0, 0, 0, 274.095446418953537, 0, 730.921190450542765, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 127.911208328844984, 0, -2558.22416657689968, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -228.412872015794614, 0, 1705.48277771793312, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 27.4095446418953537, 0, -182.730297612635691, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = 19/2
        0, 0, -1.15109267733298502, 0, 0, 0, 0, 40.2882437066544757, 0, 7.67395118221990013, 0, 0, 0, 0, 0, 0, -103.598340959968652, 0, -276.262242559916405, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -48.3458924479853708, 0, 966.917848959707416, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 86.3319507999738764, 0, -644.611899306471611, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -10.3598340959968652, 0, 69.0655606399791011, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.383697559110995006, 0, 0, 16.8826926008837803, 0, 8.44134630044189014, 0, 0, 0, 0, -63.310097253314176, 0, -379.860583519885056, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1772.68272309279693, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 63.310097253314176, 0, -1772.68272309279693, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -16.8826926008837803, 0, 379.860583519885056, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.383697559110995006, 0, -8.44134630044189014, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = 21/2
        0.115689166958762041, 0, 0, -5.09032334618552982, 0, -2.54516167309276491, 0, 0, 0, 0, 19.0887125481957368, 0, 114.532275289174421, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -534.483951349480631, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -19.0887125481957368, 0, 534.483951349480631, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5.09032334618552982, 0, -114.532275289174421, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.115689166958762041, 0, 2.54516167309276491, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -2.66085084005152695, 0, 0, 0, 0, 146.346796202833982, 0, 0, 0, 0, 0, 0, 0, 0, -878.080777217003893, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1229.31308810380545, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -439.040388608501947, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 29.2693592405667964, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = 23/2
        0, 0, 0.554825753806619302, 0, 0, 0, 0, -30.5154164593640616, 0, 0, 0, 0, 0, 0, 0, 0, 183.09249875618437, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -256.329498258658117, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 91.5462493780921848, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -6.10308329187281232, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.554825753806619302, 0, 0, -36.6184997512368739, 0, 0, 0, 0, 0, 0, 274.638748134276554, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -512.658996517316235, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 274.638748134276554, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -36.6184997512368739, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.554825753806619302, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 25/2, mj = -25/2
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.566266663742191171, 0, 0, -37.3735998069846173, 0, 0, 0, 0, 0, 0, 280.30199855238463, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -523.230397297784642, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 280.30199855238463, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -37.3735998069846173, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.566266663742191171, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 25/2, mj = -23/2
        0.113253332748438234, 0, 0, -7.47471996139692346, 0, 0, 0, 0, 0, 0, 56.0603997104769259, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -104.646079459556928, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 56.0603997104769259, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -7.47471996139692346, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.113253332748438234, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 2.71807998596251762, 0, 0, 0, 0, -149.494399227938469, 0, 0, 0, 0, 0, 0, 0, 0, 896.966395367630815, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1255.75295351468314, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 448.483197683815407, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -29.8988798455876938, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 25/2, mj = -21/2
        0, 0, 0.784642105787196884, 0, 0, 0, 0, -43.1553158182958286, 0, 0, 0, 0, 0, 0, 0, 0, 258.931894909774972, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -362.50465287368496, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 129.465947454887486, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -8.63106316365916572, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.392321052893598442, 0, 0, 17.2621263273183314, 0, 8.63106316365916572, 0, 0, 0, 0, -64.7329737274437429, 0, -388.397842364662457, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1812.5232643684248, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64.7329737274437429, 0, -1812.5232643684248, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -17.2621263273183314, 0, 388.397842364662457, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.392321052893598442, 0, -8.63106316365916572, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 25/2, mj = -19/2
        -0.141689713908309089, 0, 0, 6.23434741196559994, 0, 3.11717370598279997, 0, 0, 0, 0, -23.3788027948709998, 0, -140.272816769225999, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 654.606478256387993, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 23.3788027948709998, 0, -654.606478256387993, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -6.23434741196559994, 0, 140.272816769225999, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.141689713908309089, 0, -3.11717370598279997, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -3.11717370598279997, 0, 0, 0, 0, 109.101079709397999, 0, 20.7811580398853331, 0, 0, 0, 0, 0, 0, -280.545633538451997, 0, -748.121689435871992, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -130.921295651277599, 0, 2618.42591302555197, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 233.788027948709998, 0, -1745.61727535036798, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -28.0545633538451997, 0, 187.030422358967998, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 25/2, mj = -17/2
        0, 0, -1.32916733424081186, 0, 0, 0, 0, 46.5208566984284153, 0, 8.86111556160541243, 0, 0, 0, 0, 0, 0, -119.625060081673068, 0, -319.000160217794847, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -55.8250280381140983, 0, 1116.50056076228197, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 99.6875500680608898, 0, -744.333707174854644, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -11.9625060081673068, 0, 79.7500400544487119, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.332291833560202966, 0, 0, -8.63958767256527712, 0, -13.2916733424081186, 0, 0, 0, 0, 4.98437750340304449, 0, 358.875180245019203, 0, 39.8750200272243559, 0, 0, 0, 0, 0, 0, 27.9125140190570492, 0, -558.250280381140983, 0, -1116.50056076228197, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4.98437750340304449, 0, -558.250280381140983, 0, 2791.25140190570492, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -8.63958767256527712, 0, 358.875180245019203, 0, -1116.50056076228197, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.332291833560202966, 0, -13.2916733424081186, 0, 39.8750200272243559, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 25/2, mj = -15/2
        0.16214181230580196, 0, 0, -4.21568711995085096, 0, -6.48567249223207841, 0, 0, 0, 0, 2.4321271845870294, 0, 175.113157290266117, 0, 19.4570174766962352, 0, 0, 0, 0, 0, 0, 13.6199122336873647, 0, -272.398244673747293, 0, -544.796489347494586, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.4321271845870294, 0, -272.398244673747293, 0, 1361.99122336873647, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4.21568711995085096, 0, 175.113157290266117, 0, -544.796489347494586, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.16214181230580196, 0, -6.48567249223207841, 0, 19.4570174766962352, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 3.2428362461160392, 0, 0, 0, 0, -61.6138886762047449, 0, -38.9140349533924704, 0, 0, 0, 0, 0, 0, -19.4570174766962352, 0, 778.280699067849409, 0, 62.2624559254279527, 0, 0, 0, 0, 0, 0, 0, 0, 136.199122336873647, 0, -544.796489347494586, 0, -1307.51157443398701, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 68.0995611684368233, 0, -1089.59297869498917, 0, 2179.18595738997834, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -22.6998537228122744, 0, 272.398244673747293, 0, -435.837191477995669, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 25/2, mj = -13/2
        0, 0, 1.77617456229312929, 0, 0, 0, 0, -33.7473166835694566, 0, -21.3140947475175515, 0, 0, 0, 0, 0, 0, -10.6570473737587758, 0, 426.281894950351031, 0, 34.1025515960280824, 0, 0, 0, 0, 0, 0, 0, 0, 74.5993316163114303, 0, -298.397326465245721, 0, -716.153583516589731, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 37.2996658081557152, 0, -596.794652930491443, 0, 1193.58930586098289, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -12.4332219360519051, 0, 149.198663232622861, 0, -238.717861172196577, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.296029093715521549, 0, 0, 3.55234912458625859, 0, 15.9855710606381636, 0, 0, 0, 0, 7.99278553031908182, 0, -207.812423788296127, 0, -85.2563789900702061, 0, 0, 0, 0, 0, 0, 0, 0, -223.797994848934291, 0, 1193.58930586098289, 0, 79.5726203907321924, 0, 0, 0, 0, 0, 0, 0, 0, -7.99278553031908182, 0, 223.797994848934291, 0, 0, 0, -1193.58930586098289, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3.55234912458625859, 0, 207.812423788296127, 0, -1193.58930586098289, 0, 1193.58930586098289, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.296029093715521549, 0, -15.9855710606381636, 0, 85.2563789900702061, 0, -79.5726203907321924, 0, 0, 0, 0, 0, 0,
        // j = 25/2, mj = -11/2
        -0.179682844900943126, 0, 0, 2.15619413881131751, 0, 9.7028736246509288, 0, 0, 0, 0, 4.8514368123254644, 0, -126.137357120462074, 0, -51.7486593314716203, 0, 0, 0, 0, 0, 0, 0, 0, -135.840230745113003, 0, 724.481230640602684, 0, 48.2987487093735122, 0, 0, 0, 0, 0, 0, 0, 0, -4.8514368123254644, 0, 135.840230745113003, 0, 0, 0, -724.481230640602684, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.15619413881131751, 0, 126.137357120462074, 0, -724.481230640602684, 0, 724.481230640602684, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.179682844900943126, 0, -9.7028736246509288, 0, 51.7486593314716203, 0, -48.2987487093735122, 0, 0, 0, 0, 0, 0,
        0, 0, -3.23429120821697627, 0, 0, 0, 0, 22.6400384575188339, 0, 51.7486593314716203, 0, 0, 0, 0, 0, 0, 71.1544065807734779, 0, -413.989274651772962, 0, -144.896246128120537, 0, 0, 0, 0, 0, 0, 0, 0, 45.2800769150376677, 0, -724.481230640602684, 0, 1304.06621515308483, 0, 82.7978549303545924, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -16.1714560410848813, 0, 0, 0, 724.481230640602684, 0, -827.978549303545924, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -16.1714560410848813, 0, 258.743296657358101, 0, -724.481230640602684, 0, 413.989274651772962, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 25/2, mj = -9/2
        0, 0, -2.15619413881131751, 0, 0, 0, 0, 15.0933589716792226, 0, 34.4991062209810802, 0, 0, 0, 0, 0, 0, 47.4362710538489852, 0, -275.992849767848641, 0, -96.5974974187470245, 0, 0, 0, 0, 0, 0, 0, 0, 30.1867179433584452, 0, -482.987487093735122, 0, 869.37747676872322, 0, 55.1985699535697283, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -10.7809706940565876, 0, 0, 0, 482.987487093735122, 0, -551.985699535697283, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -10.7809706940565876, 0, 172.495531104905401, 0, -482.987487093735122, 0, 275.992849767848641, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.269524267351414689, 0, 0, -0.539048534702829378, 0, -17.2495531104905401, 0, 0, 0, 0, -4.58191254497404971, 0, 51.7486593314716203, 0, 120.746871773433781, 0, 0, 0, 0, 0, 0, -7.54667948583961129, 0, 241.493743546867561, 0, -482.987487093735122, 0, -193.194994837494049, 0, 0, 0, 0, 0, 0, 0, 0, -4.58191254497404971, 0, 241.493743546867561, 0, -1207.46871773433781, 0, 965.974974187470245, 0, 68.9982124419621604, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.539048534702829378, 0, 51.7486593314716203, 0, -482.987487093735122, 0, 965.974974187470245, 0, -413.989274651772962, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.269524267351414689, 0, -17.2495531104905401, 0, 120.746871773433781, 0, -193.194994837494049, 0, 68.9982124419621604, 0, 0, 0, 0,
        // j = 25/2, mj = -7/2
        0.196107709933605223, 0, 0, -0.392215419867210446, 0, -12.5508934357507343, 0, 0, 0, 0, -3.33383106887128879, 0, 37.6526803072522028, 0, 87.8562540502551398, 0, 0, 0, 0, 0, 0, -5.49101587814094624, 0, 175.71250810051028, 0, -351.425016201020559, 0, -140.570006480408224, 0, 0, 0, 0, 0, 0, 0, 0, -3.33383106887128879, 0, 175.71250810051028, 0, -878.562540502551398, 0, 702.850032402041119, 0, 50.203573743002937, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.392215419867210446, 0, 37.6526803072522028, 0, -351.425016201020559, 0, 702.850032402041119, 0, -301.221442458017622, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.196107709933605223, 0, -12.5508934357507343, 0, 87.8562540502551398, 0, -140.570006480408224, 0, 50.203573743002937, 0, 0, 0, 0,
        0, 0, 3.13772335893768356, 0, 0, 0, 0, 3.13772335893768356, 0, -58.5708360335034265, 0, 0, 0, 0, 0, 0, -18.8263401536261014, 0, 0, 0, 210.855009720612336, 0, 0, 0, 0, 0, 0, 0, 0, -43.9281270251275699, 0, 351.425016201020559, 0, -210.855009720612336, 0, -200.814294972011748, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -34.5149569483145192, 0, 468.566688268027412, 0, -1054.27504860306168, 0, 401.628589944023496, 0, 44.6253988826692774, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -9.41317007681305069, 0, 175.71250810051028, 0, -632.565029161837007, 0, 602.442884916035244, 0, -133.876196648007832, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 25/2, mj = -5/2
        0, 0, 2.48058812043928146, 0, 0, 0, 0, 2.48058812043928146, 0, -46.3043115815332539, 0, 0, 0, 0, 0, 0, -14.8835287226356887, 0, 0, 0, 166.695521693519714, 0, 0, 0, 0, 0, 0, 0, 0, -34.7282336861499404, 0, 277.825869489199523, 0, -166.695521693519714, 0, -158.757639708114013, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -27.286469324832096, 0, 370.434492652266031, 0, -833.47760846759857, 0, 317.515279416228027, 0, 35.279475490692003, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -7.44176436131784437, 0, 138.912934744599762, 0, -500.086565080559142, 0, 476.27291912434204, 0, -105.838426472076009, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.248058812043928146, 0, 0, -0.992235248175712583, 0, 17.3641168430749702, 0, 0, 0, 0, -1.24029406021964073, 0, 52.0923505292249106, 0, -138.912934744599762, 0, 0, 0, 0, 0, 0, 0, 0, 34.7282336861499404, 0, -277.825869489199523, 0, 277.825869489199523, 0, 0, 0, 0, 0, 0, 0, 0, 1.24029406021964073, 0, -34.7282336861499404, 0, 0, 0, 277.825869489199523, 0, -158.757639708114013, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.992235248175712583, 0, -52.0923505292249106, 0, 277.825869489199523, 0, -277.825869489199523, 0, 0, 0, 21.1676852944152018, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.248058812043928146, 0, -17.3641168430749702, 0, 138.912934744599762, 0, -277.825869489199523, 0, 158.757639708114013, 0, -21.1676852944152018, 0, 0,
        // j = 25/2, mj = -3/2
        -0.212424875592775713, 0, 0, -0.849699502371102853, 0, 14.8697412914942999, 0, 0, 0, 0, -1.06212437796387857, 0, 44.6092238744828998, 0, -118.957930331954399, 0, 0, 0, 0, 0, 0, 0, 0, 29.7394825829885999, 0, -237.915860663908799, 0, 237.915860663908799, 0, 0, 0, 0, 0, 0, 0, 0, 1.06212437796387857, 0, -29.7394825829885999, 0, 0, 0, 237.915860663908799, 0, -135.951920379376457, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.849699502371102853, 0, -44.6092238744828998, 0, 237.915860663908799, 0, -237.915860663908799, 0, 0, 0, 18.1269227172501942, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.212424875592775713, 0, -14.8697412914942999, 0, 118.957930331954399, 0, -237.915860663908799, 0, 135.951920379376457, 0, -18.1269227172501942, 0, 0,
        0, 0, -2.97394825829885999, 0, 0, 0, 0, -14.8697412914942999, 0, 59.4789651659771997, 0, 0, 0, 0, 0, 0, -29.7394825829885999, 0, 237.915860663908799, 0, -237.915860663908799, 0, 0, 0, 0, 0, 0, 0, 0, -29.7394825829885999, 0, 356.873790995863198, 0, -713.747581991726397, 0, 271.903840758752913, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -14.8697412914942999, 0, 237.915860663908799, 0, -713.747581991726397, 0, 543.807681517505826, 0, -90.634613586250971, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.97394825829885999, 0, 59.4789651659771997, 0, -237.915860663908799, 0, 271.903840758752913, 0, -90.634613586250971, 0, 6.59160826081825244, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 25/2, mj = -1/2
        0, 0, -2.7533410732166562, 0, 0, 0, 0, -13.766705366083281, 0, 55.066821464333124, 0, 0, 0, 0, 0, 0, -27.533410732166562, 0, 220.267285857332496, 0, -220.267285857332496, 0, 0, 0, 0, 0, 0, 0, 0, -27.533410732166562, 0, 330.400928785998744, 0, -660.801857571997488, 0, 251.734040979808567, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -13.766705366083281, 0, 220.267285857332496, 0, -660.801857571997488, 0, 503.468081959617133, 0, -83.9113469932695222, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.7533410732166562, 0, 55.066821464333124, 0, -220.267285857332496, 0, 251.734040979808567, 0, -83.9113469932695222, 0, 6.10264341769232889, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.22944508943472135, 0, 0, 1.3766705366083281, 0, -16.5200464392999372, 0, 0, 0, 0, 3.44167634152082025, 0, -82.6002321964996859, 0, 137.66705366083281, 0, 0, 0, 0, 0, 0, 4.588901788694427, 0, -165.200464392999372, 0, 550.66821464333124, 0, -293.689714476443328, 0, 0, 0, 0, 0, 0, 0, 0, 3.44167634152082025, 0, -165.200464392999372, 0, 826.002321964996859, 0, -881.069143429329983, 0, 188.800530734856425, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.3766705366083281, 0, -82.6002321964996859, 0, 550.66821464333124, 0, -881.069143429329983, 0, 377.60106146971285, 0, -33.5645387973078089, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.22944508943472135, 0, -16.5200464392999372, 0, 137.66705366083281, 0, -293.689714476443328, 0, 188.800530734856425, 0, -33.5645387973078089, 0, 1.01710723628205481,
        // j = 25/2, mj = 1/2
        0.22944508943472135, 0, 0, 1.3766705366083281, 0, -16.5200464392999372, 0, 0, 0, 0, 3.44167634152082025, 0, -82.6002321964996859, 0, 137.66705366083281, 0, 0, 0, 0, 0, 0, 4.588901788694427, 0, -165.200464392999372, 0, 550.66821464333124, 0, -293.689714476443328, 0, 0, 0, 0, 0, 0, 0, 0, 3.44167634152082025, 0, -165.200464392999372, 0, 826.002321964996859, 0, -881.069143429329983, 0, 188.800530734856425, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.3766705366083281, 0, -82.6002321964996859, 0, 550.66821464333124, 0, -881.069143429329983, 0, 377.60106146971285, 0, -33.5645387973078089, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.22944508943472135, 0, -16.5200464392999372, 0, 137.66705366083281, 0, -293.689714476443328, 0, 188.800530734856425, 0, -33.5645387973078089, 0, 1.01710723628205481,
        0, 0, 2.7533410732166562, 0, 0, 0, 0, 13.766705366083281, 0, -55.066821464333124, 0, 0, 0, 0, 0, 0, 27.533410732166562, 0, -220.267285857332496, 0, 220.267285857332496, 0, 0, 0, 0, 0, 0, 0, 0, 27.533410732166562, 0, -330.400928785998744, 0, 660.801857571997488, 0, -251.734040979808567, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 13.766705366083281, 0, -220.267285857332496, 0, 660.801857571997488, 0, -503.468081959617133, 0, 83.9113469932695222, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.7533410732166562, 0, -55.066821464333124, 0, 220.267285857332496, 0, -251.734040979808567, 0, 83.9113469932695222, 0, -6.10264341769232889, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 25/2, mj = 3/2
        0, 0, 2.97394825829885999, 0, 0, 0, 0, 14.8697412914942999, 0, -59.4789651659771997, 0, 0, 0, 0, 0, 0, 29.7394825829885999, 0, -237.915860663908799, 0, 237.915860663908799, 0, 0, 0, 0, 0, 0, 0, 0, 29.7394825829885999, 0, -356.873790995863198, 0, 713.747581991726397, 0, -271.903840758752913, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 14.8697412914942999, 0, -237.915860663908799, 0, 713.747581991726397, 0, -543.807681517505826, 0, 90.634613586250971, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.97394825829885999, 0, -59.4789651659771997, 0, 237.915860663908799, 0, -271.903840758752913, 0, 90.634613586250971, 0, -6.59160826081825244, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.212424875592775713, 0, 0, -0.849699502371102853, 0, 14.8697412914942999, 0, 0, 0, 0, -1.06212437796387857, 0, 44.6092238744828998, 0, -118.957930331954399, 0, 0, 0, 0, 0, 0, 0, 0, 29.7394825829885999, 0, -237.915860663908799, 0, 237.915860663908799, 0, 0, 0, 0, 0, 0, 0, 0, 1.06212437796387857, 0, -29.7394825829885999, 0, 0, 0, 237.915860663908799, 0, -135.951920379376457, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.849699502371102853, 0, -44.6092238744828998, 0, 237.915860663908799, 0, -237.915860663908799, 0, 0, 0, 18.1269227172501942, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.212424875592775713, 0, -14.8697412914942999, 0, 118.957930331954399, 0, -237.915860663908799, 0, 135.951920379376457, 0, -18.1269227172501942, 0, 0,
        // j = 25/2, mj = 5/2
        -0.248058812043928146, 0, 0, -0.992235248175712583, 0, 17.3641168430749702, 0, 0, 0, 0, -1.24029406021964073, 0, 52.0923505292249106, 0, -138.912934744599762, 0, 0, 0, 0, 0, 0, 0, 0, 34.7282336861499404, 0, -277.825869489199523, 0, 277.825869489199523, 0, 0, 0, 0, 0, 0, 0, 0, 1.24029406021964073, 0, -34.7282336861499404, 0, 0, 0, 277.825869489199523, 0, -158.757639708114013, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.992235248175712583, 0, -52.0923505292249106, 0, 277.825869489199523, 0, -277.825869489199523, 0, 0, 0, 21.1676852944152018, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.248058812043928146, 0, -17.3641168430749702, 0, 138.912934744599762, 0, -277.825869489199523, 0, 158.757639708114013, 0, -21.1676852944152018, 0, 0,
        0, 0, -2.48058812043928146, 0, 0, 0, 0, -2.48058812043928146, 0, 46.3043115815332539, 0, 0, 0, 0, 0, 0, 14.8835287226356887, 0, 0, 0, -166.695521693519714, 0, 0, 0, 0, 0, 0, 0, 0, 34.7282336861499404, 0, -277.825869489199523, 0, 166.695521693519714, 0, 158.757639708114013, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 27.286469324832096, 0, -370.434492652266031, 0, 833.47760846759857, 0, -317.515279416228027, 0, -35.279475490692003, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 7.44176436131784437, 0, -138.912934744599762, 0, 500.086565080559142, 0, -476.27291912434204, 0, 105.838426472076009, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 25/2, mj = 7/2
        0, 0, -3.13772335893768356, 0, 0, 0, 0, -3.13772335893768356, 0, 58.5708360335034265, 0, 0, 0, 0, 0, 0, 18.8263401536261014, 0, 0, 0, -210.855009720612336, 0, 0, 0, 0, 0, 0, 0, 0, 43.9281270251275699, 0, -351.425016201020559, 0, 210.855009720612336, 0, 200.814294972011748, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 34.5149569483145192, 0, -468.566688268027412, 0, 1054.27504860306168, 0, -401.628589944023496, 0, -44.6253988826692774, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 9.41317007681305069, 0, -175.71250810051028, 0, 632.565029161837007, 0, -602.442884916035244, 0, 133.876196648007832, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.196107709933605223, 0, 0, -0.392215419867210446, 0, -12.5508934357507343, 0, 0, 0, 0, -3.33383106887128879, 0, 37.6526803072522028, 0, 87.8562540502551398, 0, 0, 0, 0, 0, 0, -5.49101587814094624, 0, 175.71250810051028, 0, -351.425016201020559, 0, -140.570006480408224, 0, 0, 0, 0, 0, 0, 0, 0, -3.33383106887128879, 0, 175.71250810051028, 0, -878.562540502551398, 0, 702.850032402041119, 0, 50.203573743002937, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.392215419867210446, 0, 37.6526803072522028, 0, -351.425016201020559, 0, 702.850032402041119, 0, -301.221442458017622, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.196107709933605223, 0, -12.5508934357507343, 0, 87.8562540502551398, 0, -140.570006480408224, 0, 50.203573743002937, 0, 0, 0, 0,
        // j = 25/2, mj = 9/2
        0.269524267351414689, 0, 0, -0.539048534702829378, 0, -17.2495531104905401, 0, 0, 0, 0, -4.58191254497404971, 0, 51.7486593314716203, 0, 120.746871773433781, 0, 0, 0, 0, 0, 0, -7.54667948583961129, 0, 241.493743546867561, 0, -482.987487093735122, 0, -193.194994837494049, 0, 0, 0, 0, 0, 0, 0, 0, -4.58191254497404971, 0, 241.493743546867561, 0, -1207.46871773433781, 0, 965.974974187470245, 0, 68.9982124419621604, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.539048534702829378, 0, 51.7486593314716203, 0, -482.987487093735122, 0, 965.974974187470245, 0, -413.989274651772962, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.269524267351414689, 0, -17.2495531104905401, 0, 120.746871773433781, 0, -193.194994837494049, 0, 68.9982124419621604, 0, 0, 0, 0,
        0, 0, 2.15619413881131751, 0, 0, 0, 0, -15.0933589716792226, 0, -34.4991062209810802, 0, 0, 0, 0, 0, 0, -47.4362710538489852, 0, 275.992849767848641, 0, 96.5974974187470245, 0, 0, 0, 0, 0, 0, 0, 0, -30.1867179433584452, 0, 482.987487093735122, 0, -869.37747676872322, 0, -55.1985699535697283, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 10.7809706940565876, 0, 0, 0, -482.987487093735122, 0, 551.985699535697283, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 10.7809706940565876, 0, -172.495531104905401, 0, 482.987487093735122, 0, -275.992849767848641, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 25/2, mj = 11/2
        0, 0, 3.23429120821697627, 0, 0, 0, 0, -22.6400384575188339, 0, -51.7486593314716203, 0, 0, 0, 0, 0, 0, -71.1544065807734779, 0, 413.989274651772962, 0, 144.896246128120537, 0, 0, 0, 0, 0, 0, 0, 0, -45.2800769150376677, 0, 724.481230640602684, 0, -1304.06621515308483, 0, -82.7978549303545924, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16.1714560410848813, 0, 0, 0, -724.481230640602684, 0, 827.978549303545924, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16.1714560410848813, 0, -258.743296657358101, 0, 724.481230640602684, 0, -413.989274651772962, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.179682844900943126, 0, 0, 2.15619413881131751, 0, 9.7028736246509288, 0, 0, 0, 0, 4.8514368123254644, 0, -126.137357120462074, 0, -51.7486593314716203, 0, 0, 0, 0, 0, 0, 0, 0, -135.840230745113003, 0, 724.481230640602684, 0, 48.2987487093735122, 0, 0, 0, 0, 0, 0, 0, 0, -4.8514368123254644, 0, 135.840230745113003, 0, 0, 0, -724.481230640602684, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.15619413881131751, 0, 126.137357120462074, 0, -724.481230640602684, 0, 724.481230640602684, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.179682844900943126, 0, -9.7028736246509288, 0, 51.7486593314716203, 0, -48.2987487093735122, 0, 0, 0, 0, 0, 0,
        // j = 25/2, mj = 13/2
        -0.296029093715521549, 0, 0, 3.55234912458625859, 0, 15.9855710606381636, 0, 0, 0, 0, 7.99278553031908182, 0, -207.812423788296127, 0, -85.2563789900702061, 0, 0, 0, 0, 0, 0, 0, 0, -223.797994848934291, 0, 1193.58930586098289, 0, 79.5726203907321924, 0, 0, 0, 0, 0, 0, 0, 0, -7.99278553031908182, 0, 223.797994848934291, 0, 0, 0, -1193.58930586098289, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3.55234912458625859, 0, 207.812423788296127, 0, -1193.58930586098289, 0, 1193.58930586098289, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.296029093715521549, 0, -15.9855710606381636, 0, 85.2563789900702061, 0, -79.5726203907321924, 0, 0, 0, 0, 0, 0,
        0, 0, -1.77617456229312929, 0, 0, 0, 0, 33.7473166835694566, 0, 21.3140947475175515, 0, 0, 0, 0, 0, 0, 10.6570473737587758, 0, -426.281894950351031, 0, -34.1025515960280824, 0, 0, 0, 0, 0, 0, 0, 0, -74.5993316163114303, 0, 298.397326465245721, 0, 716.153583516589731, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -37.2996658081557152, 0, 596.794652930491443, 0, -1193.58930586098289, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12.4332219360519051, 0, -149.198663232622861, 0, 238.717861172196577, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 25/2, mj = 15/2
        0, 0, -3.2428362461160392, 0, 0, 0, 0, 61.6138886762047449, 0, 38.9140349533924704, 0, 0, 0, 0, 0, 0, 19.4570174766962352, 0, -778.280699067849409, 0, -62.2624559254279527, 0, 0, 0, 0, 0, 0, 0, 0, -136.199122336873647, 0, 544.796489347494586, 0, 1307.51157443398701, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -68.0995611684368233, 0, 1089.59297869498917, 0, -2179.18595738997834, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 22.6998537228122744, 0, -272.398244673747293, 0, 435.837191477995669, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.16214181230580196, 0, 0, -4.21568711995085096, 0, -6.48567249223207841, 0, 0, 0, 0, 2.4321271845870294, 0, 175.113157290266117, 0, 19.4570174766962352, 0, 0, 0, 0, 0, 0, 13.6199122336873647, 0, -272.398244673747293, 0, -544.796489347494586, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.4321271845870294, 0, -272.398244673747293, 0, 1361.99122336873647, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4.21568711995085096, 0, 175.113157290266117, 0, -544.796489347494586, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.16214181230580196, 0, -6.48567249223207841, 0, 19.4570174766962352, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 25/2, mj = 17/2
        0.332291833560202966, 0, 0, -8.63958767256527712, 0, -13.2916733424081186, 0, 0, 0, 0, 4.98437750340304449, 0, 358.875180245019203, 0, 39.8750200272243559, 0, 0, 0, 0, 0, 0, 27.9125140190570492, 0, -558.250280381140983, 0, -1116.50056076228197, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4.98437750340304449, 0, -558.250280381140983, 0, 2791.25140190570492, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -8.63958767256527712, 0, 358.875180245019203, 0, -1116.50056076228197, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.332291833560202966, 0, -13.2916733424081186, 0, 39.8750200272243559, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 1.32916733424081186, 0, 0, 0, 0, -46.5208566984284153, 0, -8.86111556160541243, 0, 0, 0, 0, 0, 0, 119.625060081673068, 0, 319.000160217794847, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 55.8250280381140983, 0, -1116.50056076228197, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -99.6875500680608898, 0, 744.333707174854644, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 11.9625060081673068, 0, -79.7500400544487119, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 25/2, mj = 19/2
        0, 0, 3.11717370598279997, 0, 0, 0, 0, -109.101079709397999, 0, -20.7811580398853331, 0, 0, 0, 0, 0, 0, 280.545633538451997, 0, 748.121689435871992, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 130.921295651277599, 0, -2618.42591302555197, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -233.788027948709998, 0, 1745.61727535036798, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 28.0545633538451997, 0, -187.030422358967998, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        -0.141689713908309089, 0, 0, 6.23434741196559994, 0, 3.11717370598279997, 0, 0, 0, 0, -23.3788027948709998, 0, -140.272816769225999, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 654.606478256387993, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 23.3788027948709998, 0, -654.606478256387993, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -6.23434741196559994, 0, 140.272816769225999, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.141689713908309089, 0, -3.11717370598279997, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 25/2, mj = 21/2
        -0.392321052893598442, 0, 0, 17.2621263273183314, 0, 8.63106316365916572, 0, 0, 0, 0, -64.7329737274437429, 0, -388.397842364662457, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1812.5232643684248, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64.7329737274437429, 0, -1812.5232643684248, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -17.2621263273183314, 0, 388.397842364662457, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.392321052893598442, 0, -8.63106316365916572, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, -0.784642105787196884, 0, 0, 0, 0, 43.1553158182958286, 0, 0, 0, 0, 0, 0, 0, 0, -258.931894909774972, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 362.50465287368496, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -129.465947454887486, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 8.63106316365916572, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 25/2, mj = 23/2
        0, 0, -2.71807998596251762, 0, 0, 0, 0, 149.494399227938469, 0, 0, 0, 0, 0, 0, 0, 0, -896.966395367630815, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1255.75295351468314, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -448.483197683815407, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 29.8988798455876938, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0.113253332748438234, 0, 0, -7.47471996139692346, 0, 0, 0, 0, 0, 0, 56.0603997104769259, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -104.646079459556928, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 56.0603997104769259, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -7.47471996139692346, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.113253332748438234, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 25/2, mj = 25/2
        0.566266663742191171, 0, 0, -37.3735998069846173, 0, 0, 0, 0, 0, 0, 280.30199855238463, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -523.230397297784642, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 280.30199855238463, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -37.3735998069846173, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.566266663742191171, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
};
/*
 * Imaginary part of cartesian to spinor transformation
 */
static double g_trans_cart2jI[] = {
        0,
        0,
        0,
        0,
        0,  // p1/2
        0.577350269189625764,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
       -0.577350269189625764,
        0,
        0,  // p3/2
        0,
        0,
        0,
       -0.707106781186547524,
        0,
        0,
       -0.408248290463863016,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
       -0.408248290463863016,
        0,
        0,
       -0.707106781186547524,
        0,
        0,
        0,
        0,
        0,  // d3/2
        0.690988298942670958,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
       -0.345494149471335479,
        0,
        0,
        0,
        0,
        0,
        0.598413420602149016,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
       -0.598413420602149016,
        0,
        0,
        0,
        0,
        0,
        0.345494149471335479,
        0,
        0,
        0.690988298942670958,
        0,
        0,
        0,
        0,
        0,  // d5/2
        0,
        0,
        0,
        0,
        0,
        0,
       -0.772548404046379160,
        0,
        0,
        0,
        0,
        0,
       -0.345494149471335479,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
       -0.690988298942670958,
        0,
        0,
        0,
        0,
        0,
       -0.488602511902919921,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
       -0.488602511902919921,
        0,
        0,
        0,
        0,
        0,
       -0.690988298942670958,
        0,
        0,
        0.345494149471335479,
        0,
        0,
        0,
        0,
        0,
        0.772548404046379160,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,  // f5/2
        1.158822606069568741,
        0,
        0,
        0,
        0,
       -0.386274202023189580,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
       -0.772548404046379160,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1.727470747356677396,
        0,
        0,
        0,
        0,
        0,
        0,
        0.172747074735667739,
        0,
        0,
        0,
        0,
        0.172747074735667739,
        0,
       -0.690988298942670958,
        0,
        0,
       -0.244301255951459960,
        0,
        0,
        0,
        0,
       -0.244301255951459960,
        0,
        0.977205023805839843,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0.244301255951459960,
        0,
        0,
        0,
        0,
        0.244301255951459960,
        0,
       -0.977205023805839843,
        0,
        0,
       -0.172747074735667739,
        0,
        0,
        0,
        0,
       -0.172747074735667739,
        0,
        0.690988298942670958,
        0,
        0,
        0,
        0,
        0,
        1.727470747356677396,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
       -0.772548404046379160,
        0,
        0,
        0,
        0,
        0,
        0,
       -1.158822606069568741,
        0,
        0,
        0,
        0,
        0.386274202023189580,
        0,
        0,
        0,
        0,  // f7/2
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
       -1.251671470898352269,
        0,
        0,
        0,
        0,
        0.417223823632784089,
        0,
        0,
        0,
        0,
       -0.473087347878780009,
        0,
        0,
        0,
        0,
        0.157695782626260003,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
       -1.892349391515120036,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
       -1.092548430592079070,
        0,
        0,
        0,
        0,
        0,
        0,
        0.273137107648019767,
        0,
        0,
        0,
        0,
        0.273137107648019767,
        0,
       -1.092548430592079070,
        0,
        0,
        0.211571093830408607,
        0,
        0,
        0,
        0,
        0.211571093830408607,
        0,
       -0.846284375321634430,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0.211571093830408607,
        0,
        0,
        0,
        0,
        0.211571093830408607,
        0,
       -0.846284375321634430,
        0,
        0,
        0.273137107648019767,
        0,
        0,
        0,
        0,
        0.273137107648019767,
        0,
       -1.092548430592079070,
        0,
        0,
        0,
        0,
        0,
        1.092548430592079070,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1.892349391515120036,
        0,
        0,
        0,
        0,
        0,
        0,
       -0.473087347878780009,
        0,
        0,
        0,
        0,
        0.157695782626260003,
        0,
        0,
        0,
        0,
       -1.251671470898352269,
        0,
        0,
        0,
        0,
        0.417223823632784089,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        // g7/2, -7/2
        0, 1.668895294531136358, 0, 0, 0, 0, -1.668895294531136358, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -1.251671470898352269, 0, 0, 0, 0, 0, 0, 0.417223823632784089, 0, 0, 0,
        // g7/2, -5/2
        0, 0, 0, 0, 3.311611435151460063, 0, 0, 0, 0, 0, 0, -1.103870478383820021, 0, 0, 0,
        0, 0.315391565252520006, 0, 0, 0, 0, 0.315391565252520006, 0, -1.892349391515120036, 0, 0, 0, 0, 0, 0,
        // g7/2, -3/2
        0, -0.546274215296039535, 0, 0, 0, 0, -0.546274215296039535, 0, 3.277645291776237211, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0.819411322944059302, 0, 0, 0, 0, 0, 0, 0.819411322944059302, 0, -1.092548430592079070, 0,
        // g7/2, -1/2
        0, 0, 0, 0, -1.057855469152043038, 0, 0, 0, 0, 0, 0, -1.057855469152043038, 0, 1.410473958869390717, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // g7/2, 1/2
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 1.057855469152043038, 0, 0, 0, 0, 0, 0, 1.057855469152043038, 0, -1.410473958869390717, 0,
        // g7/2, 3/2
        0, 0, 0, 0, -0.819411322944059302, 0, 0, 0, 0, 0, 0, -0.819411322944059302, 0, 1.092548430592079070, 0,
        0, -0.546274215296039535, 0, 0, 0, 0, -0.546274215296039535, 0, 3.277645291776237211, 0, 0, 0, 0, 0, 0,
        // g7/2, 5/2
        0, 0.315391565252520006, 0, 0, 0, 0, 0.315391565252520006, 0, -1.892349391515120036, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -3.311611435151460063, 0, 0, 0, 0, 0, 0, 1.103870478383820021, 0, 0, 0,
        // g7/2, 7/2
        0, 0, 0, 0, 1.251671470898352269, 0, 0, 0, 0, 0, 0, -0.417223823632784089, 0, 0, 0,
        0, 1.668895294531136358, 0, 0, 0, 0, -1.668895294531136358, 0, 0, 0, 0, 0, 0, 0, 0,
        // g9/2, -9/2
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -1.770130769779930531, 0, 0, 0, 0, 1.770130769779930531, 0, 0, 0, 0, 0, 0, 0, 0,
        // g9/2, -7/2
        0, -0.590043589926643510, 0, 0, 0, 0, 0.590043589926643510, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -3.540261539559861062, 0, 0, 0, 0, 0, 0, 1.180087179853287020, 0, 0, 0,
        // g9/2, -5/2
        0, 0, 0, 0, -1.770130769779930531, 0, 0, 0, 0, 0, 0, 0.590043589926643510, 0, 0, 0,
        0, 0.590043589926643510, 0, 0, 0, 0, 0.590043589926643510, 0, -3.540261539559861062, 0, 0, 0, 0, 0, 0,
        // g9/2, -3/2
        0, 0.386274202023189580, 0, 0, 0, 0, 0.386274202023189580, 0, -2.317645212139137482, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 1.158822606069568741, 0, 0, 0, 0, 0, 0, 1.158822606069568741, 0, -1.545096808092758321, 0,
        // g9/2, -1/2
        0, 0, 0, 0, 0.946174695757560018, 0, 0, 0, 0, 0, 0, 0.946174695757560018, 0, -1.261566261010080024, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // g9/2, 1/2
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0.946174695757560018, 0, 0, 0, 0, 0, 0, 0.946174695757560018, 0, -1.261566261010080024, 0,
        // g9/2, 3/2
        0, 0, 0, 0, 1.158822606069568741, 0, 0, 0, 0, 0, 0, 1.158822606069568741, 0, -1.545096808092758321, 0,
        0, -0.386274202023189580, 0, 0, 0, 0, -0.386274202023189580, 0, 2.317645212139137482, 0, 0, 0, 0, 0, 0,
        // g9/2, 5/2
        0, -0.590043589926643510, 0, 0, 0, 0, -0.590043589926643510, 0, 3.540261539559861062, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -1.770130769779930531, 0, 0, 0, 0, 0, 0, 0.590043589926643510, 0, 0, 0,
        // g9/2, 7/2
        0, 0, 0, 0, -3.540261539559861062, 0, 0, 0, 0, 0, 0, 1.180087179853287020, 0, 0, 0,
        0, 0.590043589926643510, 0, 0, 0, 0, -0.590043589926643510, 0, 0, 0, 0, 0, 0, 0, 0,
        // g9/2, 9/2
        0, 1.770130769779930531, 0, 0, 0, 0, -1.770130769779930531, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // h9/2, -9/2
        0, 2.212663462224913163, 0, 0, 0, 0, -4.425326924449826327, 0, 0, 0, 0, 0, 0, 0, 0, 0.442532692444982632, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -1.770130769779930531, 0, 0, 0, 0, 0, 0, 1.770130769779930531, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // h9/2, -7/2
        0, 0, 0, 0, 5.310392309339791593, 0, 0, 0, 0, 0, 0, -5.310392309339791593, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0.442532692444982632, 0, 0, 0, 0, 0.295021794963321755, 0, -3.540261539559861062, 0, 0, 0, 0, 0, 0, -0.147510897481660877, 0, 1.180087179853287020, 0, 0, 0,
        // h9/2, -5/2
        0, -0.885065384889965265, 0, 0, 0, 0, -0.590043589926643510, 0, 7.080523079119722124, 0, 0, 0, 0, 0, 0, 0.295021794963321755, 0, -2.360174359706574041, 0, 0, 0,
        0, 0, 0, 0, 1.770130769779930531, 0, 0, 0, 0, 0, 0, 1.770130769779930531, 0, -3.540261539559861062, 0, 0, 0, 0, 0, 0, 0,
        // h9/2, -3/2
        0, 0, 0, 0, -2.703919414162327062, 0, 0, 0, 0, 0, 0, -2.703919414162327062, 0, 5.407838828324654124, 0, 0, 0, 0, 0, 0, 0,
        0, -0.193137101011594790, 0, 0, 0, 0, -0.386274202023189580, 0, 2.317645212139137482, 0, 0, 0, 0, 0, 0, -0.193137101011594790, 0, 2.317645212139137482, 0, -1.545096808092758321, 0,
        // h9/2, -1/2
        0, 0.236543673939390004, 0, 0, 0, 0, 0.473087347878780009, 0, -2.838524087272680054, 0, 0, 0, 0, 0, 0, 0.236543673939390004, 0, -2.838524087272680054, 0, 1.892349391515120036, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // h9/2, 1/2
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -0.236543673939390004, 0, 0, 0, 0, -0.473087347878780009, 0, 2.838524087272680054, 0, 0, 0, 0, 0, 0, -0.236543673939390004, 0, 2.838524087272680054, 0, -1.892349391515120036, 0,
        // h9/2, 3/2
        0, 0.193137101011594790, 0, 0, 0, 0, 0.386274202023189580, 0, -2.317645212139137482, 0, 0, 0, 0, 0, 0, 0.193137101011594790, 0, -2.317645212139137482, 0, 1.545096808092758321, 0,
        0, 0, 0, 0, -2.703919414162327062, 0, 0, 0, 0, 0, 0, -2.703919414162327062, 0, 5.407838828324654124, 0, 0, 0, 0, 0, 0, 0,
        // h9/2, 5/2
        0, 0, 0, 0, 1.770130769779930531, 0, 0, 0, 0, 0, 0, 1.770130769779930531, 0, -3.540261539559861062, 0, 0, 0, 0, 0, 0, 0,
        0, 0.885065384889965265, 0, 0, 0, 0, 0.590043589926643510, 0, -7.080523079119722124, 0, 0, 0, 0, 0, 0, -0.295021794963321755, 0, 2.360174359706574041, 0, 0, 0,
        // h9/2, 7/2
        0, -0.442532692444982632, 0, 0, 0, 0, -0.295021794963321755, 0, 3.540261539559861062, 0, 0, 0, 0, 0, 0, 0.147510897481660877, 0, -1.180087179853287020, 0, 0, 0,
        0, 0, 0, 0, 5.310392309339791593, 0, 0, 0, 0, 0, 0, -5.310392309339791593, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // h9/2, 9/2
        0, 0, 0, 0, -1.770130769779930531, 0, 0, 0, 0, 0, 0, 1.770130769779930531, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -2.212663462224913163, 0, 0, 0, 0, 4.425326924449826327, 0, 0, 0, 0, 0, 0, 0, 0, -0.442532692444982632, 0, 0, 0, 0, 0,
        // h11/2, -11/2
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -2.320661017204290803, 0, 0, 0, 0, 4.641322034408581606, 0, 0, 0, 0, 0, 0, 0, 0, -0.464132203440858160, 0, 0, 0, 0, 0,
        // h11/2, -9/2
        0, -0.699705623606466358, 0, 0, 0, 0, 1.399411247212932717, 0, 0, 0, 0, 0, 0, 0, 0, -0.139941124721293271, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -5.597644988851730871, 0, 0, 0, 0, 0, 0, 5.597644988851730871, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // h11/2, -7/2
        0, 0, 0, 0, -2.503342941796704538, 0, 0, 0, 0, 0, 0, 2.503342941796704538, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0.938753603173764201, 0, 0, 0, 0, 0.625835735449176134, 0, -7.510028825390113615, 0, 0, 0, 0, 0, 0, -0.312917867724588067, 0, 2.503342941796704538, 0, 0, 0,
        // h11/2, -5/2
        0, 0.541989645495103885, 0, 0, 0, 0, 0.361326430330069256, 0, -4.335917163960831083, 0, 0, 0, 0, 0, 0, -0.180663215165034628, 0, 1.445305721320277027, 0, 0, 0,
        0, 0, 0, 0, 2.890611442640554055, 0, 0, 0, 0, 0, 0, 2.890611442640554055, 0, -5.781222885281108110, 0, 0, 0, 0, 0, 0, 0,
        // h11/2, -3/2
        0, 0, 0, 0, 2.043970952866564726, 0, 0, 0, 0, 0, 0, 2.043970952866564726, 0, -4.087941905733129453, 0, 0, 0, 0, 0, 0, 0,
        0, -0.255496369108320590, 0, 0, 0, 0, -0.510992738216641181, 0, 3.065956429299847090, 0, 0, 0, 0, 0, 0, -0.255496369108320590, 0, 3.065956429299847090, 0, -2.043970952866564726, 0,
        // h11/2, -1/2
        0, -0.215933843419584674, 0, 0, 0, 0, -0.431867686839169349, 0, 2.591206121035016094, 0, 0, 0, 0, 0, 0, -0.215933843419584674, 0, 2.591206121035016094, 0, -1.727470747356677396, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // h11/2, 1/2
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -0.215933843419584674, 0, 0, 0, 0, -0.431867686839169349, 0, 2.591206121035016094, 0, 0, 0, 0, 0, 0, -0.215933843419584674, 0, 2.591206121035016094, 0, -1.727470747356677396, 0,
        // h11/2, 3/2
        0, -0.255496369108320590, 0, 0, 0, 0, -0.510992738216641181, 0, 3.065956429299847090, 0, 0, 0, 0, 0, 0, -0.255496369108320590, 0, 3.065956429299847090, 0, -2.043970952866564726, 0,
        0, 0, 0, 0, -2.043970952866564726, 0, 0, 0, 0, 0, 0, -2.043970952866564726, 0, 4.087941905733129453, 0, 0, 0, 0, 0, 0, 0,
        // h11/2, 5/2
        0, 0, 0, 0, -2.890611442640554055, 0, 0, 0, 0, 0, 0, -2.890611442640554055, 0, 5.781222885281108110, 0, 0, 0, 0, 0, 0, 0,
        0, 0.541989645495103885, 0, 0, 0, 0, 0.361326430330069256, 0, -4.335917163960831083, 0, 0, 0, 0, 0, 0, -0.180663215165034628, 0, 1.445305721320277027, 0, 0, 0,
        // h11/2, 7/2
        0, 0.938753603173764201, 0, 0, 0, 0, 0.625835735449176134, 0, -7.510028825390113615, 0, 0, 0, 0, 0, 0, -0.312917867724588067, 0, 2.503342941796704538, 0, 0, 0,
        0, 0, 0, 0, 2.503342941796704538, 0, 0, 0, 0, 0, 0, -2.503342941796704538, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // h11/2, 9/2
        0, 0, 0, 0, 5.597644988851730871, 0, 0, 0, 0, 0, 0, -5.597644988851730871, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -0.699705623606466358, 0, 0, 0, 0, 1.399411247212932717, 0, 0, 0, 0, 0, 0, 0, 0, -0.139941124721293271, 0, 0, 0, 0, 0,
        // h11/2, 11/2
        0, -2.320661017204290803, 0, 0, 0, 0, 4.641322034408581606, 0, 0, 0, 0, 0, 0, 0, 0, -0.464132203440858160, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //i11/2,-11/2
        0, 2.7847932206451498, 0, 0, 0, 0, -9.2826440688171665, 0, 0, 0, 0, 0, 0, 0, 0, 2.7847932206451498, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -2.3206610172042916, 0, 0, 0, 0, 0, 0, 4.6413220344085833, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.4641322034408583, 0, 0, 0, 0, 0,
        //i11/2,-9/2
        0, 0, 0, 0, 7.6967618596711320, 0, 0, 0, 0, 0, 0, -15.3935237193422640, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.5393523719342264, 0, 0, 0, 0, 0,
        0, 0.5597644988851730, 0, 0, 0, 0, 0, 0, -5.5976449888517319, 0, 0, 0, 0, 0, 0, -0.5597644988851730, 0, 5.5976449888517319, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //i11/2,-7/2
        0, -1.2516714708983521, 0, 0, 0, 0, 0, 0, 12.5167147089835247, 0, 0, 0, 0, 0, 0, 1.2516714708983521, 0, -12.5167147089835247, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 2.8162608095212929, 0, 0, 0, 0, 0, 0, 1.8775072063475287, 0, -7.5100288253901146, 0, 0, 0, 0, 0, 0, 0, 0, -0.9387536031737643, 0, 2.5033429417967050, 0, 0, 0,
        //i11/2,-5/2
        0, 0, 0, 0, -4.8779068094559346, 0, 0, 0, 0, 0, 0, -3.2519378729706236, 0, 13.0077514918824946, 0, 0, 0, 0, 0, 0, 0, 0, 1.6259689364853118, 0, -4.3359171639608318, 0, 0, 0,
        0, -0.3613264303300693, 0, 0, 0, 0, -0.7226528606601387, 0, 5.7812228852811094, 0, 0, 0, 0, 0, 0, -0.3613264303300693, 0, 5.7812228852811094, 0, -5.7812228852811094, 0, 0, 0, 0, 0, 0, 0, 0,
        //i11/2,-3/2
        0, 0.5109927382166413, 0, 0, 0, 0, 1.0219854764332825, 0, -8.1758838114662602, 0, 0, 0, 0, 0, 0, 0.5109927382166413, 0, -8.1758838114662602, 0, 8.1758838114662602, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -1.2774818455416030, 0, 0, 0, 0, 0, 0, -2.5549636910832061, 0, 5.1099273821664122, 0, 0, 0, 0, 0, 0, 0, 0, -1.2774818455416030, 0, 5.1099273821664122, 0, -2.0439709528665646, 0,
        //i11/2,-1/2
        0, 0, 0, 0, 1.5115369039370925, 0, 0, 0, 0, 0, 0, 3.0230738078741850, 0, -6.0461476157483700, 0, 0, 0, 0, 0, 0, 0, 0, 1.5115369039370925, 0, -6.0461476157483700, 0, 2.4184590462993478, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //i11/2,1/2
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -1.5115369039370925, 0, 0, 0, 0, 0, 0, -3.0230738078741850, 0, 6.0461476157483700, 0, 0, 0, 0, 0, 0, 0, 0, -1.5115369039370925, 0, 6.0461476157483700, 0, -2.4184590462993478, 0,
        //i11/2,3/2
        0, 0, 0, 0, 1.2774818455416030, 0, 0, 0, 0, 0, 0, 2.5549636910832061, 0, -5.1099273821664122, 0, 0, 0, 0, 0, 0, 0, 0, 1.2774818455416030, 0, -5.1099273821664122, 0, 2.0439709528665646, 0,
        0, 0.5109927382166413, 0, 0, 0, 0, 1.0219854764332825, 0, -8.1758838114662602, 0, 0, 0, 0, 0, 0, 0.5109927382166413, 0, -8.1758838114662602, 0, 8.1758838114662602, 0, 0, 0, 0, 0, 0, 0, 0,
        //i11/2,5/2
        0, -0.3613264303300693, 0, 0, 0, 0, -0.7226528606601387, 0, 5.7812228852811094, 0, 0, 0, 0, 0, 0, -0.3613264303300693, 0, 5.7812228852811094, 0, -5.7812228852811094, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 4.8779068094559346, 0, 0, 0, 0, 0, 0, 3.2519378729706236, 0, -13.0077514918824946, 0, 0, 0, 0, 0, 0, 0, 0, -1.6259689364853118, 0, 4.3359171639608318, 0, 0, 0,
        //i11/2,7/2
        0, 0, 0, 0, -2.8162608095212929, 0, 0, 0, 0, 0, 0, -1.8775072063475287, 0, 7.5100288253901146, 0, 0, 0, 0, 0, 0, 0, 0, 0.9387536031737643, 0, -2.5033429417967050, 0, 0, 0,
        0, -1.2516714708983521, 0, 0, 0, 0, 0, 0, 12.5167147089835247, 0, 0, 0, 0, 0, 0, 1.2516714708983521, 0, -12.5167147089835247, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //i11/2,9/2
        0, 0.5597644988851730, 0, 0, 0, 0, 0, 0, -5.5976449888517319, 0, 0, 0, 0, 0, 0, -0.5597644988851730, 0, 5.5976449888517319, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -7.6967618596711320, 0, 0, 0, 0, 0, 0, 15.3935237193422640, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.5393523719342264, 0, 0, 0, 0, 0,
        //i11/2,11/2
        0, 0, 0, 0, 2.3206610172042916, 0, 0, 0, 0, 0, 0, -4.6413220344085833, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.4641322034408583, 0, 0, 0, 0, 0,
        0, 2.7847932206451498, 0, 0, 0, 0, -9.2826440688171665, 0, 0, 0, 0, 0, 0, 0, 0, 2.7847932206451498, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //i13/2,-13/2
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -2.8985046814803979, 0, 0, 0, 0, 9.6616822716013271, 0, 0, 0, 0, 0, 0, 0, 0, -2.8985046814803979, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //i13/2,-11/2
        0, -0.8039005577884609, 0, 0, 0, 0, 2.6796685259615369, 0, 0, 0, 0, 0, 0, 0, 0, -0.8039005577884609, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -8.0390055778846108, 0, 0, 0, 0, 0, 0, 16.0780111557692216, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.6078011155769223, 0, 0, 0, 0, 0,
        //i13/2,-9/2
        0, 0, 0, 0, -3.2819102842008516, 0, 0, 0, 0, 0, 0, 6.5638205684017032, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.6563820568401703, 0, 0, 0, 0, 0,
        0, 1.3127641136803401, 0, 0, 0, 0, 0, 0, -13.1276411368034029, 0, 0, 0, 0, 0, 0, -1.3127641136803401, 0, 13.1276411368034029, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //i13/2,-7/2
        0, 0.6855686991966986, 0, 0, 0, 0, 0, 0, -6.8556869919669872, 0, 0, 0, 0, 0, 0, -0.6855686991966986, 0, 6.8556869919669872, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 5.1417652439752395, 0, 0, 0, 0, 0, 0, 3.4278434959834931, 0, -13.7113739839339726, 0, 0, 0, 0, 0, 0, 0, 0, -1.7139217479917466, 0, 4.5704579946446584, 0, 0, 0,
        //i13/2,-5/2
        0, 0, 0, 0, 3.2519378729706232, 0, 0, 0, 0, 0, 0, 2.1679585819804159, 0, -8.6718343279216636, 0, 0, 0, 0, 0, 0, 0, 0, -1.0839792909902080, 0, 2.8906114426405547, 0, 0, 0,
        0, -0.5419896454951040, 0, 0, 0, 0, -1.0839792909902080, 0, 8.6718343279216636, 0, 0, 0, 0, 0, 0, -0.5419896454951040, 0, 8.6718343279216636, 0, -8.6718343279216636, 0, 0, 0, 0, 0, 0, 0, 0,
        //i13/2,-3/2
        0, -0.4039752301426884, 0, 0, 0, 0, -0.8079504602853768, 0, 6.4636036822830141, 0, 0, 0, 0, 0, 0, -0.4039752301426884, 0, 6.4636036822830141, 0, -6.4636036822830141, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -1.6159009205707533, 0, 0, 0, 0, 0, 0, -3.2318018411415066, 0, 6.4636036822830132, 0, 0, 0, 0, 0, 0, 0, 0, -1.6159009205707533, 0, 6.4636036822830132, 0, -2.5854414729132049, 0,
        //i13/2,-1/2
        0, 0, 0, 0, -1.3994112472129328, 0, 0, 0, 0, 0, 0, -2.7988224944258655, 0, 5.5976449888517310, 0, 0, 0, 0, 0, 0, 0, 0, -1.3994112472129328, 0, 5.5976449888517310, 0, -2.2390579955406920, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //i13/2,1/2
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -1.3994112472129328, 0, 0, 0, 0, 0, 0, -2.7988224944258655, 0, 5.5976449888517310, 0, 0, 0, 0, 0, 0, 0, 0, -1.3994112472129328, 0, 5.5976449888517310, 0, -2.2390579955406920, 0,
        //i13/2,3/2
        0, 0, 0, 0, -1.6159009205707533, 0, 0, 0, 0, 0, 0, -3.2318018411415066, 0, 6.4636036822830132, 0, 0, 0, 0, 0, 0, 0, 0, -1.6159009205707533, 0, 6.4636036822830132, 0, -2.5854414729132049, 0,
        0, 0.4039752301426884, 0, 0, 0, 0, 0.8079504602853768, 0, -6.4636036822830141, 0, 0, 0, 0, 0, 0, 0.4039752301426884, 0, -6.4636036822830141, 0, 6.4636036822830141, 0, 0, 0, 0, 0, 0, 0, 0,
        //i13/2,5/2
        0, 0.5419896454951040, 0, 0, 0, 0, 1.0839792909902080, 0, -8.6718343279216636, 0, 0, 0, 0, 0, 0, 0.5419896454951040, 0, -8.6718343279216636, 0, 8.6718343279216636, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 3.2519378729706232, 0, 0, 0, 0, 0, 0, 2.1679585819804159, 0, -8.6718343279216636, 0, 0, 0, 0, 0, 0, 0, 0, -1.0839792909902080, 0, 2.8906114426405547, 0, 0, 0,
        //i13/2,7/2
        0, 0, 0, 0, 5.1417652439752395, 0, 0, 0, 0, 0, 0, 3.4278434959834931, 0, -13.7113739839339726, 0, 0, 0, 0, 0, 0, 0, 0, -1.7139217479917466, 0, 4.5704579946446584, 0, 0, 0,
        0, -0.6855686991966986, 0, 0, 0, 0, 0, 0, 6.8556869919669872, 0, 0, 0, 0, 0, 0, 0.6855686991966986, 0, -6.8556869919669872, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //i13/2,9/2
        0, -1.3127641136803401, 0, 0, 0, 0, 0, 0, 13.1276411368034029, 0, 0, 0, 0, 0, 0, 1.3127641136803401, 0, -13.1276411368034029, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -3.2819102842008516, 0, 0, 0, 0, 0, 0, 6.5638205684017032, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.6563820568401703, 0, 0, 0, 0, 0,
        //i13/2,11/2
        0, 0, 0, 0, -8.0390055778846108, 0, 0, 0, 0, 0, 0, 16.0780111557692216, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.6078011155769223, 0, 0, 0, 0, 0,
        0, 0.8039005577884609, 0, 0, 0, 0, -2.6796685259615369, 0, 0, 0, 0, 0, 0, 0, 0, 0.8039005577884609, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //i13/2,13/2
        0, 2.8985046814803979, 0, 0, 0, 0, -9.6616822716013271, 0, 0, 0, 0, 0, 0, 0, 0, 2.8985046814803979, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //j13/2,-13/2
        0, 3.3815887950604644, 0, 0, 0, 0, -16.9079439753023237, 0, 0, 0, 0, 0, 0, 0, 0, 10.1447663851813932, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.4830841135800663, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -2.8985046814803970, 0, 0, 0, 0, 0, 0, 9.6616822716013235, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.8985046814803970, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //j13/2,-11/2
        0, 0, 0, 0, 10.4507072512499910, 0, 0, 0, 0, 0, 0, -34.8356908374999676, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 10.4507072512499910, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0.6699171314903840, 0, 0, 0, 0, -0.6699171314903840, 0, -8.0390055778846072, 0, 0, 0, 0, 0, 0, -1.2058508366826914, 0, 16.0780111557692145, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.1339834262980768, 0, -1.6078011155769216, 0, 0, 0, 0, 0,
        //j13/2,-9/2
        0, -1.6409551421004256, 0, 0, 0, 0, 1.6409551421004256, 0, 19.6914617052051071, 0, 0, 0, 0, 0, 0, 2.9537192557807663, 0, -39.3829234104102142, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.3281910284200851, 0, 3.9382923410410213, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 3.9382923410410213, 0, 0, 0, 0, 0, 0, 0, 0, -13.1276411368034047, 0, 0, 0, 0, 0, 0, 0, 0, -3.9382923410410213, 0, 13.1276411368034047, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //j13/2,-7/2
        0, 0, 0, 0, -7.5412556911636850, 0, 0, 0, 0, 0, 0, 0, 0, 25.1375189705456172, 0, 0, 0, 0, 0, 0, 0, 0, 7.5412556911636850, 0, -25.1375189705456172, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -0.5141765243975239, 0, 0, 0, 0, -0.8569608739958733, 0, 10.2835304879504772, 0, 0, 0, 0, 0, 0, -0.1713921747991747, 0, 6.8556869919669845, 0, -13.7113739839339726, 0, 0, 0, 0, 0, 0, 0, 0, 0.1713921747991747, 0, -3.4278434959834923, 0, 4.5704579946446575, 0, 0, 0,
        //j13/2,-5/2
        0, 0.8129844682426559, 0, 0, 0, 0, 1.3549741137377600, 0, -16.2596893648531164, 0, 0, 0, 0, 0, 0, 0.2709948227475520, 0, -10.8397929099020782, 0, 21.6795858198041600, 0, 0, 0, 0, 0, 0, 0, 0, -0.2709948227475520, 0, 5.4198964549510391, 0, -7.2265286066013861, 0, 0, 0,
        0, 0, 0, 0, -2.7099482274755200, 0, 0, 0, 0, 0, 0, -5.4198964549510400, 0, 14.4530572132027721, 0, 0, 0, 0, 0, 0, 0, 0, -2.7099482274755200, 0, 14.4530572132027721, 0, -8.6718343279216654, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //j13/2,-3/2
        0, 0, 0, 0, 3.6357770712841964, 0, 0, 0, 0, 0, 0, 7.2715541425683927, 0, -19.3908110468490449, 0, 0, 0, 0, 0, 0, 0, 0, 3.6357770712841964, 0, -19.3908110468490449, 0, 11.6344866281094284, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0.2019876150713442, 0, 0, 0, 0, 0.6059628452140327, 0, -4.8477027617122603, 0, 0, 0, 0, 0, 0, 0.6059628452140327, 0, -9.6954055234245207, 0, 9.6954055234245207, 0, 0, 0, 0, 0, 0, 0, 0, 0.2019876150713442, 0, -4.8477027617122603, 0, 9.6954055234245207, 0, -2.5854414729132063, 0,
        //j13/2,-1/2
        0, -0.2332352078688221, 0, 0, 0, 0, -0.6997056236064664, 0, 5.5976449888517301, 0, 0, 0, 0, 0, 0, -0.6997056236064664, 0, 11.1952899777034602, 0, -11.1952899777034602, 0, 0, 0, 0, 0, 0, 0, 0, -0.2332352078688221, 0, 5.5976449888517301, 0, -11.1952899777034602, 0, 2.9854106607209236, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //j13/2,1/2
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0.2332352078688221, 0, 0, 0, 0, 0.6997056236064664, 0, -5.5976449888517301, 0, 0, 0, 0, 0, 0, 0.6997056236064664, 0, -11.1952899777034602, 0, 11.1952899777034602, 0, 0, 0, 0, 0, 0, 0, 0, 0.2332352078688221, 0, -5.5976449888517301, 0, 11.1952899777034602, 0, -2.9854106607209236, 0,
        //j13/2,3/2
        0, -0.2019876150713442, 0, 0, 0, 0, -0.6059628452140327, 0, 4.8477027617122603, 0, 0, 0, 0, 0, 0, -0.6059628452140327, 0, 9.6954055234245207, 0, -9.6954055234245207, 0, 0, 0, 0, 0, 0, 0, 0, -0.2019876150713442, 0, 4.8477027617122603, 0, -9.6954055234245207, 0, 2.5854414729132063, 0,
        0, 0, 0, 0, 3.6357770712841964, 0, 0, 0, 0, 0, 0, 7.2715541425683927, 0, -19.3908110468490449, 0, 0, 0, 0, 0, 0, 0, 0, 3.6357770712841964, 0, -19.3908110468490449, 0, 11.6344866281094284, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //j13/2,5/2
        0, 0, 0, 0, -2.7099482274755200, 0, 0, 0, 0, 0, 0, -5.4198964549510400, 0, 14.4530572132027721, 0, 0, 0, 0, 0, 0, 0, 0, -2.7099482274755200, 0, 14.4530572132027721, 0, -8.6718343279216654, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -0.8129844682426559, 0, 0, 0, 0, -1.3549741137377600, 0, 16.2596893648531164, 0, 0, 0, 0, 0, 0, -0.2709948227475520, 0, 10.8397929099020782, 0, -21.6795858198041600, 0, 0, 0, 0, 0, 0, 0, 0, 0.2709948227475520, 0, -5.4198964549510391, 0, 7.2265286066013861, 0, 0, 0,
        //j13/2,7/2
        0, 0.5141765243975239, 0, 0, 0, 0, 0.8569608739958733, 0, -10.2835304879504772, 0, 0, 0, 0, 0, 0, 0.1713921747991747, 0, -6.8556869919669845, 0, 13.7113739839339726, 0, 0, 0, 0, 0, 0, 0, 0, -0.1713921747991747, 0, 3.4278434959834923, 0, -4.5704579946446575, 0, 0, 0,
        0, 0, 0, 0, -7.5412556911636850, 0, 0, 0, 0, 0, 0, 0, 0, 25.1375189705456172, 0, 0, 0, 0, 0, 0, 0, 0, 7.5412556911636850, 0, -25.1375189705456172, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //j13/2,9/2
        0, 0, 0, 0, 3.9382923410410213, 0, 0, 0, 0, 0, 0, 0, 0, -13.1276411368034047, 0, 0, 0, 0, 0, 0, 0, 0, -3.9382923410410213, 0, 13.1276411368034047, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 1.6409551421004256, 0, 0, 0, 0, -1.6409551421004256, 0, -19.6914617052051071, 0, 0, 0, 0, 0, 0, -2.9537192557807663, 0, 39.3829234104102142, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.3281910284200851, 0, -3.9382923410410213, 0, 0, 0, 0, 0,
        //j13/2,11/2
        0, -0.6699171314903840, 0, 0, 0, 0, 0.6699171314903840, 0, 8.0390055778846072, 0, 0, 0, 0, 0, 0, 1.2058508366826914, 0, -16.0780111557692145, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.1339834262980768, 0, 1.6078011155769216, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 10.4507072512499910, 0, 0, 0, 0, 0, 0, -34.8356908374999676, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 10.4507072512499910, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //j13/2,13/2
        0, 0, 0, 0, -2.8985046814803970, 0, 0, 0, 0, 0, 0, 9.6616822716013235, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.8985046814803970, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -3.3815887950604644, 0, 0, 0, 0, 16.9079439753023237, 0, 0, 0, 0, 0, 0, 0, 0, -10.1447663851813932, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.4830841135800663, 0, 0, 0, 0, 0, 0, 0,
        //j15/2,-15/2
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -3.5002769449938556, 0, 0, 0, 0, 17.5013847249692773, 0, 0, 0, 0, 0, 0, 0, 0, -10.5008308349815653, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.5000395635705508, 0, 0, 0, 0, 0, 0, 0,
        //j15/2,-13/2
        0, -0.9037676210049984, 0, 0, 0, 0, 4.5188381050249919, 0, 0, 0, 0, 0, 0, 0, 0, -2.7113028630149949, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.1291096601435712, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -10.8452114520599796, 0, 0, 0, 0, 0, 0, 36.1507048401999356, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -10.8452114520599796, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //j15/2,-11/2
        0, 0, 0, 0, -4.0991046311514854, 0, 0, 0, 0, 0, 0, 13.6636821038382852, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4.0991046311514854, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 1.7079602629797861, 0, 0, 0, 0, -1.7079602629797861, 0, -20.4955231557574322, 0, 0, 0, 0, 0, 0, -3.0743284733636154, 0, 40.9910463115148644, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.3415920525959572, 0, -4.0991046311514863, 0, 0, 0, 0, 0,
        //j15/2,-9/2
        0, 0.8204775710502128, 0, 0, 0, 0, -0.8204775710502128, 0, -9.8457308526025535, 0, 0, 0, 0, 0, 0, -1.4768596278903832, 0, 19.6914617052051071, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.1640955142100426, 0, -1.9691461705205107, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 7.8765846820820427, 0, 0, 0, 0, 0, 0, 0, 0, -26.2552822736068094, 0, 0, 0, 0, 0, 0, 0, 0, -7.8765846820820427, 0, 26.2552822736068094, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //j15/2,-7/2
        0, 0, 0, 0, 4.5475482864949495, 0, 0, 0, 0, 0, 0, 0, 0, -15.1584942883164988, 0, 0, 0, 0, 0, 0, 0, 0, -4.5475482864949495, 0, 15.1584942883164988, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -0.8526653037178029, 0, 0, 0, 0, -1.4211088395296716, 0, 17.0533060743560583, 0, 0, 0, 0, 0, 0, -0.2842217679059343, 0, 11.3688707162373710, 0, -22.7377414324747456, 0, 0, 0, 0, 0, 0, 0, 0, 0.2842217679059343, 0, -5.6844353581186855, 0, 7.5792471441582485, 0, 0, 0,
        //j15/2,-5/2
        0, -0.5748668304937213, 0, 0, 0, 0, -0.9581113841562022, 0, 11.4973366098744254, 0, 0, 0, 0, 0, 0, -0.1916222768312404, 0, 7.6648910732496169, 0, -15.3297821464992357, 0, 0, 0, 0, 0, 0, 0, 0, 0.1916222768312404, 0, -3.8324455366248085, 0, 5.1099273821664122, 0, 0, 0,
        0, 0, 0, 0, -3.8324455366248107, 0, 0, 0, 0, 0, 0, -7.6648910732496214, 0, 20.4397095286656523, 0, 0, 0, 0, 0, 0, 0, 0, -3.8324455366248107, 0, 20.4397095286656523, 0, -12.2638257171993938, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //j15/2,-3/2
        0, 0, 0, 0, -2.9685995477189673, 0, 0, 0, 0, 0, 0, -5.9371990954379346, 0, 15.8325309211678249, 0, 0, 0, 0, 0, 0, 0, 0, -2.9685995477189673, 0, 15.8325309211678249, 0, -9.4995185527006960, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0.2473832956432473, 0, 0, 0, 0, 0.7421498869297418, 0, -5.9371990954379337, 0, 0, 0, 0, 0, 0, 0.7421498869297418, 0, -11.8743981908758673, 0, 11.8743981908758673, 0, 0, 0, 0, 0, 0, 0, 0, 0.2473832956432473, 0, -5.9371990954379337, 0, 11.8743981908758673, 0, -3.1665061842335653, 0,
        //j15/2,-1/2
        0, 0.2181715595945335, 0, 0, 0, 0, 0.6545146787836006, 0, -5.2361174302688038, 0, 0, 0, 0, 0, 0, 0.6545146787836006, 0, -10.4722348605376077, 0, 10.4722348605376077, 0, 0, 0, 0, 0, 0, 0, 0, 0.2181715595945335, 0, -5.2361174302688038, 0, 10.4722348605376077, 0, -2.7925959628100294, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //j15/2,1/2
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0.2181715595945335, 0, 0, 0, 0, 0.6545146787836006, 0, -5.2361174302688038, 0, 0, 0, 0, 0, 0, 0.6545146787836006, 0, -10.4722348605376077, 0, 10.4722348605376077, 0, 0, 0, 0, 0, 0, 0, 0, 0.2181715595945335, 0, -5.2361174302688038, 0, 10.4722348605376077, 0, -2.7925959628100294, 0,
        //j15/2,3/2
        0, 0.2473832956432473, 0, 0, 0, 0, 0.7421498869297418, 0, -5.9371990954379337, 0, 0, 0, 0, 0, 0, 0.7421498869297418, 0, -11.8743981908758673, 0, 11.8743981908758673, 0, 0, 0, 0, 0, 0, 0, 0, 0.2473832956432473, 0, -5.9371990954379337, 0, 11.8743981908758673, 0, -3.1665061842335653, 0,
        0, 0, 0, 0, 2.9685995477189673, 0, 0, 0, 0, 0, 0, 5.9371990954379346, 0, -15.8325309211678249, 0, 0, 0, 0, 0, 0, 0, 0, 2.9685995477189673, 0, -15.8325309211678249, 0, 9.4995185527006960, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //j15/2,5/2
        0, 0, 0, 0, 3.8324455366248107, 0, 0, 0, 0, 0, 0, 7.6648910732496214, 0, -20.4397095286656523, 0, 0, 0, 0, 0, 0, 0, 0, 3.8324455366248107, 0, -20.4397095286656523, 0, 12.2638257171993938, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -0.5748668304937213, 0, 0, 0, 0, -0.9581113841562022, 0, 11.4973366098744254, 0, 0, 0, 0, 0, 0, -0.1916222768312404, 0, 7.6648910732496169, 0, -15.3297821464992357, 0, 0, 0, 0, 0, 0, 0, 0, 0.1916222768312404, 0, -3.8324455366248085, 0, 5.1099273821664122, 0, 0, 0,
        //j15/2,7/2
        0, -0.8526653037178029, 0, 0, 0, 0, -1.4211088395296716, 0, 17.0533060743560583, 0, 0, 0, 0, 0, 0, -0.2842217679059343, 0, 11.3688707162373710, 0, -22.7377414324747456, 0, 0, 0, 0, 0, 0, 0, 0, 0.2842217679059343, 0, -5.6844353581186855, 0, 7.5792471441582485, 0, 0, 0,
        0, 0, 0, 0, -4.5475482864949495, 0, 0, 0, 0, 0, 0, 0, 0, 15.1584942883164988, 0, 0, 0, 0, 0, 0, 0, 0, 4.5475482864949495, 0, -15.1584942883164988, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //j15/2,9/2
        0, 0, 0, 0, -7.8765846820820427, 0, 0, 0, 0, 0, 0, 0, 0, 26.2552822736068094, 0, 0, 0, 0, 0, 0, 0, 0, 7.8765846820820427, 0, -26.2552822736068094, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0.8204775710502128, 0, 0, 0, 0, -0.8204775710502128, 0, -9.8457308526025535, 0, 0, 0, 0, 0, 0, -1.4768596278903832, 0, 19.6914617052051071, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.1640955142100426, 0, -1.9691461705205107, 0, 0, 0, 0, 0,
        //j15/2,11/2
        0, 1.7079602629797861, 0, 0, 0, 0, -1.7079602629797861, 0, -20.4955231557574322, 0, 0, 0, 0, 0, 0, -3.0743284733636154, 0, 40.9910463115148644, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.3415920525959572, 0, -4.0991046311514863, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 4.0991046311514854, 0, 0, 0, 0, 0, 0, -13.6636821038382852, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4.0991046311514854, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        //j15/2,13/2
        0, 0, 0, 0, 10.8452114520599796, 0, 0, 0, 0, 0, 0, -36.1507048401999356, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 10.8452114520599796, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -0.9037676210049984, 0, 0, 0, 0, 4.5188381050249919, 0, 0, 0, 0, 0, 0, 0, 0, -2.7113028630149949, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.1291096601435712, 0, 0, 0, 0, 0, 0, 0,
        //j15/2,15/2
        0, -3.5002769449938556, 0, 0, 0, 0, 17.5013847249692773, 0, 0, 0, 0, 0, 0, 0, 0, -10.5008308349815653, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.5000395635705508, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 15/2, mj = -15/2
        0, 4.00031650856440531, 0, 0, 0, 0, -28.0022155599508372, 0, 0, 0, 0, 0, 0, 0, 0, 28.0022155599508372, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4.00031650856440531, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -3.50027694499385465, 0, 0, 0, 0, 0, 0, 17.5013847249692732, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -10.5008308349815639, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.500039563570550664, 0, 0, 0, 0, 0, 0, 0,
        // j = 15/2, mj = -13/2
        0, 0, 0, 0, 13.5565143150749737, 0, 0, 0, 0, 0, 0, -67.7825715753748683, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40.669542945224921, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.93664490215356767, 0, 0, 0, 0, 0, 0, 0,
        0, 0.774657960861427067, 0, 0, 0, 0, -1.80753524200999649, 0, -10.8452114520599789, 0, 0, 0, 0, 0, 0, -1.80753524200999649, 0, 36.1507048401999298, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.774657960861427067, 0, -10.8452114520599789, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 15/2, mj = -11/2
        0, -2.04955231557574297, 0, 0, 0, 0, 4.78228873634340025, 0, 28.6937324180604015, 0, 0, 0, 0, 0, 0, 4.78228873634340025, 0, -95.6457747268680051, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.04955231557574297, 0, 28.6937324180604015, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 5.12388078893935741, 0, 0, 0, 0, 0, 0, -5.12388078893935741, 0, -20.4955231557574297, 0, 0, 0, 0, 0, 0, 0, 0, -9.22298542009084335, 0, 40.9910463115148593, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.02477615778787148, 0, -4.09910463115148593, 0, 0, 0, 0, 0,
        // j = 15/2, mj = -9/2
        0, 0, 0, 0, -10.6662084236527642, 0, 0, 0, 0, 0, 0, 10.6662084236527642, 0, 42.6648336946110567, 0, 0, 0, 0, 0, 0, 0, 0, 19.1991751625749755, 0, -85.3296673892221134, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.13324168473055283, 0, 8.53296673892221134, 0, 0, 0, 0, 0,
        0, -0.656382056840170103, 0, 0, 0, 0, -0.656382056840170103, 0, 15.7531693641640825, 0, 0, 0, 0, 0, 0, 0.656382056840170103, 0, 0, 0, -26.2552822736068041, 0, 0, 0, 0, 0, 0, 0, 0, 0.656382056840170103, 0, -15.7531693641640825, 0, 26.2552822736068041, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 15/2, mj = -7/2
        0, 1.13688707162373734, 0, 0, 0, 0, 1.13688707162373734, 0, -27.2852897189696963, 0, 0, 0, 0, 0, 0, -1.13688707162373734, 0, 0, 0, 45.4754828649494938, 0, 0, 0, 0, 0, 0, 0, 0, -1.13688707162373734, 0, 27.2852897189696963, 0, -45.4754828649494938, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -4.26332651858901504, 0, 0, 0, 0, 0, 0, -7.1055441976483584, 0, 28.4221767905934336, 0, 0, 0, 0, 0, 0, 0, 0, -1.42110883952967168, 0, 18.9481178603956224, 0, -22.7377414324747469, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.42110883952967168, 0, -9.47405893019781121, 0, 7.57924714415824896, 0, 0, 0,
        // j = 15/2, mj = -5/2
        0, 0, 0, 0, 6.32353513543093462, 0, 0, 0, 0, 0, 0, 10.5392252257182244, 0, -42.1569009028728975, 0, 0, 0, 0, 0, 0, 0, 0, 2.10784504514364487, 0, -28.104600601915265, 0, 33.725520722298318, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.10784504514364487, 0, 14.0523003009576325, 0, -11.241840240766106, 0, 0, 0,
        0, 0.383244553662480886, 0, 0, 0, 0, 1.14973366098744266, 0, -11.4973366098744266, 0, 0, 0, 0, 0, 0, 1.14973366098744266, 0, -22.9946732197488532, 0, 30.6595642929984709, 0, 0, 0, 0, 0, 0, 0, 0, 0.383244553662480886, 0, -11.4973366098744266, 0, 30.6595642929984709, 0, -12.2638257171993884, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 15/2, mj = -3/2
        0, -0.49476659128649439, 0, 0, 0, 0, -1.48429977385948317, 0, 14.8429977385948317, 0, 0, 0, 0, 0, 0, -1.48429977385948317, 0, 29.6859954771896634, 0, -39.5813273029195512, 0, 0, 0, 0, 0, 0, 0, 0, -0.49476659128649439, 0, 14.8429977385948317, 0, -39.5813273029195512, 0, 15.8325309211678205, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 1.73168306950273036, 0, 0, 0, 0, 0, 0, 5.19504920850819109, 0, -13.8534645560218429, 0, 0, 0, 0, 0, 0, 0, 0, 5.19504920850819109, 0, -27.7069291120436858, 0, 16.6241574672262115, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.73168306950273036, 0, -13.8534645560218429, 0, 16.6241574672262115, 0, -3.16650618423356409, 0,
        // j = 15/2, mj = -1/2
        0, 0, 0, 0, -1.96354403635080146, 0, 0, 0, 0, 0, 0, -5.89063210905240439, 0, 15.7083522908064117, 0, 0, 0, 0, 0, 0, 0, 0, -5.89063210905240439, 0, 31.4167045816128234, 0, -18.850022748967694, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.96354403635080146, 0, 15.7083522908064117, 0, -18.850022748967694, 0, 3.5904805236128941, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 15/2, mj = 1/2
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 1.96354403635080146, 0, 0, 0, 0, 0, 0, 5.89063210905240439, 0, -15.7083522908064117, 0, 0, 0, 0, 0, 0, 0, 0, 5.89063210905240439, 0, -31.4167045816128234, 0, 18.850022748967694, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.96354403635080146, 0, -15.7083522908064117, 0, 18.850022748967694, 0, -3.5904805236128941, 0,
        // j = 15/2, mj = 3/2
        0, 0, 0, 0, -1.73168306950273036, 0, 0, 0, 0, 0, 0, -5.19504920850819109, 0, 13.8534645560218429, 0, 0, 0, 0, 0, 0, 0, 0, -5.19504920850819109, 0, 27.7069291120436858, 0, -16.6241574672262115, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.73168306950273036, 0, 13.8534645560218429, 0, -16.6241574672262115, 0, 3.16650618423356409, 0,
        0, -0.49476659128649439, 0, 0, 0, 0, -1.48429977385948317, 0, 14.8429977385948317, 0, 0, 0, 0, 0, 0, -1.48429977385948317, 0, 29.6859954771896634, 0, -39.5813273029195512, 0, 0, 0, 0, 0, 0, 0, 0, -0.49476659128649439, 0, 14.8429977385948317, 0, -39.5813273029195512, 0, 15.8325309211678205, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 15/2, mj = 5/2
        0, 0.383244553662480886, 0, 0, 0, 0, 1.14973366098744266, 0, -11.4973366098744266, 0, 0, 0, 0, 0, 0, 1.14973366098744266, 0, -22.9946732197488532, 0, 30.6595642929984709, 0, 0, 0, 0, 0, 0, 0, 0, 0.383244553662480886, 0, -11.4973366098744266, 0, 30.6595642929984709, 0, -12.2638257171993884, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -6.32353513543093462, 0, 0, 0, 0, 0, 0, -10.5392252257182244, 0, 42.1569009028728975, 0, 0, 0, 0, 0, 0, 0, 0, -2.10784504514364487, 0, 28.104600601915265, 0, -33.725520722298318, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.10784504514364487, 0, -14.0523003009576325, 0, 11.241840240766106, 0, 0, 0,
        // j = 15/2, mj = 7/2
        0, 0, 0, 0, 4.26332651858901504, 0, 0, 0, 0, 0, 0, 7.1055441976483584, 0, -28.4221767905934336, 0, 0, 0, 0, 0, 0, 0, 0, 1.42110883952967168, 0, -18.9481178603956224, 0, 22.7377414324747469, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.42110883952967168, 0, 9.47405893019781121, 0, -7.57924714415824896, 0, 0, 0,
        0, 1.13688707162373734, 0, 0, 0, 0, 1.13688707162373734, 0, -27.2852897189696963, 0, 0, 0, 0, 0, 0, -1.13688707162373734, 0, 0, 0, 45.4754828649494938, 0, 0, 0, 0, 0, 0, 0, 0, -1.13688707162373734, 0, 27.2852897189696963, 0, -45.4754828649494938, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 15/2, mj = 9/2
        0, -0.656382056840170103, 0, 0, 0, 0, -0.656382056840170103, 0, 15.7531693641640825, 0, 0, 0, 0, 0, 0, 0.656382056840170103, 0, 0, 0, -26.2552822736068041, 0, 0, 0, 0, 0, 0, 0, 0, 0.656382056840170103, 0, -15.7531693641640825, 0, 26.2552822736068041, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 10.6662084236527642, 0, 0, 0, 0, 0, 0, -10.6662084236527642, 0, -42.6648336946110567, 0, 0, 0, 0, 0, 0, 0, 0, -19.1991751625749755, 0, 85.3296673892221134, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.13324168473055283, 0, -8.53296673892221134, 0, 0, 0, 0, 0,
        // j = 15/2, mj = 11/2
        0, 0, 0, 0, -5.12388078893935741, 0, 0, 0, 0, 0, 0, 5.12388078893935741, 0, 20.4955231557574297, 0, 0, 0, 0, 0, 0, 0, 0, 9.22298542009084335, 0, -40.9910463115148593, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.02477615778787148, 0, 4.09910463115148593, 0, 0, 0, 0, 0,
        0, -2.04955231557574297, 0, 0, 0, 0, 4.78228873634340025, 0, 28.6937324180604015, 0, 0, 0, 0, 0, 0, 4.78228873634340025, 0, -95.6457747268680051, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.04955231557574297, 0, 28.6937324180604015, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 15/2, mj = 13/2
        0, 0.774657960861427067, 0, 0, 0, 0, -1.80753524200999649, 0, -10.8452114520599789, 0, 0, 0, 0, 0, 0, -1.80753524200999649, 0, 36.1507048401999298, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.774657960861427067, 0, -10.8452114520599789, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -13.5565143150749737, 0, 0, 0, 0, 0, 0, 67.7825715753748683, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -40.669542945224921, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.93664490215356767, 0, 0, 0, 0, 0, 0, 0,
        // j = 15/2, mj = 15/2
        0, 0, 0, 0, 3.50027694499385465, 0, 0, 0, 0, 0, 0, -17.5013847249692732, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 10.5008308349815639, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.500039563570550664, 0, 0, 0, 0, 0, 0, 0,
        0, 4.00031650856440531, 0, 0, 0, 0, -28.0022155599508372, 0, 0, 0, 0, 0, 0, 0, 0, 28.0022155599508372, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4.00031650856440531, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = -17/2
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -4.12343187517827448, 0, 0, 0, 0, 28.8640231262479213, 0, 0, 0, 0, 0, 0, 0, 0, -28.8640231262479213, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4.12343187517827448, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = -15/2
        0, -1.00007912714110133, 0, 0, 0, 0, 7.0005538899877093, 0, 0, 0, 0, 0, 0, 0, 0, -7.0005538899877093, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.00007912714110133, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -14.0011077799754186, 0, 0, 0, 0, 0, 0, 70.005538899877093, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -42.0033233399262558, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.00015825428220266, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = -13/2
        0, 0, 0, 0, -4.95013912767217325, 0, 0, 0, 0, 0, 0, 24.7506956383608662, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -14.8504173830165197, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.707162732524596178, 0, 0, 0, 0, 0, 0, 0,
        0, 2.12148819757378853, 0, 0, 0, 0, -4.95013912767217325, 0, -29.7008347660330395, 0, 0, 0, 0, 0, 0, -4.95013912767217325, 0, 99.002782553443465, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.12148819757378853, 0, -29.7008347660330395, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = -11/2
        0, 0.948758364647699116, 0, 0, 0, 0, -2.21376951751129794, 0, -13.2826171050677876, 0, 0, 0, 0, 0, 0, -2.21376951751129794, 0, 44.2753903502259587, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.948758364647699116, 0, -13.2826171050677876, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 11.0688475875564897, 0, 0, 0, 0, 0, 0, -11.0688475875564897, 0, -44.2753903502259587, 0, 0, 0, 0, 0, 0, 0, 0, -19.9239256576016814, 0, 88.5507807004519175, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.21376951751129794, 0, -8.85507807004519175, 0, 0, 0, 0, 0,
        // j = 17/2, mj = -9/2
        0, 0, 0, 0, 5.91654790557938008, 0, 0, 0, 0, 0, 0, -5.91654790557938008, 0, -23.6661916223175203, 0, 0, 0, 0, 0, 0, 0, 0, -10.6497862300428841, 0, 47.3323832446350406, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.18330958111587602, 0, -4.73323832446350406, 0, 0, 0, 0, 0,
        0, -1.18330958111587602, 0, 0, 0, 0, -1.18330958111587602, 0, 28.3994299467810244, 0, 0, 0, 0, 0, 0, 1.18330958111587602, 0, 0, 0, -47.3323832446350406, 0, 0, 0, 0, 0, 0, 0, 0, 1.18330958111587602, 0, -28.3994299467810244, 0, 47.3323832446350406, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = -7/2
        0, -0.733857449152875582, 0, 0, 0, 0, -0.733857449152875582, 0, 17.612578779669014, 0, 0, 0, 0, 0, 0, 0.733857449152875582, 0, 0, 0, -29.3542979661150233, 0, 0, 0, 0, 0, 0, 0, 0, 0.733857449152875582, 0, -17.612578779669014, 0, 29.3542979661150233, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -6.60471704237588023, 0, 0, 0, 0, 0, 0, -11.0078617372931337, 0, 44.0314469491725349, 0, 0, 0, 0, 0, 0, 0, 0, -2.20157234745862674, 0, 29.3542979661150233, 0, -35.2251575593380279, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.20157234745862674, 0, -14.6771489830575116, 0, 11.7417191864460093, 0, 0, 0,
        // j = 17/2, mj = -5/2
        0, 0, 0, 0, -4.67024020848234288, 0, 0, 0, 0, 0, 0, -7.7837336808039048, 0, 31.1349347232156192, 0, 0, 0, 0, 0, 0, 0, 0, -1.55674673616078096, 0, 20.7566231488104128, 0, -24.9079477785724953, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.55674673616078096, 0, -10.3783115744052064, 0, 8.30264925952416512, 0, 0, 0,
        0, 0.51891557872026032, 0, 0, 0, 0, 1.55674673616078096, 0, -15.5674673616078096, 0, 0, 0, 0, 0, 0, 1.55674673616078096, 0, -31.1349347232156192, 0, 41.5132462976208256, 0, 0, 0, 0, 0, 0, 0, 0, 0.51891557872026032, 0, -15.5674673616078096, 0, 41.5132462976208256, 0, -16.6052985190483302, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = -3/2
        0, 0.413951429393932508, 0, 0, 0, 0, 1.24185428818179752, 0, -12.4185428818179752, 0, 0, 0, 0, 0, 0, 1.24185428818179752, 0, -24.8370857636359505, 0, 33.1161143515146006, 0, 0, 0, 0, 0, 0, 0, 0, 0.413951429393932508, 0, -12.4185428818179752, 0, 33.1161143515146006, 0, -13.2464457406058403, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 2.06975714696966254, 0, 0, 0, 0, 0, 0, 6.20927144090898762, 0, -16.5580571757573003, 0, 0, 0, 0, 0, 0, 0, 0, 6.20927144090898762, 0, -33.1161143515146006, 0, 19.8696686109087604, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.06975714696966254, 0, -16.5580571757573003, 0, 19.8696686109087604, 0, -3.78469878303024007, 0,
        // j = 17/2, mj = -1/2
        0, 0, 0, 0, 1.85124707101607532, 0, 0, 0, 0, 0, 0, 5.55374121304822595, 0, -14.8099765681286025, 0, 0, 0, 0, 0, 0, 0, 0, 5.55374121304822595, 0, -29.6199531362572051, 0, 17.771971881754323, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.85124707101607532, 0, -14.8099765681286025, 0, 17.771971881754323, 0, -3.38513750128653772, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = 1/2
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 1.85124707101607532, 0, 0, 0, 0, 0, 0, 5.55374121304822595, 0, -14.8099765681286025, 0, 0, 0, 0, 0, 0, 0, 0, 5.55374121304822595, 0, -29.6199531362572051, 0, 17.771971881754323, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.85124707101607532, 0, -14.8099765681286025, 0, 17.771971881754323, 0, -3.38513750128653772, 0,
        // j = 17/2, mj = 3/2
        0, 0, 0, 0, 2.06975714696966254, 0, 0, 0, 0, 0, 0, 6.20927144090898762, 0, -16.5580571757573003, 0, 0, 0, 0, 0, 0, 0, 0, 6.20927144090898762, 0, -33.1161143515146006, 0, 19.8696686109087604, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.06975714696966254, 0, -16.5580571757573003, 0, 19.8696686109087604, 0, -3.78469878303024007, 0,
        0, -0.413951429393932508, 0, 0, 0, 0, -1.24185428818179752, 0, 12.4185428818179752, 0, 0, 0, 0, 0, 0, -1.24185428818179752, 0, 24.8370857636359505, 0, -33.1161143515146006, 0, 0, 0, 0, 0, 0, 0, 0, -0.413951429393932508, 0, 12.4185428818179752, 0, -33.1161143515146006, 0, 13.2464457406058403, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = 5/2
        0, -0.51891557872026032, 0, 0, 0, 0, -1.55674673616078096, 0, 15.5674673616078096, 0, 0, 0, 0, 0, 0, -1.55674673616078096, 0, 31.1349347232156192, 0, -41.5132462976208256, 0, 0, 0, 0, 0, 0, 0, 0, -0.51891557872026032, 0, 15.5674673616078096, 0, -41.5132462976208256, 0, 16.6052985190483302, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -4.67024020848234288, 0, 0, 0, 0, 0, 0, -7.7837336808039048, 0, 31.1349347232156192, 0, 0, 0, 0, 0, 0, 0, 0, -1.55674673616078096, 0, 20.7566231488104128, 0, -24.9079477785724953, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.55674673616078096, 0, -10.3783115744052064, 0, 8.30264925952416512, 0, 0, 0,
        // j = 17/2, mj = 7/2
        0, 0, 0, 0, -6.60471704237588023, 0, 0, 0, 0, 0, 0, -11.0078617372931337, 0, 44.0314469491725349, 0, 0, 0, 0, 0, 0, 0, 0, -2.20157234745862674, 0, 29.3542979661150233, 0, -35.2251575593380279, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.20157234745862674, 0, -14.6771489830575116, 0, 11.7417191864460093, 0, 0, 0,
        0, 0.733857449152875582, 0, 0, 0, 0, 0.733857449152875582, 0, -17.612578779669014, 0, 0, 0, 0, 0, 0, -0.733857449152875582, 0, 0, 0, 29.3542979661150233, 0, 0, 0, 0, 0, 0, 0, 0, -0.733857449152875582, 0, 17.612578779669014, 0, -29.3542979661150233, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = 9/2
        0, 1.18330958111587602, 0, 0, 0, 0, 1.18330958111587602, 0, -28.3994299467810244, 0, 0, 0, 0, 0, 0, -1.18330958111587602, 0, 0, 0, 47.3323832446350406, 0, 0, 0, 0, 0, 0, 0, 0, -1.18330958111587602, 0, 28.3994299467810244, 0, -47.3323832446350406, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 5.91654790557938008, 0, 0, 0, 0, 0, 0, -5.91654790557938008, 0, -23.6661916223175203, 0, 0, 0, 0, 0, 0, 0, 0, -10.6497862300428841, 0, 47.3323832446350406, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.18330958111587602, 0, -4.73323832446350406, 0, 0, 0, 0, 0,
        // j = 17/2, mj = 11/2
        0, 0, 0, 0, 11.0688475875564897, 0, 0, 0, 0, 0, 0, -11.0688475875564897, 0, -44.2753903502259587, 0, 0, 0, 0, 0, 0, 0, 0, -19.9239256576016814, 0, 88.5507807004519175, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.21376951751129794, 0, -8.85507807004519175, 0, 0, 0, 0, 0,
        0, -0.948758364647699116, 0, 0, 0, 0, 2.21376951751129794, 0, 13.2826171050677876, 0, 0, 0, 0, 0, 0, 2.21376951751129794, 0, -44.2753903502259587, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.948758364647699116, 0, 13.2826171050677876, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = 13/2
        0, -2.12148819757378853, 0, 0, 0, 0, 4.95013912767217325, 0, 29.7008347660330395, 0, 0, 0, 0, 0, 0, 4.95013912767217325, 0, -99.002782553443465, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.12148819757378853, 0, 29.7008347660330395, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -4.95013912767217325, 0, 0, 0, 0, 0, 0, 24.7506956383608662, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -14.8504173830165197, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.707162732524596178, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = 15/2
        0, 0, 0, 0, -14.0011077799754186, 0, 0, 0, 0, 0, 0, 70.005538899877093, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -42.0033233399262558, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.00015825428220266, 0, 0, 0, 0, 0, 0, 0,
        0, 1.00007912714110133, 0, 0, 0, 0, -7.0005538899877093, 0, 0, 0, 0, 0, 0, 0, 0, 7.0005538899877093, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.00007912714110133, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = 17/2
        0, 4.12343187517827448, 0, 0, 0, 0, -28.8640231262479213, 0, 0, 0, 0, 0, 0, 0, 0, 28.8640231262479213, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4.12343187517827448, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = -17/2
        0, 4.63886085957555879, 0, 0, 0, 0, -43.296034689371882, 0, 0, 0, 0, 0, 0, 0, 0, 64.944052034057823, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -18.5554434383022352, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.51542898439728431, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -4.12343187517827448, 0, 0, 0, 0, 0, 0, 28.8640231262479213, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -28.8640231262479213, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4.12343187517827448, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = -15/2
        0, 0, 0, 0, 17.0013451613987226, 0, 0, 0, 0, 0, 0, -119.009416129791058, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 119.009416129791058, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -17.0013451613987226, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0.875069236248463662, 0, 0, 0, 0, -3.50027694499385465, 0, -14.0011077799754186, 0, 0, 0, 0, 0, 0, -1.75013847249692732, 0, 70.005538899877093, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.50019781785275332, 0, -42.0033233399262558, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.125009890892637666, 0, 2.00015825428220266, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = -13/2
        0, -2.47506956383608662, 0, 0, 0, 0, 9.9002782553443465, 0, 39.601113021377386, 0, 0, 0, 0, 0, 0, 4.95013912767217325, 0, -198.00556510688693, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -7.07162732524596178, 0, 118.803339064132158, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.353581366262298089, 0, -5.65730186019676943, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 6.3644645927213656, 0, 0, 0, 0, 0, 0, -14.8504173830165197, 0, -29.7008347660330395, 0, 0, 0, 0, 0, 0, 0, 0, -14.8504173830165197, 0, 99.002782553443465, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6.3644645927213656, 0, -29.7008347660330395, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = -11/2
        0, 0, 0, 0, -14.2313754697154867, 0, 0, 0, 0, 0, 0, 33.2065427626694691, 0, 66.4130855253389381, 0, 0, 0, 0, 0, 0, 0, 0, 33.2065427626694691, 0, -221.376951751129794, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -14.2313754697154867, 0, 66.4130855253389381, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -0.790631970539749263, 0, 0, 0, 0, 0, 0, 22.1376951751129794, 0, 0, 0, 0, 0, 0, 2.21376951751129794, 0, -22.1376951751129794, 0, -44.2753903502259587, 0, 0, 0, 0, 0, 0, 0, 0, 1.26501115286359882, 0, -39.8478513152033629, 0, 88.5507807004519175, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.158126394107949853, 0, 4.42753903502259587, 0, -8.85507807004519175, 0, 0, 0, 0, 0,
        // j = 17/2, mj = -9/2
        0, 1.47913697639484502, 0, 0, 0, 0, 0, 0, -41.4158353390556606, 0, 0, 0, 0, 0, 0, -4.14158353390556606, 0, 41.4158353390556606, 0, 82.8316706781113211, 0, 0, 0, 0, 0, 0, 0, 0, -2.36661916223175203, 0, 74.548503610300189, 0, -165.663341356222642, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.295827395278969004, 0, -8.28316706781113211, 0, 16.5663341356222642, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -5.91654790557938008, 0, 0, 0, 0, 0, 0, -5.91654790557938008, 0, 47.3323832446350406, 0, 0, 0, 0, 0, 0, 0, 0, 5.91654790557938008, 0, 0, 0, -47.3323832446350406, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5.91654790557938008, 0, -47.3323832446350406, 0, 47.3323832446350406, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = -7/2
        0, 0, 0, 0, 9.54014683898738256, 0, 0, 0, 0, 0, 0, 9.54014683898738256, 0, -76.3211747118990605, 0, 0, 0, 0, 0, 0, 0, 0, -9.54014683898738256, 0, 0, 0, 76.3211747118990605, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -9.54014683898738256, 0, 76.3211747118990605, 0, -76.3211747118990605, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0.550393086864656686, 0, 0, 0, 0, 1.46771489830575116, 0, -19.8141511271276407, 0, 0, 0, 0, 0, 0, 1.10078617372931337, 0, -33.0235852118794012, 0, 66.0471704237588023, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -6.60471704237588023, 0, 44.0314469491725349, 0, -35.2251575593380279, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.183464362288218895, 0, 6.60471704237588023, 0, -22.0157234745862674, 0, 11.7417191864460093, 0, 0, 0,
        // j = 17/2, mj = -5/2
        0, -0.77837336808039048, 0, 0, 0, 0, -2.07566231488104128, 0, 28.0214412508940573, 0, 0, 0, 0, 0, 0, -1.55674673616078096, 0, 46.7024020848234288, 0, -93.4048041696468576, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 9.34048041696468576, 0, -62.2698694464312384, 0, 49.8158955571449907, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.25945778936013016, 0, -9.34048041696468576, 0, 31.1349347232156192, 0, -16.6052985190483302, 0, 0, 0,
        0, 0, 0, 0, 3.63240905104182224, 0, 0, 0, 0, 0, 0, 10.8972271531254667, 0, -36.3240905104182224, 0, 0, 0, 0, 0, 0, 0, 0, 10.8972271531254667, 0, -72.6481810208364448, 0, 58.1185448166691558, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.63240905104182224, 0, -36.3240905104182224, 0, 58.1185448166691558, 0, -16.6052985190483302, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = -3/2
        0, 0, 0, 0, -4.55346572333325759, 0, 0, 0, 0, 0, 0, -13.6603971699997728, 0, 45.5346572333325759, 0, 0, 0, 0, 0, 0, 0, 0, -13.6603971699997728, 0, 91.0693144666651517, 0, -72.8554515733321214, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4.55346572333325759, 0, 45.5346572333325759, 0, -72.8554515733321214, 0, 20.8158433066663204, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -0.206975714696966254, 0, 0, 0, 0, -0.827902858787865016, 0, 8.27902858787865016, 0, 0, 0, 0, 0, 0, -1.24185428818179752, 0, 24.8370857636359505, 0, -33.1161143515146006, 0, 0, 0, 0, 0, 0, 0, 0, -0.827902858787865016, 0, 24.8370857636359505, 0, -66.2322287030292013, 0, 26.4928914812116805, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.206975714696966254, 0, 8.27902858787865016, 0, -33.1161143515146006, 0, 26.4928914812116805, 0, -3.78469878303024007, 0,
        // j = 17/2, mj = -1/2
        0, 0.231405883877009415, 0, 0, 0, 0, 0.925623535508037658, 0, -9.25623535508037658, 0, 0, 0, 0, 0, 0, 1.38843530326205649, 0, -27.7687060652411297, 0, 37.0249414203215063, 0, 0, 0, 0, 0, 0, 0, 0, 0.925623535508037658, 0, -27.7687060652411297, 0, 74.0498828406430127, 0, -29.6199531362572051, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.231405883877009415, 0, -9.25623535508037658, 0, 37.0249414203215063, 0, -29.6199531362572051, 0, 4.23142187660817215, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = 1/2
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -0.231405883877009415, 0, 0, 0, 0, -0.925623535508037658, 0, 9.25623535508037658, 0, 0, 0, 0, 0, 0, -1.38843530326205649, 0, 27.7687060652411297, 0, -37.0249414203215063, 0, 0, 0, 0, 0, 0, 0, 0, -0.925623535508037658, 0, 27.7687060652411297, 0, -74.0498828406430127, 0, 29.6199531362572051, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.231405883877009415, 0, 9.25623535508037658, 0, -37.0249414203215063, 0, 29.6199531362572051, 0, -4.23142187660817215, 0,
        // j = 17/2, mj = 3/2
        0, 0.206975714696966254, 0, 0, 0, 0, 0.827902858787865016, 0, -8.27902858787865016, 0, 0, 0, 0, 0, 0, 1.24185428818179752, 0, -24.8370857636359505, 0, 33.1161143515146006, 0, 0, 0, 0, 0, 0, 0, 0, 0.827902858787865016, 0, -24.8370857636359505, 0, 66.2322287030292013, 0, -26.4928914812116805, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.206975714696966254, 0, -8.27902858787865016, 0, 33.1161143515146006, 0, -26.4928914812116805, 0, 3.78469878303024007, 0,
        0, 0, 0, 0, -4.55346572333325759, 0, 0, 0, 0, 0, 0, -13.6603971699997728, 0, 45.5346572333325759, 0, 0, 0, 0, 0, 0, 0, 0, -13.6603971699997728, 0, 91.0693144666651517, 0, -72.8554515733321214, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4.55346572333325759, 0, 45.5346572333325759, 0, -72.8554515733321214, 0, 20.8158433066663204, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = 5/2
        0, 0, 0, 0, 3.63240905104182224, 0, 0, 0, 0, 0, 0, 10.8972271531254667, 0, -36.3240905104182224, 0, 0, 0, 0, 0, 0, 0, 0, 10.8972271531254667, 0, -72.6481810208364448, 0, 58.1185448166691558, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.63240905104182224, 0, -36.3240905104182224, 0, 58.1185448166691558, 0, -16.6052985190483302, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0.77837336808039048, 0, 0, 0, 0, 2.07566231488104128, 0, -28.0214412508940573, 0, 0, 0, 0, 0, 0, 1.55674673616078096, 0, -46.7024020848234288, 0, 93.4048041696468576, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -9.34048041696468576, 0, 62.2698694464312384, 0, -49.8158955571449907, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.25945778936013016, 0, 9.34048041696468576, 0, -31.1349347232156192, 0, 16.6052985190483302, 0, 0, 0,
        // j = 17/2, mj = 7/2
        0, -0.550393086864656686, 0, 0, 0, 0, -1.46771489830575116, 0, 19.8141511271276407, 0, 0, 0, 0, 0, 0, -1.10078617372931337, 0, 33.0235852118794012, 0, -66.0471704237588023, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6.60471704237588023, 0, -44.0314469491725349, 0, 35.2251575593380279, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.183464362288218895, 0, -6.60471704237588023, 0, 22.0157234745862674, 0, -11.7417191864460093, 0, 0, 0,
        0, 0, 0, 0, 9.54014683898738256, 0, 0, 0, 0, 0, 0, 9.54014683898738256, 0, -76.3211747118990605, 0, 0, 0, 0, 0, 0, 0, 0, -9.54014683898738256, 0, 0, 0, 76.3211747118990605, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -9.54014683898738256, 0, 76.3211747118990605, 0, -76.3211747118990605, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = 9/2
        0, 0, 0, 0, -5.91654790557938008, 0, 0, 0, 0, 0, 0, -5.91654790557938008, 0, 47.3323832446350406, 0, 0, 0, 0, 0, 0, 0, 0, 5.91654790557938008, 0, 0, 0, -47.3323832446350406, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5.91654790557938008, 0, -47.3323832446350406, 0, 47.3323832446350406, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -1.47913697639484502, 0, 0, 0, 0, 0, 0, 41.4158353390556606, 0, 0, 0, 0, 0, 0, 4.14158353390556606, 0, -41.4158353390556606, 0, -82.8316706781113211, 0, 0, 0, 0, 0, 0, 0, 0, 2.36661916223175203, 0, -74.548503610300189, 0, 165.663341356222642, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.295827395278969004, 0, 8.28316706781113211, 0, -16.5663341356222642, 0, 0, 0, 0, 0,
        // j = 17/2, mj = 11/2
        0, 0.790631970539749263, 0, 0, 0, 0, 0, 0, -22.1376951751129794, 0, 0, 0, 0, 0, 0, -2.21376951751129794, 0, 22.1376951751129794, 0, 44.2753903502259587, 0, 0, 0, 0, 0, 0, 0, 0, -1.26501115286359882, 0, 39.8478513152033629, 0, -88.5507807004519175, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.158126394107949853, 0, -4.42753903502259587, 0, 8.85507807004519175, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -14.2313754697154867, 0, 0, 0, 0, 0, 0, 33.2065427626694691, 0, 66.4130855253389381, 0, 0, 0, 0, 0, 0, 0, 0, 33.2065427626694691, 0, -221.376951751129794, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -14.2313754697154867, 0, 66.4130855253389381, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = 13/2
        0, 0, 0, 0, 6.3644645927213656, 0, 0, 0, 0, 0, 0, -14.8504173830165197, 0, -29.7008347660330395, 0, 0, 0, 0, 0, 0, 0, 0, -14.8504173830165197, 0, 99.002782553443465, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6.3644645927213656, 0, -29.7008347660330395, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 2.47506956383608662, 0, 0, 0, 0, -9.9002782553443465, 0, -39.601113021377386, 0, 0, 0, 0, 0, 0, -4.95013912767217325, 0, 198.00556510688693, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 7.07162732524596178, 0, -118.803339064132158, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.353581366262298089, 0, 5.65730186019676943, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = 15/2
        0, -0.875069236248463662, 0, 0, 0, 0, 3.50027694499385465, 0, 14.0011077799754186, 0, 0, 0, 0, 0, 0, 1.75013847249692732, 0, -70.005538899877093, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.50019781785275332, 0, 42.0033233399262558, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.125009890892637666, 0, -2.00015825428220266, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 17.0013451613987226, 0, 0, 0, 0, 0, 0, -119.009416129791058, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 119.009416129791058, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -17.0013451613987226, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 17/2, mj = 17/2
        0, 0, 0, 0, -4.12343187517827448, 0, 0, 0, 0, 0, 0, 28.8640231262479213, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -28.8640231262479213, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4.12343187517827448, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -4.63886085957555879, 0, 0, 0, 0, 43.296034689371882, 0, 0, 0, 0, 0, 0, 0, 0, -64.944052034057823, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 18.5554434383022352, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.51542898439728431, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = -19/2
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -4.76597647343204622, 0, 0, 0, 0, 44.4824470853657647, 0, 0, 0, 0, 0, 0, 0, 0, -66.723670628048647, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 19.0639058937281849, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.52955294149244958, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = -17/2
        0, -1.0933899902622448, 0, 0, 0, 0, 10.2049732424476182, 0, 0, 0, 0, 0, 0, 0, 0, -15.3074598636714272, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4.37355996104897921, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.121487776695804978, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -17.4942398441959169, 0, 0, 0, 0, 0, 0, 122.459678909371418, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -122.459678909371418, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 17.4942398441959169, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = -15/2
        0, 0, 0, 0, -5.83141328139863895, 0, 0, 0, 0, 0, 0, 40.8198929697904727, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -40.8198929697904727, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5.83141328139863895, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 2.55124331061190454, 0, 0, 0, 0, -10.2049732424476182, 0, -40.8198929697904727, 0, 0, 0, 0, 0, 0, -5.10248662122380908, 0, 204.099464848952363, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 7.28926660174829869, 0, -122.459678909371418, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.364463330087414934, 0, 5.83141328139863895, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = -13/2
        0, 1.07173655920786068, 0, 0, 0, 0, -4.28694623683144273, 0, -17.1477849473257709, 0, 0, 0, 0, 0, 0, -2.14347311841572136, 0, 85.7389247366288545, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.06210445487960195, 0, -51.4433548419773127, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.153105222743980097, 0, 2.44968356390368156, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 14.6981013834220894, 0, 0, 0, 0, 0, 0, -34.2955698946515418, 0, -68.5911397893030836, 0, 0, 0, 0, 0, 0, 0, 0, -34.2955698946515418, 0, 228.637132631010279, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 14.6981013834220894, 0, -68.5911397893030836, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = -11/2
        0, 0, 0, 0, 7.34905069171104468, 0, 0, 0, 0, 0, 0, -17.1477849473257709, 0, -34.2955698946515418, 0, 0, 0, 0, 0, 0, 0, 0, -17.1477849473257709, 0, 114.318566315505139, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 7.34905069171104468, 0, -34.2955698946515418, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -1.53105222743980097, 0, 0, 0, 0, 0, 0, 42.8694623683144273, 0, 0, 0, 0, 0, 0, 4.28694623683144273, 0, -42.8694623683144273, 0, -85.7389247366288545, 0, 0, 0, 0, 0, 0, 0, 0, 2.44968356390368156, 0, -77.1650322629659691, 0, 171.477849473257709, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.306210445487960195, 0, 8.57389247366288545, 0, -17.1477849473257709, 0, 0, 0, 0, 0,
        // j = 19/2, mj = -9/2
        0, -0.883953415655745223, 0, 0, 0, 0, 0, 0, 24.7506956383608662, 0, 0, 0, 0, 0, 0, 2.47506956383608662, 0, -24.7506956383608662, 0, -49.5013912767217325, 0, 0, 0, 0, 0, 0, 0, 0, 1.41432546504919236, 0, -44.5512521490495592, 0, 99.002782553443465, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.176790683131149045, 0, 4.95013912767217325, 0, -9.9002782553443465, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -9.9002782553443465, 0, 0, 0, 0, 0, 0, -9.9002782553443465, 0, 79.202226042754772, 0, 0, 0, 0, 0, 0, 0, 0, 9.9002782553443465, 0, 0, 0, -79.202226042754772, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 9.9002782553443465, 0, -79.202226042754772, 0, 79.202226042754772, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = -7/2
        0, 0, 0, 0, -6.48125350089154427, 0, 0, 0, 0, 0, 0, -6.48125350089154427, 0, 51.8500280071323542, 0, 0, 0, 0, 0, 0, 0, 0, 6.48125350089154427, 0, 0, 0, -51.8500280071323542, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6.48125350089154427, 0, -51.8500280071323542, 0, 51.8500280071323542, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0.810156687611443034, 0, 0, 0, 0, 2.16041783363051476, 0, -29.1656407540119492, 0, 0, 0, 0, 0, 0, 1.62031337522288607, 0, -48.609401256686582, 0, 97.218802513373164, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -9.7218802513373164, 0, 64.8125350089154427, 0, -51.8500280071323542, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.270052229203814345, 0, 9.7218802513373164, 0, -32.4062675044577213, 0, 17.2833426690441181, 0, 0, 0,
        // j = 19/2, mj = -5/2
        0, 0.594492479694352209, 0, 0, 0, 0, 1.58531327918493922, 0, -21.4017292689966795, 0, 0, 0, 0, 0, 0, 1.18898495938870442, 0, -35.6695487816611326, 0, 71.3390975633222651, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -7.13390975633222651, 0, 47.5593983755481767, 0, -38.0475187004385414, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.198164159898117403, 0, 7.13390975633222651, 0, -23.7796991877740884, 0, 12.6825062334795138, 0, 0, 0,
        0, 0, 0, 0, 4.75593983755481767, 0, 0, 0, 0, 0, 0, 14.267819512664453, 0, -47.5593983755481767, 0, 0, 0, 0, 0, 0, 0, 0, 14.267819512664453, 0, -95.1187967510963535, 0, 76.0950374008770828, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4.75593983755481767, 0, -47.5593983755481767, 0, 76.0950374008770828, 0, -21.7414392573934522, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = -3/2
        0, 0, 0, 0, 3.88320861646147339, 0, 0, 0, 0, 0, 0, 11.6496258493844202, 0, -38.8320861646147339, 0, 0, 0, 0, 0, 0, 0, 0, 11.6496258493844202, 0, -77.6641723292294678, 0, 62.1313378633835742, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.88320861646147339, 0, -38.8320861646147339, 0, 62.1313378633835742, 0, -17.7518108181095926, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -0.242700538528842087, 0, 0, 0, 0, -0.970802154115368348, 0, 9.70802154115368348, 0, 0, 0, 0, 0, 0, -1.45620323117305252, 0, 29.1240646234610504, 0, -38.8320861646147339, 0, 0, 0, 0, 0, 0, 0, 0, -0.970802154115368348, 0, 29.1240646234610504, 0, -77.6641723292294678, 0, 31.0656689316917871, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.242700538528842087, 0, 9.70802154115368348, 0, -38.8320861646147339, 0, 31.0656689316917871, 0, -4.43795270452739816, 0,
        // j = 19/2, mj = -1/2
        0, -0.219530897104735508, 0, 0, 0, 0, -0.878123588418942033, 0, 8.78123588418942033, 0, 0, 0, 0, 0, 0, -1.31718538262841305, 0, 26.343707652568261, 0, -35.1249435367576813, 0, 0, 0, 0, 0, 0, 0, 0, -0.878123588418942033, 0, 26.343707652568261, 0, -70.2498870735153626, 0, 28.0999548294061451, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.219530897104735508, 0, 8.78123588418942033, 0, -35.1249435367576813, 0, 28.0999548294061451, 0, -4.01427926134373501, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = 1/2
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -0.219530897104735508, 0, 0, 0, 0, -0.878123588418942033, 0, 8.78123588418942033, 0, 0, 0, 0, 0, 0, -1.31718538262841305, 0, 26.343707652568261, 0, -35.1249435367576813, 0, 0, 0, 0, 0, 0, 0, 0, -0.878123588418942033, 0, 26.343707652568261, 0, -70.2498870735153626, 0, 28.0999548294061451, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.219530897104735508, 0, 8.78123588418942033, 0, -35.1249435367576813, 0, 28.0999548294061451, 0, -4.01427926134373501, 0,
        // j = 19/2, mj = 3/2
        0, -0.242700538528842087, 0, 0, 0, 0, -0.970802154115368348, 0, 9.70802154115368348, 0, 0, 0, 0, 0, 0, -1.45620323117305252, 0, 29.1240646234610504, 0, -38.8320861646147339, 0, 0, 0, 0, 0, 0, 0, 0, -0.970802154115368348, 0, 29.1240646234610504, 0, -77.6641723292294678, 0, 31.0656689316917871, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.242700538528842087, 0, 9.70802154115368348, 0, -38.8320861646147339, 0, 31.0656689316917871, 0, -4.43795270452739816, 0,
        0, 0, 0, 0, -3.88320861646147339, 0, 0, 0, 0, 0, 0, -11.6496258493844202, 0, 38.8320861646147339, 0, 0, 0, 0, 0, 0, 0, 0, -11.6496258493844202, 0, 77.6641723292294678, 0, -62.1313378633835742, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3.88320861646147339, 0, 38.8320861646147339, 0, -62.1313378633835742, 0, 17.7518108181095926, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = 5/2
        0, 0, 0, 0, -4.75593983755481767, 0, 0, 0, 0, 0, 0, -14.267819512664453, 0, 47.5593983755481767, 0, 0, 0, 0, 0, 0, 0, 0, -14.267819512664453, 0, 95.1187967510963535, 0, -76.0950374008770828, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4.75593983755481767, 0, 47.5593983755481767, 0, -76.0950374008770828, 0, 21.7414392573934522, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0.594492479694352209, 0, 0, 0, 0, 1.58531327918493922, 0, -21.4017292689966795, 0, 0, 0, 0, 0, 0, 1.18898495938870442, 0, -35.6695487816611326, 0, 71.3390975633222651, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -7.13390975633222651, 0, 47.5593983755481767, 0, -38.0475187004385414, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.198164159898117403, 0, 7.13390975633222651, 0, -23.7796991877740884, 0, 12.6825062334795138, 0, 0, 0,
        // j = 19/2, mj = 7/2
        0, 0.810156687611443034, 0, 0, 0, 0, 2.16041783363051476, 0, -29.1656407540119492, 0, 0, 0, 0, 0, 0, 1.62031337522288607, 0, -48.609401256686582, 0, 97.218802513373164, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -9.7218802513373164, 0, 64.8125350089154427, 0, -51.8500280071323542, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.270052229203814345, 0, 9.7218802513373164, 0, -32.4062675044577213, 0, 17.2833426690441181, 0, 0, 0,
        0, 0, 0, 0, 6.48125350089154427, 0, 0, 0, 0, 0, 0, 6.48125350089154427, 0, -51.8500280071323542, 0, 0, 0, 0, 0, 0, 0, 0, -6.48125350089154427, 0, 0, 0, 51.8500280071323542, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -6.48125350089154427, 0, 51.8500280071323542, 0, -51.8500280071323542, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = 9/2
        0, 0, 0, 0, 9.9002782553443465, 0, 0, 0, 0, 0, 0, 9.9002782553443465, 0, -79.202226042754772, 0, 0, 0, 0, 0, 0, 0, 0, -9.9002782553443465, 0, 0, 0, 79.202226042754772, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -9.9002782553443465, 0, 79.202226042754772, 0, -79.202226042754772, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -0.883953415655745223, 0, 0, 0, 0, 0, 0, 24.7506956383608662, 0, 0, 0, 0, 0, 0, 2.47506956383608662, 0, -24.7506956383608662, 0, -49.5013912767217325, 0, 0, 0, 0, 0, 0, 0, 0, 1.41432546504919236, 0, -44.5512521490495592, 0, 99.002782553443465, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.176790683131149045, 0, 4.95013912767217325, 0, -9.9002782553443465, 0, 0, 0, 0, 0,
        // j = 19/2, mj = 11/2
        0, -1.53105222743980097, 0, 0, 0, 0, 0, 0, 42.8694623683144273, 0, 0, 0, 0, 0, 0, 4.28694623683144273, 0, -42.8694623683144273, 0, -85.7389247366288545, 0, 0, 0, 0, 0, 0, 0, 0, 2.44968356390368156, 0, -77.1650322629659691, 0, 171.477849473257709, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.306210445487960195, 0, 8.57389247366288545, 0, -17.1477849473257709, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -7.34905069171104468, 0, 0, 0, 0, 0, 0, 17.1477849473257709, 0, 34.2955698946515418, 0, 0, 0, 0, 0, 0, 0, 0, 17.1477849473257709, 0, -114.318566315505139, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -7.34905069171104468, 0, 34.2955698946515418, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = 13/2
        0, 0, 0, 0, -14.6981013834220894, 0, 0, 0, 0, 0, 0, 34.2955698946515418, 0, 68.5911397893030836, 0, 0, 0, 0, 0, 0, 0, 0, 34.2955698946515418, 0, -228.637132631010279, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -14.6981013834220894, 0, 68.5911397893030836, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 1.07173655920786068, 0, 0, 0, 0, -4.28694623683144273, 0, -17.1477849473257709, 0, 0, 0, 0, 0, 0, -2.14347311841572136, 0, 85.7389247366288545, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.06210445487960195, 0, -51.4433548419773127, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.153105222743980097, 0, 2.44968356390368156, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = 15/2
        0, 2.55124331061190454, 0, 0, 0, 0, -10.2049732424476182, 0, -40.8198929697904727, 0, 0, 0, 0, 0, 0, -5.10248662122380908, 0, 204.099464848952363, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 7.28926660174829869, 0, -122.459678909371418, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.364463330087414934, 0, 5.83141328139863895, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 5.83141328139863895, 0, 0, 0, 0, 0, 0, -40.8198929697904727, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40.8198929697904727, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -5.83141328139863895, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = 17/2
        0, 0, 0, 0, 17.4942398441959169, 0, 0, 0, 0, 0, 0, -122.459678909371418, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 122.459678909371418, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -17.4942398441959169, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -1.0933899902622448, 0, 0, 0, 0, 10.2049732424476182, 0, 0, 0, 0, 0, 0, 0, 0, -15.3074598636714272, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4.37355996104897921, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.121487776695804978, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = 19/2
        0, -4.76597647343204622, 0, 0, 0, 0, 44.4824470853657647, 0, 0, 0, 0, 0, 0, 0, 0, -66.723670628048647, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 19.0639058937281849, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.52955294149244958, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = -19/2
        0, 5.2955294149244958, 0, 0, 0, 0, -63.5463529790939495, 0, 0, 0, 0, 0, 0, 0, 0, 133.447341256097294, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -63.5463529790939495, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5.2955294149244958, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -4.76597647343204622, 0, 0, 0, 0, 0, 0, 44.4824470853657647, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -66.723670628048647, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 19.0639058937281849, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.52955294149244958, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = -17/2
        0, 0, 0, 0, 20.7744098149826513, 0, 0, 0, 0, 0, 0, -193.894491606504745, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 290.841737409757118, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -83.0976392599306051, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.30826775722029458, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0.971902213566439825, 0, 0, 0, 0, -5.83141328139863895, 0, -17.4942398441959169, 0, 0, 0, 0, 0, 0, 0, 0, 122.459678909371418, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5.83141328139863895, 0, -122.459678909371418, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.971902213566439825, 0, 17.4942398441959169, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = -15/2
        0, -2.91570664069931948, 0, 0, 0, 0, 17.4942398441959169, 0, 52.4827195325877506, 0, 0, 0, 0, 0, 0, 0, 0, -367.379036728114254, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -17.4942398441959169, 0, 367.379036728114254, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.91570664069931948, 0, -52.4827195325877506, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 7.65372993183571362, 0, 0, 0, 0, 0, 0, -30.6149197273428545, 0, -40.8198929697904727, 0, 0, 0, 0, 0, 0, 0, 0, -15.3074598636714272, 0, 204.099464848952363, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 21.8677998052448961, 0, -122.459678909371418, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.0933899902622448, 0, 5.83141328139863895, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = -13/2
        0, 0, 0, 0, -18.2195215065336316, 0, 0, 0, 0, 0, 0, 72.8780860261345264, 0, 97.1707813681793685, 0, 0, 0, 0, 0, 0, 0, 0, 36.4390430130672632, 0, -485.853906840896842, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -52.0557757329532331, 0, 291.512344104538105, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.60278878664766166, 0, -13.8815401954541955, 0, 0, 0, 0, 0, 0, 0,
        0, -0.918631336463880584, 0, 0, 0, 0, 1.22484178195184078, 0, 29.3962027668441787, 0, 0, 0, 0, 0, 0, 4.28694623683144273, 0, -68.5911397893030836, 0, -68.5911397893030836, 0, 0, 0, 0, 0, 0, 0, 0, 1.22484178195184078, 0, -68.5911397893030836, 0, 228.637132631010279, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.918631336463880584, 0, 29.3962027668441787, 0, -68.5911397893030836, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = -11/2
        0, 1.83726267292776117, 0, 0, 0, 0, -2.44968356390368156, 0, -58.7924055336883574, 0, 0, 0, 0, 0, 0, -8.57389247366288545, 0, 137.182279578606167, 0, 137.182279578606167, 0, 0, 0, 0, 0, 0, 0, 0, -2.44968356390368156, 0, 137.182279578606167, 0, -457.274265262020558, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.83726267292776117, 0, -58.7924055336883574, 0, 137.182279578606167, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -7.65526113719900487, 0, 0, 0, 0, 0, 0, 0, 0, 71.4491039471907121, 0, 0, 0, 0, 0, 0, 0, 0, 21.4347311841572136, 0, -71.4491039471907121, 0, -85.7389247366288545, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12.2484178195184078, 0, -128.608387104943282, 0, 171.477849473257709, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.53105222743980097, 0, 14.2898207894381424, 0, -17.1477849473257709, 0, 0, 0, 0, 0,
        // j = 19/2, mj = -9/2
        0, 0, 0, 0, 13.2593012348361783, 0, 0, 0, 0, 0, 0, 0, 0, -123.753478191804331, 0, 0, 0, 0, 0, 0, 0, 0, -37.1260434575412994, 0, 123.753478191804331, 0, 148.504173830165197, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -21.2148819757378853, 0, 222.756260745247796, 0, -297.008347660330395, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.65186024696723567, 0, -24.7506956383608662, 0, 29.7008347660330395, 0, 0, 0, 0, 0,
        0, 0.707162732524596178, 0, 0, 0, 0, 1.41432546504919236, 0, -29.7008347660330395, 0, 0, 0, 0, 0, 0, 0, 0, -29.7008347660330395, 0, 118.803339064132158, 0, 0, 0, 0, 0, 0, 0, 0, -1.41432546504919236, 0, 29.7008347660330395, 0, 0, 0, -79.202226042754772, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.707162732524596178, 0, 29.7008347660330395, 0, -118.803339064132158, 0, 79.202226042754772, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = -7/2
        0, -1.08020891681525738, 0, 0, 0, 0, -2.16041783363051476, 0, 45.3687745062408099, 0, 0, 0, 0, 0, 0, 0, 0, 45.3687745062408099, 0, -181.47509802496324, 0, 0, 0, 0, 0, 0, 0, 0, 2.16041783363051476, 0, -45.3687745062408099, 0, 0, 0, 120.983398683308826, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.08020891681525738, 0, -45.3687745062408099, 0, 181.47509802496324, 0, -120.983398683308826, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 5.67109681328010124, 0, 0, 0, 0, 0, 0, 15.1229248354136033, 0, -68.0531617593612148, 0, 0, 0, 0, 0, 0, 0, 0, 11.3421936265602025, 0, -113.421936265602025, 0, 136.10632351872243, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -22.6843872531204049, 0, 90.7375490124816198, 0, -51.8500280071323542, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.89036560442670041, 0, 22.6843872531204049, 0, -45.3687745062408099, 0, 17.2833426690441181, 0, 0, 0,
        // j = 19/2, mj = -5/2
        0, 0, 0, 0, -7.72840223602657872, 0, 0, 0, 0, 0, 0, -20.6090726294042099, 0, 92.7408268323189446, 0, 0, 0, 0, 0, 0, 0, 0, -15.4568044720531574, 0, 154.568044720531574, 0, -185.481653664637889, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 30.9136089441063149, 0, -123.65443577642526, 0, 70.6596775865287197, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.57613407867552624, 0, -30.9136089441063149, 0, 61.8272178882126298, 0, -23.5532258621762399, 0, 0, 0,
        0, -0.396328319796234806, 0, 0, 0, 0, -1.58531327918493922, 0, 19.0237593502192707, 0, 0, 0, 0, 0, 0, -2.37796991877740884, 0, 57.0712780506578121, 0, -95.1187967510963535, 0, 0, 0, 0, 0, 0, 0, 0, -1.58531327918493922, 0, 57.0712780506578121, 0, -190.237593502192707, 0, 101.46004986783611, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.396328319796234806, 0, 19.0237593502192707, 0, -95.1187967510963535, 0, 101.46004986783611, 0, -21.7414392573934522, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = -3/2
        0, 0.485401077057684174, 0, 0, 0, 0, 1.9416043082307367, 0, -23.2992516987688403, 0, 0, 0, 0, 0, 0, 2.91240646234610504, 0, -69.897755096306521, 0, 116.496258493844202, 0, 0, 0, 0, 0, 0, 0, 0, 1.9416043082307367, 0, -69.897755096306521, 0, 232.992516987688403, 0, -124.262675726767148, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.485401077057684174, 0, -23.2992516987688403, 0, 116.496258493844202, 0, -124.262675726767148, 0, 26.627716227164389, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -2.18430484675957878, 0, 0, 0, 0, 0, 0, -8.73721938703831513, 0, 29.1240646234610504, 0, 0, 0, 0, 0, 0, 0, 0, -13.1058290805574727, 0, 87.3721938703831513, 0, -69.897755096306521, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -8.73721938703831513, 0, 87.3721938703831513, 0, -139.795510192613042, 0, 39.9415743407465834, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.18430484675957878, 0, 29.1240646234610504, 0, -69.897755096306521, 0, 39.9415743407465834, 0, -4.43795270452739816, 0,
        // j = 19/2, mj = -1/2
        0, 0, 0, 0, 2.41483986815209059, 0, 0, 0, 0, 0, 0, 9.65935947260836236, 0, -32.1978649086945412, 0, 0, 0, 0, 0, 0, 0, 0, 14.4890392089125435, 0, -96.5935947260836236, 0, 77.2748757808668989, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 9.65935947260836236, 0, -96.5935947260836236, 0, 154.549751561733798, 0, -44.1570718747810851, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.41483986815209059, 0, -32.1978649086945412, 0, 77.2748757808668989, 0, -44.1570718747810851, 0, 4.90634131942012056, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = 1/2
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -2.41483986815209059, 0, 0, 0, 0, 0, 0, -9.65935947260836236, 0, 32.1978649086945412, 0, 0, 0, 0, 0, 0, 0, 0, -14.4890392089125435, 0, 96.5935947260836236, 0, -77.2748757808668989, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -9.65935947260836236, 0, 96.5935947260836236, 0, -154.549751561733798, 0, 44.1570718747810851, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.41483986815209059, 0, 32.1978649086945412, 0, -77.2748757808668989, 0, 44.1570718747810851, 0, -4.90634131942012056, 0,
        // j = 19/2, mj = 3/2
        0, 0, 0, 0, 2.18430484675957878, 0, 0, 0, 0, 0, 0, 8.73721938703831513, 0, -29.1240646234610504, 0, 0, 0, 0, 0, 0, 0, 0, 13.1058290805574727, 0, -87.3721938703831513, 0, 69.897755096306521, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 8.73721938703831513, 0, -87.3721938703831513, 0, 139.795510192613042, 0, -39.9415743407465834, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.18430484675957878, 0, -29.1240646234610504, 0, 69.897755096306521, 0, -39.9415743407465834, 0, 4.43795270452739816, 0,
        0, 0.485401077057684174, 0, 0, 0, 0, 1.9416043082307367, 0, -23.2992516987688403, 0, 0, 0, 0, 0, 0, 2.91240646234610504, 0, -69.897755096306521, 0, 116.496258493844202, 0, 0, 0, 0, 0, 0, 0, 0, 1.9416043082307367, 0, -69.897755096306521, 0, 232.992516987688403, 0, -124.262675726767148, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.485401077057684174, 0, -23.2992516987688403, 0, 116.496258493844202, 0, -124.262675726767148, 0, 26.627716227164389, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = 5/2
        0, -0.396328319796234806, 0, 0, 0, 0, -1.58531327918493922, 0, 19.0237593502192707, 0, 0, 0, 0, 0, 0, -2.37796991877740884, 0, 57.0712780506578121, 0, -95.1187967510963535, 0, 0, 0, 0, 0, 0, 0, 0, -1.58531327918493922, 0, 57.0712780506578121, 0, -190.237593502192707, 0, 101.46004986783611, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.396328319796234806, 0, 19.0237593502192707, 0, -95.1187967510963535, 0, 101.46004986783611, 0, -21.7414392573934522, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 7.72840223602657872, 0, 0, 0, 0, 0, 0, 20.6090726294042099, 0, -92.7408268323189446, 0, 0, 0, 0, 0, 0, 0, 0, 15.4568044720531574, 0, -154.568044720531574, 0, 185.481653664637889, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -30.9136089441063149, 0, 123.65443577642526, 0, -70.6596775865287197, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.57613407867552624, 0, 30.9136089441063149, 0, -61.8272178882126298, 0, 23.5532258621762399, 0, 0, 0,
        // j = 19/2, mj = 7/2
        0, 0, 0, 0, -5.67109681328010124, 0, 0, 0, 0, 0, 0, -15.1229248354136033, 0, 68.0531617593612148, 0, 0, 0, 0, 0, 0, 0, 0, -11.3421936265602025, 0, 113.421936265602025, 0, -136.10632351872243, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 22.6843872531204049, 0, -90.7375490124816198, 0, 51.8500280071323542, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.89036560442670041, 0, -22.6843872531204049, 0, 45.3687745062408099, 0, -17.2833426690441181, 0, 0, 0,
        0, -1.08020891681525738, 0, 0, 0, 0, -2.16041783363051476, 0, 45.3687745062408099, 0, 0, 0, 0, 0, 0, 0, 0, 45.3687745062408099, 0, -181.47509802496324, 0, 0, 0, 0, 0, 0, 0, 0, 2.16041783363051476, 0, -45.3687745062408099, 0, 0, 0, 120.983398683308826, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.08020891681525738, 0, -45.3687745062408099, 0, 181.47509802496324, 0, -120.983398683308826, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = 9/2
        0, 0.707162732524596178, 0, 0, 0, 0, 1.41432546504919236, 0, -29.7008347660330395, 0, 0, 0, 0, 0, 0, 0, 0, -29.7008347660330395, 0, 118.803339064132158, 0, 0, 0, 0, 0, 0, 0, 0, -1.41432546504919236, 0, 29.7008347660330395, 0, 0, 0, -79.202226042754772, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.707162732524596178, 0, 29.7008347660330395, 0, -118.803339064132158, 0, 79.202226042754772, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -13.2593012348361783, 0, 0, 0, 0, 0, 0, 0, 0, 123.753478191804331, 0, 0, 0, 0, 0, 0, 0, 0, 37.1260434575412994, 0, -123.753478191804331, 0, -148.504173830165197, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 21.2148819757378853, 0, -222.756260745247796, 0, 297.008347660330395, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.65186024696723567, 0, 24.7506956383608662, 0, -29.7008347660330395, 0, 0, 0, 0, 0,
        // j = 19/2, mj = 11/2
        0, 0, 0, 0, 7.65526113719900487, 0, 0, 0, 0, 0, 0, 0, 0, -71.4491039471907121, 0, 0, 0, 0, 0, 0, 0, 0, -21.4347311841572136, 0, 71.4491039471907121, 0, 85.7389247366288545, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -12.2484178195184078, 0, 128.608387104943282, 0, -171.477849473257709, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.53105222743980097, 0, -14.2898207894381424, 0, 17.1477849473257709, 0, 0, 0, 0, 0,
        0, 1.83726267292776117, 0, 0, 0, 0, -2.44968356390368156, 0, -58.7924055336883574, 0, 0, 0, 0, 0, 0, -8.57389247366288545, 0, 137.182279578606167, 0, 137.182279578606167, 0, 0, 0, 0, 0, 0, 0, 0, -2.44968356390368156, 0, 137.182279578606167, 0, -457.274265262020558, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.83726267292776117, 0, -58.7924055336883574, 0, 137.182279578606167, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = 13/2
        0, -0.918631336463880584, 0, 0, 0, 0, 1.22484178195184078, 0, 29.3962027668441787, 0, 0, 0, 0, 0, 0, 4.28694623683144273, 0, -68.5911397893030836, 0, -68.5911397893030836, 0, 0, 0, 0, 0, 0, 0, 0, 1.22484178195184078, 0, -68.5911397893030836, 0, 228.637132631010279, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.918631336463880584, 0, 29.3962027668441787, 0, -68.5911397893030836, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 18.2195215065336316, 0, 0, 0, 0, 0, 0, -72.8780860261345264, 0, -97.1707813681793685, 0, 0, 0, 0, 0, 0, 0, 0, -36.4390430130672632, 0, 485.853906840896842, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 52.0557757329532331, 0, -291.512344104538105, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.60278878664766166, 0, 13.8815401954541955, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = 15/2
        0, 0, 0, 0, -7.65372993183571362, 0, 0, 0, 0, 0, 0, 30.6149197273428545, 0, 40.8198929697904727, 0, 0, 0, 0, 0, 0, 0, 0, 15.3074598636714272, 0, -204.099464848952363, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -21.8677998052448961, 0, 122.459678909371418, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.0933899902622448, 0, -5.83141328139863895, 0, 0, 0, 0, 0, 0, 0,
        0, -2.91570664069931948, 0, 0, 0, 0, 17.4942398441959169, 0, 52.4827195325877506, 0, 0, 0, 0, 0, 0, 0, 0, -367.379036728114254, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -17.4942398441959169, 0, 367.379036728114254, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.91570664069931948, 0, -52.4827195325877506, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = 17/2
        0, 0.971902213566439825, 0, 0, 0, 0, -5.83141328139863895, 0, -17.4942398441959169, 0, 0, 0, 0, 0, 0, 0, 0, 122.459678909371418, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5.83141328139863895, 0, -122.459678909371418, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.971902213566439825, 0, 17.4942398441959169, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -20.7744098149826513, 0, 0, 0, 0, 0, 0, 193.894491606504745, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -290.841737409757118, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 83.0976392599306051, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.30826775722029458, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 19/2, mj = 19/2
        0, 0, 0, 0, 4.76597647343204622, 0, 0, 0, 0, 0, 0, -44.4824470853657647, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66.723670628048647, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -19.0639058937281849, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.52955294149244958, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 5.2955294149244958, 0, 0, 0, 0, -63.5463529790939495, 0, 0, 0, 0, 0, 0, 0, 0, 133.447341256097294, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -63.5463529790939495, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5.2955294149244958, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = -21/2
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -5.42630291944221461, 0, 0, 0, 0, 65.1156350333065753, 0, 0, 0, 0, 0, 0, 0, 0, -136.742833569943808, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 65.1156350333065753, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -5.42630291944221461, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = -19/2
        0, -1.1841163748620862, 0, 0, 0, 0, 14.2093964983450344, 0, 0, 0, 0, 0, 0, 0, 0, -29.8397326465245721, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 14.2093964983450344, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.1841163748620862, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -21.3140947475175515, 0, 0, 0, 0, 0, 0, 198.931550976830481, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -298.397326465245721, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 85.2563789900702061, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.36823274972417239, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = -17/2
        0, 0, 0, 0, -6.74010856667869467, 0, 0, 0, 0, 0, 0, 62.9076799556678169, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -94.3615199335017254, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 26.9604342667147787, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.748900951853188297, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 2.99560380741275319, 0, 0, 0, 0, -17.9736228444765191, 0, -53.9208685334295574, 0, 0, 0, 0, 0, 0, 0, 0, 377.446079734006901, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 17.9736228444765191, 0, -377.446079734006901, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.99560380741275319, 0, 53.9208685334295574, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = -15/2
        0, 1.19033225155963006, 0, 0, 0, 0, -7.14199350935778034, 0, -21.425980528073341, 0, 0, 0, 0, 0, 0, 0, 0, 149.981863696513387, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 7.14199350935778034, 0, -149.981863696513387, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.19033225155963006, 0, 21.425980528073341, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 18.7477329620641734, 0, 0, 0, 0, 0, 0, -74.9909318482566936, 0, -99.9879091310089248, 0, 0, 0, 0, 0, 0, 0, 0, -37.4954659241283468, 0, 499.939545655044624, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 53.5649513201833526, 0, -299.963727393026774, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.67824756600916763, 0, 14.2839870187155607, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = -13/2
        0, 0, 0, 0, 8.83776607290009063, 0, 0, 0, 0, 0, 0, -35.3510642916003625, 0, -47.1347523888004833, 0, 0, 0, 0, 0, 0, 0, 0, -17.6755321458001813, 0, 235.673761944002417, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 25.2507602082859732, 0, -141.40425716640145, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.26253801041429866, 0, 6.73353605554292619, 0, 0, 0, 0, 0, 0, 0,
        0, -1.89380701562144799, 0, 0, 0, 0, 2.52507602082859732, 0, 60.6018244998863357, 0, 0, 0, 0, 0, 0, 8.83776607290009063, 0, -141.40425716640145, 0, -141.40425716640145, 0, 0, 0, 0, 0, 0, 0, 0, 2.52507602082859732, 0, -141.40425716640145, 0, 471.347523888004833, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.89380701562144799, 0, 60.6018244998863357, 0, -141.40425716640145, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = -11/2
        0, -1.02706105729735913, 0, 0, 0, 0, 1.36941474306314551, 0, 32.8659538335154923, 0, 0, 0, 0, 0, 0, 4.79295160072100929, 0, -76.6872256115361487, 0, -76.6872256115361487, 0, 0, 0, 0, 0, 0, 0, 0, 1.36941474306314551, 0, -76.6872256115361487, 0, 255.624085371787162, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.02706105729735913, 0, 32.8659538335154923, 0, -76.6872256115361487, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -13.6941474306314551, 0, 0, 0, 0, 0, 0, 0, 0, 127.812042685893581, 0, 0, 0, 0, 0, 0, 0, 0, 38.3436128057680743, 0, -127.812042685893581, 0, -153.374451223072297, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 21.9106358890103282, 0, -230.061676834608446, 0, 306.748902446144595, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.73882948612629102, 0, 25.5624085371787162, 0, -30.6748902446144595, 0, 0, 0, 0, 0,
        // j = 21/2, mj = -9/2
        0, 0, 0, 0, -8.38591841687309056, 0, 0, 0, 0, 0, 0, 0, 0, 78.2685718908155119, 0, 0, 0, 0, 0, 0, 0, 0, 23.4805715672446536, 0, -78.2685718908155119, 0, -93.9222862689786143, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 13.4174694669969449, 0, -140.883429403467921, 0, 187.844572537957229, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.67718368337461811, 0, 15.6537143781631024, 0, -18.7844572537957229, 0, 0, 0, 0, 0,
        0, 1.11812245558307874, 0, 0, 0, 0, 2.23624491116615748, 0, -46.9611431344893072, 0, 0, 0, 0, 0, 0, 0, 0, -46.9611431344893072, 0, 187.844572537957229, 0, 0, 0, 0, 0, 0, 0, 0, -2.23624491116615748, 0, 46.9611431344893072, 0, 0, 0, -125.229715025304819, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.11812245558307874, 0, 46.9611431344893072, 0, -187.844572537957229, 0, 125.229715025304819, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = -7/2
        0, 0.763823050178243716, 0, 0, 0, 0, 1.52764610035648743, 0, -32.0805681074862361, 0, 0, 0, 0, 0, 0, 0, 0, -32.0805681074862361, 0, 128.322272429944944, 0, 0, 0, 0, 0, 0, 0, 0, -1.52764610035648743, 0, 32.0805681074862361, 0, 0, 0, -85.5481816199632961, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.763823050178243716, 0, 32.0805681074862361, 0, -128.322272429944944, 0, 85.5481816199632961, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 8.02014202687155901, 0, 0, 0, 0, 0, 0, 21.387045404990824, 0, -96.2417043224587082, 0, 0, 0, 0, 0, 0, 0, 0, 16.040284053743118, 0, -160.40284053743118, 0, 192.483408644917416, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -32.0805681074862361, 0, 128.322272429944944, 0, -73.3270128171113967, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.673380675623853, 0, 32.0805681074862361, 0, -64.1611362149724721, 0, 24.4423376057037989, 0, 0, 0,
        // j = 21/2, mj = -5/2
        0, 0, 0, 0, 6.06265750929132863, 0, 0, 0, 0, 0, 0, 16.167086691443543, 0, -72.7518901114959435, 0, 0, 0, 0, 0, 0, 0, 0, 12.1253150185826573, 0, -121.253150185826573, 0, 145.503780222991887, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -24.2506300371653145, 0, 97.0025201486612581, 0, -55.4300115135207189, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.02088583643044288, 0, 24.2506300371653145, 0, -48.501260074330629, 0, 18.4766705045069063, 0, 0, 0,
        0, -0.505221459107610719, 0, 0, 0, 0, -2.02088583643044288, 0, 24.2506300371653145, 0, 0, 0, 0, 0, 0, -3.03132875464566431, 0, 72.7518901114959435, 0, -121.253150185826573, 0, 0, 0, 0, 0, 0, 0, 0, -2.02088583643044288, 0, 72.7518901114959435, 0, -242.506300371653145, 0, 129.336693531548344, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.505221459107610719, 0, 24.2506300371653145, 0, -121.253150185826573, 0, 129.336693531548344, 0, -27.7150057567603594, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = -3/2
        0, -0.420369663756282355, 0, 0, 0, 0, -1.68147865502512942, 0, 20.177743860301553, 0, 0, 0, 0, 0, 0, -2.52221798253769413, 0, 60.5332315809046591, 0, -100.888719301507765, 0, 0, 0, 0, 0, 0, 0, 0, -1.68147865502512942, 0, 60.5332315809046591, 0, -201.77743860301553, 0, 107.614633921608283, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.420369663756282355, 0, 20.177743860301553, 0, -100.888719301507765, 0, 107.614633921608283, 0, -23.0602786974874892, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -2.52221798253769413, 0, 0, 0, 0, 0, 0, -10.0888719301507765, 0, 33.6295731005025884, 0, 0, 0, 0, 0, 0, 0, 0, -15.1333078952261648, 0, 100.888719301507765, 0, -80.7109754412062121, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -10.0888719301507765, 0, 100.888719301507765, 0, -161.421950882412424, 0, 46.1205573949749784, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.52221798253769413, 0, 33.6295731005025884, 0, -80.7109754412062121, 0, 46.1205573949749784, 0, -5.12450637721944204, 0,
        // j = 21/2, mj = -1/2
        0, 0, 0, 0, -2.30245947330177705, 0, 0, 0, 0, 0, 0, -9.20983789320710822, 0, 30.6994596440236941, 0, 0, 0, 0, 0, 0, 0, 0, -13.8147568398106623, 0, 92.0983789320710822, 0, -73.6787031456568658, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -9.20983789320710822, 0, 92.0983789320710822, 0, -147.357406291313732, 0, 42.1021160832324947, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.30245947330177705, 0, 30.6994596440236941, 0, -73.6787031456568658, 0, 42.1021160832324947, 0, -4.67801289813694386, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = 1/2
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -2.30245947330177705, 0, 0, 0, 0, 0, 0, -9.20983789320710822, 0, 30.6994596440236941, 0, 0, 0, 0, 0, 0, 0, 0, -13.8147568398106623, 0, 92.0983789320710822, 0, -73.6787031456568658, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -9.20983789320710822, 0, 92.0983789320710822, 0, -147.357406291313732, 0, 42.1021160832324947, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.30245947330177705, 0, 30.6994596440236941, 0, -73.6787031456568658, 0, 42.1021160832324947, 0, -4.67801289813694386, 0,
        // j = 21/2, mj = 3/2
        0, 0, 0, 0, -2.52221798253769413, 0, 0, 0, 0, 0, 0, -10.0888719301507765, 0, 33.6295731005025884, 0, 0, 0, 0, 0, 0, 0, 0, -15.1333078952261648, 0, 100.888719301507765, 0, -80.7109754412062121, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -10.0888719301507765, 0, 100.888719301507765, 0, -161.421950882412424, 0, 46.1205573949749784, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.52221798253769413, 0, 33.6295731005025884, 0, -80.7109754412062121, 0, 46.1205573949749784, 0, -5.12450637721944204, 0,
        0, 0.420369663756282355, 0, 0, 0, 0, 1.68147865502512942, 0, -20.177743860301553, 0, 0, 0, 0, 0, 0, 2.52221798253769413, 0, -60.5332315809046591, 0, 100.888719301507765, 0, 0, 0, 0, 0, 0, 0, 0, 1.68147865502512942, 0, -60.5332315809046591, 0, 201.77743860301553, 0, -107.614633921608283, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.420369663756282355, 0, -20.177743860301553, 0, 100.888719301507765, 0, -107.614633921608283, 0, 23.0602786974874892, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = 5/2
        0, 0.505221459107610719, 0, 0, 0, 0, 2.02088583643044288, 0, -24.2506300371653145, 0, 0, 0, 0, 0, 0, 3.03132875464566431, 0, -72.7518901114959435, 0, 121.253150185826573, 0, 0, 0, 0, 0, 0, 0, 0, 2.02088583643044288, 0, -72.7518901114959435, 0, 242.506300371653145, 0, -129.336693531548344, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.505221459107610719, 0, -24.2506300371653145, 0, 121.253150185826573, 0, -129.336693531548344, 0, 27.7150057567603594, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 6.06265750929132863, 0, 0, 0, 0, 0, 0, 16.167086691443543, 0, -72.7518901114959435, 0, 0, 0, 0, 0, 0, 0, 0, 12.1253150185826573, 0, -121.253150185826573, 0, 145.503780222991887, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -24.2506300371653145, 0, 97.0025201486612581, 0, -55.4300115135207189, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.02088583643044288, 0, 24.2506300371653145, 0, -48.501260074330629, 0, 18.4766705045069063, 0, 0, 0,
        // j = 21/2, mj = 7/2
        0, 0, 0, 0, 8.02014202687155901, 0, 0, 0, 0, 0, 0, 21.387045404990824, 0, -96.2417043224587082, 0, 0, 0, 0, 0, 0, 0, 0, 16.040284053743118, 0, -160.40284053743118, 0, 192.483408644917416, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -32.0805681074862361, 0, 128.322272429944944, 0, -73.3270128171113967, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.673380675623853, 0, 32.0805681074862361, 0, -64.1611362149724721, 0, 24.4423376057037989, 0, 0, 0,
        0, -0.763823050178243716, 0, 0, 0, 0, -1.52764610035648743, 0, 32.0805681074862361, 0, 0, 0, 0, 0, 0, 0, 0, 32.0805681074862361, 0, -128.322272429944944, 0, 0, 0, 0, 0, 0, 0, 0, 1.52764610035648743, 0, -32.0805681074862361, 0, 0, 0, 85.5481816199632961, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.763823050178243716, 0, -32.0805681074862361, 0, 128.322272429944944, 0, -85.5481816199632961, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = 9/2
        0, -1.11812245558307874, 0, 0, 0, 0, -2.23624491116615748, 0, 46.9611431344893072, 0, 0, 0, 0, 0, 0, 0, 0, 46.9611431344893072, 0, -187.844572537957229, 0, 0, 0, 0, 0, 0, 0, 0, 2.23624491116615748, 0, -46.9611431344893072, 0, 0, 0, 125.229715025304819, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.11812245558307874, 0, -46.9611431344893072, 0, 187.844572537957229, 0, -125.229715025304819, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -8.38591841687309056, 0, 0, 0, 0, 0, 0, 0, 0, 78.2685718908155119, 0, 0, 0, 0, 0, 0, 0, 0, 23.4805715672446536, 0, -78.2685718908155119, 0, -93.9222862689786143, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 13.4174694669969449, 0, -140.883429403467921, 0, 187.844572537957229, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.67718368337461811, 0, 15.6537143781631024, 0, -18.7844572537957229, 0, 0, 0, 0, 0,
        // j = 21/2, mj = 11/2
        0, 0, 0, 0, -13.6941474306314551, 0, 0, 0, 0, 0, 0, 0, 0, 127.812042685893581, 0, 0, 0, 0, 0, 0, 0, 0, 38.3436128057680743, 0, -127.812042685893581, 0, -153.374451223072297, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 21.9106358890103282, 0, -230.061676834608446, 0, 306.748902446144595, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.73882948612629102, 0, 25.5624085371787162, 0, -30.6748902446144595, 0, 0, 0, 0, 0,
        0, 1.02706105729735913, 0, 0, 0, 0, -1.36941474306314551, 0, -32.8659538335154923, 0, 0, 0, 0, 0, 0, -4.79295160072100929, 0, 76.6872256115361487, 0, 76.6872256115361487, 0, 0, 0, 0, 0, 0, 0, 0, -1.36941474306314551, 0, 76.6872256115361487, 0, -255.624085371787162, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.02706105729735913, 0, -32.8659538335154923, 0, 76.6872256115361487, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = 13/2
        0, 1.89380701562144799, 0, 0, 0, 0, -2.52507602082859732, 0, -60.6018244998863357, 0, 0, 0, 0, 0, 0, -8.83776607290009063, 0, 141.40425716640145, 0, 141.40425716640145, 0, 0, 0, 0, 0, 0, 0, 0, -2.52507602082859732, 0, 141.40425716640145, 0, -471.347523888004833, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.89380701562144799, 0, -60.6018244998863357, 0, 141.40425716640145, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 8.83776607290009063, 0, 0, 0, 0, 0, 0, -35.3510642916003625, 0, -47.1347523888004833, 0, 0, 0, 0, 0, 0, 0, 0, -17.6755321458001813, 0, 235.673761944002417, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 25.2507602082859732, 0, -141.40425716640145, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.26253801041429866, 0, 6.73353605554292619, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = 15/2
        0, 0, 0, 0, 18.7477329620641734, 0, 0, 0, 0, 0, 0, -74.9909318482566936, 0, -99.9879091310089248, 0, 0, 0, 0, 0, 0, 0, 0, -37.4954659241283468, 0, 499.939545655044624, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 53.5649513201833526, 0, -299.963727393026774, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.67824756600916763, 0, 14.2839870187155607, 0, 0, 0, 0, 0, 0, 0,
        0, -1.19033225155963006, 0, 0, 0, 0, 7.14199350935778034, 0, 21.425980528073341, 0, 0, 0, 0, 0, 0, 0, 0, -149.981863696513387, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -7.14199350935778034, 0, 149.981863696513387, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.19033225155963006, 0, -21.425980528073341, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = 17/2
        0, -2.99560380741275319, 0, 0, 0, 0, 17.9736228444765191, 0, 53.9208685334295574, 0, 0, 0, 0, 0, 0, 0, 0, -377.446079734006901, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -17.9736228444765191, 0, 377.446079734006901, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.99560380741275319, 0, -53.9208685334295574, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -6.74010856667869467, 0, 0, 0, 0, 0, 0, 62.9076799556678169, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -94.3615199335017254, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 26.9604342667147787, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.748900951853188297, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = 19/2
        0, 0, 0, 0, -21.3140947475175515, 0, 0, 0, 0, 0, 0, 198.931550976830481, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -298.397326465245721, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 85.2563789900702061, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.36823274972417239, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 1.1841163748620862, 0, 0, 0, 0, -14.2093964983450344, 0, 0, 0, 0, 0, 0, 0, 0, 29.8397326465245721, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -14.2093964983450344, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.1841163748620862, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = 21/2
        0, 5.42630291944221461, 0, 0, 0, 0, -65.1156350333065753, 0, 0, 0, 0, 0, 0, 0, 0, 136.742833569943808, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -65.1156350333065753, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5.42630291944221461, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = -21/2
        0, 5.96893321138643607, 0, 0, 0, 0, -89.533998170796541, 0, 0, 0, 0, 0, 0, 0, 0, 250.695194878230315, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -179.067996341593082, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 29.8446660569321803, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.542630291944221461, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -5.42630291944221461, 0, 0, 0, 0, 0, 0, 65.1156350333065753, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -136.742833569943808, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 65.1156350333065753, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -5.42630291944221461, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = -19/2
        0, 0, 0, 0, 24.8664438721038101, 0, 0, 0, 0, 0, 0, -298.397326465245721, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 626.634385577016015, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -298.397326465245721, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 24.8664438721038101, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 1.06570473737587758, 0, 0, 0, 0, -8.88087281146564647, 0, -21.3140947475175515, 0, 0, 0, 0, 0, 0, 4.97328877442076202, 0, 198.931550976830481, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 10.6570473737587758, 0, -298.397326465245721, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4.14440731201730169, 0, 85.2563789900702061, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.11841163748620862, 0, -2.36823274972417239, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = -17/2
        0, -3.37005428333934733, 0, 0, 0, 0, 28.0837856944945611, 0, 67.4010856667869467, 0, 0, 0, 0, 0, 0, -15.7269199889169542, 0, -629.076799556678169, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -33.7005428333934733, 0, 943.615199335017254, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 13.1057666574307952, 0, -269.604342667147787, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.374450475926594148, 0, 7.48900951853188297, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 8.98681142223825956, 0, 0, 0, 0, 0, 0, -53.9208685334295574, 0, -53.9208685334295574, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 377.446079734006901, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 53.9208685334295574, 0, -377.446079734006901, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -8.98681142223825956, 0, 53.9208685334295574, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = -15/2
        0, 0, 0, 0, -22.6163127796329711, 0, 0, 0, 0, 0, 0, 135.697876677797827, 0, 135.697876677797827, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -949.885136744584786, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -135.697876677797827, 0, 949.885136744584786, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 22.6163127796329711, 0, -135.697876677797827, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -1.0415407201146763, 0, 0, 0, 0, 3.1246221603440289, 0, 37.4954659241283468, 0, 0, 0, 0, 0, 0, 6.2492443206880578, 0, -149.981863696513387, 0, -99.9879091310089248, 0, 0, 0, 0, 0, 0, 0, 0, -0.892749188669722543, 0, -74.9909318482566936, 0, 499.939545655044624, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.82703909745412139, 0, 107.129902640366705, 0, -299.963727393026774, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.148791531444953757, 0, -5.35649513201833526, 0, 14.2839870187155607, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = -13/2
        0, 2.20944151822502266, 0, 0, 0, 0, -6.62832455467506797, 0, -79.5398946561008156, 0, 0, 0, 0, 0, 0, -13.2566491093501359, 0, 318.159578624403263, 0, 212.106385749602175, 0, 0, 0, 0, 0, 0, 0, 0, 1.89380701562144799, 0, 159.079789312201631, 0, -1060.53192874801088, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5.99705554946791864, 0, -227.256841874573759, 0, 636.319157248806525, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.315634502603574665, 0, 11.3628420937286879, 0, -30.3009122499431679, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -9.46903507810723996, 0, 0, 0, 0, 0, 0, 12.6253801041429866, 0, 101.003040833143893, 0, 0, 0, 0, 0, 0, 0, 0, 44.1888303645004531, 0, -235.673761944002417, 0, -141.40425716640145, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12.6253801041429866, 0, -235.673761944002417, 0, 471.347523888004833, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -9.46903507810723996, 0, 101.003040833143893, 0, -141.40425716640145, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = -11/2
        0, 0, 0, 0, 17.4600379740551053, 0, 0, 0, 0, 0, 0, -23.2800506320734737, 0, -186.24040505658779, 0, 0, 0, 0, 0, 0, 0, 0, -81.480177212257158, 0, 434.560945132038176, 0, 260.736567079222906, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -23.2800506320734737, 0, 434.560945132038176, 0, -869.121890264076352, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 17.4600379740551053, 0, -186.24040505658779, 0, 260.736567079222906, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0.855884214414465945, 0, 0, 0, 0, 0.855884214414465945, 0, -41.0824422918943654, 0, 0, 0, 0, 0, 0, -2.39647580036050465, 0, 0, 0, 191.718064028840372, 0, 0, 0, 0, 0, 0, 0, 0, -3.76589054342365016, 0, 115.030838417304223, 0, -191.718064028840372, 0, -153.374451223072297, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.19823790018025232, 0, 65.7319076670309846, 0, -345.092515251912669, 0, 306.748902446144595, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.171176842882893189, 0, -8.21648845837887307, 0, 38.3436128057680743, 0, -30.6748902446144595, 0, 0, 0, 0, 0,
        // j = 21/2, mj = -9/2
        0, -1.39765306947884843, 0, 0, 0, 0, -1.39765306947884843, 0, 67.0873473349847245, 0, 0, 0, 0, 0, 0, 3.9134285945407756, 0, 0, 0, -313.074287563262048, 0, 0, 0, 0, 0, 0, 0, 0, 6.14967350570693308, 0, -187.844572537957229, 0, 313.074287563262048, 0, 250.459430050609638, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.9567142972703878, 0, -107.339755735975559, 0, 563.533717613871686, 0, -500.918860101219276, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.279530613895769685, 0, 13.4174694669969449, 0, -62.6148575126524095, 0, 50.0918860101219276, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 7.82685718908155119, 0, 0, 0, 0, 0, 0, 15.6537143781631024, 0, -109.576000647141717, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -109.576000647141717, 0, 262.98240155314012, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -15.6537143781631024, 0, 109.576000647141717, 0, 0, 0, -125.229715025304819, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -7.82685718908155119, 0, 109.576000647141717, 0, -262.98240155314012, 0, 125.229715025304819, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = -7/2
        0, 0, 0, 0, -11.4573457526736557, 0, 0, 0, 0, 0, 0, -22.9146915053473115, 0, 160.40284053743118, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 160.40284053743118, 0, -384.966817289834833, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 22.9146915053473115, 0, -160.40284053743118, 0, 0, 0, 183.317532042778492, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 11.4573457526736557, 0, -160.40284053743118, 0, 384.966817289834833, 0, -183.317532042778492, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -0.572867287633682787, 0, 0, 0, 0, -2.10051338799017022, 0, 32.0805681074862361, 0, 0, 0, 0, 0, 0, -2.673380675623853, 0, 85.5481816199632961, 0, -192.483408644917416, 0, 0, 0, 0, 0, 0, 0, 0, -1.14573457526736557, 0, 64.1611362149724721, 0, -320.805681074862361, 0, 256.644544859889888, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.190955762544560929, 0, 0, 0, -64.1611362149724721, 0, 171.096363239926592, 0, -73.3270128171113967, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.190955762544560929, 0, -10.693522702495412, 0, 64.1611362149724721, 0, -85.5481816199632961, 0, 24.4423376057037989, 0, 0, 0,
        // j = 21/2, mj = -5/2
        0, 0.757832188661416079, 0, 0, 0, 0, 2.77871802509185895, 0, -42.4386025650393004, 0, 0, 0, 0, 0, 0, 3.53655021375327503, 0, -113.169606840104801, 0, 254.631615390235802, 0, 0, 0, 0, 0, 0, 0, 0, 1.51566437732283216, 0, -84.8772051300786008, 0, 424.386025650393004, 0, -339.508820520314403, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.25261072955380536, 0, 0, 0, 84.8772051300786008, 0, -226.339213680209602, 0, 97.0025201486612581, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.25261072955380536, 0, 14.1462008550131001, 0, -84.8772051300786008, 0, 113.169606840104801, 0, -32.334173382887086, 0, 0, 0,
        0, 0, 0, 0, -4.54699313196849647, 0, 0, 0, 0, 0, 0, -18.1879725278739859, 0, 72.7518901114959435, 0, 0, 0, 0, 0, 0, 0, 0, -27.2819587918109788, 0, 218.255670334487831, 0, -218.255670334487831, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -18.1879725278739859, 0, 218.255670334487831, 0, -436.511340668975661, 0, 166.290034540562157, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4.54699313196849647, 0, 72.7518901114959435, 0, -218.255670334487831, 0, 166.290034540562157, 0, -27.7150057567603594, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = -3/2
        0, 0, 0, 0, 5.46480562883167061, 0, 0, 0, 0, 0, 0, 21.8592225153266825, 0, -87.4368900613067298, 0, 0, 0, 0, 0, 0, 0, 0, 32.7888337729900237, 0, -262.310670183920189, 0, 262.310670183920189, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 21.8592225153266825, 0, -262.310670183920189, 0, 524.621340367840379, 0, -199.85574871155824, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5.46480562883167061, 0, -87.4368900613067298, 0, 262.310670183920189, 0, -199.85574871155824, 0, 33.3092914519263733, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0.210184831878141177, 0, 0, 0, 0, 1.05092415939070589, 0, -12.6110899126884706, 0, 0, 0, 0, 0, 0, 2.10184831878141177, 0, -50.4443596507538826, 0, 84.073932751256471, 0, 0, 0, 0, 0, 0, 0, 0, 2.10184831878141177, 0, -75.6665394761308239, 0, 252.221798253769413, 0, -134.518292402010354, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.05092415939070589, 0, -50.4443596507538826, 0, 252.221798253769413, 0, -269.036584804020707, 0, 57.650696743718723, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.210184831878141177, 0, -12.6110899126884706, 0, 84.073932751256471, 0, -134.518292402010354, 0, 57.650696743718723, 0, -5.12450637721944204, 0,
        // j = 21/2, mj = -1/2
        0, -0.230245947330177705, 0, 0, 0, 0, -1.15122973665088853, 0, 13.8147568398106623, 0, 0, 0, 0, 0, 0, -2.30245947330177705, 0, 55.2590273592426493, 0, -92.0983789320710822, 0, 0, 0, 0, 0, 0, 0, 0, -2.30245947330177705, 0, 82.888541038863974, 0, -276.295136796213247, 0, 147.357406291313732, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.15122973665088853, 0, 55.2590273592426493, 0, -276.295136796213247, 0, 294.714812582627463, 0, -63.1531741248487421, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.230245947330177705, 0, 13.8147568398106623, 0, -92.0983789320710822, 0, 147.357406291313732, 0, -63.1531741248487421, 0, 5.61361547776433263, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = 1/2
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0.230245947330177705, 0, 0, 0, 0, 1.15122973665088853, 0, -13.8147568398106623, 0, 0, 0, 0, 0, 0, 2.30245947330177705, 0, -55.2590273592426493, 0, 92.0983789320710822, 0, 0, 0, 0, 0, 0, 0, 0, 2.30245947330177705, 0, -82.888541038863974, 0, 276.295136796213247, 0, -147.357406291313732, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.15122973665088853, 0, -55.2590273592426493, 0, 276.295136796213247, 0, -294.714812582627463, 0, 63.1531741248487421, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.230245947330177705, 0, -13.8147568398106623, 0, 92.0983789320710822, 0, -147.357406291313732, 0, 63.1531741248487421, 0, -5.61361547776433263, 0,
        // j = 21/2, mj = 3/2
        0, -0.210184831878141177, 0, 0, 0, 0, -1.05092415939070589, 0, 12.6110899126884706, 0, 0, 0, 0, 0, 0, -2.10184831878141177, 0, 50.4443596507538826, 0, -84.073932751256471, 0, 0, 0, 0, 0, 0, 0, 0, -2.10184831878141177, 0, 75.6665394761308239, 0, -252.221798253769413, 0, 134.518292402010354, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.05092415939070589, 0, 50.4443596507538826, 0, -252.221798253769413, 0, 269.036584804020707, 0, -57.650696743718723, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.210184831878141177, 0, 12.6110899126884706, 0, -84.073932751256471, 0, 134.518292402010354, 0, -57.650696743718723, 0, 5.12450637721944204, 0,
        0, 0, 0, 0, 5.46480562883167061, 0, 0, 0, 0, 0, 0, 21.8592225153266825, 0, -87.4368900613067298, 0, 0, 0, 0, 0, 0, 0, 0, 32.7888337729900237, 0, -262.310670183920189, 0, 262.310670183920189, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 21.8592225153266825, 0, -262.310670183920189, 0, 524.621340367840379, 0, -199.85574871155824, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5.46480562883167061, 0, -87.4368900613067298, 0, 262.310670183920189, 0, -199.85574871155824, 0, 33.3092914519263733, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = 5/2
        0, 0, 0, 0, -4.54699313196849647, 0, 0, 0, 0, 0, 0, -18.1879725278739859, 0, 72.7518901114959435, 0, 0, 0, 0, 0, 0, 0, 0, -27.2819587918109788, 0, 218.255670334487831, 0, -218.255670334487831, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -18.1879725278739859, 0, 218.255670334487831, 0, -436.511340668975661, 0, 166.290034540562157, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4.54699313196849647, 0, 72.7518901114959435, 0, -218.255670334487831, 0, 166.290034540562157, 0, -27.7150057567603594, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -0.757832188661416079, 0, 0, 0, 0, -2.77871802509185895, 0, 42.4386025650393004, 0, 0, 0, 0, 0, 0, -3.53655021375327503, 0, 113.169606840104801, 0, -254.631615390235802, 0, 0, 0, 0, 0, 0, 0, 0, -1.51566437732283216, 0, 84.8772051300786008, 0, -424.386025650393004, 0, 339.508820520314403, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.25261072955380536, 0, 0, 0, -84.8772051300786008, 0, 226.339213680209602, 0, -97.0025201486612581, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.25261072955380536, 0, -14.1462008550131001, 0, 84.8772051300786008, 0, -113.169606840104801, 0, 32.334173382887086, 0, 0, 0,
        // j = 21/2, mj = 7/2
        0, 0.572867287633682787, 0, 0, 0, 0, 2.10051338799017022, 0, -32.0805681074862361, 0, 0, 0, 0, 0, 0, 2.673380675623853, 0, -85.5481816199632961, 0, 192.483408644917416, 0, 0, 0, 0, 0, 0, 0, 0, 1.14573457526736557, 0, -64.1611362149724721, 0, 320.805681074862361, 0, -256.644544859889888, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.190955762544560929, 0, 0, 0, 64.1611362149724721, 0, -171.096363239926592, 0, 73.3270128171113967, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.190955762544560929, 0, 10.693522702495412, 0, -64.1611362149724721, 0, 85.5481816199632961, 0, -24.4423376057037989, 0, 0, 0,
        0, 0, 0, 0, -11.4573457526736557, 0, 0, 0, 0, 0, 0, -22.9146915053473115, 0, 160.40284053743118, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 160.40284053743118, 0, -384.966817289834833, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 22.9146915053473115, 0, -160.40284053743118, 0, 0, 0, 183.317532042778492, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 11.4573457526736557, 0, -160.40284053743118, 0, 384.966817289834833, 0, -183.317532042778492, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = 9/2
        0, 0, 0, 0, 7.82685718908155119, 0, 0, 0, 0, 0, 0, 15.6537143781631024, 0, -109.576000647141717, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -109.576000647141717, 0, 262.98240155314012, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -15.6537143781631024, 0, 109.576000647141717, 0, 0, 0, -125.229715025304819, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -7.82685718908155119, 0, 109.576000647141717, 0, -262.98240155314012, 0, 125.229715025304819, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 1.39765306947884843, 0, 0, 0, 0, 1.39765306947884843, 0, -67.0873473349847245, 0, 0, 0, 0, 0, 0, -3.9134285945407756, 0, 0, 0, 313.074287563262048, 0, 0, 0, 0, 0, 0, 0, 0, -6.14967350570693308, 0, 187.844572537957229, 0, -313.074287563262048, 0, -250.459430050609638, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.9567142972703878, 0, 107.339755735975559, 0, -563.533717613871686, 0, 500.918860101219276, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.279530613895769685, 0, -13.4174694669969449, 0, 62.6148575126524095, 0, -50.0918860101219276, 0, 0, 0, 0, 0,
        // j = 21/2, mj = 11/2
        0, -0.855884214414465945, 0, 0, 0, 0, -0.855884214414465945, 0, 41.0824422918943654, 0, 0, 0, 0, 0, 0, 2.39647580036050465, 0, 0, 0, -191.718064028840372, 0, 0, 0, 0, 0, 0, 0, 0, 3.76589054342365016, 0, -115.030838417304223, 0, 191.718064028840372, 0, 153.374451223072297, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.19823790018025232, 0, -65.7319076670309846, 0, 345.092515251912669, 0, -306.748902446144595, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.171176842882893189, 0, 8.21648845837887307, 0, -38.3436128057680743, 0, 30.6748902446144595, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 17.4600379740551053, 0, 0, 0, 0, 0, 0, -23.2800506320734737, 0, -186.24040505658779, 0, 0, 0, 0, 0, 0, 0, 0, -81.480177212257158, 0, 434.560945132038176, 0, 260.736567079222906, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -23.2800506320734737, 0, 434.560945132038176, 0, -869.121890264076352, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 17.4600379740551053, 0, -186.24040505658779, 0, 260.736567079222906, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = 13/2
        0, 0, 0, 0, -9.46903507810723996, 0, 0, 0, 0, 0, 0, 12.6253801041429866, 0, 101.003040833143893, 0, 0, 0, 0, 0, 0, 0, 0, 44.1888303645004531, 0, -235.673761944002417, 0, -141.40425716640145, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12.6253801041429866, 0, -235.673761944002417, 0, 471.347523888004833, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -9.46903507810723996, 0, 101.003040833143893, 0, -141.40425716640145, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -2.20944151822502266, 0, 0, 0, 0, 6.62832455467506797, 0, 79.5398946561008156, 0, 0, 0, 0, 0, 0, 13.2566491093501359, 0, -318.159578624403263, 0, -212.106385749602175, 0, 0, 0, 0, 0, 0, 0, 0, -1.89380701562144799, 0, -159.079789312201631, 0, 1060.53192874801088, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -5.99705554946791864, 0, 227.256841874573759, 0, -636.319157248806525, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.315634502603574665, 0, -11.3628420937286879, 0, 30.3009122499431679, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = 15/2
        0, 1.0415407201146763, 0, 0, 0, 0, -3.1246221603440289, 0, -37.4954659241283468, 0, 0, 0, 0, 0, 0, -6.2492443206880578, 0, 149.981863696513387, 0, 99.9879091310089248, 0, 0, 0, 0, 0, 0, 0, 0, 0.892749188669722543, 0, 74.9909318482566936, 0, -499.939545655044624, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.82703909745412139, 0, -107.129902640366705, 0, 299.963727393026774, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.148791531444953757, 0, 5.35649513201833526, 0, -14.2839870187155607, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -22.6163127796329711, 0, 0, 0, 0, 0, 0, 135.697876677797827, 0, 135.697876677797827, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -949.885136744584786, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -135.697876677797827, 0, 949.885136744584786, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 22.6163127796329711, 0, -135.697876677797827, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = 17/2
        0, 0, 0, 0, 8.98681142223825956, 0, 0, 0, 0, 0, 0, -53.9208685334295574, 0, -53.9208685334295574, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 377.446079734006901, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 53.9208685334295574, 0, -377.446079734006901, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -8.98681142223825956, 0, 53.9208685334295574, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 3.37005428333934733, 0, 0, 0, 0, -28.0837856944945611, 0, -67.4010856667869467, 0, 0, 0, 0, 0, 0, 15.7269199889169542, 0, 629.076799556678169, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 33.7005428333934733, 0, -943.615199335017254, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -13.1057666574307952, 0, 269.604342667147787, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.374450475926594148, 0, -7.48900951853188297, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = 19/2
        0, -1.06570473737587758, 0, 0, 0, 0, 8.88087281146564647, 0, 21.3140947475175515, 0, 0, 0, 0, 0, 0, -4.97328877442076202, 0, -198.931550976830481, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -10.6570473737587758, 0, 298.397326465245721, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4.14440731201730169, 0, -85.2563789900702061, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.11841163748620862, 0, 2.36823274972417239, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 24.8664438721038101, 0, 0, 0, 0, 0, 0, -298.397326465245721, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 626.634385577016015, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -298.397326465245721, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 24.8664438721038101, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 21/2, mj = 21/2
        0, 0, 0, 0, -5.42630291944221461, 0, 0, 0, 0, 0, 0, 65.1156350333065753, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -136.742833569943808, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 65.1156350333065753, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -5.42630291944221461, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -5.96893321138643607, 0, 0, 0, 0, 89.533998170796541, 0, 0, 0, 0, 0, 0, 0, 0, -250.695194878230315, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 179.067996341593082, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -29.8446660569321803, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.542630291944221461, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = -23/2
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -6.10308329187281232, 0, 0, 0, 0, 91.5462493780921848, 0, 0, 0, 0, 0, 0, 0, 0, -256.329498258658117, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 183.09249875618437, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -30.5154164593640616, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.554825753806619302, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = -21/2
        0, -1.27258083654638245, 0, 0, 0, 0, 19.0887125481957368, 0, 0, 0, 0, 0, 0, 0, 0, -53.4483951349480631, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 38.1774250963914736, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -6.36290418273191227, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.115689166958762041, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -25.4516167309276491, 0, 0, 0, 0, 0, 0, 305.419400771131789, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -641.380741619376757, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 305.419400771131789, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -25.4516167309276491, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = -19/2
        0, 0, 0, 0, -7.67395118221990013, 0, 0, 0, 0, 0, 0, 92.0874141866388015, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -193.383569791941483, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 92.0874141866388015, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -7.67395118221990013, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 3.45327803199895506, 0, 0, 0, 0, -28.7773169333246255, 0, -69.0655606399791011, 0, 0, 0, 0, 0, 0, 16.1152974826617903, 0, 644.611899306471611, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 34.5327803199895506, 0, -966.917848959707416, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -13.4294145688848252, 0, 276.262242559916405, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.383697559110995006, 0, -7.67395118221990013, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = -17/2
        0, 1.30521641151882637, 0, 0, 0, 0, -10.876803429323553, 0, -26.1043282303765273, 0, 0, 0, 0, 0, 0, 6.09100992042118971, 0, 243.640396816847588, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 13.0521641151882637, 0, -365.460595225271382, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -5.07584160035099142, 0, 104.417312921506109, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.145024045724314041, 0, -2.90048091448628081, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 23.2038473158902465, 0, 0, 0, 0, 0, 0, -139.223083895341479, 0, -139.223083895341479, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 974.561587267390353, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 139.223083895341479, 0, -974.561587267390353, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -23.2038473158902465, 0, 139.223083895341479, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = -15/2
        0, 0, 0, 0, 10.3770759875713254, 0, 0, 0, 0, 0, 0, -62.2624559254279527, 0, -62.2624559254279527, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 435.837191477995669, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 62.2624559254279527, 0, -435.837191477995669, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -10.3770759875713254, 0, 62.2624559254279527, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -2.26998537228122744, 0, 0, 0, 0, 6.80995611684368233, 0, 81.7194734021241879, 0, 0, 0, 0, 0, 0, 13.6199122336873647, 0, -326.877893608496752, 0, -217.918595738997834, 0, 0, 0, 0, 0, 0, 0, 0, -1.94570174766962352, 0, -163.438946804248376, 0, 1089.59297869498917, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -6.16138886762047449, 0, 233.484209720354823, 0, -653.755787216993503, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.32428362461160392, 0, -11.6742104860177411, 0, 31.1312279627139763, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = -13/2
        0, -1.16447792575524938, 0, 0, 0, 0, 3.49343377726574814, 0, 41.9212053271889777, 0, 0, 0, 0, 0, 0, 6.98686755453149629, 0, -167.684821308755911, 0, -111.789880872503941, 0, 0, 0, 0, 0, 0, 0, 0, -0.998123936361642327, 0, -83.8424106543779555, 0, 558.949404362519703, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3.16072579847853403, 0, 119.774872363397079, 0, -335.369642617511822, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.166353989393607054, 0, -5.98874361816985396, 0, 15.9699829817862772, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -17.9662308545095619, 0, 0, 0, 0, 0, 0, 23.9549744726794158, 0, 191.639795781435327, 0, 0, 0, 0, 0, 0, 0, 0, 83.8424106543779555, 0, -447.159523490015762, 0, -268.295714094009457, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 23.9549744726794158, 0, -447.159523490015762, 0, 894.319046980031525, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -17.9662308545095619, 0, 191.639795781435327, 0, -268.295714094009457, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = -11/2
        0, 0, 0, 0, -10.3728082201740557, 0, 0, 0, 0, 0, 0, 13.8304109602320742, 0, 110.643287681856594, 0, 0, 0, 0, 0, 0, 0, 0, 48.4064383608122598, 0, -258.167671257665386, 0, -154.900602754599231, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 13.8304109602320742, 0, -258.167671257665386, 0, 516.335342515330771, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -10.3728082201740557, 0, 110.643287681856594, 0, -154.900602754599231, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 1.44066780835750773, 0, 0, 0, 0, 1.44066780835750773, 0, -69.1520548011603712, 0, 0, 0, 0, 0, 0, -4.03386986340102165, 0, 0, 0, 322.709589072081732, 0, 0, 0, 0, 0, 0, 0, 0, -6.33893835677303403, 0, 193.625753443249039, 0, -322.709589072081732, 0, -258.167671257665386, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.01693493170051083, 0, 110.643287681856594, 0, -580.877260329747118, 0, 516.335342515330771, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.288133561671501547, 0, -13.8304109602320742, 0, 64.5419178144163464, 0, -51.6335342515330771, 0, 0, 0, 0, 0,
        // j = 23/2, mj = -9/2
        0, 0.924460610246778137, 0, 0, 0, 0, 0.924460610246778137, 0, -44.3741092918453506, 0, 0, 0, 0, 0, 0, -2.58848970869097878, 0, 0, 0, 207.079176695278303, 0, 0, 0, 0, 0, 0, 0, 0, -4.0676266850858238, 0, 124.247506017166982, 0, -207.079176695278303, 0, -165.663341356222642, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.29424485434548939, 0, 70.998574866952561, 0, -372.742518051500945, 0, 331.326682712445284, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.184892122049355627, 0, -8.87482185836907012, 0, 41.4158353390556606, 0, -33.1326682712445284, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 11.8330958111587602, 0, 0, 0, 0, 0, 0, 23.6661916223175203, 0, -165.663341356222642, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -165.663341356222642, 0, 397.592019254934341, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -23.6661916223175203, 0, 165.663341356222642, 0, 0, 0, -189.329532978540163, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -11.8330958111587602, 0, 165.663341356222642, 0, -397.592019254934341, 0, 189.329532978540163, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = -7/2
        0, 0, 0, 0, 8.36726229050048951, 0, 0, 0, 0, 0, 0, 16.734524581000979, 0, -117.141672067006853, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -117.141672067006853, 0, 281.140012960816447, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -16.734524581000979, 0, 117.141672067006853, 0, 0, 0, -133.876196648007832, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -8.36726229050048951, 0, 117.141672067006853, 0, -281.140012960816447, 0, 133.876196648007832, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -0.784430839734420891, 0, 0, 0, 0, -2.87624641235954327, 0, 43.9281270251275699, 0, 0, 0, 0, 0, 0, -3.66067725209396416, 0, 117.141672067006853, 0, -263.568762150765419, 0, 0, 0, 0, 0, 0, 0, 0, -1.56886167946884178, 0, 87.8562540502551398, 0, -439.281270251275699, 0, 351.425016201020559, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.261476946578140297, 0, 0, 0, -87.8562540502551398, 0, 234.283344134013706, 0, -100.407147486005874, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.261476946578140297, 0, -14.6427090083758566, 0, 87.8562540502551398, 0, -117.141672067006853, 0, 33.469049162001958, 0, 0, 0,
        // j = 23/2, mj = -5/2
        0, -0.607617515708582275, 0, 0, 0, 0, -2.22793089093146834, 0, 34.0265808796806074, 0, 0, 0, 0, 0, 0, -2.83554840664005062, 0, 90.7375490124816198, 0, -204.159485278083644, 0, 0, 0, 0, 0, 0, 0, 0, -1.21523503141716455, 0, 68.0531617593612148, 0, -340.265808796806074, 0, 272.212647037444859, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.202539171902860758, 0, 0, 0, -68.0531617593612148, 0, 181.47509802496324, 0, -77.7750420106985312, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.202539171902860758, 0, -11.3421936265602025, 0, 68.0531617593612148, 0, -90.7375490124816198, 0, 25.9250140035661771, 0, 0, 0,
        0, 0, 0, 0, -5.67109681328010124, 0, 0, 0, 0, 0, 0, -22.6843872531204049, 0, 90.7375490124816198, 0, 0, 0, 0, 0, 0, 0, 0, -34.0265808796806074, 0, 272.212647037444859, 0, -272.212647037444859, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -22.6843872531204049, 0, 272.212647037444859, 0, -544.425294074889719, 0, 207.400112028529417, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -5.67109681328010124, 0, 90.7375490124816198, 0, -272.212647037444859, 0, 207.400112028529417, 0, -34.5666853380882361, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = -3/2
        0, 0, 0, 0, -4.79295160072100929, 0, 0, 0, 0, 0, 0, -19.1718064028840372, 0, 76.6872256115361487, 0, 0, 0, 0, 0, 0, 0, 0, -28.7577096043260558, 0, 230.061676834608446, 0, -230.061676834608446, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -19.1718064028840372, 0, 230.061676834608446, 0, -460.123353669216892, 0, 175.285087112082626, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4.79295160072100929, 0, 76.6872256115361487, 0, -230.061676834608446, 0, 175.285087112082626, 0, -29.2141811853471043, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0.239647580036050465, 0, 0, 0, 0, 1.19823790018025232, 0, -14.3788548021630279, 0, 0, 0, 0, 0, 0, 2.39647580036050465, 0, -57.5154192086521115, 0, 95.8590320144201859, 0, 0, 0, 0, 0, 0, 0, 0, 2.39647580036050465, 0, -86.2731288129781673, 0, 287.577096043260558, 0, -153.374451223072297, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.19823790018025232, 0, -57.5154192086521115, 0, 287.577096043260558, 0, -306.748902446144595, 0, 65.7319076670309846, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.239647580036050465, 0, -14.3788548021630279, 0, 95.8590320144201859, 0, -153.374451223072297, 0, 65.7319076670309846, 0, -5.84283623706942085, 0,
        // j = 23/2, mj = -1/2
        0, 0.220443711424950199, 0, 0, 0, 0, 1.10221855712475099, 0, -13.2266226854970119, 0, 0, 0, 0, 0, 0, 2.20443711424950199, 0, -52.9064907419880478, 0, 88.1774845699800796, 0, 0, 0, 0, 0, 0, 0, 0, 2.20443711424950199, 0, -79.3597361129820716, 0, 264.532453709940239, 0, -141.083975311968127, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.10221855712475099, 0, -52.9064907419880478, 0, 264.532453709940239, 0, -282.167950623936255, 0, 60.4645608479863403, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.220443711424950199, 0, -13.2266226854970119, 0, 88.1774845699800796, 0, -141.083975311968127, 0, 60.4645608479863403, 0, -5.37462763093211914, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = 1/2
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0.220443711424950199, 0, 0, 0, 0, 1.10221855712475099, 0, -13.2266226854970119, 0, 0, 0, 0, 0, 0, 2.20443711424950199, 0, -52.9064907419880478, 0, 88.1774845699800796, 0, 0, 0, 0, 0, 0, 0, 0, 2.20443711424950199, 0, -79.3597361129820716, 0, 264.532453709940239, 0, -141.083975311968127, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.10221855712475099, 0, -52.9064907419880478, 0, 264.532453709940239, 0, -282.167950623936255, 0, 60.4645608479863403, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.220443711424950199, 0, -13.2266226854970119, 0, 88.1774845699800796, 0, -141.083975311968127, 0, 60.4645608479863403, 0, -5.37462763093211914, 0,
        // j = 23/2, mj = 3/2
        0, 0.239647580036050465, 0, 0, 0, 0, 1.19823790018025232, 0, -14.3788548021630279, 0, 0, 0, 0, 0, 0, 2.39647580036050465, 0, -57.5154192086521115, 0, 95.8590320144201859, 0, 0, 0, 0, 0, 0, 0, 0, 2.39647580036050465, 0, -86.2731288129781673, 0, 287.577096043260558, 0, -153.374451223072297, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.19823790018025232, 0, -57.5154192086521115, 0, 287.577096043260558, 0, -306.748902446144595, 0, 65.7319076670309846, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.239647580036050465, 0, -14.3788548021630279, 0, 95.8590320144201859, 0, -153.374451223072297, 0, 65.7319076670309846, 0, -5.84283623706942085, 0,
        0, 0, 0, 0, 4.79295160072100929, 0, 0, 0, 0, 0, 0, 19.1718064028840372, 0, -76.6872256115361487, 0, 0, 0, 0, 0, 0, 0, 0, 28.7577096043260558, 0, -230.061676834608446, 0, 230.061676834608446, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 19.1718064028840372, 0, -230.061676834608446, 0, 460.123353669216892, 0, -175.285087112082626, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4.79295160072100929, 0, -76.6872256115361487, 0, 230.061676834608446, 0, -175.285087112082626, 0, 29.2141811853471043, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = 5/2
        0, 0, 0, 0, 5.67109681328010124, 0, 0, 0, 0, 0, 0, 22.6843872531204049, 0, -90.7375490124816198, 0, 0, 0, 0, 0, 0, 0, 0, 34.0265808796806074, 0, -272.212647037444859, 0, 272.212647037444859, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 22.6843872531204049, 0, -272.212647037444859, 0, 544.425294074889719, 0, -207.400112028529417, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5.67109681328010124, 0, -90.7375490124816198, 0, 272.212647037444859, 0, -207.400112028529417, 0, 34.5666853380882361, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -0.607617515708582275, 0, 0, 0, 0, -2.22793089093146834, 0, 34.0265808796806074, 0, 0, 0, 0, 0, 0, -2.83554840664005062, 0, 90.7375490124816198, 0, -204.159485278083644, 0, 0, 0, 0, 0, 0, 0, 0, -1.21523503141716455, 0, 68.0531617593612148, 0, -340.265808796806074, 0, 272.212647037444859, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.202539171902860758, 0, 0, 0, -68.0531617593612148, 0, 181.47509802496324, 0, -77.7750420106985312, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.202539171902860758, 0, -11.3421936265602025, 0, 68.0531617593612148, 0, -90.7375490124816198, 0, 25.9250140035661771, 0, 0, 0,
        // j = 23/2, mj = 7/2
        0, -0.784430839734420891, 0, 0, 0, 0, -2.87624641235954327, 0, 43.9281270251275699, 0, 0, 0, 0, 0, 0, -3.66067725209396416, 0, 117.141672067006853, 0, -263.568762150765419, 0, 0, 0, 0, 0, 0, 0, 0, -1.56886167946884178, 0, 87.8562540502551398, 0, -439.281270251275699, 0, 351.425016201020559, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.261476946578140297, 0, 0, 0, -87.8562540502551398, 0, 234.283344134013706, 0, -100.407147486005874, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.261476946578140297, 0, -14.6427090083758566, 0, 87.8562540502551398, 0, -117.141672067006853, 0, 33.469049162001958, 0, 0, 0,
        0, 0, 0, 0, -8.36726229050048951, 0, 0, 0, 0, 0, 0, -16.734524581000979, 0, 117.141672067006853, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 117.141672067006853, 0, -281.140012960816447, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16.734524581000979, 0, -117.141672067006853, 0, 0, 0, 133.876196648007832, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 8.36726229050048951, 0, -117.141672067006853, 0, 281.140012960816447, 0, -133.876196648007832, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = 9/2
        0, 0, 0, 0, -11.8330958111587602, 0, 0, 0, 0, 0, 0, -23.6661916223175203, 0, 165.663341356222642, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 165.663341356222642, 0, -397.592019254934341, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 23.6661916223175203, 0, -165.663341356222642, 0, 0, 0, 189.329532978540163, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 11.8330958111587602, 0, -165.663341356222642, 0, 397.592019254934341, 0, -189.329532978540163, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0.924460610246778137, 0, 0, 0, 0, 0.924460610246778137, 0, -44.3741092918453506, 0, 0, 0, 0, 0, 0, -2.58848970869097878, 0, 0, 0, 207.079176695278303, 0, 0, 0, 0, 0, 0, 0, 0, -4.0676266850858238, 0, 124.247506017166982, 0, -207.079176695278303, 0, -165.663341356222642, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.29424485434548939, 0, 70.998574866952561, 0, -372.742518051500945, 0, 331.326682712445284, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.184892122049355627, 0, -8.87482185836907012, 0, 41.4158353390556606, 0, -33.1326682712445284, 0, 0, 0, 0, 0,
        // j = 23/2, mj = 11/2
        0, 1.44066780835750773, 0, 0, 0, 0, 1.44066780835750773, 0, -69.1520548011603712, 0, 0, 0, 0, 0, 0, -4.03386986340102165, 0, 0, 0, 322.709589072081732, 0, 0, 0, 0, 0, 0, 0, 0, -6.33893835677303403, 0, 193.625753443249039, 0, -322.709589072081732, 0, -258.167671257665386, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.01693493170051083, 0, 110.643287681856594, 0, -580.877260329747118, 0, 516.335342515330771, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.288133561671501547, 0, -13.8304109602320742, 0, 64.5419178144163464, 0, -51.6335342515330771, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 10.3728082201740557, 0, 0, 0, 0, 0, 0, -13.8304109602320742, 0, -110.643287681856594, 0, 0, 0, 0, 0, 0, 0, 0, -48.4064383608122598, 0, 258.167671257665386, 0, 154.900602754599231, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -13.8304109602320742, 0, 258.167671257665386, 0, -516.335342515330771, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 10.3728082201740557, 0, -110.643287681856594, 0, 154.900602754599231, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = 13/2
        0, 0, 0, 0, 17.9662308545095619, 0, 0, 0, 0, 0, 0, -23.9549744726794158, 0, -191.639795781435327, 0, 0, 0, 0, 0, 0, 0, 0, -83.8424106543779555, 0, 447.159523490015762, 0, 268.295714094009457, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -23.9549744726794158, 0, 447.159523490015762, 0, -894.319046980031525, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 17.9662308545095619, 0, -191.639795781435327, 0, 268.295714094009457, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -1.16447792575524938, 0, 0, 0, 0, 3.49343377726574814, 0, 41.9212053271889777, 0, 0, 0, 0, 0, 0, 6.98686755453149629, 0, -167.684821308755911, 0, -111.789880872503941, 0, 0, 0, 0, 0, 0, 0, 0, -0.998123936361642327, 0, -83.8424106543779555, 0, 558.949404362519703, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3.16072579847853403, 0, 119.774872363397079, 0, -335.369642617511822, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.166353989393607054, 0, -5.98874361816985396, 0, 15.9699829817862772, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = 15/2
        0, -2.26998537228122744, 0, 0, 0, 0, 6.80995611684368233, 0, 81.7194734021241879, 0, 0, 0, 0, 0, 0, 13.6199122336873647, 0, -326.877893608496752, 0, -217.918595738997834, 0, 0, 0, 0, 0, 0, 0, 0, -1.94570174766962352, 0, -163.438946804248376, 0, 1089.59297869498917, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -6.16138886762047449, 0, 233.484209720354823, 0, -653.755787216993503, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.32428362461160392, 0, -11.6742104860177411, 0, 31.1312279627139763, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -10.3770759875713254, 0, 0, 0, 0, 0, 0, 62.2624559254279527, 0, 62.2624559254279527, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -435.837191477995669, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -62.2624559254279527, 0, 435.837191477995669, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 10.3770759875713254, 0, -62.2624559254279527, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = 17/2
        0, 0, 0, 0, -23.2038473158902465, 0, 0, 0, 0, 0, 0, 139.223083895341479, 0, 139.223083895341479, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -974.561587267390353, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -139.223083895341479, 0, 974.561587267390353, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 23.2038473158902465, 0, -139.223083895341479, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 1.30521641151882637, 0, 0, 0, 0, -10.876803429323553, 0, -26.1043282303765273, 0, 0, 0, 0, 0, 0, 6.09100992042118971, 0, 243.640396816847588, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 13.0521641151882637, 0, -365.460595225271382, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -5.07584160035099142, 0, 104.417312921506109, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.145024045724314041, 0, -2.90048091448628081, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = 19/2
        0, 3.45327803199895506, 0, 0, 0, 0, -28.7773169333246255, 0, -69.0655606399791011, 0, 0, 0, 0, 0, 0, 16.1152974826617903, 0, 644.611899306471611, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 34.5327803199895506, 0, -966.917848959707416, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -13.4294145688848252, 0, 276.262242559916405, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.383697559110995006, 0, -7.67395118221990013, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 7.67395118221990013, 0, 0, 0, 0, 0, 0, -92.0874141866388015, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 193.383569791941483, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -92.0874141866388015, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 7.67395118221990013, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = 21/2
        0, 0, 0, 0, 25.4516167309276491, 0, 0, 0, 0, 0, 0, -305.419400771131789, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 641.380741619376757, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -305.419400771131789, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 25.4516167309276491, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -1.27258083654638245, 0, 0, 0, 0, 19.0887125481957368, 0, 0, 0, 0, 0, 0, 0, 0, -53.4483951349480631, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 38.1774250963914736, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -6.36290418273191227, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.115689166958762041, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = 23/2
        0, -6.10308329187281232, 0, 0, 0, 0, 91.5462493780921848, 0, 0, 0, 0, 0, 0, 0, 0, -256.329498258658117, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 183.09249875618437, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -30.5154164593640616, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.554825753806619302, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = -23/2
        0, 6.65790904567943162, 0, 0, 0, 0, -122.061665837456246, 0, 0, 0, 0, 0, 0, 0, 0, 439.421997014842487, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -439.421997014842487, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 122.061665837456246, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -6.65790904567943162, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -6.10308329187281232, 0, 0, 0, 0, 0, 0, 91.5462493780921848, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -256.329498258658117, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 183.09249875618437, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -30.5154164593640616, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.554825753806619302, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = -21/2
        0, 0, 0, 0, 29.2693592405667964, 0, 0, 0, 0, 0, 0, -439.040388608501947, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1229.31308810380545, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -878.080777217003893, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 146.346796202833982, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.66085084005152695, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 1.15689166958762041, 0, 0, 0, 0, -12.7258083654638245, 0, -25.4516167309276491, 0, 0, 0, 0, 0, 0, 15.2709700385565894, 0, 305.419400771131789, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 15.2709700385565894, 0, -641.380741619376757, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -12.7258083654638245, 0, 305.419400771131789, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.15689166958762041, 0, -25.4516167309276491, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = -19/2
        0, -3.83697559110995006, 0, 0, 0, 0, 42.2067315022094507, 0, 84.4134630044189014, 0, 0, 0, 0, 0, 0, -50.6480778026513408, 0, -1012.96155605302682, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -50.6480778026513408, 0, 2127.21926771135631, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 42.2067315022094507, 0, -1012.96155605302682, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3.83697559110995006, 0, 84.4134630044189014, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 10.3598340959968652, 0, 0, 0, 0, 0, 0, -86.3319507999738764, 0, -69.0655606399791011, 0, 0, 0, 0, 0, 0, 0, 0, 48.3458924479853708, 0, 644.611899306471611, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 103.598340959968652, 0, -966.917848959707416, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -40.2882437066544757, 0, 276.262242559916405, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.15109267733298502, 0, -7.67395118221990013, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = -17/2
        0, 0, 0, 0, -27.4095446418953537, 0, 0, 0, 0, 0, 0, 228.412872015794614, 0, 182.730297612635691, 0, 0, 0, 0, 0, 0, 0, 0, -127.911208328844984, 0, -1705.48277771793312, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -274.095446418953537, 0, 2558.22416657689968, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 106.59267360737082, 0, -730.921190450542765, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3.04550496021059485, 0, 20.3033664014039657, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -1.16019236579451232, 0, 0, 0, 0, 5.80096182897256162, 0, 46.407694631780493, 0, 0, 0, 0, 0, 0, 6.96115419476707395, 0, -278.446167790682958, 0, -139.223083895341479, 0, 0, 0, 0, 0, 0, 0, 0, -6.96115419476707395, 0, 0, 0, 974.561587267390353, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -5.80096182897256162, 0, 278.446167790682958, 0, -974.561587267390353, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.16019236579451232, 0, -46.407694631780493, 0, 139.223083895341479, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = -15/2
        0, 2.59426899689283136, 0, 0, 0, 0, -12.9713449844641568, 0, -103.770759875713254, 0, 0, 0, 0, 0, 0, -15.5656139813569882, 0, 622.624559254279527, 0, 311.312279627139763, 0, 0, 0, 0, 0, 0, 0, 0, 15.5656139813569882, 0, 0, 0, -2179.18595738997834, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12.9713449844641568, 0, -622.624559254279527, 0, 2179.18595738997834, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.59426899689283136, 0, 103.770759875713254, 0, -311.312279627139763, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -11.3499268614061372, 0, 0, 0, 0, 0, 0, 34.0497805842184116, 0, 136.199122336873647, 0, 0, 0, 0, 0, 0, 0, 0, 68.0995611684368233, 0, -544.796489347494586, 0, -217.918595738997834, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -9.72850873834811761, 0, -272.398244673747293, 0, 1089.59297869498917, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -30.8069443381023724, 0, 389.140349533924704, 0, -653.755787216993503, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.6214181230580196, 0, -19.4570174766962352, 0, 31.1312279627139763, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = -13/2
        0, 0, 0, 0, 22.1250805893497382, 0, 0, 0, 0, 0, 0, -66.3752417680492147, 0, -265.500967072196859, 0, 0, 0, 0, 0, 0, 0, 0, -132.750483536098429, 0, 1062.00386828878744, 0, 424.801547315514974, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 18.9643547908712042, 0, 531.001934144393718, 0, -2124.00773657757487, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 60.0537901710921467, 0, -758.574191634848168, 0, 1274.40464194654492, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3.16072579847853403, 0, 37.9287095817424084, 0, -60.6859353307878535, 0, 0, 0, 0, 0, 0, 0,
        0, 0.998123936361642327, 0, 0, 0, 0, -0.332707978787214109, 0, -53.8986925635286856, 0, 0, 0, 0, 0, 0, -5.98874361816985396, 0, 71.8649234180382475, 0, 287.45969367215299, 0, 0, 0, 0, 0, 0, 0, 0, -5.98874361816985396, 0, 251.527231963133866, 0, -670.739285235023644, 0, -268.295714094009457, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.332707978787214109, 0, 71.8649234180382475, 0, -670.739285235023644, 0, 894.319046980031525, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.998123936361642327, 0, -53.8986925635286856, 0, 287.45969367215299, 0, -268.295714094009457, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = -11/2
        0, -1.72880137002900928, 0, 0, 0, 0, 0.576267123343003093, 0, 93.3552739815665011, 0, 0, 0, 0, 0, 0, 10.3728082201740557, 0, -124.473698642088668, 0, -497.894794568354673, 0, 0, 0, 0, 0, 0, 0, 0, 10.3728082201740557, 0, -435.657945247310338, 0, 1161.75452065949424, 0, 464.701808263797694, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.576267123343003093, 0, -124.473698642088668, 0, 1161.75452065949424, 0, -1549.00602754599231, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.72880137002900928, 0, 93.3552739815665011, 0, -497.894794568354673, 0, 464.701808263797694, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 10.0846746585025541, 0, 0, 0, 0, 0, 0, 10.0846746585025541, 0, -161.354794536040866, 0, 0, 0, 0, 0, 0, 0, 0, -28.2370890438071516, 0, 0, 0, 451.793424700914425, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -44.3725684974112382, 0, 451.793424700914425, 0, -451.793424700914425, 0, -258.167671257665386, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -14.1185445219035758, 0, 258.167671257665386, 0, -813.228164461645965, 0, 516.335342515330771, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.01693493170051083, 0, -32.2709589072081732, 0, 90.358684940182885, 0, -51.6335342515330771, 0, 0, 0, 0, 0,
        // j = 23/2, mj = -9/2
        0, 0, 0, 0, -15.7158303741952283, 0, 0, 0, 0, 0, 0, -15.7158303741952283, 0, 251.453285987123653, 0, 0, 0, 0, 0, 0, 0, 0, 44.0043250477466393, 0, 0, 0, -704.06920076394623, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 69.1496536464590047, 0, -704.06920076394623, 0, 704.06920076394623, 0, 402.325257579397845, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 22.0021625238733197, 0, -402.325257579397845, 0, 1267.32456137510321, 0, -804.650515158795691, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3.14316607483904567, 0, 50.2906571974247307, 0, -140.813840152789246, 0, 80.4650515158795691, 0, 0, 0, 0, 0,
        0, -0.73956848819742251, 0, 0, 0, 0, -2.21870546459226753, 0, 47.3323832446350406, 0, 0, 0, 0, 0, 0, -1.47913697639484502, 0, 94.6647664892700813, 0, -331.326682712445284, 0, 0, 0, 0, 0, 0, 0, 0, 1.47913697639484502, 0, 0, 0, -331.326682712445284, 0, 530.122692339912455, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.21870546459226753, 0, -94.6647664892700813, 0, 331.326682712445284, 0, 0, 0, -189.329532978540163, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.73956848819742251, 0, -47.3323832446350406, 0, 331.326682712445284, 0, -530.122692339912455, 0, 189.329532978540163, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = -7/2
        0, 1.04590778631256119, 0, 0, 0, 0, 3.13772335893768356, 0, -66.938098324003916, 0, 0, 0, 0, 0, 0, 2.09181557262512238, 0, -133.876196648007832, 0, 468.566688268027412, 0, 0, 0, 0, 0, 0, 0, 0, -2.09181557262512238, 0, 0, 0, 468.566688268027412, 0, -749.70670122884386, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3.13772335893768356, 0, 133.876196648007832, 0, -468.566688268027412, 0, 0, 0, 267.752393296015664, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.04590778631256119, 0, 66.938098324003916, 0, -468.566688268027412, 0, 749.70670122884386, 0, -267.752393296015664, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -7.05987755760978802, 0, 0, 0, 0, 0, 0, -25.8862177112358894, 0, 131.78438107538271, 0, 0, 0, 0, 0, 0, 0, 0, -32.9460952688456774, 0, 351.425016201020559, 0, -474.423771871377755, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -14.119755115219576, 0, 263.568762150765419, 0, -790.706286452296258, 0, 451.832163687026433, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.35329251920326267, 0, 0, 0, -158.141257290459252, 0, 301.221442458017622, 0, -100.407147486005874, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.35329251920326267, 0, -43.9281270251275699, 0, 158.141257290459252, 0, -150.610721229008811, 0, 33.469049162001958, 0, 0, 0,
        // j = 23/2, mj = -5/2
        0, 0, 0, 0, 9.11426273562873413, 0, 0, 0, 0, 0, 0, 33.4189633639720251, 0, -170.132904398403037, 0, 0, 0, 0, 0, 0, 0, 0, 42.5332260996007593, 0, -453.687745062408099, 0, 612.478455834250933, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 18.2285254712574683, 0, -340.265808796806074, 0, 1020.79742639041822, 0, -583.312815080238984, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3.03808757854291138, 0, 0, 0, 204.159485278083644, 0, -388.875210053492656, 0, 129.625070017830885, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3.03808757854291138, 0, 56.7109681328010124, 0, -204.159485278083644, 0, 194.437605026746328, 0, -43.2083566726102951, 0, 0, 0,
        0, 0.405078343805721517, 0, 0, 0, 0, 2.02539171902860758, 0, -28.3554840664005062, 0, 0, 0, 0, 0, 0, 4.05078343805721517, 0, -113.421936265602025, 0, 226.843872531204049, 0, 0, 0, 0, 0, 0, 0, 0, 4.05078343805721517, 0, -170.132904398403037, 0, 680.531617593612148, 0, -453.687745062408099, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.02539171902860758, 0, -113.421936265602025, 0, 680.531617593612148, 0, -907.375490124816198, 0, 259.250140035661771, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.405078343805721517, 0, -28.3554840664005062, 0, 226.843872531204049, 0, -453.687745062408099, 0, 259.250140035661771, 0, -34.5666853380882361, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = -3/2
        0, -0.479295160072100929, 0, 0, 0, 0, -2.39647580036050465, 0, 33.5506612050470651, 0, 0, 0, 0, 0, 0, -4.79295160072100929, 0, 134.20264482018826, 0, -268.40528964037652, 0, 0, 0, 0, 0, 0, 0, 0, -4.79295160072100929, 0, 201.30396723028239, 0, -805.215868921129561, 0, 536.810579280753041, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.39647580036050465, 0, 134.20264482018826, 0, -805.215868921129561, 0, 1073.62115856150608, 0, -306.748902446144595, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.479295160072100929, 0, 33.5506612050470651, 0, -268.40528964037652, 0, 536.810579280753041, 0, -306.748902446144595, 0, 40.899853659485946, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 2.63612338039655511, 0, 0, 0, 0, 0, 0, 13.1806169019827756, 0, -52.7224676079311022, 0, 0, 0, 0, 0, 0, 0, 0, 26.3612338039655511, 0, -210.889870431724409, 0, 210.889870431724409, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 26.3612338039655511, 0, -316.334805647586613, 0, 632.669611295173227, 0, -241.01699477911361, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 13.1806169019827756, 0, -210.889870431724409, 0, 632.669611295173227, 0, -482.03398955822722, 0, 80.3389982597045367, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.63612338039655511, 0, -52.7224676079311022, 0, 210.889870431724409, 0, -241.01699477911361, 0, 80.3389982597045367, 0, -5.84283623706942085, 0,
        // j = 23/2, mj = -1/2
        0, 0, 0, 0, -2.86576824852435259, 0, 0, 0, 0, 0, 0, -14.3288412426217629, 0, 57.3153649704870517, 0, 0, 0, 0, 0, 0, 0, 0, -28.6576824852435259, 0, 229.261459881948207, 0, -229.261459881948207, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -28.6576824852435259, 0, 343.89218982292231, 0, -687.784379645844621, 0, 262.013097007940808, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -14.3288412426217629, 0, 229.261459881948207, 0, -687.784379645844621, 0, 524.026194015881616, 0, -87.337699002646936, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.86576824852435259, 0, 57.3153649704870517, 0, -229.261459881948207, 0, 262.013097007940808, 0, -87.337699002646936, 0, 6.35183265473795898, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = 1/2
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 2.86576824852435259, 0, 0, 0, 0, 0, 0, 14.3288412426217629, 0, -57.3153649704870517, 0, 0, 0, 0, 0, 0, 0, 0, 28.6576824852435259, 0, -229.261459881948207, 0, 229.261459881948207, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 28.6576824852435259, 0, -343.89218982292231, 0, 687.784379645844621, 0, -262.013097007940808, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 14.3288412426217629, 0, -229.261459881948207, 0, 687.784379645844621, 0, -524.026194015881616, 0, 87.337699002646936, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.86576824852435259, 0, -57.3153649704870517, 0, 229.261459881948207, 0, -262.013097007940808, 0, 87.337699002646936, 0, -6.35183265473795898, 0,
        // j = 23/2, mj = 3/2
        0, 0, 0, 0, -2.63612338039655511, 0, 0, 0, 0, 0, 0, -13.1806169019827756, 0, 52.7224676079311022, 0, 0, 0, 0, 0, 0, 0, 0, -26.3612338039655511, 0, 210.889870431724409, 0, -210.889870431724409, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -26.3612338039655511, 0, 316.334805647586613, 0, -632.669611295173227, 0, 241.01699477911361, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -13.1806169019827756, 0, 210.889870431724409, 0, -632.669611295173227, 0, 482.03398955822722, 0, -80.3389982597045367, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.63612338039655511, 0, 52.7224676079311022, 0, -210.889870431724409, 0, 241.01699477911361, 0, -80.3389982597045367, 0, 5.84283623706942085, 0,
        0, -0.479295160072100929, 0, 0, 0, 0, -2.39647580036050465, 0, 33.5506612050470651, 0, 0, 0, 0, 0, 0, -4.79295160072100929, 0, 134.20264482018826, 0, -268.40528964037652, 0, 0, 0, 0, 0, 0, 0, 0, -4.79295160072100929, 0, 201.30396723028239, 0, -805.215868921129561, 0, 536.810579280753041, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.39647580036050465, 0, 134.20264482018826, 0, -805.215868921129561, 0, 1073.62115856150608, 0, -306.748902446144595, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.479295160072100929, 0, 33.5506612050470651, 0, -268.40528964037652, 0, 536.810579280753041, 0, -306.748902446144595, 0, 40.899853659485946, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = 5/2
        0, 0.405078343805721517, 0, 0, 0, 0, 2.02539171902860758, 0, -28.3554840664005062, 0, 0, 0, 0, 0, 0, 4.05078343805721517, 0, -113.421936265602025, 0, 226.843872531204049, 0, 0, 0, 0, 0, 0, 0, 0, 4.05078343805721517, 0, -170.132904398403037, 0, 680.531617593612148, 0, -453.687745062408099, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.02539171902860758, 0, -113.421936265602025, 0, 680.531617593612148, 0, -907.375490124816198, 0, 259.250140035661771, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.405078343805721517, 0, -28.3554840664005062, 0, 226.843872531204049, 0, -453.687745062408099, 0, 259.250140035661771, 0, -34.5666853380882361, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -9.11426273562873413, 0, 0, 0, 0, 0, 0, -33.4189633639720251, 0, 170.132904398403037, 0, 0, 0, 0, 0, 0, 0, 0, -42.5332260996007593, 0, 453.687745062408099, 0, -612.478455834250933, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -18.2285254712574683, 0, 340.265808796806074, 0, -1020.79742639041822, 0, 583.312815080238984, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.03808757854291138, 0, 0, 0, -204.159485278083644, 0, 388.875210053492656, 0, -129.625070017830885, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.03808757854291138, 0, -56.7109681328010124, 0, 204.159485278083644, 0, -194.437605026746328, 0, 43.2083566726102951, 0, 0, 0,
        // j = 23/2, mj = 7/2
        0, 0, 0, 0, 7.05987755760978802, 0, 0, 0, 0, 0, 0, 25.8862177112358894, 0, -131.78438107538271, 0, 0, 0, 0, 0, 0, 0, 0, 32.9460952688456774, 0, -351.425016201020559, 0, 474.423771871377755, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 14.119755115219576, 0, -263.568762150765419, 0, 790.706286452296258, 0, -451.832163687026433, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.35329251920326267, 0, 0, 0, 158.141257290459252, 0, -301.221442458017622, 0, 100.407147486005874, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.35329251920326267, 0, 43.9281270251275699, 0, -158.141257290459252, 0, 150.610721229008811, 0, -33.469049162001958, 0, 0, 0,
        0, 1.04590778631256119, 0, 0, 0, 0, 3.13772335893768356, 0, -66.938098324003916, 0, 0, 0, 0, 0, 0, 2.09181557262512238, 0, -133.876196648007832, 0, 468.566688268027412, 0, 0, 0, 0, 0, 0, 0, 0, -2.09181557262512238, 0, 0, 0, 468.566688268027412, 0, -749.70670122884386, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3.13772335893768356, 0, 133.876196648007832, 0, -468.566688268027412, 0, 0, 0, 267.752393296015664, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.04590778631256119, 0, 66.938098324003916, 0, -468.566688268027412, 0, 749.70670122884386, 0, -267.752393296015664, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = 9/2
        0, -0.73956848819742251, 0, 0, 0, 0, -2.21870546459226753, 0, 47.3323832446350406, 0, 0, 0, 0, 0, 0, -1.47913697639484502, 0, 94.6647664892700813, 0, -331.326682712445284, 0, 0, 0, 0, 0, 0, 0, 0, 1.47913697639484502, 0, 0, 0, -331.326682712445284, 0, 530.122692339912455, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.21870546459226753, 0, -94.6647664892700813, 0, 331.326682712445284, 0, 0, 0, -189.329532978540163, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.73956848819742251, 0, -47.3323832446350406, 0, 331.326682712445284, 0, -530.122692339912455, 0, 189.329532978540163, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 15.7158303741952283, 0, 0, 0, 0, 0, 0, 15.7158303741952283, 0, -251.453285987123653, 0, 0, 0, 0, 0, 0, 0, 0, -44.0043250477466393, 0, 0, 0, 704.06920076394623, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -69.1496536464590047, 0, 704.06920076394623, 0, -704.06920076394623, 0, -402.325257579397845, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -22.0021625238733197, 0, 402.325257579397845, 0, -1267.32456137510321, 0, 804.650515158795691, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.14316607483904567, 0, -50.2906571974247307, 0, 140.813840152789246, 0, -80.4650515158795691, 0, 0, 0, 0, 0,
        // j = 23/2, mj = 11/2
        0, 0, 0, 0, -10.0846746585025541, 0, 0, 0, 0, 0, 0, -10.0846746585025541, 0, 161.354794536040866, 0, 0, 0, 0, 0, 0, 0, 0, 28.2370890438071516, 0, 0, 0, -451.793424700914425, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 44.3725684974112382, 0, -451.793424700914425, 0, 451.793424700914425, 0, 258.167671257665386, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 14.1185445219035758, 0, -258.167671257665386, 0, 813.228164461645965, 0, -516.335342515330771, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.01693493170051083, 0, 32.2709589072081732, 0, -90.358684940182885, 0, 51.6335342515330771, 0, 0, 0, 0, 0,
        0, -1.72880137002900928, 0, 0, 0, 0, 0.576267123343003093, 0, 93.3552739815665011, 0, 0, 0, 0, 0, 0, 10.3728082201740557, 0, -124.473698642088668, 0, -497.894794568354673, 0, 0, 0, 0, 0, 0, 0, 0, 10.3728082201740557, 0, -435.657945247310338, 0, 1161.75452065949424, 0, 464.701808263797694, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.576267123343003093, 0, -124.473698642088668, 0, 1161.75452065949424, 0, -1549.00602754599231, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.72880137002900928, 0, 93.3552739815665011, 0, -497.894794568354673, 0, 464.701808263797694, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = 13/2
        0, 0.998123936361642327, 0, 0, 0, 0, -0.332707978787214109, 0, -53.8986925635286856, 0, 0, 0, 0, 0, 0, -5.98874361816985396, 0, 71.8649234180382475, 0, 287.45969367215299, 0, 0, 0, 0, 0, 0, 0, 0, -5.98874361816985396, 0, 251.527231963133866, 0, -670.739285235023644, 0, -268.295714094009457, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.332707978787214109, 0, 71.8649234180382475, 0, -670.739285235023644, 0, 894.319046980031525, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.998123936361642327, 0, -53.8986925635286856, 0, 287.45969367215299, 0, -268.295714094009457, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -22.1250805893497382, 0, 0, 0, 0, 0, 0, 66.3752417680492147, 0, 265.500967072196859, 0, 0, 0, 0, 0, 0, 0, 0, 132.750483536098429, 0, -1062.00386828878744, 0, -424.801547315514974, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -18.9643547908712042, 0, -531.001934144393718, 0, 2124.00773657757487, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -60.0537901710921467, 0, 758.574191634848168, 0, -1274.40464194654492, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.16072579847853403, 0, -37.9287095817424084, 0, 60.6859353307878535, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = 15/2
        0, 0, 0, 0, 11.3499268614061372, 0, 0, 0, 0, 0, 0, -34.0497805842184116, 0, -136.199122336873647, 0, 0, 0, 0, 0, 0, 0, 0, -68.0995611684368233, 0, 544.796489347494586, 0, 217.918595738997834, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 9.72850873834811761, 0, 272.398244673747293, 0, -1089.59297869498917, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 30.8069443381023724, 0, -389.140349533924704, 0, 653.755787216993503, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.6214181230580196, 0, 19.4570174766962352, 0, -31.1312279627139763, 0, 0, 0, 0, 0, 0, 0,
        0, 2.59426899689283136, 0, 0, 0, 0, -12.9713449844641568, 0, -103.770759875713254, 0, 0, 0, 0, 0, 0, -15.5656139813569882, 0, 622.624559254279527, 0, 311.312279627139763, 0, 0, 0, 0, 0, 0, 0, 0, 15.5656139813569882, 0, 0, 0, -2179.18595738997834, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12.9713449844641568, 0, -622.624559254279527, 0, 2179.18595738997834, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.59426899689283136, 0, 103.770759875713254, 0, -311.312279627139763, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = 17/2
        0, -1.16019236579451232, 0, 0, 0, 0, 5.80096182897256162, 0, 46.407694631780493, 0, 0, 0, 0, 0, 0, 6.96115419476707395, 0, -278.446167790682958, 0, -139.223083895341479, 0, 0, 0, 0, 0, 0, 0, 0, -6.96115419476707395, 0, 0, 0, 974.561587267390353, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -5.80096182897256162, 0, 278.446167790682958, 0, -974.561587267390353, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.16019236579451232, 0, -46.407694631780493, 0, 139.223083895341479, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 27.4095446418953537, 0, 0, 0, 0, 0, 0, -228.412872015794614, 0, -182.730297612635691, 0, 0, 0, 0, 0, 0, 0, 0, 127.911208328844984, 0, 1705.48277771793312, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 274.095446418953537, 0, -2558.22416657689968, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -106.59267360737082, 0, 730.921190450542765, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.04550496021059485, 0, -20.3033664014039657, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = 19/2
        0, 0, 0, 0, -10.3598340959968652, 0, 0, 0, 0, 0, 0, 86.3319507999738764, 0, 69.0655606399791011, 0, 0, 0, 0, 0, 0, 0, 0, -48.3458924479853708, 0, -644.611899306471611, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -103.598340959968652, 0, 966.917848959707416, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40.2882437066544757, 0, -276.262242559916405, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.15109267733298502, 0, 7.67395118221990013, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -3.83697559110995006, 0, 0, 0, 0, 42.2067315022094507, 0, 84.4134630044189014, 0, 0, 0, 0, 0, 0, -50.6480778026513408, 0, -1012.96155605302682, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -50.6480778026513408, 0, 2127.21926771135631, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 42.2067315022094507, 0, -1012.96155605302682, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3.83697559110995006, 0, 84.4134630044189014, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = 21/2
        0, 1.15689166958762041, 0, 0, 0, 0, -12.7258083654638245, 0, -25.4516167309276491, 0, 0, 0, 0, 0, 0, 15.2709700385565894, 0, 305.419400771131789, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 15.2709700385565894, 0, -641.380741619376757, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -12.7258083654638245, 0, 305.419400771131789, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.15689166958762041, 0, -25.4516167309276491, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -29.2693592405667964, 0, 0, 0, 0, 0, 0, 439.040388608501947, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1229.31308810380545, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 878.080777217003893, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -146.346796202833982, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.66085084005152695, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 23/2, mj = 23/2
        0, 0, 0, 0, 6.10308329187281232, 0, 0, 0, 0, 0, 0, -91.5462493780921848, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 256.329498258658117, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -183.09249875618437, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 30.5154164593640616, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.554825753806619302, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 6.65790904567943162, 0, 0, 0, 0, -122.061665837456246, 0, 0, 0, 0, 0, 0, 0, 0, 439.421997014842487, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -439.421997014842487, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 122.061665837456246, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -6.65790904567943162, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 25/2, mj = -25/2
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -6.79519996490629405, 0, 0, 0, 0, 124.578666023282058, 0, 0, 0, 0, 0, 0, 0, 0, -448.483197683815407, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 448.483197683815407, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -124.578666023282058, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6.79519996490629405, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 25/2, mj = -23/2
        0, -1.35903999298125881, 0, 0, 0, 0, 24.9157332046564115, 0, 0, 0, 0, 0, 0, 0, 0, -89.6966395367630815, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 89.6966395367630815, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -24.9157332046564115, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.35903999298125881, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -29.8988798455876938, 0, 0, 0, 0, 0, 0, 448.483197683815407, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1255.75295351468314, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 896.966395367630815, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -149.494399227938469, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.71807998596251762, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 25/2, mj = -21/2
        0, 0, 0, 0, -8.63106316365916572, 0, 0, 0, 0, 0, 0, 129.465947454887486, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -362.50465287368496, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 258.931894909774972, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -43.1553158182958286, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.784642105787196884, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 3.92321052893598442, 0, 0, 0, 0, -43.1553158182958286, 0, -86.3106316365916572, 0, 0, 0, 0, 0, 0, 51.7863789819549943, 0, 1035.72757963909989, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 51.7863789819549943, 0, -2175.02791724210976, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -43.1553158182958286, 0, 1035.72757963909989, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.92321052893598442, 0, -86.3106316365916572, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 25/2, mj = -19/2
        0, 1.41689713908309089, 0, 0, 0, 0, -15.5858685299139998, 0, -31.1717370598279997, 0, 0, 0, 0, 0, 0, 18.7030422358967998, 0, 374.060844717935996, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 18.7030422358967998, 0, -785.527773907665592, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -15.5858685299139998, 0, 374.060844717935996, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.41689713908309089, 0, -31.1717370598279997, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 28.0545633538451997, 0, 0, 0, 0, 0, 0, -233.788027948709998, 0, -187.030422358967998, 0, 0, 0, 0, 0, 0, 0, 0, 130.921295651277599, 0, 1745.61727535036798, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 280.545633538451997, 0, -2618.42591302555197, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -109.101079709397999, 0, 748.121689435871992, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.11717370598279997, 0, -20.7811580398853331, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 25/2, mj = -17/2
        0, 0, 0, 0, 11.9625060081673068, 0, 0, 0, 0, 0, 0, -99.6875500680608898, 0, -79.7500400544487119, 0, 0, 0, 0, 0, 0, 0, 0, 55.8250280381140983, 0, 744.333707174854644, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 119.625060081673068, 0, -1116.50056076228197, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -46.5208566984284153, 0, 319.000160217794847, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.32916733424081186, 0, -8.86111556160541243, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -2.65833466848162373, 0, 0, 0, 0, 13.2916733424081186, 0, 106.333386739264949, 0, 0, 0, 0, 0, 0, 15.9500080108897424, 0, -638.000320435589695, 0, -319.000160217794847, 0, 0, 0, 0, 0, 0, 0, 0, -15.9500080108897424, 0, 0, 0, 2233.00112152456393, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -13.2916733424081186, 0, 638.000320435589695, 0, -2233.00112152456393, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.65833466848162373, 0, -106.333386739264949, 0, 319.000160217794847, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 25/2, mj = -15/2
        0, -1.29713449844641568, 0, 0, 0, 0, 6.48567249223207841, 0, 51.8853799378566272, 0, 0, 0, 0, 0, 0, 7.78280699067849409, 0, -311.312279627139763, 0, -155.656139813569882, 0, 0, 0, 0, 0, 0, 0, 0, -7.78280699067849409, 0, 0, 0, 1089.59297869498917, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -6.48567249223207841, 0, 311.312279627139763, 0, -1089.59297869498917, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.29713449844641568, 0, -51.8853799378566272, 0, 155.656139813569882, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -22.6998537228122744, 0, 0, 0, 0, 0, 0, 68.0995611684368233, 0, 272.398244673747293, 0, 0, 0, 0, 0, 0, 0, 0, 136.199122336873647, 0, -1089.59297869498917, 0, -435.837191477995669, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -19.4570174766962352, 0, -544.796489347494586, 0, 2179.18595738997834, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -61.6138886762047449, 0, 778.280699067849409, 0, -1307.51157443398701, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.2428362461160392, 0, -38.9140349533924704, 0, 62.2624559254279527, 0, 0, 0, 0, 0, 0, 0,
        // j = 25/2, mj = -13/2
        0, 0, 0, 0, -12.4332219360519051, 0, 0, 0, 0, 0, 0, 37.2996658081557152, 0, 149.198663232622861, 0, 0, 0, 0, 0, 0, 0, 0, 74.5993316163114303, 0, -596.794652930491443, 0, -238.717861172196577, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -10.6570473737587758, 0, -298.397326465245721, 0, 1193.58930586098289, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -33.7473166835694566, 0, 426.281894950351031, 0, -716.153583516589731, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.77617456229312929, 0, -21.3140947475175515, 0, 34.1025515960280824, 0, 0, 0, 0, 0, 0, 0,
        0, 1.77617456229312929, 0, 0, 0, 0, -0.592058187431043098, 0, -95.9134263638289819, 0, 0, 0, 0, 0, 0, -10.6570473737587758, 0, 127.884568485105309, 0, 511.538273940421237, 0, 0, 0, 0, 0, 0, 0, 0, -10.6570473737587758, 0, 447.595989697868582, 0, -1193.58930586098289, 0, -477.435722344393154, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.592058187431043098, 0, 127.884568485105309, 0, -1193.58930586098289, 0, 1591.45240781464385, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.77617456229312929, 0, -95.9134263638289819, 0, 511.538273940421237, 0, -477.435722344393154, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 25/2, mj = -11/2
        0, 1.07809706940565876, 0, 0, 0, 0, -0.359365689801886252, 0, -58.2172417479055728, 0, 0, 0, 0, 0, 0, -6.46858241643395253, 0, 77.6229889972074304, 0, 310.491955988829722, 0, 0, 0, 0, 0, 0, 0, 0, -6.46858241643395253, 0, 271.680461490226006, 0, -724.481230640602684, 0, -289.792492256241073, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.359365689801886252, 0, 77.6229889972074304, 0, -724.481230640602684, 0, 965.974974187470245, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.07809706940565876, 0, -58.2172417479055728, 0, 310.491955988829722, 0, -289.792492256241073, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 16.1714560410848813, 0, 0, 0, 0, 0, 0, 16.1714560410848813, 0, -258.743296657358101, 0, 0, 0, 0, 0, 0, 0, 0, -45.2800769150376677, 0, 0, 0, 724.481230640602684, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -71.1544065807734779, 0, 724.481230640602684, 0, -724.481230640602684, 0, -413.989274651772962, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -22.6400384575188339, 0, 413.989274651772962, 0, -1304.06621515308483, 0, 827.978549303545924, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.23429120821697627, 0, -51.7486593314716203, 0, 144.896246128120537, 0, -82.7978549303545924, 0, 0, 0, 0, 0,
        // j = 25/2, mj = -9/2
        0, 0, 0, 0, 10.7809706940565876, 0, 0, 0, 0, 0, 0, 10.7809706940565876, 0, -172.495531104905401, 0, 0, 0, 0, 0, 0, 0, 0, -30.1867179433584452, 0, 0, 0, 482.987487093735122, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -47.4362710538489852, 0, 482.987487093735122, 0, -482.987487093735122, 0, -275.992849767848641, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -15.0933589716792226, 0, 275.992849767848641, 0, -869.37747676872322, 0, 551.985699535697283, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.15619413881131751, 0, -34.4991062209810802, 0, 96.5974974187470245, 0, -55.1985699535697283, 0, 0, 0, 0, 0,
        0, -1.07809706940565876, 0, 0, 0, 0, -3.23429120821697627, 0, 68.9982124419621604, 0, 0, 0, 0, 0, 0, -2.15619413881131751, 0, 137.996424883924321, 0, -482.987487093735122, 0, 0, 0, 0, 0, 0, 0, 0, 2.15619413881131751, 0, 0, 0, -482.987487093735122, 0, 772.779979349976196, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.23429120821697627, 0, -137.996424883924321, 0, 482.987487093735122, 0, 0, 0, -275.992849767848641, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.07809706940565876, 0, -68.9982124419621604, 0, 482.987487093735122, 0, -772.779979349976196, 0, 275.992849767848641, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 25/2, mj = -7/2
        0, -0.784430839734420891, 0, 0, 0, 0, -2.35329251920326267, 0, 50.203573743002937, 0, 0, 0, 0, 0, 0, -1.56886167946884178, 0, 100.407147486005874, 0, -351.425016201020559, 0, 0, 0, 0, 0, 0, 0, 0, 1.56886167946884178, 0, 0, 0, -351.425016201020559, 0, 562.280025921632895, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.35329251920326267, 0, -100.407147486005874, 0, 351.425016201020559, 0, 0, 0, -200.814294972011748, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.784430839734420891, 0, -50.203573743002937, 0, 351.425016201020559, 0, -562.280025921632895, 0, 200.814294972011748, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -9.41317007681305069, 0, 0, 0, 0, 0, 0, -34.5149569483145192, 0, 175.71250810051028, 0, 0, 0, 0, 0, 0, 0, 0, -43.9281270251275699, 0, 468.566688268027412, 0, -632.565029161837007, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -18.8263401536261014, 0, 351.425016201020559, 0, -1054.27504860306168, 0, 602.442884916035244, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.13772335893768356, 0, 0, 0, -210.855009720612336, 0, 401.628589944023496, 0, -133.876196648007832, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.13772335893768356, 0, -58.5708360335034265, 0, 210.855009720612336, 0, -200.814294972011748, 0, 44.6253988826692774, 0, 0, 0,
        // j = 25/2, mj = -5/2
        0, 0, 0, 0, -7.44176436131784437, 0, 0, 0, 0, 0, 0, -27.286469324832096, 0, 138.912934744599762, 0, 0, 0, 0, 0, 0, 0, 0, -34.7282336861499404, 0, 370.434492652266031, 0, -500.086565080559142, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -14.8835287226356887, 0, 277.825869489199523, 0, -833.47760846759857, 0, 476.27291912434204, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.48058812043928146, 0, 0, 0, -166.695521693519714, 0, 317.515279416228027, 0, -105.838426472076009, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.48058812043928146, 0, -46.3043115815332539, 0, 166.695521693519714, 0, -158.757639708114013, 0, 35.279475490692003, 0, 0, 0,
        0, 0.496117624087856292, 0, 0, 0, 0, 2.48058812043928146, 0, -34.7282336861499404, 0, 0, 0, 0, 0, 0, 4.96117624087856292, 0, -138.912934744599762, 0, 277.825869489199523, 0, 0, 0, 0, 0, 0, 0, 0, 4.96117624087856292, 0, -208.369402116899642, 0, 833.47760846759857, 0, -555.651738978399047, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.48058812043928146, 0, -138.912934744599762, 0, 833.47760846759857, 0, -1111.30347795679809, 0, 317.515279416228027, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.496117624087856292, 0, -34.7282336861499404, 0, 277.825869489199523, 0, -555.651738978399047, 0, 317.515279416228027, 0, -42.3353705888304036, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 25/2, mj = -3/2
        0, 0.424849751185551427, 0, 0, 0, 0, 2.12424875592775713, 0, -29.7394825829885999, 0, 0, 0, 0, 0, 0, 4.24849751185551427, 0, -118.957930331954399, 0, 237.915860663908799, 0, 0, 0, 0, 0, 0, 0, 0, 4.24849751185551427, 0, -178.436895497931599, 0, 713.747581991726397, 0, -475.831721327817598, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.12424875592775713, 0, -118.957930331954399, 0, 713.747581991726397, 0, -951.663442655635196, 0, 271.903840758752913, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.424849751185551427, 0, -29.7394825829885999, 0, 237.915860663908799, 0, -475.831721327817598, 0, 271.903840758752913, 0, -36.2538454345003884, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 2.97394825829885999, 0, 0, 0, 0, 0, 0, 14.8697412914942999, 0, -59.4789651659771997, 0, 0, 0, 0, 0, 0, 0, 0, 29.7394825829885999, 0, -237.915860663908799, 0, 237.915860663908799, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 29.7394825829885999, 0, -356.873790995863198, 0, 713.747581991726397, 0, -271.903840758752913, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 14.8697412914942999, 0, -237.915860663908799, 0, 713.747581991726397, 0, -543.807681517505826, 0, 90.634613586250971, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.97394825829885999, 0, -59.4789651659771997, 0, 237.915860663908799, 0, -271.903840758752913, 0, 90.634613586250971, 0, -6.59160826081825244, 0,
        // j = 25/2, mj = -1/2
        0, 0, 0, 0, 2.7533410732166562, 0, 0, 0, 0, 0, 0, 13.766705366083281, 0, -55.066821464333124, 0, 0, 0, 0, 0, 0, 0, 0, 27.533410732166562, 0, -220.267285857332496, 0, 220.267285857332496, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 27.533410732166562, 0, -330.400928785998744, 0, 660.801857571997488, 0, -251.734040979808567, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 13.766705366083281, 0, -220.267285857332496, 0, 660.801857571997488, 0, -503.468081959617133, 0, 83.9113469932695222, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.7533410732166562, 0, -55.066821464333124, 0, 220.267285857332496, 0, -251.734040979808567, 0, 83.9113469932695222, 0, -6.10264341769232889, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 25/2, mj = 1/2
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 2.7533410732166562, 0, 0, 0, 0, 0, 0, 13.766705366083281, 0, -55.066821464333124, 0, 0, 0, 0, 0, 0, 0, 0, 27.533410732166562, 0, -220.267285857332496, 0, 220.267285857332496, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 27.533410732166562, 0, -330.400928785998744, 0, 660.801857571997488, 0, -251.734040979808567, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 13.766705366083281, 0, -220.267285857332496, 0, 660.801857571997488, 0, -503.468081959617133, 0, 83.9113469932695222, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.7533410732166562, 0, -55.066821464333124, 0, 220.267285857332496, 0, -251.734040979808567, 0, 83.9113469932695222, 0, -6.10264341769232889, 0,
        // j = 25/2, mj = 3/2
        0, 0, 0, 0, 2.97394825829885999, 0, 0, 0, 0, 0, 0, 14.8697412914942999, 0, -59.4789651659771997, 0, 0, 0, 0, 0, 0, 0, 0, 29.7394825829885999, 0, -237.915860663908799, 0, 237.915860663908799, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 29.7394825829885999, 0, -356.873790995863198, 0, 713.747581991726397, 0, -271.903840758752913, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 14.8697412914942999, 0, -237.915860663908799, 0, 713.747581991726397, 0, -543.807681517505826, 0, 90.634613586250971, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.97394825829885999, 0, -59.4789651659771997, 0, 237.915860663908799, 0, -271.903840758752913, 0, 90.634613586250971, 0, -6.59160826081825244, 0,
        0, -0.424849751185551427, 0, 0, 0, 0, -2.12424875592775713, 0, 29.7394825829885999, 0, 0, 0, 0, 0, 0, -4.24849751185551427, 0, 118.957930331954399, 0, -237.915860663908799, 0, 0, 0, 0, 0, 0, 0, 0, -4.24849751185551427, 0, 178.436895497931599, 0, -713.747581991726397, 0, 475.831721327817598, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.12424875592775713, 0, 118.957930331954399, 0, -713.747581991726397, 0, 951.663442655635196, 0, -271.903840758752913, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.424849751185551427, 0, 29.7394825829885999, 0, -237.915860663908799, 0, 475.831721327817598, 0, -271.903840758752913, 0, 36.2538454345003884, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 25/2, mj = 5/2
        0, -0.496117624087856292, 0, 0, 0, 0, -2.48058812043928146, 0, 34.7282336861499404, 0, 0, 0, 0, 0, 0, -4.96117624087856292, 0, 138.912934744599762, 0, -277.825869489199523, 0, 0, 0, 0, 0, 0, 0, 0, -4.96117624087856292, 0, 208.369402116899642, 0, -833.47760846759857, 0, 555.651738978399047, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.48058812043928146, 0, 138.912934744599762, 0, -833.47760846759857, 0, 1111.30347795679809, 0, -317.515279416228027, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.496117624087856292, 0, 34.7282336861499404, 0, -277.825869489199523, 0, 555.651738978399047, 0, -317.515279416228027, 0, 42.3353705888304036, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -7.44176436131784437, 0, 0, 0, 0, 0, 0, -27.286469324832096, 0, 138.912934744599762, 0, 0, 0, 0, 0, 0, 0, 0, -34.7282336861499404, 0, 370.434492652266031, 0, -500.086565080559142, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -14.8835287226356887, 0, 277.825869489199523, 0, -833.47760846759857, 0, 476.27291912434204, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.48058812043928146, 0, 0, 0, -166.695521693519714, 0, 317.515279416228027, 0, -105.838426472076009, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.48058812043928146, 0, -46.3043115815332539, 0, 166.695521693519714, 0, -158.757639708114013, 0, 35.279475490692003, 0, 0, 0,
        // j = 25/2, mj = 7/2
        0, 0, 0, 0, -9.41317007681305069, 0, 0, 0, 0, 0, 0, -34.5149569483145192, 0, 175.71250810051028, 0, 0, 0, 0, 0, 0, 0, 0, -43.9281270251275699, 0, 468.566688268027412, 0, -632.565029161837007, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -18.8263401536261014, 0, 351.425016201020559, 0, -1054.27504860306168, 0, 602.442884916035244, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.13772335893768356, 0, 0, 0, -210.855009720612336, 0, 401.628589944023496, 0, -133.876196648007832, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.13772335893768356, 0, -58.5708360335034265, 0, 210.855009720612336, 0, -200.814294972011748, 0, 44.6253988826692774, 0, 0, 0,
        0, 0.784430839734420891, 0, 0, 0, 0, 2.35329251920326267, 0, -50.203573743002937, 0, 0, 0, 0, 0, 0, 1.56886167946884178, 0, -100.407147486005874, 0, 351.425016201020559, 0, 0, 0, 0, 0, 0, 0, 0, -1.56886167946884178, 0, 0, 0, 351.425016201020559, 0, -562.280025921632895, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.35329251920326267, 0, 100.407147486005874, 0, -351.425016201020559, 0, 0, 0, 200.814294972011748, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.784430839734420891, 0, 50.203573743002937, 0, -351.425016201020559, 0, 562.280025921632895, 0, -200.814294972011748, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 25/2, mj = 9/2
        0, 1.07809706940565876, 0, 0, 0, 0, 3.23429120821697627, 0, -68.9982124419621604, 0, 0, 0, 0, 0, 0, 2.15619413881131751, 0, -137.996424883924321, 0, 482.987487093735122, 0, 0, 0, 0, 0, 0, 0, 0, -2.15619413881131751, 0, 0, 0, 482.987487093735122, 0, -772.779979349976196, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3.23429120821697627, 0, 137.996424883924321, 0, -482.987487093735122, 0, 0, 0, 275.992849767848641, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.07809706940565876, 0, 68.9982124419621604, 0, -482.987487093735122, 0, 772.779979349976196, 0, -275.992849767848641, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 10.7809706940565876, 0, 0, 0, 0, 0, 0, 10.7809706940565876, 0, -172.495531104905401, 0, 0, 0, 0, 0, 0, 0, 0, -30.1867179433584452, 0, 0, 0, 482.987487093735122, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -47.4362710538489852, 0, 482.987487093735122, 0, -482.987487093735122, 0, -275.992849767848641, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -15.0933589716792226, 0, 275.992849767848641, 0, -869.37747676872322, 0, 551.985699535697283, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.15619413881131751, 0, -34.4991062209810802, 0, 96.5974974187470245, 0, -55.1985699535697283, 0, 0, 0, 0, 0,
        // j = 25/2, mj = 11/2
        0, 0, 0, 0, 16.1714560410848813, 0, 0, 0, 0, 0, 0, 16.1714560410848813, 0, -258.743296657358101, 0, 0, 0, 0, 0, 0, 0, 0, -45.2800769150376677, 0, 0, 0, 724.481230640602684, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -71.1544065807734779, 0, 724.481230640602684, 0, -724.481230640602684, 0, -413.989274651772962, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -22.6400384575188339, 0, 413.989274651772962, 0, -1304.06621515308483, 0, 827.978549303545924, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.23429120821697627, 0, -51.7486593314716203, 0, 144.896246128120537, 0, -82.7978549303545924, 0, 0, 0, 0, 0,
        0, -1.07809706940565876, 0, 0, 0, 0, 0.359365689801886252, 0, 58.2172417479055728, 0, 0, 0, 0, 0, 0, 6.46858241643395253, 0, -77.6229889972074304, 0, -310.491955988829722, 0, 0, 0, 0, 0, 0, 0, 0, 6.46858241643395253, 0, -271.680461490226006, 0, 724.481230640602684, 0, 289.792492256241073, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.359365689801886252, 0, -77.6229889972074304, 0, 724.481230640602684, 0, -965.974974187470245, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.07809706940565876, 0, 58.2172417479055728, 0, -310.491955988829722, 0, 289.792492256241073, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 25/2, mj = 13/2
        0, -1.77617456229312929, 0, 0, 0, 0, 0.592058187431043098, 0, 95.9134263638289819, 0, 0, 0, 0, 0, 0, 10.6570473737587758, 0, -127.884568485105309, 0, -511.538273940421237, 0, 0, 0, 0, 0, 0, 0, 0, 10.6570473737587758, 0, -447.595989697868582, 0, 1193.58930586098289, 0, 477.435722344393154, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.592058187431043098, 0, -127.884568485105309, 0, 1193.58930586098289, 0, -1591.45240781464385, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.77617456229312929, 0, 95.9134263638289819, 0, -511.538273940421237, 0, 477.435722344393154, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -12.4332219360519051, 0, 0, 0, 0, 0, 0, 37.2996658081557152, 0, 149.198663232622861, 0, 0, 0, 0, 0, 0, 0, 0, 74.5993316163114303, 0, -596.794652930491443, 0, -238.717861172196577, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -10.6570473737587758, 0, -298.397326465245721, 0, 1193.58930586098289, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -33.7473166835694566, 0, 426.281894950351031, 0, -716.153583516589731, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.77617456229312929, 0, -21.3140947475175515, 0, 34.1025515960280824, 0, 0, 0, 0, 0, 0, 0,
        // j = 25/2, mj = 15/2
        0, 0, 0, 0, -22.6998537228122744, 0, 0, 0, 0, 0, 0, 68.0995611684368233, 0, 272.398244673747293, 0, 0, 0, 0, 0, 0, 0, 0, 136.199122336873647, 0, -1089.59297869498917, 0, -435.837191477995669, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -19.4570174766962352, 0, -544.796489347494586, 0, 2179.18595738997834, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -61.6138886762047449, 0, 778.280699067849409, 0, -1307.51157443398701, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.2428362461160392, 0, -38.9140349533924704, 0, 62.2624559254279527, 0, 0, 0, 0, 0, 0, 0,
        0, 1.29713449844641568, 0, 0, 0, 0, -6.48567249223207841, 0, -51.8853799378566272, 0, 0, 0, 0, 0, 0, -7.78280699067849409, 0, 311.312279627139763, 0, 155.656139813569882, 0, 0, 0, 0, 0, 0, 0, 0, 7.78280699067849409, 0, 0, 0, -1089.59297869498917, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6.48567249223207841, 0, -311.312279627139763, 0, 1089.59297869498917, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.29713449844641568, 0, 51.8853799378566272, 0, -155.656139813569882, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 25/2, mj = 17/2
        0, 2.65833466848162373, 0, 0, 0, 0, -13.2916733424081186, 0, -106.333386739264949, 0, 0, 0, 0, 0, 0, -15.9500080108897424, 0, 638.000320435589695, 0, 319.000160217794847, 0, 0, 0, 0, 0, 0, 0, 0, 15.9500080108897424, 0, 0, 0, -2233.00112152456393, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 13.2916733424081186, 0, -638.000320435589695, 0, 2233.00112152456393, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.65833466848162373, 0, 106.333386739264949, 0, -319.000160217794847, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 11.9625060081673068, 0, 0, 0, 0, 0, 0, -99.6875500680608898, 0, -79.7500400544487119, 0, 0, 0, 0, 0, 0, 0, 0, 55.8250280381140983, 0, 744.333707174854644, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 119.625060081673068, 0, -1116.50056076228197, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -46.5208566984284153, 0, 319.000160217794847, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.32916733424081186, 0, -8.86111556160541243, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 25/2, mj = 19/2
        0, 0, 0, 0, 28.0545633538451997, 0, 0, 0, 0, 0, 0, -233.788027948709998, 0, -187.030422358967998, 0, 0, 0, 0, 0, 0, 0, 0, 130.921295651277599, 0, 1745.61727535036798, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 280.545633538451997, 0, -2618.42591302555197, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -109.101079709397999, 0, 748.121689435871992, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.11717370598279997, 0, -20.7811580398853331, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, -1.41689713908309089, 0, 0, 0, 0, 15.5858685299139998, 0, 31.1717370598279997, 0, 0, 0, 0, 0, 0, -18.7030422358967998, 0, -374.060844717935996, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -18.7030422358967998, 0, 785.527773907665592, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 15.5858685299139998, 0, -374.060844717935996, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.41689713908309089, 0, 31.1717370598279997, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 25/2, mj = 21/2
        0, -3.92321052893598442, 0, 0, 0, 0, 43.1553158182958286, 0, 86.3106316365916572, 0, 0, 0, 0, 0, 0, -51.7863789819549943, 0, -1035.72757963909989, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -51.7863789819549943, 0, 2175.02791724210976, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 43.1553158182958286, 0, -1035.72757963909989, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3.92321052893598442, 0, 86.3106316365916572, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, -8.63106316365916572, 0, 0, 0, 0, 0, 0, 129.465947454887486, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -362.50465287368496, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 258.931894909774972, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -43.1553158182958286, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.784642105787196884, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 25/2, mj = 23/2
        0, 0, 0, 0, -29.8988798455876938, 0, 0, 0, 0, 0, 0, 448.483197683815407, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1255.75295351468314, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 896.966395367630815, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -149.494399227938469, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.71807998596251762, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 1.35903999298125881, 0, 0, 0, 0, -24.9157332046564115, 0, 0, 0, 0, 0, 0, 0, 0, 89.6966395367630815, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -89.6966395367630815, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 24.9157332046564115, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.35903999298125881, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        // j = 25/2, mj = 25/2
        0, 6.79519996490629405, 0, 0, 0, 0, -124.578666023282058, 0, 0, 0, 0, 0, 0, 0, 0, 448.483197683815407, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -448.483197683815407, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 124.578666023282058, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -6.79519996490629405, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
};

static FINT _len_spinor(FINT kappa, FINT l)
{
        if (0 == kappa) {
                return 4 * l + 2;
        } else if (kappa < 0) {
                return 2 * l + 2;
        } else {
                return 2 * l;
        }
}

static int _len_cart[] = {
        1, 3, 6, 10, 15, 21, 28, 36, 45, 55, 66, 78, 91, 105, 120, 136
};

struct cart2sp_t {
        double *cart2sph;
        double *cart2j_lt_lR; // j < kappa, l > 0
        double *cart2j_lt_lI; // j < kappa, l > 0
        double *cart2j_gt_lR; // j > kappa, l < 0
        double *cart2j_gt_lI; // j > kappa, l < 0
};

// [*] = n(n+1)(n+2)(n+3)/4+(n+1)(n+2)(n+3)/6
static struct cart2sp_t g_c2s[] = {
        {g_trans_cart2sph     ,g_trans_cart2jR      , g_trans_cart2jI      , g_trans_cart2jR      , g_trans_cart2jI      },
        {g_trans_cart2sph+1   ,g_trans_cart2jR+4    , g_trans_cart2jI+4    , g_trans_cart2jR+16   , g_trans_cart2jI+16   },
        {g_trans_cart2sph+10  ,g_trans_cart2jR+40   , g_trans_cart2jI+40   , g_trans_cart2jR+88   , g_trans_cart2jI+88   },
        {g_trans_cart2sph+40  ,g_trans_cart2jR+160  , g_trans_cart2jI+160  , g_trans_cart2jR+280  , g_trans_cart2jI+280  },
        {g_trans_cart2sph+110 ,g_trans_cart2jR+440  , g_trans_cart2jI+440  , g_trans_cart2jR+680  , g_trans_cart2jI+680  },
        {g_trans_cart2sph+245 ,g_trans_cart2jR+980  , g_trans_cart2jI+980  , g_trans_cart2jR+1400 , g_trans_cart2jI+1400 },
        {g_trans_cart2sph+476 ,g_trans_cart2jR+1904 , g_trans_cart2jI+1904 , g_trans_cart2jR+2576 , g_trans_cart2jI+2576 },
        {g_trans_cart2sph+840 ,g_trans_cart2jR+3360 , g_trans_cart2jI+3360 , g_trans_cart2jR+4368 , g_trans_cart2jI+4368 },
        {g_trans_cart2sph+1380,g_trans_cart2jR+5520 , g_trans_cart2jI+5520 , g_trans_cart2jR+6960 , g_trans_cart2jI+6960 },
        {g_trans_cart2sph+2145,g_trans_cart2jR+8580 , g_trans_cart2jI+8580 , g_trans_cart2jR+10560, g_trans_cart2jI+10560},
        {g_trans_cart2sph+3190,g_trans_cart2jR+12760, g_trans_cart2jI+12760, g_trans_cart2jR+15400, g_trans_cart2jI+15400},
        {g_trans_cart2sph+4576,g_trans_cart2jR+18304, g_trans_cart2jI+18304, g_trans_cart2jR+21736, g_trans_cart2jI+21736},
        {g_trans_cart2sph+6370,g_trans_cart2jR+25480, g_trans_cart2jI+25480, g_trans_cart2jR+29848, g_trans_cart2jI+29848},
        {g_trans_cart2sph+8645,  NULL, NULL, NULL, NULL},
        {g_trans_cart2sph+11480, NULL, NULL, NULL, NULL},
        {g_trans_cart2sph+14960, NULL, NULL, NULL, NULL},
};

// transform integrals from cartesian to spheric
static double *a_bra_cart2spheric(double *gsph, FINT nket, double *gcart, FINT l)
{
        FINT nf = _len_cart[l];
        FINT nd = l * 2 + 1;
        CINTdgemm_TN(nd, nket, nf, g_c2s[l].cart2sph, gcart, gsph);
        return gsph;
}

static double *a_ket_cart2spheric(double *gsph, double *gcart,
                                  FINT lds, FINT nbra, FINT l)
{
        FINT nf = _len_cart[l];
        FINT nd = l * 2 + 1;
        CINTdgemm_NN1(nbra, nd, nf, gcart, g_c2s[l].cart2sph, gsph, lds);
        return gsph;
}

// transform s function from cartesian to spheric
static double *s_bra_cart2spheric(double *gsph, FINT nket, double *gcart, FINT l)
{
        /*
        FINT i;
        for (i = 0; i < nket; i++) {
                *gsph = gcart[i];
        }*/
        return gcart;
}
static double *s_ket_cart2spheric(double *gsph, double *gcart,
                                  FINT lds, FINT nbra, FINT l)
{
        /*
        FINT i;
        for (i = 0; i < nbra; i++) {
                gsph[i] = gcart[i];
        }*/
        return gcart;
}
static double *s_ket_cart2spheric_copy(double *gsph, double *gcart,
                                       FINT lds, FINT nbra, FINT l)
{
        FINT i;
        for (i = 0; i < nbra; i++) {
                gsph[i] = gcart[i];
        }
        return gsph;
}

// transform p function from cartesian to spheric
static double *p_bra_cart2spheric(double *gsph, FINT nket, double *gcart, FINT l)
{
#ifdef PYPZPX
        FINT i;
        for (i = 0; i < nket; i++) {
                gsph[i*3+0] = gcart[i*3+1];  // py
                gsph[i*3+1] = gcart[i*3+2];  // pz
                gsph[i*3+2] = gcart[i*3+0];  // px
        }
        return gsph;
#else
        return gcart;
#endif
}
static double *p_ket_cart2spheric(double *gsph, double *gcart,
                                  FINT lds, FINT nbra, FINT l)
{
#ifdef PYPZPX
        FINT i;
        for (i = 0; i < nbra; i++) {
                gsph[0*nbra+i] = gcart[1*nbra+i];  // py
                gsph[1*nbra+i] = gcart[2*nbra+i];  // pz
                gsph[2*nbra+i] = gcart[0*nbra+i];  // px
        }
        return gsph;
#else
        return gcart;
#endif
}
static double *p_ket_cart2spheric_copy(double *gsph, double *gcart,
                                       FINT lds, FINT nbra, FINT l)
{
        FINT i;
#ifdef PYPZPX
        for (i = 0; i < nbra; i++) {
                gsph[0*nbra+i] = gcart[1*nbra+i];  // py
                gsph[1*nbra+i] = gcart[2*nbra+i];  // pz
                gsph[2*nbra+i] = gcart[0*nbra+i];  // px
        }
#else
        for (i = 0; i < nbra; i++) {
                gsph[0*lds+i] = gcart[0*nbra+i];
                gsph[1*lds+i] = gcart[1*nbra+i];
                gsph[2*lds+i] = gcart[2*nbra+i];
        }
#endif
        return gsph;
}

// transform d function from cartesian to spheric
static double *d_bra_cart2spheric(double *gsph, FINT nket, double *gcart, FINT l)
{
        double *coeff_c2s = g_c2s[2].cart2sph;
        double *pgsph = gsph;
        FINT i;
        for (i = 0; i < nket; i++) {
                gsph[0] = coeff_c2s[ 1] * gcart[1];
                gsph[1] = coeff_c2s[10] * gcart[4];
                gsph[2] = coeff_c2s[12] * gcart[0]
                        + coeff_c2s[15] * gcart[3]
                        + coeff_c2s[17] * gcart[5];
                gsph[3] = coeff_c2s[20] * gcart[2];
                gsph[4] = coeff_c2s[24] * gcart[0]
                        + coeff_c2s[27] * gcart[3];
                gsph += 5;
                gcart += 6;
        }
        return pgsph;
}
static double *d_ket_cart2spheric(double *gsph, double *gcart,
                                  FINT lds, FINT nbra, FINT l)
{
        double *coeff_c2s = g_c2s[2].cart2sph;
        double *pgsph = gsph;
        FINT i;
        for (i = 0; i < nbra; i++) {
                gsph[0*lds+i] = coeff_c2s[ 1] * gcart[1*nbra+i];
        }
        for (i = 0; i < nbra; i++) {
                gsph[1*lds+i] = coeff_c2s[10] * gcart[4*nbra+i];
        }
        for (i = 0; i < nbra; i++) {
                gsph[2*lds+i] = coeff_c2s[12] * gcart[0*nbra+i]
                              + coeff_c2s[15] * gcart[3*nbra+i]
                              + coeff_c2s[17] * gcart[5*nbra+i];
        }
        for (i = 0; i < nbra; i++) {
                gsph[3*lds+i] = coeff_c2s[20] * gcart[2*nbra+i];
        }
        for (i = 0; i < nbra; i++) {
                gsph[4*lds+i] = coeff_c2s[24] * gcart[0*nbra+i]
                              + coeff_c2s[27] * gcart[3*nbra+i];
        }
        return pgsph;
}

// transform f function from cartesian to spheric
static double *f_bra_cart2spheric(double *gsph, FINT nket, double *gcart, FINT l)
{
        double *coeff_c2s = g_c2s[3].cart2sph;
        double *pgsph = gsph;
        FINT i;
        for (i = 0; i < nket; i++) {
                gsph[0] = coeff_c2s[ 1] * gcart[1]
                        + coeff_c2s[ 6] * gcart[6];
                gsph[1] = coeff_c2s[14] * gcart[4];
                gsph[2] = coeff_c2s[21] * gcart[1]
                        + coeff_c2s[26] * gcart[6]
                        + coeff_c2s[28] * gcart[8];
                gsph[3] = coeff_c2s[32] * gcart[2]
                        + coeff_c2s[37] * gcart[7]
                        + coeff_c2s[39] * gcart[9];
                gsph[4] = coeff_c2s[40] * gcart[0]
                        + coeff_c2s[43] * gcart[3]
                        + coeff_c2s[45] * gcart[5];
                gsph[5] = coeff_c2s[52] * gcart[2]
                        + coeff_c2s[57] * gcart[7];
                gsph[6] = coeff_c2s[60] * gcart[0]
                        + coeff_c2s[63] * gcart[3];
                gsph += 7;
                gcart += 10;
        }
        return pgsph;
}
static double *f_ket_cart2spheric(double *gsph, double *gcart,
                                  FINT lds, FINT nbra, FINT l)
{
        double *coeff_c2s = g_c2s[3].cart2sph;
        double *pgsph = gsph;
        FINT i;
        for (i = 0; i < nbra; i++) {
                gsph[0*lds+i] = coeff_c2s[ 1] * gcart[1*nbra+i]
                              + coeff_c2s[ 6] * gcart[6*nbra+i];
        }
        for (i = 0; i < nbra; i++) {
                gsph[1*lds+i] = coeff_c2s[14] * gcart[4*nbra+i];
        }
        for (i = 0; i < nbra; i++) {
                gsph[2*lds+i] = coeff_c2s[21] * gcart[1*nbra+i]
                              + coeff_c2s[26] * gcart[6*nbra+i]
                              + coeff_c2s[28] * gcart[8*nbra+i];
        }
        for (i = 0; i < nbra; i++) {
                gsph[3*lds+i] = coeff_c2s[32] * gcart[2*nbra+i]
                              + coeff_c2s[37] * gcart[7*nbra+i]
                              + coeff_c2s[39] * gcart[9*nbra+i];
        }
        for (i = 0; i < nbra; i++) {
                gsph[4*lds+i] = coeff_c2s[40] * gcart[0*nbra+i]
                              + coeff_c2s[43] * gcart[3*nbra+i]
                              + coeff_c2s[45] * gcart[5*nbra+i];
        }
        for (i = 0; i < nbra; i++) {
                gsph[5*lds+i] = coeff_c2s[52] * gcart[2*nbra+i]
                              + coeff_c2s[57] * gcart[7*nbra+i];
        }
        for (i = 0; i < nbra; i++) {
                gsph[6*lds+i] = coeff_c2s[60] * gcart[0*nbra+i]
                              + coeff_c2s[63] * gcart[3*nbra+i];
        }
        return pgsph;
}

// transform g function from cartesian to spheric
static double *g_bra_cart2spheric(double *gsph, FINT nket, double *gcart, FINT l)
{
        double *coeff_c2s = g_c2s[4].cart2sph;
        double *pgsph = gsph;
        FINT i;
        for (i = 0; i < nket; i++) {
                gsph[0] = coeff_c2s[  1] * gcart[ 1]
                        + coeff_c2s[  6] * gcart[ 6];
                gsph[1] = coeff_c2s[ 19] * gcart[ 4]
                        + coeff_c2s[ 26] * gcart[11];
                gsph[2] = coeff_c2s[ 31] * gcart[ 1]
                        + coeff_c2s[ 36] * gcart[ 6]
                        + coeff_c2s[ 38] * gcart[ 8];
                gsph[3] = coeff_c2s[ 49] * gcart[ 4]
                        + coeff_c2s[ 56] * gcart[11]
                        + coeff_c2s[ 58] * gcart[13];
                gsph[4] = coeff_c2s[ 60] * gcart[ 0]
                        + coeff_c2s[ 63] * gcart[ 3]
                        + coeff_c2s[ 65] * gcart[ 5]
                        + coeff_c2s[ 70] * gcart[10]
                        + coeff_c2s[ 72] * gcart[12]
                        + coeff_c2s[ 74] * gcart[14];
                gsph[5] = coeff_c2s[ 77] * gcart[ 2]
                        + coeff_c2s[ 82] * gcart[ 7]
                        + coeff_c2s[ 84] * gcart[ 9];
                gsph[6] = coeff_c2s[ 90] * gcart[ 0]
                        + coeff_c2s[ 95] * gcart[ 5]
                        + coeff_c2s[100] * gcart[10]
                        + coeff_c2s[102] * gcart[12];
                gsph[7] = coeff_c2s[107] * gcart[ 2]
                        + coeff_c2s[112] * gcart[ 7];
                gsph[8] = coeff_c2s[120] * gcart[ 0]
                        + coeff_c2s[123] * gcart[ 3]
                        + coeff_c2s[130] * gcart[10];
                gsph += 9;
                gcart += 15;
        }
        return pgsph;
}
static double *g_ket_cart2spheric(double *gsph, double *gcart,
                                  FINT lds, FINT nbra, FINT l)
{
        double *coeff_c2s = g_c2s[4].cart2sph;
        double *pgsph = gsph;
        FINT i;
        for (i = 0; i < nbra; i++) {
                gsph[0*lds+i] = coeff_c2s[  1] * gcart[ 1*nbra+i]
                              + coeff_c2s[  6] * gcart[ 6*nbra+i];
        }
        for (i = 0; i < nbra; i++) {
                gsph[1*lds+i] = coeff_c2s[ 19] * gcart[ 4*nbra+i]
                              + coeff_c2s[ 26] * gcart[11*nbra+i];
        }
        for (i = 0; i < nbra; i++) {
                gsph[2*lds+i] = coeff_c2s[ 31] * gcart[ 1*nbra+i]
                              + coeff_c2s[ 36] * gcart[ 6*nbra+i]
                              + coeff_c2s[ 38] * gcart[ 8*nbra+i];
        }
        for (i = 0; i < nbra; i++) {
                gsph[3*lds+i] = coeff_c2s[ 49] * gcart[ 4*nbra+i]
                              + coeff_c2s[ 56] * gcart[11*nbra+i]
                              + coeff_c2s[ 58] * gcart[13*nbra+i];
        }
        for (i = 0; i < nbra; i++) {
                gsph[4*lds+i] = coeff_c2s[ 60] * gcart[ 0*nbra+i]
                              + coeff_c2s[ 63] * gcart[ 3*nbra+i]
                              + coeff_c2s[ 65] * gcart[ 5*nbra+i]
                              + coeff_c2s[ 70] * gcart[10*nbra+i]
                              + coeff_c2s[ 72] * gcart[12*nbra+i]
                              + coeff_c2s[ 74] * gcart[14*nbra+i];
        }
        for (i = 0; i < nbra; i++) {
                gsph[5*lds+i] = coeff_c2s[ 77] * gcart[ 2*nbra+i]
                              + coeff_c2s[ 82] * gcart[ 7*nbra+i]
                              + coeff_c2s[ 84] * gcart[ 9*nbra+i];
        }
        for (i = 0; i < nbra; i++) {
                gsph[6*lds+i] = coeff_c2s[ 90] * gcart[ 0*nbra+i]
                              + coeff_c2s[ 95] * gcart[ 5*nbra+i]
                              + coeff_c2s[100] * gcart[10*nbra+i]
                              + coeff_c2s[102] * gcart[12*nbra+i];
        }
        for (i = 0; i < nbra; i++) {
                gsph[7*lds+i] = coeff_c2s[107] * gcart[ 2*nbra+i]
                              + coeff_c2s[112] * gcart[ 7*nbra+i];
        }
        for (i = 0; i < nbra; i++) {
                gsph[8*lds+i] = coeff_c2s[120] * gcart[ 0*nbra+i]
                              + coeff_c2s[123] * gcart[ 3*nbra+i]
                              + coeff_c2s[130] * gcart[10*nbra+i];
        }
        return pgsph;
}

/*
 * return the address of gemm results, for s,p function, results ==
 * input, so return the input address optimize
 */
double *(*c2s_bra_sph[])() = {
        s_bra_cart2spheric,
        p_bra_cart2spheric,
        d_bra_cart2spheric,
        f_bra_cart2spheric,
        g_bra_cart2spheric,
        a_bra_cart2spheric,
        a_bra_cart2spheric,
        a_bra_cart2spheric,
        a_bra_cart2spheric,
        a_bra_cart2spheric,
        a_bra_cart2spheric,
        a_bra_cart2spheric,
        a_bra_cart2spheric,
        a_bra_cart2spheric,
        a_bra_cart2spheric,
        a_bra_cart2spheric,
};

double *(*c2s_ket_sph[])(double *gsph, double *gcart,
                         FINT lds, FINT nbra, FINT l) = {
        s_ket_cart2spheric,
        p_ket_cart2spheric,
        d_ket_cart2spheric,
        f_ket_cart2spheric,
        g_ket_cart2spheric,
        a_ket_cart2spheric,
        a_ket_cart2spheric,
        a_ket_cart2spheric,
        a_ket_cart2spheric,
        a_ket_cart2spheric,
        a_ket_cart2spheric,
        a_ket_cart2spheric,
        a_ket_cart2spheric,
        a_ket_cart2spheric,
        a_ket_cart2spheric,
        a_ket_cart2spheric,
};

double *(*c2s_ket_sph1[])(double *gsph, double *gcart,
                          FINT lds, FINT nbra, FINT l) = {
        s_ket_cart2spheric_copy,
        p_ket_cart2spheric_copy,
        d_ket_cart2spheric,
        f_ket_cart2spheric,
        g_ket_cart2spheric,
        a_ket_cart2spheric,
        a_ket_cart2spheric,
        a_ket_cart2spheric,
        a_ket_cart2spheric,
        a_ket_cart2spheric,
        a_ket_cart2spheric,
        a_ket_cart2spheric,
        a_ket_cart2spheric,
        a_ket_cart2spheric,
        a_ket_cart2spheric,
        a_ket_cart2spheric,
};


// [ca, cb] * [ 1+1j*z, y+1j*x]
//            [-y+1j*x, 1-1j*z]
static void a_bra_cart2spinor_si(double *gspR, double *gspI,
                                 double *gx, double *gy, double *gz, double *g1,
                                 FINT nket, FINT kappa, FINT l)
{
        FINT nf = _len_cart[l];
        FINT nd = _len_spinor(kappa, l);
        double *gspaR = gspR;
        double *gspaI = gspI;
        double *gspbR = gspR + nket * nd;
        double *gspbI = gspI + nket * nd;
        double *coeffR, *coeffI;
        if (kappa < 0) { // j = l + 1/2
                coeffR = g_c2s[l].cart2j_gt_lR;
                coeffI = g_c2s[l].cart2j_gt_lI;
        } else {
                coeffR = g_c2s[l].cart2j_lt_lR;
                coeffI = g_c2s[l].cart2j_lt_lI;
        }

        FINT i, j, n;
        double saR, saI, sbR, sbI, caR, caI, cbR, cbI, v1, vx, vy, vz;

        for (j = 0; j < nket; j++) {
        for (i = 0; i < nd; i++) {
                saR = 0;
                saI = 0;
                sbR = 0;
                sbI = 0;
#pragma GCC ivdep
                for (n = 0; n < nf; n++) {
                        v1 = g1[j*nf+n];
                        vx = gx[j*nf+n];
                        vy = gy[j*nf+n];
                        vz = gz[j*nf+n];
                        caR = coeffR[i*nf*2   +n];
                        caI = coeffI[i*nf*2   +n];
                        cbR = coeffR[i*nf*2+nf+n];
                        cbI = coeffI[i*nf*2+nf+n];
                        saR += caR * v1 + caI * vz - cbR * vy + cbI * vx;
                        saI +=-caI * v1 + caR * vz + cbI * vy + cbR * vx;
                        sbR += cbR * v1 - cbI * vz + caR * vy + caI * vx;
                        sbI +=-cbI * v1 - cbR * vz - caI * vy + caR * vx;
                }
                gspaR[j*nd+i] = saR;
                gspaI[j*nd+i] = saI;
                gspbR[j*nd+i] = sbR;
                gspbI[j*nd+i] = sbI;
        } }
}

static void a_bra_cart2spinor_sf(double *gspR, double *gspI,
                                 double *gx, double *gy, double *gz, double *g1,
                                 FINT nket, FINT kappa, FINT l)
{
        FINT nf = _len_cart[l];
        FINT nd = _len_spinor(kappa, l);
        double *gspaR = gspR;
        double *gspaI = gspI;
        double *gspbR = gspR + nket * nd;
        double *gspbI = gspI + nket * nd;
        double *coeffR, *coeffI;
        if (kappa < 0) { // j = l + 1/2
                coeffR = g_c2s[l].cart2j_gt_lR;
                coeffI = g_c2s[l].cart2j_gt_lI;
        } else {
                coeffR = g_c2s[l].cart2j_lt_lR;
                coeffI = g_c2s[l].cart2j_lt_lI;
        }

        FINT i, j, n;
        double saR, saI, sbR, sbI, caR, caI, cbR, cbI, v1;

        for (j = 0; j < nket; j++) {
        for (i = 0; i < nd; i++) {
                saR = 0;
                saI = 0;
                sbR = 0;
                sbI = 0;
#pragma GCC ivdep
                for (n = 0; n < nf; n++) {
                        v1 = g1[j*nf+n];
                        caR = coeffR[i*nf*2   +n];
                        caI = coeffI[i*nf*2   +n];
                        cbR = coeffR[i*nf*2+nf+n];
                        cbI = coeffI[i*nf*2+nf+n];
                        saR += caR * v1;
                        saI +=-caI * v1;
                        sbR += cbR * v1;
                        sbI +=-cbI * v1;
                }
                gspaR[j*nd+i] = saR;
                gspaI[j*nd+i] = saI;
                gspbR[j*nd+i] = sbR;
                gspbI[j*nd+i] = sbI;
        } }
}

static void a_bra1_cart2spinor_si(double *gspR, double *gspI,
                                  double *gx, double *gy, double *gz, double *g1,
                                  FINT ngrids, FINT nket, FINT kappa, FINT l)
{
        FINT nf = _len_cart[l];
        FINT nd = _len_spinor(kappa, l);
        FINT ndg = nd * ngrids;
        double *gspaR = gspR;
        double *gspaI = gspI;
        double *gspbR = gspR + nket * ndg;
        double *gspbI = gspI + nket * ndg;
        double *coeffR, *coeffI;
        if (kappa < 0) { // j = l + 1/2
                coeffR = g_c2s[l].cart2j_gt_lR;
                coeffI = g_c2s[l].cart2j_gt_lI;
        } else {
                coeffR = g_c2s[l].cart2j_lt_lR;
                coeffI = g_c2s[l].cart2j_lt_lI;
        }

        FINT i, j, n, m;
        double caR, caI, cbR, cbI, v1, vx, vy, vz;

        for (j = 0; j < nket; j++) {
#pragma GCC ivdep
                for (i = 0; i < ndg; i++) {
                        gspaR[j*ndg+i] = 0;
                        gspaI[j*ndg+i] = 0;
                        gspbR[j*ndg+i] = 0;
                        gspbI[j*ndg+i] = 0;
                }
                for (i = 0; i < nd; i++) {
                for (n = 0; n < nf; n++) {
                        caR = coeffR[i*nf*2   +n];
                        caI = coeffI[i*nf*2   +n];
                        cbR = coeffR[i*nf*2+nf+n];
                        cbI = coeffI[i*nf*2+nf+n];
#pragma GCC ivdep
                        for (m = 0; m < ngrids; m++) {
                                v1 = g1[(j*nf+n)*ngrids+m];
                                vx = gx[(j*nf+n)*ngrids+m];
                                vy = gy[(j*nf+n)*ngrids+m];
                                vz = gz[(j*nf+n)*ngrids+m];
                                gspaR[(j*nd+i)*ngrids+m] += caR * v1 + caI * vz - cbR * vy + cbI * vx;
                                gspaI[(j*nd+i)*ngrids+m] +=-caI * v1 + caR * vz + cbI * vy + cbR * vx;
                                gspbR[(j*nd+i)*ngrids+m] += cbR * v1 - cbI * vz + caR * vy + caI * vx;
                                gspbI[(j*nd+i)*ngrids+m] +=-cbI * v1 - cbR * vz - caI * vy + caR * vx;
                        }
                } }
        }
}

static void a_bra1_cart2spinor_sf(double *gspR, double *gspI,
                                  double *gx, double *gy, double *gz, double *g1,
                                  FINT ngrids, FINT nket, FINT kappa, FINT l)
{
        FINT nf = _len_cart[l];
        FINT nd = _len_spinor(kappa, l);
        FINT ndg = nd * ngrids;
        double *gspaR = gspR;
        double *gspaI = gspI;
        double *gspbR = gspR + nket * ndg;
        double *gspbI = gspI + nket * ndg;
        double *coeffR, *coeffI;
        if (kappa < 0) { // j = l + 1/2
                coeffR = g_c2s[l].cart2j_gt_lR;
                coeffI = g_c2s[l].cart2j_gt_lI;
        } else {
                coeffR = g_c2s[l].cart2j_lt_lR;
                coeffI = g_c2s[l].cart2j_lt_lI;
        }

        FINT i, j, n, m;
        double caR, caI, cbR, cbI, v1;

        for (j = 0; j < nket; j++) {
#pragma GCC ivdep
                for (i = 0; i < ndg; i++) {
                        gspaR[j*ndg+i] = 0;
                        gspaI[j*ndg+i] = 0;
                        gspbR[j*ndg+i] = 0;
                        gspbI[j*ndg+i] = 0;
                }
                for (i = 0; i < nd; i++) {
                for (n = 0; n < nf; n++) {
                        caR = coeffR[i*nf*2   +n];
                        caI = coeffI[i*nf*2   +n];
                        cbR = coeffR[i*nf*2+nf+n];
                        cbI = coeffI[i*nf*2+nf+n];
#pragma GCC ivdep
                        for (m = 0; m < ngrids; m++) {
                                v1 = g1[(j*nf+n)*ngrids+m];
                                gspaR[(j*nd+i)*ngrids+m] += caR * v1;
                                gspaI[(j*nd+i)*ngrids+m] +=-caI * v1;
                                gspbR[(j*nd+i)*ngrids+m] += cbR * v1;
                                gspbI[(j*nd+i)*ngrids+m] +=-cbI * v1;
                        }
                } }
        }
}

static void a_bra1_cart2spinor_zi(double *gspR, double *gspI,
                                  double *gx, double *gy, double *gz, double *g1,
                                  FINT ngrids, FINT nket, FINT kappa, FINT l)
{
        FINT nf = _len_cart[l];
        FINT nd = _len_spinor(kappa, l);
        FINT ndg = nd * ngrids;
        double *gspaR = gspR;
        double *gspaI = gspI;
        double *gspbR = gspR + nket * ndg;
        double *gspbI = gspI + nket * ndg;
        double *gxR = gx;
        double *gyR = gy;
        double *gzR = gz;
        double *g1R = g1;
        double *gxI = gx + nket * nf * ngrids;
        double *gyI = gy + nket * nf * ngrids;
        double *gzI = gz + nket * nf * ngrids;
        double *g1I = g1 + nket * nf * ngrids;
        double *coeffR, *coeffI;
        if (kappa < 0) { // j = l + 1/2
                coeffR = g_c2s[l].cart2j_gt_lR;
                coeffI = g_c2s[l].cart2j_gt_lI;
        } else {
                coeffR = g_c2s[l].cart2j_lt_lR;
                coeffI = g_c2s[l].cart2j_lt_lI;
        }

        FINT i, j, n, m;
        double caR, caI, cbR, cbI;
        double v1R, vxR, vyR, vzR;
        double v1I, vxI, vyI, vzI;
        double v11R, v12R, v21R, v22R;
        double v11I, v12I, v21I, v22I;

        for (j = 0; j < nket; j++) {
#pragma GCC ivdep
                for (i = 0; i < ndg; i++) {
                        gspaR[j*ndg+i] = 0;
                        gspaI[j*ndg+i] = 0;
                        gspbR[j*ndg+i] = 0;
                        gspbI[j*ndg+i] = 0;
                }
                for (i = 0; i < nd; i++) {
                for (n = 0; n < nf; n++) {
                        caR = coeffR[i*nf*2   +n];
                        caI = coeffI[i*nf*2   +n];
                        cbR = coeffR[i*nf*2+nf+n];
                        cbI = coeffI[i*nf*2+nf+n];
#pragma GCC ivdep
                        for (m = 0; m < ngrids; m++) {
                                // [ 1+1j*z, y+1j*x]
                                // [-y+1j*x, 1-1j*z]
                                v11R = g1R[(j*nf+n)*ngrids+m] - gzI[(j*nf+n)*ngrids+m];
                                v11I = g1I[(j*nf+n)*ngrids+m] + gzR[(j*nf+n)*ngrids+m];
                                v12R = gyR[(j*nf+n)*ngrids+m] - gxI[(j*nf+n)*ngrids+m];
                                v12I = gyI[(j*nf+n)*ngrids+m] + gxR[(j*nf+n)*ngrids+m];
                                v21R =-gyR[(j*nf+n)*ngrids+m] - gxI[(j*nf+n)*ngrids+m];
                                v21I =-gyI[(j*nf+n)*ngrids+m] + gxR[(j*nf+n)*ngrids+m];
                                v22R = g1R[(j*nf+n)*ngrids+m] + gzI[(j*nf+n)*ngrids+m];
                                v22I = g1I[(j*nf+n)*ngrids+m] - gzR[(j*nf+n)*ngrids+m];
                                gspaR[(j*nd+i)*ngrids+m] += caR * v11R + caI * v11I + cbR * v21R + cbI * v21I;
                                gspaI[(j*nd+i)*ngrids+m] += caR * v11I - caI * v11R + cbR * v21I - cbI * v21R;
                                gspbR[(j*nd+i)*ngrids+m] += caR * v12R + caI * v12I + cbR * v22R + cbI * v22I;
                                gspbI[(j*nd+i)*ngrids+m] += caR * v12I - caI * v12R + cbR * v22I - cbI * v22R;
                        }
                } }
        }
}

static void a_bra1_cart2spinor_zf(double *gspR, double *gspI,
                                  double *gx, double *gy, double *gz, double *g1,
                                  FINT ngrids, FINT nket, FINT kappa, FINT l)
{
        FINT nf = _len_cart[l];
        FINT nd = _len_spinor(kappa, l);
        FINT ndg = nd * ngrids;
        double *gspaR = gspR;
        double *gspaI = gspI;
        double *gspbR = gspR + nket * ndg;
        double *gspbI = gspI + nket * ndg;
        double *g1R = g1;
        double *g1I = g1 + nket * nf * ngrids;
        double *coeffR, *coeffI;
        if (kappa < 0) { // j = l + 1/2
                coeffR = g_c2s[l].cart2j_gt_lR;
                coeffI = g_c2s[l].cart2j_gt_lI;
        } else {
                coeffR = g_c2s[l].cart2j_lt_lR;
                coeffI = g_c2s[l].cart2j_lt_lI;
        }

        FINT i, j, n, m;
        double caR, caI, cbR, cbI, v1R, v1I;

        for (j = 0; j < nket; j++) {
#pragma GCC ivdep
                for (i = 0; i < ndg; i++) {
                        gspaR[j*ndg+i] = 0;
                        gspaI[j*ndg+i] = 0;
                        gspbR[j*ndg+i] = 0;
                        gspbI[j*ndg+i] = 0;
                }
                for (i = 0; i < nd; i++) {
                for (n = 0; n < nf; n++) {
                        caR = coeffR[i*nf*2   +n];
                        caI = coeffI[i*nf*2   +n];
                        cbR = coeffR[i*nf*2+nf+n];
                        cbI = coeffI[i*nf*2+nf+n];
#pragma GCC ivdep
                        for (m = 0; m < ngrids; m++) {
                                v1R = g1R[(j*nf+n)*ngrids+m];
                                v1I = g1I[(j*nf+n)*ngrids+m];
                                gspaR[(j*nd+i)*ngrids+m] += caR * v1R + caI * v1I;
                                gspaI[(j*nd+i)*ngrids+m] += caR * v1I - caI * v1R;
                                gspbR[(j*nd+i)*ngrids+m] += cbR * v1R + cbI * v1I;
                                gspbI[(j*nd+i)*ngrids+m] += cbR * v1I - cbI * v1R;
                        }
                } }
        }
}

#if 0
// [ 1+1j*z, y+1j*x] * [ca]
// [-y+1j*x, 1-1j*z]   [cb]
static void a_ket_cart2spinor_si(double *gspR, double *gspI,
                                 double *gx, double *gy, double *gz, double *g1,
                                 FINT lds, FINT nbra, FINT kappa, FINT l)
{
        FINT nf = _len_cart[l];
        FINT nd = _len_spinor(kappa, l);
        double *gspaR = gspR;
        double *gspaI = gspI;
        double *gspbR = gspR + lds * nd;
        double *gspbI = gspI + lds * nd;
        double *coeffR, *coeffI;
        if (kappa < 0) { // j = l + 1/2
                coeffR = g_c2s[l].cart2j_gt_lR;
                coeffI = g_c2s[l].cart2j_gt_lI;
        } else {
                coeffR = g_c2s[l].cart2j_lt_lR;
                coeffI = g_c2s[l].cart2j_lt_lI;
        }

        FINT i, j, n;
        double caR, caI, cbR, cbI, v1, vx, vy, vz;

        for (i = 0; i < nd; i++) {
#pragma GCC ivdep
                for (j = 0; j < nbra; j++) {
                        gspaR[j+i*lds] = 0;
                        gspaI[j+i*lds] = 0;
                        gspbR[j+i*lds] = 0;
                        gspbI[j+i*lds] = 0;
                }
                for (n = 0; n < nf; n++) {
                        caR = coeffR[i*nf*2   +n];
                        caI = coeffI[i*nf*2   +n];
                        cbR = coeffR[i*nf*2+nf+n];
                        cbI = coeffI[i*nf*2+nf+n];
#pragma GCC ivdep
                        for (j = 0; j < nbra; j++) {
                                v1 = g1[j+n*nbra];
                                vx = gx[j+n*nbra];
                                vy = gy[j+n*nbra];
                                vz = gz[j+n*nbra];
// [ 1+1j*z, y+1j*x] * [ca]
// [-y+1j*x, 1-1j*z]   [cb]
                                gspaR[j+i*lds] += caR * v1 - caI * vz + cbR * vy - cbI * vx;
                                gspaI[j+i*lds] += caI * v1 + caR * vz + cbI * vy + cbR * vx;
                                gspbR[j+i*lds] += cbR * v1 + cbI * vz - caR * vy - caI * vx;
                                gspbI[j+i*lds] += cbI * v1 - cbR * vz - caI * vy + caR * vx;
                        }
                }
        }
}

static void a_ket_cart2spinor_sf(double *gspR, double *gspI,
                                 double *gx, double *gy, double *gz, double *g1,
                                 FINT lds, FINT nbra, FINT kappa, FINT l)
{
        FINT nf = _len_cart[l];
        FINT nd = _len_spinor(kappa, l);
        double *gspaR = gspR;
        double *gspaI = gspI;
        double *gspbR = gspR + lds * nd;
        double *gspbI = gspI + lds * nd;
        double *coeffR, *coeffI;
        if (kappa < 0) { // j = l + 1/2
                coeffR = g_c2s[l].cart2j_gt_lR;
                coeffI = g_c2s[l].cart2j_gt_lI;
        } else {
                coeffR = g_c2s[l].cart2j_lt_lR;
                coeffI = g_c2s[l].cart2j_lt_lI;
        }

        FINT i, j, n;
        double caR, caI, cbR, cbI, v1;

        for (i = 0; i < nd; i++) {
#pragma GCC ivdep
                for (j = 0; j < nbra; j++) {
                        gspaR[j+i*lds] = 0;
                        gspaI[j+i*lds] = 0;
                        gspbR[j+i*lds] = 0;
                        gspbI[j+i*lds] = 0;
                }
                for (n = 0; n < nf; n++) {
                        caR = coeffR[i*nf*2   +n];
                        caI = coeffI[i*nf*2   +n];
                        cbR = coeffR[i*nf*2+nf+n];
                        cbI = coeffI[i*nf*2+nf+n];
#pragma GCC ivdep
                        for (j = 0; j < nbra; j++) {
                                v1 = g1[j+n*nbra];
                                gspaR[j+i*lds] += caR * v1;
                                gspaI[j+i*lds] += caI * v1;
                                gspbR[j+i*lds] += cbR * v1;
                                gspbI[j+i*lds] += cbI * v1;
                        }
                }
        }
}
#endif

static void a_ket_cart2spinor(double *gspR, double *gspI,
                              double *gcartR, double *gcartI,
                              FINT nbra, FINT kappa, FINT l)
{
        FINT nf = _len_cart[l];
        FINT nf2 = nf * 2;
        FINT nd = _len_spinor(kappa, l);
        double *coeffR, *coeffI;
        if (kappa < 0) { // j = l + 1/2
                coeffR = g_c2s[l].cart2j_gt_lR;
                coeffI = g_c2s[l].cart2j_gt_lI;
        } else {
                coeffR = g_c2s[l].cart2j_lt_lR;
                coeffI = g_c2s[l].cart2j_lt_lI;
        }

        FINT i, j, n;
        double cR, cI, gR, gI;

        for (i = 0; i < nd; i++) {
#pragma GCC ivdep
                for (j = 0; j < nbra; j++) {
                        gspR[j+i*nbra] = 0;
                        gspI[j+i*nbra] = 0;
                }
                for (n = 0; n < nf2; n++) {
                        cR = coeffR[i*nf2+n];
                        cI = coeffI[i*nf2+n];
                        if (cR != 0) {
                                if (cI != 0) {
#pragma GCC ivdep
                                        for (j = 0; j < nbra; j++) {
                                                gR = gcartR[j+n*nbra];
                                                gI = gcartI[j+n*nbra];
                                                gspR[j+i*nbra] += cR * gR - cI * gI;
                                                gspI[j+i*nbra] += cI * gR + cR * gI;
                                        }
                                } else {
#pragma GCC ivdep
                                        for (j = 0; j < nbra; j++) {
                                                gR = gcartR[j+n*nbra];
                                                gI = gcartI[j+n*nbra];
                                                gspR[j+i*nbra] += cR * gR;
                                                gspI[j+i*nbra] += cR * gI;
                                        }
                                }
                        } else {
                                if (cI != 0) {
#pragma GCC ivdep
                                        for (j = 0; j < nbra; j++) {
                                                gR = gcartR[j+n*nbra];
                                                gI = gcartI[j+n*nbra];
                                                gspR[j+i*nbra] +=-cI * gI;
                                                gspI[j+i*nbra] += cI * gR;
                                        }
                                }
                        }
                }
        }
}

// with phase "i"
static void a_iket_cart2spinor(double *gspR, double *gspI,
                               double *gcartR, double *gcartI,
                               FINT nbra, FINT kappa, FINT l)
{
        a_ket_cart2spinor(gspI, gspR, gcartR, gcartI, nbra, kappa, l);
        FINT size = _len_spinor(kappa, l) * nbra;
        FINT i;
        for (i = 0; i < size; i++) {
                gspR[i] = -gspR[i];
        }
}

static void a_ket1_cart2spinor(double *gspR, double *gspI,
                               double *gcartR, double *gcartI,
                               FINT nbra, FINT counts, FINT kappa, FINT l)
{
        FINT nf = _len_cart[l];
        FINT nf2 = nf * 2;
        FINT nd = _len_spinor(kappa, l);
        FINT nds = nd * nbra;
        FINT nfs = nf * nbra;
        double *gcartaR = gcartR;
        double *gcartaI = gcartI;
        double *gcartbR = gcartaR + nfs * counts;
        double *gcartbI = gcartaI + nfs * counts;
        double *coeffR, *coeffI;
        if (kappa < 0) { // j = l + 1/2
                coeffR = g_c2s[l].cart2j_gt_lR;
                coeffI = g_c2s[l].cart2j_gt_lI;
        } else {
                coeffR = g_c2s[l].cart2j_lt_lR;
                coeffI = g_c2s[l].cart2j_lt_lI;
        }

        FINT i, j, k, n;
        double caR, caI, cbR, cbI, gaR, gaI, gbR, gbI;

        for (i = 0; i < nd; i++) {
                for (k = 0; k < counts; k++) {
#pragma GCC ivdep
                for (j = 0; j < nbra; j++) {
                        gspR[k*nds+j+i*nbra] = 0;
                        gspI[k*nds+j+i*nbra] = 0;
                } }
                for (n = 0; n < nf; n++) {
                        caR = coeffR[i*nf2   +n];
                        caI = coeffI[i*nf2   +n];
                        cbR = coeffR[i*nf2+nf+n];
                        cbI = coeffI[i*nf2+nf+n];
                        for (k = 0; k < counts; k++) {
#pragma GCC ivdep
                        for (j = 0; j < nbra; j++) {
                                gaR = gcartaR[k*nfs+j+n*nbra];
                                gaI = gcartaI[k*nfs+j+n*nbra];
                                gbR = gcartbR[k*nfs+j+n*nbra];
                                gbI = gcartbI[k*nfs+j+n*nbra];
                                gspR[k*nds+j+i*nbra] += caR * gaR - caI * gaI + cbR * gbR - cbI * gbI;
                                gspI[k*nds+j+i*nbra] += caR * gaI + caI * gaR + cbR * gbI + cbI * gbR;
                        } }
                }
        }
}

// with phase "i"
static void a_iket1_cart2spinor(double *gspR, double *gspI,
                                double *gcartR, double *gcartI,
                                FINT nbra, FINT counts, FINT kappa, FINT l)
{
        a_ket1_cart2spinor(gspI, gspR, gcartR, gcartI, nbra, counts, kappa, l);
        FINT size = _len_spinor(kappa, l) * nbra * counts;
        FINT i;
        for (i = 0; i < size; i++) {
                gspR[i] = -gspR[i];
        }
}


/*************************************************
 *
 * transform matrices
 *
 *************************************************/

static void dcopy_ij(double *out, double *gctr,
                     const FINT ni, const FINT nj, const FINT mi, const FINT mj)
{
        FINT i, j;

        for (j = 0; j < mj; j++) {
        for (i = 0; i < mi; i++) {
                out[j*ni+i] = gctr[j*mi+i];
        } }
}

static void zcopy_ij(double complex *out, double *gctrR, double *gctrI,
                     const FINT ni, const FINT nj, const FINT mi, const FINT mj)
{
        double *dout = (double *)out;
        FINT i, j;

        for (j = 0; j < mj; j++) {
        for (i = 0; i < mi; i++) {
                dout[(j*ni+i)*OF_CMPLX  ] = gctrR[j*mi+i];
                dout[(j*ni+i)*OF_CMPLX+1] = gctrI[j*mi+i];
        } }
}

static void dcopy_grids_ij(double *out, const double *gctr,
                           const FINT ngrids, const FINT ni, const FINT nj,
                           const FINT mgrids, const FINT mi, const FINT mj)
{
        const size_t ngi = ngrids * ni;
        const size_t mgi = mgrids * mi;
        FINT i, j, m;

        for (j = 0; j < mj; j++) {
                for (i = 0; i < mi; i++) {
#pragma GCC ivdep
                for (m = 0; m < mgrids; m++) {
                        out[i*ngrids+m] = gctr[i*mgrids+m];
                } }
                out += ngi;
                gctr += mgi;
        }
}

static void zcopy_grids_ij(double complex *out, double *gctrR, double *gctrI,
                           const FINT ngrids, const FINT ni, const FINT nj,
                           const FINT mgrids, const FINT mi, const FINT mj)
{
        size_t ngi = ngrids * ni * OF_CMPLX;
        size_t mgi = mgrids * mi;
        double *dout = (double *)out;
        FINT i, j, m;

        for (j = 0; j < mj; j++) {
                for (i = 0; i < mi; i++) {
#pragma GCC ivdep
                for (m = 0; m < mgrids; m++) {
                        dout[(i*ngrids+m)*OF_CMPLX  ] = gctrR[i*mgrids+m];
                        dout[(i*ngrids+m)*OF_CMPLX+1] = gctrI[i*mgrids+m];
                } }
                dout += ngi;
                gctrR += mgi;
                gctrI += mgi;
        }
}

/*
 * gctr(i,k,l,j) -> fijkl(i,j,k,l)
 * fijkl(ic:ic-1+di,jc:jc-1+dj,kc:kc-1+dk,lc:lc-1+dl)
 * fijkl(ni,nj,nk,nl), gctr(mi,mk,ml,mj)
 */
static void dcopy_iklj(double *fijkl, const double *gctr,
                       const FINT ni, const FINT nj, const FINT nk, const FINT nl,
                       const FINT mi, const FINT mj, const FINT mk, const FINT ml)
{
        const size_t nij = ni * nj;
        const size_t nijk = nij * nk;
        const size_t mik = mi * mk;
        const size_t mikl = mik * ml;
        FINT i, j, k, l;
        double *pijkl;
        const double *pgctr;

        switch (mi) {
        case 1:
                for (l = 0; l < ml; l++) {
                        for (k = 0; k < mk; k++) {
                                pijkl = fijkl + k * nij;
                                pgctr = gctr + k * mi;
#pragma GCC ivdep
                                for (j = 0; j < mj; j++) {
                                        pijkl[ni*j] = pgctr[mikl*j];
                                }
                        }
                        fijkl += nijk;
                        gctr += mik;
                }
                break;
        case 3:
                for (l = 0; l < ml; l++) {
                        for (k = 0; k < mk; k++) {
                                pijkl = fijkl + k * nij;
                                pgctr = gctr + k * mi;
#pragma GCC ivdep
                                for (j = 0; j < mj; j++) {
                                        pijkl[ni*j+0] = pgctr[mikl*j+0];
                                        pijkl[ni*j+1] = pgctr[mikl*j+1];
                                        pijkl[ni*j+2] = pgctr[mikl*j+2];
                                }
                        }
                        fijkl += nijk;
                        gctr += mik;
                }
                break;
        case 5:
                for (l = 0; l < ml; l++) {
                        for (k = 0; k < mk; k++) {
                                pijkl = fijkl + k * nij;
                                pgctr = gctr + k * mi;
#pragma GCC ivdep
                                for (j = 0; j < mj; j++) {
                                        pijkl[ni*j+0] = pgctr[mikl*j+0];
                                        pijkl[ni*j+1] = pgctr[mikl*j+1];
                                        pijkl[ni*j+2] = pgctr[mikl*j+2];
                                        pijkl[ni*j+3] = pgctr[mikl*j+3];
                                        pijkl[ni*j+4] = pgctr[mikl*j+4];
                                }
                        }
                        fijkl += nijk;
                        gctr += mik;
                }
                break;
        case 6:
                for (l = 0; l < ml; l++) {
                        for (k = 0; k < mk; k++) {
                                pijkl = fijkl + k * nij;
                                pgctr = gctr + k * mi;
#pragma GCC ivdep
                                for (j = 0; j < mj; j++) {
                                        pijkl[ni*j+0] = pgctr[mikl*j+0];
                                        pijkl[ni*j+1] = pgctr[mikl*j+1];
                                        pijkl[ni*j+2] = pgctr[mikl*j+2];
                                        pijkl[ni*j+3] = pgctr[mikl*j+3];
                                        pijkl[ni*j+4] = pgctr[mikl*j+4];
                                        pijkl[ni*j+5] = pgctr[mikl*j+5];
                                }
                        }
                        fijkl += nijk;
                        gctr += mik;
                }
                break;
        case 7:
                for (l = 0; l < ml; l++) {
                        for (k = 0; k < mk; k++) {
                                pijkl = fijkl + k * nij;
                                pgctr = gctr + k * mi;
#pragma GCC ivdep
                                for (j = 0; j < mj; j++) {
                                        pijkl[ni*j+0] = pgctr[mikl*j+0];
                                        pijkl[ni*j+1] = pgctr[mikl*j+1];
                                        pijkl[ni*j+2] = pgctr[mikl*j+2];
                                        pijkl[ni*j+3] = pgctr[mikl*j+3];
                                        pijkl[ni*j+4] = pgctr[mikl*j+4];
                                        pijkl[ni*j+5] = pgctr[mikl*j+5];
                                        pijkl[ni*j+6] = pgctr[mikl*j+6];
                                }
                        }
                        fijkl += nijk;
                        gctr += mik;
                }
                break;
        default:
                for (l = 0; l < ml; l++) {
                        for (k = 0; k < mk; k++) {
                                pijkl = fijkl + k * nij;
                                pgctr = gctr + k * mi;
                                for (j = 0; j < mj; j++) {
#pragma GCC ivdep
                                        for (i = 0; i < mi; i++) {
                                                pijkl[ni*j+i] = pgctr[mikl*j+i];
                                        }
                                }
                        }
                        fijkl += nijk;
                        gctr += mik;
                }
        }
}

static void zcopy_iklj(double complex *fijkl, double *gctrR, double *gctrI,
                       const FINT ni, const FINT nj, const FINT nk, const FINT nl,
                       const FINT mi, const FINT mj, const FINT mk, const FINT ml)
{
        size_t nij = ni * nj;
        size_t nijk = nij * nk;
        size_t mik = mi * mk;
        size_t mikl = mik * ml;
        FINT i, j, k, l;
        double *pijkl;
        double *pgctrR, *pgctrI;

        for (l = 0; l < ml; l++) {
                for (k = 0; k < mk; k++) {
                        pijkl = (double *)(fijkl + k * nij);
                        pgctrR = gctrR + k * mi;
                        pgctrI = gctrI + k * mi;
                        for (j = 0; j < mj; j++) {
#pragma GCC ivdep
                                for (i = 0; i < mi; i++) {
                                        pijkl[(j*ni+i)*OF_CMPLX  ] = pgctrR[j*mikl+i];
                                        pijkl[(j*ni+i)*OF_CMPLX+1] = pgctrI[j*mikl+i];
                                }
                        }
                }
                fijkl += nijk;
                gctrR += mik;
                gctrI += mik;
        }
}

void c2s_dset0(double *out, FINT *dims, FINT *counts)
{
        FINT ni = dims[0];
        FINT nj = dims[1];
        FINT nk = dims[2];
        size_t nij = ni * nj;
        size_t nijk = nij * nk;
        FINT i, j, k, l;
        if (dims == counts) {
                for (i = 0; i < nijk * counts[3]; i++) {
                        out[i] = 0;
                }
                return;
        }
        FINT di = counts[0];
        FINT dj = counts[1];
        FINT dk = counts[2];
        FINT dl = counts[3];
        double *pout;
        for (l = 0; l < dl; l++) {
                for (k = 0; k < dk; k++) {
                        pout = out + k * nij;
                        for (j = 0; j < dj; j++) {
                        for (i = 0; i < di; i++) {
                                pout[j*ni+i] = 0;
                        } }
                }
                out += nijk;
        }
}
void c2s_zset0(double complex *out, FINT *dims, FINT *counts)
{
        FINT ni = dims[0];
        FINT nj = dims[1];
        FINT nk = dims[2];
        size_t nij = ni * nj;
        size_t nijk = nij * nk;
        FINT i, j, k, l;
        if (dims == counts) {
                for (i = 0; i < nijk * counts[3]; i++) {
                        out[i] = 0;
                }
                return;
        }
        FINT di = counts[0];
        FINT dj = counts[1];
        FINT dk = counts[2];
        FINT dl = counts[3];
        double complex *pout;
        for (l = 0; l < dl; l++) {
                for (k = 0; k < dk; k++) {
                        pout = out + k * nij;
                        for (j = 0; j < dj; j++) {
                        for (i = 0; i < di; i++) {
                                pout[j*ni+i] = 0;
                        } }
                }
                out += nijk;
        }
}

void c2s_grids_dset0(double *out, FINT *dims, FINT *counts)
{
        FINT counts1[4] = {counts[2], counts[0], counts[1], counts[3]};
        FINT dims1[4] = {dims[2], dims[0], dims[1],dims [3]};
        c2s_dset0(out, dims1, counts1);
}

void c2s_grids_zset0(double complex *out, FINT *dims, FINT *counts)
{
        FINT counts1[4] = {counts[2], counts[0], counts[1], counts[3]};
        FINT dims1[4] = {dims[2], dims[0], dims[1],dims [3]};
        c2s_zset0(out, dims1, counts1);
}


/*
 * use f_ket to transform k,l for gctr(i,j,k,l), where
 * sizsph = nbra * (2*l+1)
 * sizcart = nbra * (l*(l+1)/2)
 * and return the pointer to the buffer which holds the transformed gctr
 */
static double *sph2e_inner(double *gsph, double *gcart,
                           FINT l, FINT nbra, FINT ncall, FINT sizsph, FINT sizcart)
{
        FINT n;
        switch (l) {
#ifdef PYPZPX
        case 0:
                return gcart;
        case 1:
                for (n = 0; n < ncall; n++) {
                        p_ket_cart2spheric(gsph+n*sizsph, gcart+n*sizcart, nbra, nbra, l);
                }
                break;
#else
        case 0: case 1:
                return gcart;
#endif
        case 2:
                for (n = 0; n < ncall; n++) {
                        d_ket_cart2spheric(gsph+n*sizsph, gcart+n*sizcart, nbra, nbra, l);
                }
                break;
        case 3:
                for (n = 0; n < ncall; n++) {
                        f_ket_cart2spheric(gsph+n*sizsph, gcart+n*sizcart, nbra, nbra, l);
                }
                break;
        case 4:
                for (n = 0; n < ncall; n++) {
                        g_ket_cart2spheric(gsph+n*sizsph, gcart+n*sizcart, nbra, nbra, l);
                }
                break;
        default:
                for (n = 0; n < ncall; n++) {
                        a_ket_cart2spheric(gsph+n*sizsph, gcart+n*sizcart, nbra, nbra, l);
                }
        }
        return gsph;
}


/*
 * 1e integrals, cartesian to real spheric.
 */
void c2s_sph_1e(double *opij, double *gctr, FINT *dims,
                CINTEnvVars *envs, double *cache)
{
        FINT i_l = envs->i_l;
        FINT j_l = envs->j_l;
        FINT i_ctr = envs->x_ctr[0];
        FINT j_ctr = envs->x_ctr[1];
        FINT di = i_l * 2 + 1;
        FINT dj = j_l * 2 + 1;
        FINT ni = dims[0];
        FINT nj = dims[1];
        FINT ofj = ni * dj;
        FINT nfi = envs->nfi;
        FINT nf = envs->nf;
        FINT ic, jc;
        FINT buflen = nfi*dj;
        double *buf1, *buf2;
        MALLOC_INSTACK(buf1, buflen);
        MALLOC_INSTACK(buf2, buflen);
        double *pij;
        double *tmp1;

        for (jc = 0; jc < j_ctr; jc++) {
        for (ic = 0; ic < i_ctr; ic++) {
                tmp1 = (c2s_ket_sph[j_l])(buf1, gctr, nfi, nfi, j_l);
                tmp1 = (c2s_bra_sph[i_l])(buf2, dj, tmp1, i_l);
                pij = opij + ofj * jc + di * ic;
                dcopy_ij(pij, tmp1, ni, nj, di, dj);
                gctr += nf;
        } }
}


/*
 * 1e integrals, cartesian to spin-free spinor.
 */
void c2s_sf_1e(double complex *opij, double *gctr, FINT *dims,
               CINTEnvVars *envs, double *cache)
{
        FINT *shls = envs->shls;
        FINT *bas = envs->bas;
        FINT i_sh = shls[0];
        FINT j_sh = shls[1];
        FINT i_l = envs->i_l;
        FINT j_l = envs->j_l;
        FINT i_kp = bas(KAPPA_OF, i_sh);
        FINT j_kp = bas(KAPPA_OF, j_sh);
        FINT i_ctr = envs->x_ctr[0];
        FINT j_ctr = envs->x_ctr[1];
        FINT di = _len_spinor(i_kp, i_l);
        FINT dj = _len_spinor(j_kp, j_l);
        FINT ni = dims[0];
        FINT nj = dims[1];
        FINT ofj = ni * dj;
        FINT nfj = envs->nfj;
        FINT nf2j = nfj + nfj;
        FINT nf = envs->nf;
        FINT ic, jc;
        double *tmp1R, *tmp1I, *tmp2R, *tmp2I;
        MALLOC_INSTACK(tmp1R, di*nf2j);
        MALLOC_INSTACK(tmp1I, di*nf2j);
        MALLOC_INSTACK(tmp2R, di*dj);
        MALLOC_INSTACK(tmp2I, di*dj);

        for (jc = 0; jc < j_ctr; jc++) {
        for (ic = 0; ic < i_ctr; ic++) {
                a_bra_cart2spinor_sf(tmp1R, tmp1I, NULL, NULL, NULL, gctr, nfj, i_kp, i_l);
                a_ket_cart2spinor(tmp2R, tmp2I, tmp1R, tmp1I, di, j_kp, j_l);
                zcopy_ij(opij+ofj*jc+di*ic, tmp2R, tmp2I, ni, nj, di, dj);
                gctr += nf;
        } }
}

void c2s_sf_1ei(double complex *opij, double *gctr, FINT *dims,
                CINTEnvVars *envs, double *cache)
{
        FINT *shls = envs->shls;
        FINT *bas = envs->bas;
        FINT i_sh = shls[0];
        FINT j_sh = shls[1];
        FINT i_l = envs->i_l;
        FINT j_l = envs->j_l;
        FINT i_kp = bas(KAPPA_OF, i_sh);
        FINT j_kp = bas(KAPPA_OF, j_sh);
        FINT i_ctr = envs->x_ctr[0];
        FINT j_ctr = envs->x_ctr[1];
        FINT di = _len_spinor(i_kp, i_l);
        FINT dj = _len_spinor(j_kp, j_l);
        FINT ni = dims[0];
        FINT nj = dims[1];
        FINT ofj = ni * dj;
        FINT nfj = envs->nfj;
        FINT nf2j = nfj + nfj;
        FINT nf = envs->nf;
        FINT ic, jc;
        double *tmp1R, *tmp1I, *tmp2R, *tmp2I;
        MALLOC_INSTACK(tmp1R, di*nf2j);
        MALLOC_INSTACK(tmp1I, di*nf2j);
        MALLOC_INSTACK(tmp2R, di*dj);
        MALLOC_INSTACK(tmp2I, di*dj);

        for (jc = 0; jc < j_ctr; jc++) {
        for (ic = 0; ic < i_ctr; ic++) {
                a_bra_cart2spinor_sf(tmp1R, tmp1I, NULL, NULL, NULL, gctr, nfj, i_kp, i_l);
                a_iket_cart2spinor(tmp2R, tmp2I, tmp1R, tmp1I, di, j_kp, j_l);
                zcopy_ij(opij+ofj*jc+di*ic, tmp2R, tmp2I, ni, nj, di, dj);
                gctr += nf;
        } }
}


/*
 * 1e integrals, cartesian to spinor.
 */
void c2s_si_1e(double complex *opij, double *gctr, FINT *dims,
               CINTEnvVars *envs, double *cache)
{
        FINT *shls = envs->shls;
        FINT *bas = envs->bas;
        FINT i_sh = shls[0];
        FINT j_sh = shls[1];
        FINT i_l = envs->i_l;
        FINT j_l = envs->j_l;
        FINT i_kp = bas(KAPPA_OF, i_sh);
        FINT j_kp = bas(KAPPA_OF, j_sh);
        FINT i_ctr = envs->x_ctr[0];
        FINT j_ctr = envs->x_ctr[1];
        FINT di = _len_spinor(i_kp, i_l);
        FINT dj = _len_spinor(j_kp, j_l);
        FINT ni = dims[0];
        FINT nj = dims[1];
        FINT ofj = ni * dj;
        FINT nfi = envs->nfi;
        FINT nfj = envs->nfj;
        FINT nf2i = nfi + nfi;
        FINT nf2j = nfj + nfj;
        FINT nf = envs->nf;
        FINT ic, jc;
        double *gc_x = gctr;
        double *gc_y = gc_x + nf * i_ctr * j_ctr;
        double *gc_z = gc_y + nf * i_ctr * j_ctr;
        double *gc_1 = gc_z + nf * i_ctr * j_ctr;
        double *tmp1R, *tmp1I, *tmp2R, *tmp2I;
        MALLOC_INSTACK(tmp1R, di*nf2j);
        MALLOC_INSTACK(tmp1I, di*nf2j);
        MALLOC_INSTACK(tmp2R, di*dj);
        MALLOC_INSTACK(tmp2I, di*dj);

        for (jc = 0; jc < j_ctr; jc++) {
        for (ic = 0; ic < i_ctr; ic++) {
                a_bra_cart2spinor_si(tmp1R, tmp1I, gc_x, gc_y, gc_z, gc_1, nfj, i_kp, i_l);
                a_ket_cart2spinor(tmp2R, tmp2I, tmp1R, tmp1I, di, j_kp, j_l);
                zcopy_ij(opij+ofj*jc+di*ic, tmp2R, tmp2I, ni, nj, di, dj);

                gc_x += nf;
                gc_y += nf;
                gc_z += nf;
                gc_1 += nf;
        } }
}
void c2s_si_1ei(double complex *opij, double *gctr, FINT *dims,
                CINTEnvVars *envs, double *cache)
{
        FINT *shls = envs->shls;
        FINT *bas = envs->bas;
        FINT i_sh = shls[0];
        FINT j_sh = shls[1];
        FINT i_l = envs->i_l;
        FINT j_l = envs->j_l;
        FINT i_kp = bas(KAPPA_OF, i_sh);
        FINT j_kp = bas(KAPPA_OF, j_sh);
        FINT i_ctr = envs->x_ctr[0];
        FINT j_ctr = envs->x_ctr[1];
        FINT di = _len_spinor(i_kp, i_l);
        FINT dj = _len_spinor(j_kp, j_l);
        FINT ni = dims[0];
        FINT nj = dims[1];
        FINT ofj = ni * dj;
        FINT nfi = envs->nfi;
        FINT nfj = envs->nfj;
        FINT nf2i = nfi + nfi;
        FINT nf2j = nfj + nfj;
        FINT nf = envs->nf;
        FINT ic, jc;
        double *gc_x = gctr;
        double *gc_y = gc_x + nf * i_ctr * j_ctr;
        double *gc_z = gc_y + nf * i_ctr * j_ctr;
        double *gc_1 = gc_z + nf * i_ctr * j_ctr;
        double *tmp1R, *tmp1I, *tmp2R, *tmp2I;
        MALLOC_INSTACK(tmp1R, di*nf2j);
        MALLOC_INSTACK(tmp1I, di*nf2j);
        MALLOC_INSTACK(tmp2R, di*dj);
        MALLOC_INSTACK(tmp2I, di*dj);

        for (jc = 0; jc < j_ctr; jc++) {
        for (ic = 0; ic < i_ctr; ic++) {
                a_bra_cart2spinor_si(tmp1R, tmp1I, gc_x, gc_y, gc_z, gc_1, nfj, i_kp, i_l);
                a_iket_cart2spinor(tmp2R, tmp2I, tmp1R, tmp1I, di, j_kp, j_l);
                zcopy_ij(opij+ofj*jc+di*ic, tmp2R, tmp2I, ni, nj, di, dj);

                gc_x += nf;
                gc_y += nf;
                gc_z += nf;
                gc_1 += nf;
        } }
}

void c2s_sph_1e_grids(double *out, double *gctr, FINT *dims,
                      CINTEnvVars *envs, double *cache)
{
        FINT ngrids = envs->ngrids;
        FINT i_l = envs->i_l;
        FINT j_l = envs->j_l;
        FINT i_ctr = envs->x_ctr[0];
        FINT j_ctr = envs->x_ctr[1];
        FINT di = i_l * 2 + 1;
        FINT dj = j_l * 2 + 1;
        FINT ni = dims[0];
        FINT nj = dims[1];
        size_t Ng = dims[2];
        FINT ofj = ni * dj;
        FINT nfi = envs->nfi;
        FINT nf = envs->nf;
        FINT ic, jc, grids_offset;
        FINT bgrids, bgrids_di, bgrids_nfi;
        FINT buflen = GRID_BLKSIZE * nfi * dj;
        double *buf1, *buf2;
        MALLOC_ALIGN8_INSTACK(buf1, buflen);
        MALLOC_ALIGN8_INSTACK(buf2, buflen);
        double *pij;
        double *tmp1;

        for (grids_offset = 0; grids_offset < ngrids; grids_offset += GRID_BLKSIZE) {
                bgrids = MIN(ngrids - grids_offset, GRID_BLKSIZE);
                bgrids_di = bgrids * di;
                bgrids_nfi = bgrids * nfi;
                for (jc = 0; jc < j_ctr; jc++) {
                for (ic = 0; ic < i_ctr; ic++) {
                        tmp1 = (c2s_ket_sph[j_l])(buf1, gctr, bgrids_nfi, bgrids_nfi, j_l);
                        tmp1 = sph2e_inner(buf2, tmp1, i_l, bgrids, dj, bgrids_di, bgrids_nfi);
                        pij = out + Ng * (ofj * jc + di * ic) + grids_offset;
                        dcopy_grids_ij(pij, tmp1, Ng, ni, nj, bgrids, di, dj);
                        gctr += bgrids * nf;
                } }
        }
}

void c2s_cart_1e_grids(double *out, double *gctr, FINT *dims,
                       CINTEnvVars *envs, double *cache)
{
        FINT ngrids = envs->ngrids;
        FINT i_ctr = envs->x_ctr[0];
        FINT j_ctr = envs->x_ctr[1];
        FINT ni = dims[0];
        FINT nj = dims[1];
        size_t Ng = dims[2];
        FINT nfi = envs->nfi;
        FINT nfj = envs->nfj;
        FINT nf = envs->nf;
        FINT ofj = ni * nfj;
        FINT ic, jc, grids_offset;
        FINT bgrids;
        double *pij;

        for (grids_offset = 0; grids_offset < ngrids; grids_offset += GRID_BLKSIZE) {
                bgrids = MIN(ngrids - grids_offset, GRID_BLKSIZE);
                for (jc = 0; jc < j_ctr; jc++) {
                for (ic = 0; ic < i_ctr; ic++) {
                        pij = out + Ng * (ofj * jc + nfi * ic) + grids_offset;
                        dcopy_grids_ij(pij, gctr, Ng, ni, nj, bgrids, nfi, nfj);
                        gctr += bgrids * nf;
                } }
        }
}

/*
 * 1e-grids integrals, cartesian to spin-free spinor.
 */
void c2s_sf_1e_grids(double complex *out, double *gctr, FINT *dims,
                     CINTEnvVars *envs, double *cache)
{
        FINT ngrids = envs->ngrids;
        FINT *shls = envs->shls;
        FINT *bas = envs->bas;
        FINT i_sh = shls[0];
        FINT j_sh = shls[1];
        FINT i_l = envs->i_l;
        FINT j_l = envs->j_l;
        FINT i_kp = bas(KAPPA_OF, i_sh);
        FINT j_kp = bas(KAPPA_OF, j_sh);
        FINT i_ctr = envs->x_ctr[0];
        FINT j_ctr = envs->x_ctr[1];
        FINT di = _len_spinor(i_kp, i_l);
        FINT dj = _len_spinor(j_kp, j_l);
        FINT ni = dims[0];
        FINT nj = dims[1];
        size_t Ng = dims[2];
        FINT ofj = ni * dj;
        FINT nfi = envs->nfi;
        FINT nfj = envs->nfj;
        FINT nf2j = nfj + nfj;
        FINT nf = envs->nf;
        FINT ic, jc, grids_offset;
        FINT bgrids, bgrids_di;
        FINT buflen = GRID_BLKSIZE * di * nf2j;
        double *tmp1R, *tmp1I, *tmp2R, *tmp2I;
        MALLOC_ALIGN8_INSTACK(tmp1R, buflen);
        MALLOC_ALIGN8_INSTACK(tmp1I, buflen);
        MALLOC_ALIGN8_INSTACK(tmp2R, buflen);
        MALLOC_ALIGN8_INSTACK(tmp2I, buflen);
        double complex *pij;

        for (grids_offset = 0; grids_offset < ngrids; grids_offset += GRID_BLKSIZE) {
                bgrids = MIN(ngrids - grids_offset, GRID_BLKSIZE);
                bgrids_di = bgrids * di;
                for (jc = 0; jc < j_ctr; jc++) {
                for (ic = 0; ic < i_ctr; ic++) {
                        a_bra1_cart2spinor_sf(tmp1R, tmp1I, NULL, NULL, NULL, gctr, bgrids, nfj, i_kp, i_l);
                        a_ket_cart2spinor(tmp2R, tmp2I, tmp1R, tmp1I, bgrids_di, j_kp, j_l);
                        pij = out + Ng * (ofj * jc + di * ic) + grids_offset;
                        zcopy_grids_ij(pij, tmp2R, tmp2I, Ng, ni, nj, bgrids, di, dj);
                        gctr += bgrids * nf;
                } }
        }
}
void c2s_sf_1e_gridsi(double complex *out, double *gctr, FINT *dims,
                      CINTEnvVars *envs, double *cache)
{
        FINT ngrids = envs->ngrids;
        FINT *shls = envs->shls;
        FINT *bas = envs->bas;
        FINT i_sh = shls[0];
        FINT j_sh = shls[1];
        FINT i_l = envs->i_l;
        FINT j_l = envs->j_l;
        FINT i_kp = bas(KAPPA_OF, i_sh);
        FINT j_kp = bas(KAPPA_OF, j_sh);
        FINT i_ctr = envs->x_ctr[0];
        FINT j_ctr = envs->x_ctr[1];
        FINT di = _len_spinor(i_kp, i_l);
        FINT dj = _len_spinor(j_kp, j_l);
        FINT ni = dims[0];
        FINT nj = dims[1];
        size_t Ng = dims[2];
        FINT ofj = ni * dj;
        FINT nfi = envs->nfi;
        FINT nfj = envs->nfj;
        FINT nf2j = nfj + nfj;
        FINT nf = envs->nf;
        FINT ic, jc, grids_offset;
        FINT bgrids, bgrids_di;
        FINT buflen = GRID_BLKSIZE * di * nf2j;
        double *tmp1R, *tmp1I, *tmp2R, *tmp2I;
        MALLOC_ALIGN8_INSTACK(tmp1R, buflen);
        MALLOC_ALIGN8_INSTACK(tmp1I, buflen);
        MALLOC_ALIGN8_INSTACK(tmp2R, buflen);
        MALLOC_ALIGN8_INSTACK(tmp2I, buflen);
        double complex *pij;

        for (grids_offset = 0; grids_offset < ngrids; grids_offset += GRID_BLKSIZE) {
                bgrids = MIN(ngrids - grids_offset, GRID_BLKSIZE);
                bgrids_di = bgrids * di;
                for (jc = 0; jc < j_ctr; jc++) {
                for (ic = 0; ic < i_ctr; ic++) {
                        a_bra1_cart2spinor_sf(tmp1R, tmp1I, NULL, NULL, NULL, gctr, bgrids, nfj, i_kp, i_l);
                        a_iket_cart2spinor(tmp2R, tmp2I, tmp1R, tmp1I, bgrids_di, j_kp, j_l);
                        pij = out + Ng * (ofj * jc + di * ic) + grids_offset;
                        zcopy_grids_ij(pij, tmp2R, tmp2I, Ng, ni, nj, bgrids, di, dj);
                        gctr += bgrids * nf;
                } }
        }
}
void c2s_si_1e_grids(double complex *out, double *gctr, FINT *dims,
                     CINTEnvVars *envs, double *cache)
{
        FINT ngrids = envs->ngrids;
        FINT *shls = envs->shls;
        FINT *bas = envs->bas;
        FINT i_sh = shls[0];
        FINT j_sh = shls[1];
        FINT i_l = envs->i_l;
        FINT j_l = envs->j_l;
        FINT i_kp = bas(KAPPA_OF, i_sh);
        FINT j_kp = bas(KAPPA_OF, j_sh);
        FINT i_ctr = envs->x_ctr[0];
        FINT j_ctr = envs->x_ctr[1];
        FINT di = _len_spinor(i_kp, i_l);
        FINT dj = _len_spinor(j_kp, j_l);
        FINT ni = dims[0];
        FINT nj = dims[1];
        size_t Ng = dims[2];
        FINT ofj = ni * dj;
        FINT nfi = envs->nfi;
        FINT nfj = envs->nfj;
        FINT nf2i = nfi + nfi;
        FINT nf2j = nfj + nfj;
        FINT nf = envs->nf;
        FINT ic, jc, grids_offset;
        FINT bgrids, bgrids_di, bgrids_nf;
        double *gc_x = gctr;
        double *gc_y = gc_x + ngrids * nf * i_ctr * j_ctr;
        double *gc_z = gc_y + ngrids * nf * i_ctr * j_ctr;
        double *gc_1 = gc_z + ngrids * nf * i_ctr * j_ctr;
        FINT buflen = GRID_BLKSIZE * di * nf2j;
        double *tmp1R, *tmp1I, *tmp2R, *tmp2I;
        MALLOC_ALIGN8_INSTACK(tmp1R, buflen);
        MALLOC_ALIGN8_INSTACK(tmp1I, buflen);
        MALLOC_ALIGN8_INSTACK(tmp2R, buflen);
        MALLOC_ALIGN8_INSTACK(tmp2I, buflen);
        double complex *pij;

        for (grids_offset = 0; grids_offset < ngrids; grids_offset += GRID_BLKSIZE) {
                bgrids = MIN(ngrids - grids_offset, GRID_BLKSIZE);
                bgrids_di = bgrids * di;
                bgrids_nf = bgrids * nf;
                for (jc = 0; jc < j_ctr; jc++) {
                for (ic = 0; ic < i_ctr; ic++) {
                        a_bra1_cart2spinor_si(tmp1R, tmp1I, gc_x, gc_y, gc_z, gc_1, bgrids, nfj, i_kp, i_l);
                        a_ket_cart2spinor(tmp2R, tmp2I, tmp1R, tmp1I, bgrids_di, j_kp, j_l);
                        pij = out + Ng * (ofj * jc + di * ic) + grids_offset;
                        zcopy_grids_ij(pij, tmp2R, tmp2I, Ng, ni, nj, bgrids, di, dj);
                        gc_x += bgrids_nf;
                        gc_y += bgrids_nf;
                        gc_z += bgrids_nf;
                        gc_1 += bgrids_nf;
                } }
        }
}
void c2s_si_1e_gridsi(double complex *out, double *gctr, FINT *dims,
                      CINTEnvVars *envs, double *cache)
{
        FINT ngrids = envs->ngrids;
        FINT *shls = envs->shls;
        FINT *bas = envs->bas;
        FINT i_sh = shls[0];
        FINT j_sh = shls[1];
        FINT i_l = envs->i_l;
        FINT j_l = envs->j_l;
        FINT i_kp = bas(KAPPA_OF, i_sh);
        FINT j_kp = bas(KAPPA_OF, j_sh);
        FINT i_ctr = envs->x_ctr[0];
        FINT j_ctr = envs->x_ctr[1];
        FINT di = _len_spinor(i_kp, i_l);
        FINT dj = _len_spinor(j_kp, j_l);
        FINT ni = dims[0];
        FINT nj = dims[1];
        size_t Ng = dims[2];
        FINT ofj = ni * dj;
        FINT nfi = envs->nfi;
        FINT nfj = envs->nfj;
        FINT nf2i = nfi + nfi;
        FINT nf2j = nfj + nfj;
        FINT nf = envs->nf;
        FINT ic, jc, grids_offset;
        FINT bgrids, bgrids_di, bgrids_nf;
        double *gc_x = gctr;
        double *gc_y = gc_x + ngrids * nf * i_ctr * j_ctr;
        double *gc_z = gc_y + ngrids * nf * i_ctr * j_ctr;
        double *gc_1 = gc_z + ngrids * nf * i_ctr * j_ctr;
        FINT buflen = GRID_BLKSIZE * di * nf2j;
        double *tmp1R, *tmp1I, *tmp2R, *tmp2I;
        MALLOC_ALIGN8_INSTACK(tmp1R, buflen);
        MALLOC_ALIGN8_INSTACK(tmp1I, buflen);
        MALLOC_ALIGN8_INSTACK(tmp2R, buflen);
        MALLOC_ALIGN8_INSTACK(tmp2I, buflen);
        double complex *pij;

        for (grids_offset = 0; grids_offset < ngrids; grids_offset += GRID_BLKSIZE) {
                bgrids = MIN(ngrids - grids_offset, GRID_BLKSIZE);
                bgrids_di = bgrids * di;
                bgrids_nf = bgrids * nf;
                for (jc = 0; jc < j_ctr; jc++) {
                for (ic = 0; ic < i_ctr; ic++) {
                        a_bra1_cart2spinor_si(tmp1R, tmp1I, gc_x, gc_y, gc_z, gc_1, bgrids, nfj, i_kp, i_l);
                        a_iket_cart2spinor(tmp2R, tmp2I, tmp1R, tmp1I, bgrids_di, j_kp, j_l);
                        pij = out + Ng * (ofj * jc + di * ic) + grids_offset;
                        zcopy_grids_ij(pij, tmp2R, tmp2I, Ng, ni, nj, bgrids, di, dj);
                        gc_x += bgrids_nf;
                        gc_y += bgrids_nf;
                        gc_z += bgrids_nf;
                        gc_1 += bgrids_nf;
                } }
        }
}


/*
 * 2e integrals, cartesian to real spherical functions.
 *
 * gctr: Cartesian GTO integrals, ordered as <ik|lj>
 */
void c2s_sph_2e1(double *out, double *gctr, FINT *dims,
                 CINTEnvVars *envs, double *cache)
{
        FINT i_l = envs->i_l;
        FINT j_l = envs->j_l;
        FINT k_l = envs->k_l;
        FINT l_l = envs->l_l;
        FINT i_ctr = envs->x_ctr[0];
        FINT j_ctr = envs->x_ctr[1];
        FINT k_ctr = envs->x_ctr[2];
        FINT l_ctr = envs->x_ctr[3];
        FINT di = i_l * 2 + 1;
        FINT dj = j_l * 2 + 1;
        FINT dk = k_l * 2 + 1;
        FINT dl = l_l * 2 + 1;
        FINT ni = dims[0];
        FINT nj = dims[1];
        FINT nk = dims[2];
        FINT nl = dims[3];
        FINT nfi = envs->nfi;
        FINT nfk = envs->nfk;
        FINT nfl = envs->nfl;
        FINT nfik = nfi * nfk;
        FINT nfikl = nfik * nfl;
        FINT dlj = dl * dj;
        FINT nf = envs->nf;
        FINT ofj = ni * dj;
        FINT ofk = ni * nj * dk;
        FINT ofl = ni * nj * nk * dl;
        FINT ic, jc, kc, lc;
        FINT buflen = nfikl*dj;
        double *buf1;
        MALLOC_INSTACK(buf1, buflen*4);
        double *buf2 = buf1 + buflen;
        double *buf3 = buf2 + buflen;
        double *buf4 = buf3 + buflen;
        double *pout, *tmp1;

        for (lc = 0; lc < l_ctr; lc++) {
        for (kc = 0; kc < k_ctr; kc++) {
        for (jc = 0; jc < j_ctr; jc++) {
        for (ic = 0; ic < i_ctr; ic++) {
                tmp1 = (c2s_ket_sph[j_l])(buf1, gctr, nfikl, nfikl, j_l);
                tmp1 = sph2e_inner(buf2, tmp1, l_l, nfik, dj, nfik*dl, nfikl);
                tmp1 = sph2e_inner(buf3, tmp1, k_l, nfi, dlj, nfi*dk, nfik);

                tmp1 = (c2s_bra_sph[i_l])(buf4, dk*dlj, tmp1, i_l);

                pout = out + ofl * lc + ofk * kc + ofj * jc + di * ic;
                dcopy_iklj(pout, tmp1, ni, nj, nk, nl, di, dj, dk, dl);
                gctr += nf;
        } } } }
}


/*
 * 2e integrals, cartesian to spin-free spinor for electron 1.
 *
 * gctr: Cartesian GTO integrals, ordered as <ik|lj>
 * opij: partially transformed GTO integrals, ordered as <ik|lj>
 */
void c2s_sf_2e1(double *opij, double *gctr, FINT *dims,
                CINTEnvVars *envs, double *cache)
{
        FINT *shls = envs->shls;
        FINT *bas = envs->bas;
        FINT i_sh = shls[0];
        FINT j_sh = shls[1];
        FINT i_l = envs->i_l;
        FINT j_l = envs->j_l;
        FINT i_kp = bas(KAPPA_OF, i_sh);
        FINT j_kp = bas(KAPPA_OF, j_sh);
        FINT i_ctr = envs->x_ctr[0];
        FINT j_ctr = envs->x_ctr[1];
        FINT k_ctr = envs->x_ctr[2];
        FINT l_ctr = envs->x_ctr[3];
        FINT di = _len_spinor(i_kp, i_l);
        FINT dj = _len_spinor(j_kp, j_l);
        FINT nfj = envs->nfj;
        FINT nfk = envs->nfk;
        FINT nfl = envs->nfl;
        FINT nf2j = nfj + nfj;
        FINT nf = envs->nf;
        FINT no = di * nfk * nfl * dj;
        FINT d_i = di * nfk * nfl;
        FINT d_j = nfk * nfl * nfj;
        FINT i;
        FINT buflen = di * nfk * nfl * nf2j;
        double *tmp1R, *tmp1I;
        MALLOC_INSTACK(tmp1R, buflen);
        MALLOC_INSTACK(tmp1I, buflen);

        for (i = 0; i < i_ctr * j_ctr * k_ctr * l_ctr; i++) {
                a_bra_cart2spinor_sf(tmp1R, tmp1I, NULL, NULL, NULL, gctr, d_j, i_kp, i_l);
                a_ket_cart2spinor(opij, opij+no, tmp1R, tmp1I, d_i, j_kp, j_l);
                gctr += nf;
                opij += no * OF_CMPLX;
        }
}

void c2s_sf_2e1i(double *opij, double *gctr, FINT *dims,
                 CINTEnvVars *envs, double *cache)
{
        FINT *shls = envs->shls;
        FINT *bas = envs->bas;
        FINT i_sh = shls[0];
        FINT j_sh = shls[1];
        FINT i_l = envs->i_l;
        FINT j_l = envs->j_l;
        FINT i_kp = bas(KAPPA_OF, i_sh);
        FINT j_kp = bas(KAPPA_OF, j_sh);
        FINT i_ctr = envs->x_ctr[0];
        FINT j_ctr = envs->x_ctr[1];
        FINT k_ctr = envs->x_ctr[2];
        FINT l_ctr = envs->x_ctr[3];
        FINT di = _len_spinor(i_kp, i_l);
        FINT dj = _len_spinor(j_kp, j_l);
        FINT nfj = envs->nfj;
        FINT nfk = envs->nfk;
        FINT nfl = envs->nfl;
        FINT nf2j = nfj + nfj;
        FINT nf = envs->nf;
        FINT no = di * nfk * nfl * dj;
        FINT d_i = di * nfk * nfl;
        FINT d_j = nfk * nfl * nfj;
        FINT i;
        FINT len1 = di * nfk * nfl * nf2j;
        double *tmp1R, *tmp1I;
        MALLOC_INSTACK(tmp1R, len1);
        MALLOC_INSTACK(tmp1I, len1);

        for (i = 0; i < i_ctr * j_ctr * k_ctr * l_ctr; i++) {
                a_bra_cart2spinor_sf(tmp1R, tmp1I, NULL, NULL, NULL, gctr, d_j, i_kp, i_l);
                a_iket_cart2spinor(opij, opij+no, tmp1R, tmp1I, d_i, j_kp, j_l);
                gctr += nf;
                opij += no * OF_CMPLX;
        }
}


/*
 * 2e integrals, cartesian to spin-free spinor for electron 2.
 *
 * opij: partial transformed GTO integrals, ordered as <ik|lj>
 */
void c2s_sf_2e2(double complex *fijkl, double *opij, FINT *dims,
                CINTEnvVars *envs, double *cache)
{
        FINT *shls = envs->shls;
        FINT *bas = envs->bas;
        FINT i_sh = shls[0];
        FINT j_sh = shls[1];
        FINT k_sh = shls[2];
        FINT l_sh = shls[3];
        FINT i_l = envs->i_l;
        FINT j_l = envs->j_l;
        FINT k_l = envs->k_l;
        FINT l_l = envs->l_l;
        FINT i_kp = bas(KAPPA_OF, i_sh);
        FINT j_kp = bas(KAPPA_OF, j_sh);
        FINT k_kp = bas(KAPPA_OF, k_sh);
        FINT l_kp = bas(KAPPA_OF, l_sh);
        FINT i_ctr = envs->x_ctr[0];
        FINT j_ctr = envs->x_ctr[1];
        FINT k_ctr = envs->x_ctr[2];
        FINT l_ctr = envs->x_ctr[3];
        FINT di = _len_spinor(i_kp, i_l);
        FINT dj = _len_spinor(j_kp, j_l);
        FINT dk = _len_spinor(k_kp, k_l);
        FINT dl = _len_spinor(l_kp, l_l);
        FINT ni = dims[0];
        FINT nj = dims[1];
        FINT nk = dims[2];
        FINT nl = dims[3];
        FINT nfk = envs->nfk;
        FINT nfl = envs->nfl;
        FINT nf2k = nfk + nfk;
        FINT nf2l = nfl + nfl;
        FINT nop = di * nfk * nfl * dj;
        FINT ofj = ni * dj;
        FINT ofk = ni * nj * dk;
        FINT ofl = ni * nj * nk * dl;
        FINT ic, jc, kc, lc;
        FINT len1 = di * dk * nf2l * dj;
        FINT len2 = di * dk * dl * dj;
        double *tmp1R, *tmp1I, *tmp2R, *tmp2I;
        MALLOC_INSTACK(tmp1R, len1);
        MALLOC_INSTACK(tmp1I, len1);
        MALLOC_INSTACK(tmp2R, len2);
        MALLOC_INSTACK(tmp2I, len2);
        double complex *pfijkl;

        for (lc = 0; lc < l_ctr; lc++) {
        for (kc = 0; kc < k_ctr; kc++) {
        for (jc = 0; jc < j_ctr; jc++) {
        for (ic = 0; ic < i_ctr; ic++) {
                a_bra1_cart2spinor_zf(tmp1R, tmp1I, NULL, NULL, NULL, opij, di, nfl*dj, k_kp, k_l);
                a_ket1_cart2spinor(tmp2R, tmp2I, tmp1R, tmp1I, di*dk, dj, l_kp, l_l);
                pfijkl = fijkl + (ofl * lc + ofk * kc + ofj * jc + di * ic);
                zcopy_iklj(pfijkl, tmp2R, tmp2I, ni, nj, nk, nl, di, dj, dk, dl);
                opij += nop * OF_CMPLX;
        } } } }
}
void c2s_sf_2e2i(double complex *fijkl, double *opij, FINT *dims,
                 CINTEnvVars *envs, double *cache)
{
        FINT *shls = envs->shls;
        FINT *bas = envs->bas;
        FINT i_sh = shls[0];
        FINT j_sh = shls[1];
        FINT k_sh = shls[2];
        FINT l_sh = shls[3];
        FINT i_l = envs->i_l;
        FINT j_l = envs->j_l;
        FINT k_l = envs->k_l;
        FINT l_l = envs->l_l;
        FINT i_kp = bas(KAPPA_OF, i_sh);
        FINT j_kp = bas(KAPPA_OF, j_sh);
        FINT k_kp = bas(KAPPA_OF, k_sh);
        FINT l_kp = bas(KAPPA_OF, l_sh);
        FINT i_ctr = envs->x_ctr[0];
        FINT j_ctr = envs->x_ctr[1];
        FINT k_ctr = envs->x_ctr[2];
        FINT l_ctr = envs->x_ctr[3];
        FINT di = _len_spinor(i_kp, i_l);
        FINT dj = _len_spinor(j_kp, j_l);
        FINT dk = _len_spinor(k_kp, k_l);
        FINT dl = _len_spinor(l_kp, l_l);
        FINT ni = dims[0];
        FINT nj = dims[1];
        FINT nk = dims[2];
        FINT nl = dims[3];
        FINT nfk = envs->nfk;
        FINT nfl = envs->nfl;
        FINT nf2k = nfk + nfk;
        FINT nf2l = nfl + nfl;
        FINT nop = di * nfk * nfl * dj;
        FINT ofj = ni * dj;
        FINT ofk = ni * nj * dk;
        FINT ofl = ni * nj * nk * dl;
        FINT ic, jc, kc, lc;
        FINT len2 = di * dk * dl * dj;
        FINT len1 = di * dk * nf2l * dj;
        double *tmp1R, *tmp1I, *tmp2R, *tmp2I;
        MALLOC_INSTACK(tmp1R, len1);
        MALLOC_INSTACK(tmp1I, len1);
        MALLOC_INSTACK(tmp2R, len2);
        MALLOC_INSTACK(tmp2I, len2);
        double complex *pfijkl;

        for (lc = 0; lc < l_ctr; lc++) {
        for (kc = 0; kc < k_ctr; kc++) {
        for (jc = 0; jc < j_ctr; jc++) {
        for (ic = 0; ic < i_ctr; ic++) {
                a_bra1_cart2spinor_zf(tmp1R, tmp1I, NULL, NULL, NULL, opij, di, nfl*dj, k_kp, k_l);
                a_iket1_cart2spinor(tmp2R, tmp2I, tmp1R, tmp1I, di*dk, dj, l_kp, l_l);
                pfijkl = fijkl + (ofl * lc + ofk * kc + ofj * jc + di * ic);
                zcopy_iklj(pfijkl, tmp2R, tmp2I, ni, nj, nk, nl, di, dj, dk, dl);
                opij += nop * OF_CMPLX;
        } } } }
}

/*
 * 2e integrals, cartesian to spinor for electron 1.
 *
 * gctr: Cartesian GTO integrals, ordered as <ik|lj>
 * opij: partial transformed GTO integrals, ordered as <ik|lj>
 */
void c2s_si_2e1(double *opij, double *gctr, FINT *dims,
                CINTEnvVars *envs, double *cache)
{
        FINT *shls = envs->shls;
        FINT *bas = envs->bas;
        FINT i_sh = shls[0];
        FINT j_sh = shls[1];
        FINT i_l = envs->i_l;
        FINT j_l = envs->j_l;
        FINT i_kp = bas(KAPPA_OF, i_sh);
        FINT j_kp = bas(KAPPA_OF, j_sh);
        FINT i_ctr = envs->x_ctr[0];
        FINT j_ctr = envs->x_ctr[1];
        FINT k_ctr = envs->x_ctr[2];
        FINT l_ctr = envs->x_ctr[3];
        FINT di = _len_spinor(i_kp, i_l);
        FINT dj = _len_spinor(j_kp, j_l);
        FINT nfi = envs->nfi;
        FINT nfj = envs->nfj;
        FINT nfk = envs->nfk;
        FINT nfl = envs->nfl;
        FINT nf2i = nfi + nfi;
        FINT nf2j = nfj + nfj;
        FINT nf = envs->nf;
        FINT no = di * nfk * nfl * dj;
        FINT d_i = di * nfk * nfl;
        FINT d_j = nfk * nfl * nfj;
        FINT i;
        double *gc_x = gctr;
        double *gc_y = gc_x + nf * i_ctr * j_ctr * k_ctr * l_ctr;
        double *gc_z = gc_y + nf * i_ctr * j_ctr * k_ctr * l_ctr;
        double *gc_1 = gc_z + nf * i_ctr * j_ctr * k_ctr * l_ctr;
        FINT len1 = di * nfk * nfl * nf2j;
        double *tmp1R, *tmp1I;
        MALLOC_INSTACK(tmp1R, len1);
        MALLOC_INSTACK(tmp1I, len1);

        for (i = 0; i < i_ctr * j_ctr * k_ctr * l_ctr; i++) {
                a_bra_cart2spinor_si(tmp1R, tmp1I, gc_x, gc_y, gc_z, gc_1, d_j, i_kp, i_l);
                a_ket_cart2spinor(opij, opij+no, tmp1R, tmp1I, d_i, j_kp, j_l);
                gc_x += nf;
                gc_y += nf;
                gc_z += nf;
                gc_1 += nf;
                opij += no * OF_CMPLX;
        }
}
void c2s_si_2e1i(double *opij, double *gctr, FINT *dims,
                 CINTEnvVars *envs, double *cache)
{
        FINT *shls = envs->shls;
        FINT *bas = envs->bas;
        FINT i_sh = shls[0];
        FINT j_sh = shls[1];
        FINT i_l = envs->i_l;
        FINT j_l = envs->j_l;
        FINT i_kp = bas(KAPPA_OF, i_sh);
        FINT j_kp = bas(KAPPA_OF, j_sh);
        FINT i_ctr = envs->x_ctr[0];
        FINT j_ctr = envs->x_ctr[1];
        FINT k_ctr = envs->x_ctr[2];
        FINT l_ctr = envs->x_ctr[3];
        FINT di = _len_spinor(i_kp, i_l);
        FINT dj = _len_spinor(j_kp, j_l);
        FINT nfi = envs->nfi;
        FINT nfj = envs->nfj;
        FINT nfk = envs->nfk;
        FINT nfl = envs->nfl;
        FINT nf2i = nfi + nfi;
        FINT nf2j = nfj + nfj;
        FINT nf = envs->nf;
        FINT no = di * nfk * nfl * dj;
        FINT d_i = di * nfk * nfl;
        FINT d_j = nfk * nfl * nfj;
        FINT i;
        double *gc_x = gctr;
        double *gc_y = gc_x + nf * i_ctr * j_ctr * k_ctr * l_ctr;
        double *gc_z = gc_y + nf * i_ctr * j_ctr * k_ctr * l_ctr;
        double *gc_1 = gc_z + nf * i_ctr * j_ctr * k_ctr * l_ctr;
        FINT len1 = di * nfk * nfl * nf2j;
        double *tmp1R, *tmp1I;
        MALLOC_INSTACK(tmp1R, len1);
        MALLOC_INSTACK(tmp1I, len1);

        for (i = 0; i < i_ctr * j_ctr * k_ctr * l_ctr; i++) {
                a_bra_cart2spinor_si(tmp1R, tmp1I, gc_x, gc_y, gc_z, gc_1, d_j, i_kp, i_l);
                a_iket_cart2spinor(opij, opij+no, tmp1R, tmp1I, d_i, j_kp, j_l);
                gc_x += nf;
                gc_y += nf;
                gc_z += nf;
                gc_1 += nf;
                opij += no * OF_CMPLX;
        }
}

void c2s_si_2e2(double complex *fijkl, double *opij, FINT *dims,
                CINTEnvVars *envs, double *cache)
{
        FINT *shls = envs->shls;
        FINT *bas = envs->bas;
        FINT i_sh = shls[0];
        FINT j_sh = shls[1];
        FINT k_sh = shls[2];
        FINT l_sh = shls[3];
        FINT i_l = envs->i_l;
        FINT j_l = envs->j_l;
        FINT k_l = envs->k_l;
        FINT l_l = envs->l_l;
        FINT i_kp = bas(KAPPA_OF, i_sh);
        FINT j_kp = bas(KAPPA_OF, j_sh);
        FINT k_kp = bas(KAPPA_OF, k_sh);
        FINT l_kp = bas(KAPPA_OF, l_sh);
        FINT i_ctr = envs->x_ctr[0];
        FINT j_ctr = envs->x_ctr[1];
        FINT k_ctr = envs->x_ctr[2];
        FINT l_ctr = envs->x_ctr[3];
        FINT di = _len_spinor(i_kp, i_l);
        FINT dj = _len_spinor(j_kp, j_l);
        FINT dk = _len_spinor(k_kp, k_l);
        FINT dl = _len_spinor(l_kp, l_l);
        FINT ni = dims[0];
        FINT nj = dims[1];
        FINT nk = dims[2];
        FINT nl = dims[3];
        FINT nfk = envs->nfk;
        FINT nfl = envs->nfl;
        FINT nf2k = nfk + nfk;
        FINT nf2l = nfl + nfl;
        FINT nop = di * nfk * nfl * dj;
        FINT ofj = ni * dj;
        FINT ofk = ni * nj * dk;
        FINT ofl = ni * nj * nk * dl;
        FINT ic, jc, kc, lc;
        double complex *pfijkl;
        double *ox = opij;
        double *oy = ox + nop * OF_CMPLX * i_ctr * j_ctr * k_ctr * l_ctr;
        double *oz = oy + nop * OF_CMPLX * i_ctr * j_ctr * k_ctr * l_ctr;
        double *o1 = oz + nop * OF_CMPLX * i_ctr * j_ctr * k_ctr * l_ctr;
        FINT buflen = di * dk * nf2l * dj;
        double *tmp1R, *tmp1I, *tmp2R, *tmp2I;
        MALLOC_INSTACK(tmp1R, buflen);
        MALLOC_INSTACK(tmp1I, buflen);
        MALLOC_INSTACK(tmp2R, buflen);
        MALLOC_INSTACK(tmp2I, buflen);

        for (lc = 0; lc < l_ctr; lc++) {
        for (kc = 0; kc < k_ctr; kc++) {
        for (jc = 0; jc < j_ctr; jc++) {
        for (ic = 0; ic < i_ctr; ic++) {
                a_bra1_cart2spinor_zi(tmp1R, tmp1I, ox, oy, oz, o1, di, nfl*dj, k_kp, k_l);
                a_ket1_cart2spinor(tmp2R, tmp2I, tmp1R, tmp1I, di*dk, dj, l_kp, l_l);
                pfijkl = fijkl + (ofl * lc + ofk * kc + ofj * jc + di * ic);
                zcopy_iklj(pfijkl, tmp2R, tmp2I, ni, nj, nk, nl, di, dj, dk, dl);
                ox += nop * OF_CMPLX;
                oy += nop * OF_CMPLX;
                oz += nop * OF_CMPLX;
                o1 += nop * OF_CMPLX;
        } } } }
}

void c2s_si_2e2i(double complex *fijkl, double *opij, FINT *dims,
                 CINTEnvVars *envs, double *cache)
{
        FINT *shls = envs->shls;
        FINT *bas = envs->bas;
        FINT i_sh = shls[0];
        FINT j_sh = shls[1];
        FINT k_sh = shls[2];
        FINT l_sh = shls[3];
        FINT i_l = envs->i_l;
        FINT j_l = envs->j_l;
        FINT k_l = envs->k_l;
        FINT l_l = envs->l_l;
        FINT i_kp = bas(KAPPA_OF, i_sh);
        FINT j_kp = bas(KAPPA_OF, j_sh);
        FINT k_kp = bas(KAPPA_OF, k_sh);
        FINT l_kp = bas(KAPPA_OF, l_sh);
        FINT i_ctr = envs->x_ctr[0];
        FINT j_ctr = envs->x_ctr[1];
        FINT k_ctr = envs->x_ctr[2];
        FINT l_ctr = envs->x_ctr[3];
        FINT di = _len_spinor(i_kp, i_l);
        FINT dj = _len_spinor(j_kp, j_l);
        FINT dk = _len_spinor(k_kp, k_l);
        FINT dl = _len_spinor(l_kp, l_l);
        FINT ni = dims[0];
        FINT nj = dims[1];
        FINT nk = dims[2];
        FINT nl = dims[3];
        FINT nfk = envs->nfk;
        FINT nfl = envs->nfl;
        FINT nf2k = nfk + nfk;
        FINT nf2l = nfl + nfl;
        FINT nop = di * nfk * nfl * dj;
        FINT ofj = ni * dj;
        FINT ofk = ni * nj * dk;
        FINT ofl = ni * nj * nk * dl;
        FINT ic, jc, kc, lc;
        double complex *pfijkl;
        double *ox = opij;
        double *oy = ox + nop * OF_CMPLX * i_ctr * j_ctr * k_ctr * l_ctr;
        double *oz = oy + nop * OF_CMPLX * i_ctr * j_ctr * k_ctr * l_ctr;
        double *o1 = oz + nop * OF_CMPLX * i_ctr * j_ctr * k_ctr * l_ctr;
        FINT buflen = di * dk * nf2l * dj;
        double *tmp1R, *tmp1I, *tmp2R, *tmp2I;
        MALLOC_INSTACK(tmp2R, buflen);
        MALLOC_INSTACK(tmp2I, buflen);
        MALLOC_INSTACK(tmp1R, buflen);
        MALLOC_INSTACK(tmp1I, buflen);

        for (lc = 0; lc < l_ctr; lc++) {
        for (kc = 0; kc < k_ctr; kc++) {
        for (jc = 0; jc < j_ctr; jc++) {
        for (ic = 0; ic < i_ctr; ic++) {
                a_bra1_cart2spinor_zi(tmp1R, tmp1I, ox, oy, oz, o1, di, nfl*dj, k_kp, k_l);
                a_iket1_cart2spinor(tmp2R, tmp2I, tmp1R, tmp1I, di*dk, dj, l_kp, l_l);
                pfijkl = fijkl + (ofl * lc + ofk * kc + ofj * jc + di * ic);
                zcopy_iklj(pfijkl, tmp2R, tmp2I, ni, nj, nk, nl, di, dj, dk, dl);
                ox += nop * OF_CMPLX;
                oy += nop * OF_CMPLX;
                oz += nop * OF_CMPLX;
                o1 += nop * OF_CMPLX;
        } } } }
}

/*
 * 1e integrals, reorder cartesian integrals.
 */
void c2s_cart_1e(double *opij, double *gctr, FINT *dims,
                 CINTEnvVars *envs, double *cache)
{
        FINT i_ctr = envs->x_ctr[0];
        FINT j_ctr = envs->x_ctr[1];
        FINT nfi = envs->nfi;
        FINT nfj = envs->nfj;
        FINT nf = envs->nf;
        FINT ni = dims[0];
        FINT nj = dims[1];
        FINT ofj = ni * nfj;
        FINT ic, jc;
        double *popij;

        for (jc = 0; jc < j_ctr; jc++) {
        for (ic = 0; ic < i_ctr; ic++) {
                popij = opij + ofj * jc + nfi * ic;
                dcopy_ij(popij, gctr, ni, nj, nfi, nfj);
                gctr += nf;
        } }
}

/*
 * 2e integrals, reorder cartesian integrals.
 */
void c2s_cart_2e1(double *fijkl, double *gctr, FINT *dims, CINTEnvVars *envs,
                  double *cache)
{
        FINT i_ctr = envs->x_ctr[0];
        FINT j_ctr = envs->x_ctr[1];
        FINT k_ctr = envs->x_ctr[2];
        FINT l_ctr = envs->x_ctr[3];
        FINT nfi = envs->nfi;
        FINT nfj = envs->nfj;
        FINT nfk = envs->nfk;
        FINT nfl = envs->nfl;
        FINT nf = envs->nf;
        FINT ni = dims[0];
        FINT nj = dims[1];
        FINT nk = dims[2];
        FINT nl = dims[3];
        FINT ofj = ni * nfj;
        FINT ofk = ni * nj * nfk;
        FINT ofl = ni * nj * nk * nfl;
        FINT ic, jc, kc, lc;
        double *pfijkl;

        for (lc = 0; lc < l_ctr; lc++) {
        for (kc = 0; kc < k_ctr; kc++) {
        for (jc = 0; jc < j_ctr; jc++) {
        for (ic = 0; ic < i_ctr; ic++) {
                pfijkl = fijkl + ofl * lc + ofk * kc + ofj * jc + nfi * ic;
                dcopy_iklj(pfijkl, gctr, ni, nj, nk, nl, nfi, nfj, nfk, nfl);
                gctr += nf;
        } } } }
}
void c2s_cart_2e2() {};


/*************************************************
 *
 * 3-center 2-electron integral transformation
 *
 *************************************************/
void c2s_sph_3c2e1(double *bufijk, double *gctr, FINT *dims,
                   CINTEnvVars *envs, double *cache)
{
        FINT i_l = envs->i_l;
        FINT j_l = envs->j_l;
        FINT k_l = envs->k_l;
        FINT i_ctr = envs->x_ctr[0];
        FINT j_ctr = envs->x_ctr[1];
        FINT k_ctr = envs->x_ctr[2];
        FINT di = i_l * 2 + 1;
        FINT dj = j_l * 2 + 1;
        FINT dk = k_l * 2 + 1;
        FINT ni = dims[0];
        FINT nj = dims[1];
        FINT nk = dims[2];
        FINT nfi = envs->nfi;
        FINT nfk = envs->nfk;
        FINT nf = envs->nf;
        FINT nfik = nfi * nfk;
        FINT ofj = ni * dj;
        FINT ofk = ni * nj * dk;
        FINT ic, jc, kc;
        FINT buflen = nfi*nfk*dj;
        double *buf1;
        MALLOC_INSTACK(buf1, buflen*3);
        double *buf2 = buf1 + buflen;
        double *buf3 = buf2 + buflen;
        double *pijk;
        double *tmp1;

        for (kc = 0; kc < k_ctr; kc++) {
        for (jc = 0; jc < j_ctr; jc++) {
        for (ic = 0; ic < i_ctr; ic++) {
                tmp1 = (c2s_ket_sph[j_l])(buf1, gctr, nfik, nfik, j_l);
                tmp1 = sph2e_inner(buf2, tmp1, k_l, nfi, dj, nfi*dk, nfik);
                tmp1 = (c2s_bra_sph[i_l])(buf3, dk*dj, tmp1, i_l);
                pijk = bufijk + ofk * kc + ofj * jc + di * ic;
                dcopy_iklj(pijk, tmp1, ni, nj, nk, 1, di, dj, dk, 1);
                gctr += nf;
        } } }
}

void c2s_cart_3c2e1(double *bufijk, double *gctr, FINT *dims,
                    CINTEnvVars *envs, double *cache)
{
        FINT i_ctr = envs->x_ctr[0];
        FINT j_ctr = envs->x_ctr[1];
        FINT k_ctr = envs->x_ctr[2];
        FINT nfi = envs->nfi;
        FINT nfj = envs->nfj;
        FINT nfk = envs->nfk;
        FINT ni = dims[0];
        FINT nj = dims[1];
        FINT nk = dims[2];
        FINT nf = envs->nf;
        FINT ofj = ni * nfj;
        FINT ofk = ni * nj * nfk;
        FINT ic, jc, kc;
        double *pijk;

        for (kc = 0; kc < k_ctr; kc++) {
        for (jc = 0; jc < j_ctr; jc++) {
        for (ic = 0; ic < i_ctr; ic++) {
                pijk = bufijk + ofk * kc + ofj * jc + nfi * ic;
                dcopy_iklj(pijk, gctr, ni, nj, nk, 1, nfi, nfj, nfk, 1);
                gctr += nf;
        } } }
}

/*
 * ssc ~ (spheric,spheric|cartesian)
 */
void c2s_sph_3c2e1_ssc(double *bufijk, double *gctr, FINT *dims,
                       CINTEnvVars *envs, double *cache)
{
        FINT i_l = envs->i_l;
        FINT j_l = envs->j_l;
        FINT i_ctr = envs->x_ctr[0];
        FINT j_ctr = envs->x_ctr[1];
        FINT k_ctr = envs->x_ctr[2];
        FINT di = i_l * 2 + 1;
        FINT dj = j_l * 2 + 1;
        FINT nfi = envs->nfi;
        FINT nfk = envs->nfk;
        FINT ni = di * i_ctr;
        FINT nj = dj * j_ctr;
        FINT nk = nfk * k_ctr;
        FINT nf = envs->nf;
        FINT nfik = nfi * nfk;
        FINT ofj = ni * dj;
        FINT ofk = ni * nj * nfk;
        FINT ic, jc, kc;
        FINT buflen = nfi*nfk*dj;
        double *buf1, *buf2;
        MALLOC_INSTACK(buf1, buflen);
        MALLOC_INSTACK(buf2, buflen);
        double *pijk;
        double *tmp1;

        for (kc = 0; kc < k_ctr; kc++) {
        for (jc = 0; jc < j_ctr; jc++) {
        for (ic = 0; ic < i_ctr; ic++) {
                tmp1 = (c2s_ket_sph[j_l])(buf1, gctr, nfik, nfik, j_l);
                tmp1 = (c2s_bra_sph[i_l])(buf2, nfk*dj, tmp1, i_l);
                pijk = bufijk + ofk * kc + ofj * jc + nfi * ic;
                dcopy_iklj(pijk, tmp1, ni, nj, nk, 1, di, dj, nfk, 1);
                gctr += nf;
        } } }
}

/*
 * 3c2e spinor integrals, cartesian to spin-free spinor for electron 1.
 */
void c2s_sf_3c2e1(double complex *opijk, double *gctr, FINT *dims,
                  CINTEnvVars *envs, double *cache)
{
        FINT *shls = envs->shls;
        FINT *bas = envs->bas;
        FINT i_sh = shls[0];
        FINT j_sh = shls[1];
        FINT i_l = envs->i_l;
        FINT j_l = envs->j_l;
        FINT k_l = envs->k_l;
        FINT i_kp = bas(KAPPA_OF, i_sh);
        FINT j_kp = bas(KAPPA_OF, j_sh);
        FINT i_ctr = envs->x_ctr[0];
        FINT j_ctr = envs->x_ctr[1];
        FINT k_ctr = envs->x_ctr[2];
        FINT di = _len_spinor(i_kp, i_l);
        FINT dj = _len_spinor(j_kp, j_l);
        FINT dk = k_l * 2 + 1;
        FINT nfi = envs->nfi;
        FINT nfj = envs->nfj;
        FINT nfk = envs->nfk;
        FINT nf2j = nfj + nfj;
        FINT nf = envs->nf;
        FINT nfik = nfi * nfk;
        FINT ni = dims[0];
        FINT nj = dims[1];
        FINT nk = dims[2];
        FINT ofj = ni * dj;
        FINT ofk = ni * nj * dk;
        FINT ic, jc, kc;
        FINT buflen = nfi*dk*nfj;
        double *buf, *pbuf;
        MALLOC_INSTACK(buf, buflen);
        FINT len1 = di * dk * nf2j;
        FINT len2 = di * dk * dj;
        double *tmp1R, *tmp1I, *tmp2R, *tmp2I;
        MALLOC_INSTACK(tmp1R, len1);
        MALLOC_INSTACK(tmp1I, len1);
        MALLOC_INSTACK(tmp2R, len2);
        MALLOC_INSTACK(tmp2I, len2);
        double complex *pijk;

        for (kc = 0; kc < k_ctr; kc++) {
        for (jc = 0; jc < j_ctr; jc++) {
        for (ic = 0; ic < i_ctr; ic++) {
                pbuf = sph2e_inner(buf, gctr, k_l, nfi, nfj, nfi*dk, nfik);
                a_bra_cart2spinor_sf(tmp1R, tmp1I, NULL, NULL, NULL, pbuf, dk*nfj, i_kp, i_l);
                a_ket_cart2spinor(tmp2R, tmp2I, tmp1R, tmp1I, di*dk, j_kp, j_l);
                pijk = opijk + ofk * kc + ofj * jc + di * ic;
                zcopy_iklj(pijk, tmp2R, tmp2I, ni, nj, nk, 1, di, dj, dk, 1);
                gctr += nf;
        } } }
}
void c2s_sf_3c2e1i(double complex *opijk, double *gctr, FINT *dims,
                   CINTEnvVars *envs, double *cache)
{
        FINT *shls = envs->shls;
        FINT *bas = envs->bas;
        FINT i_sh = shls[0];
        FINT j_sh = shls[1];
        FINT i_l = envs->i_l;
        FINT j_l = envs->j_l;
        FINT k_l = envs->k_l;
        FINT i_kp = bas(KAPPA_OF, i_sh);
        FINT j_kp = bas(KAPPA_OF, j_sh);
        FINT i_ctr = envs->x_ctr[0];
        FINT j_ctr = envs->x_ctr[1];
        FINT k_ctr = envs->x_ctr[2];
        FINT di = _len_spinor(i_kp, i_l);
        FINT dj = _len_spinor(j_kp, j_l);
        FINT dk = k_l * 2 + 1;
        FINT nfi = envs->nfi;
        FINT nfj = envs->nfj;
        FINT nfk = envs->nfk;
        FINT nf2j = nfj + nfj;
        FINT nf = envs->nf;
        FINT nfik = nfi * nfk;
        FINT ni = dims[0];
        FINT nj = dims[1];
        FINT nk = dims[2];
        FINT ofj = ni * dj;
        FINT ofk = ni * nj * dk;
        FINT d_i = di * dk;
        FINT d_j = dk * nfj;
        FINT ic, jc, kc;
        FINT buflen = nfi*dk*nfj;
        double *buf, *pbuf;
        MALLOC_INSTACK(buf, buflen);
        FINT len1 = di*dk*nf2j;
        FINT len2 = di*dk*dj;
        double *tmp1R, *tmp1I, *tmp2R, *tmp2I;
        MALLOC_INSTACK(tmp1R, len1);
        MALLOC_INSTACK(tmp1I, len1);
        MALLOC_INSTACK(tmp2R, len2);
        MALLOC_INSTACK(tmp2I, len2);
        double complex *pijk;

        for (kc = 0; kc < k_ctr; kc++) {
        for (jc = 0; jc < j_ctr; jc++) {
        for (ic = 0; ic < i_ctr; ic++) {
                pbuf = sph2e_inner(buf, gctr, k_l, nfi, nfj, nfi*dk, nfik);
                a_bra_cart2spinor_sf(tmp1R, tmp1I, NULL, NULL, NULL, pbuf, dk*nfj, i_kp, i_l);
                a_iket_cart2spinor(tmp2R, tmp2I, tmp1R, tmp1I, di*dk, j_kp, j_l);
                pijk = opijk + ofk * kc + ofj * jc + di * ic;
                zcopy_iklj(pijk, tmp2R, tmp2I, ni, nj, nk, 1, di, dj, dk, 1);
                gctr += nf;
        } } }
}
/*
 * 3c2e integrals, cartesian to spinor for electron 1.
 */
void c2s_si_3c2e1(double complex *opijk, double *gctr, FINT *dims,
                  CINTEnvVars *envs, double *cache)
{
        FINT *shls = envs->shls;
        FINT *bas = envs->bas;
        FINT i_sh = shls[0];
        FINT j_sh = shls[1];
        FINT i_l = envs->i_l;
        FINT j_l = envs->j_l;
        FINT k_l = envs->k_l;
        FINT i_kp = bas(KAPPA_OF, i_sh);
        FINT j_kp = bas(KAPPA_OF, j_sh);
        FINT i_ctr = envs->x_ctr[0];
        FINT j_ctr = envs->x_ctr[1];
        FINT k_ctr = envs->x_ctr[2];
        FINT di = _len_spinor(i_kp, i_l);
        FINT dj = _len_spinor(j_kp, j_l);
        FINT dk = k_l * 2 + 1;
        FINT nfi = envs->nfi;
        FINT nfj = envs->nfj;
        FINT nfk = envs->nfk;
        FINT nf2i = nfi + nfi;
        FINT nf2j = nfj + nfj;
        FINT nf = envs->nf;
        FINT nfik = nfi * nfk;
        FINT nfijdk = nfi * nfj * dk;
        FINT ni = dims[0];
        FINT nj = dims[1];
        FINT nk = dims[2];
        FINT ofj = ni * dj;
        FINT ofk = ni * nj * dk;
        FINT d_i = di * dk;
        FINT d_j = dk * nf2j;
        FINT ic, jc, kc;
        double *gc_x = gctr;
        double *gc_y = gc_x + nf * i_ctr * j_ctr * k_ctr;
        double *gc_z = gc_y + nf * i_ctr * j_ctr * k_ctr;
        double *gc_1 = gc_z + nf * i_ctr * j_ctr * k_ctr;
        FINT buflen = nfi*dk*nfj;
        double *bufx;
        MALLOC_INSTACK(bufx, buflen*4);
        double *bufy = bufx + buflen;
        double *bufz = bufy + buflen;
        double *buf1 = bufz + buflen;
        double *pgx, *pgy, *pgz, *pg1;
        FINT len1 = di * dk * nf2j;
        FINT len2 = di * dk * dj;
        double *tmp1R, *tmp1I, *tmp2R, *tmp2I;
        MALLOC_INSTACK(tmp1R, len1);
        MALLOC_INSTACK(tmp1I, len1);
        MALLOC_INSTACK(tmp2R, len2);
        MALLOC_INSTACK(tmp2I, len2);
        double complex *pijk;

        for (kc = 0; kc < k_ctr; kc++) {
        for (jc = 0; jc < j_ctr; jc++) {
        for (ic = 0; ic < i_ctr; ic++) {
                pgx = sph2e_inner(bufx, gc_x, k_l, nfi, nfj, nfi*dk, nfik);
                pgy = sph2e_inner(bufy, gc_y, k_l, nfi, nfj, nfi*dk, nfik);
                pgz = sph2e_inner(bufz, gc_z, k_l, nfi, nfj, nfi*dk, nfik);
                pg1 = sph2e_inner(buf1, gc_1, k_l, nfi, nfj, nfi*dk, nfik);
                a_bra_cart2spinor_si(tmp1R, tmp1I, pgx, pgy, pgz, pg1, dk*nfj, i_kp, i_l);
                a_ket_cart2spinor(tmp2R, tmp2I, tmp1R, tmp1I, di*dk, j_kp, j_l);
                pijk = opijk + ofk * kc + ofj * jc + di * ic;
                zcopy_iklj(pijk, tmp2R, tmp2I, ni, nj, nk, 1, di, dj, dk, 1);
                gc_x += nf;
                gc_y += nf;
                gc_z += nf;
                gc_1 += nf;
        } } }
}

void c2s_si_3c2e1i(double complex *opijk, double *gctr, FINT *dims,
                   CINTEnvVars *envs, double *cache)
{
        FINT *shls = envs->shls;
        FINT *bas = envs->bas;
        FINT i_sh = shls[0];
        FINT j_sh = shls[1];
        FINT i_l = envs->i_l;
        FINT j_l = envs->j_l;
        FINT k_l = envs->k_l;
        FINT i_kp = bas(KAPPA_OF, i_sh);
        FINT j_kp = bas(KAPPA_OF, j_sh);
        FINT i_ctr = envs->x_ctr[0];
        FINT j_ctr = envs->x_ctr[1];
        FINT k_ctr = envs->x_ctr[2];
        FINT di = _len_spinor(i_kp, i_l);
        FINT dj = _len_spinor(j_kp, j_l);
        FINT dk = k_l * 2 + 1;
        FINT nfi = envs->nfi;
        FINT nfj = envs->nfj;
        FINT nfk = envs->nfk;
        FINT nf2i = nfi + nfi;
        FINT nf2j = nfj + nfj;
        FINT nf = envs->nf;
        FINT nfik = nfi * nfk;
        FINT nfijdk = nfi * nfj * dk;
        FINT ni = dims[0];
        FINT nj = dims[1];
        FINT nk = dims[2];
        FINT ofj = ni * dj;
        FINT ofk = ni * nj * dk;
        FINT ic, jc, kc;
        double *gc_x = gctr;
        double *gc_y = gc_x + nf * i_ctr * j_ctr * k_ctr;
        double *gc_z = gc_y + nf * i_ctr * j_ctr * k_ctr;
        double *gc_1 = gc_z + nf * i_ctr * j_ctr * k_ctr;
        double *bufx;
        FINT buflen = nfi*dk*nfj;
        MALLOC_INSTACK(bufx, buflen*4);
        double *bufy = bufx + buflen;
        double *bufz = bufy + buflen;
        double *buf1 = bufz + buflen;
        double *pgx, *pgy, *pgz, *pg1;
        FINT len1 = di * dk * nf2j;
        FINT len2 = di * dk * dj;
        double *tmp1R, *tmp1I, *tmp2R, *tmp2I;
        MALLOC_INSTACK(tmp1R, len1);
        MALLOC_INSTACK(tmp1I, len1);
        MALLOC_INSTACK(tmp2R, len2);
        MALLOC_INSTACK(tmp2I, len2);
        double complex *pijk;

        for (kc = 0; kc < k_ctr; kc++) {
        for (jc = 0; jc < j_ctr; jc++) {
        for (ic = 0; ic < i_ctr; ic++) {
                pgx = sph2e_inner(bufx, gc_x, k_l, nfi, nfj, nfi*dk, nfik);
                pgy = sph2e_inner(bufy, gc_y, k_l, nfi, nfj, nfi*dk, nfik);
                pgz = sph2e_inner(bufz, gc_z, k_l, nfi, nfj, nfi*dk, nfik);
                pg1 = sph2e_inner(buf1, gc_1, k_l, nfi, nfj, nfi*dk, nfik);
                a_bra_cart2spinor_si(tmp1R, tmp1I, pgx, pgy, pgz, pg1, dk*nfj, i_kp, i_l);
                a_iket_cart2spinor(tmp2R, tmp2I, tmp1R, tmp1I, di*dk, j_kp, j_l);
                pijk = opijk + ofk * kc + ofj * jc + di * ic;
                zcopy_iklj(pijk, tmp2R, tmp2I, ni, nj, nk, 1, di, dj, dk, 1);
                gc_x += nf;
                gc_y += nf;
                gc_z += nf;
                gc_1 += nf;
        } } }
}

void c2s_sf_3c2e1_ssc(double complex *opijk, double *gctr, FINT *dims,
                      CINTEnvVars *envs, double *cache)
{
        FINT *shls = envs->shls;
        FINT *bas = envs->bas;
        FINT i_sh = shls[0];
        FINT j_sh = shls[1];
        FINT i_l = envs->i_l;
        FINT j_l = envs->j_l;
        FINT i_kp = bas(KAPPA_OF, i_sh);
        FINT j_kp = bas(KAPPA_OF, j_sh);
        FINT i_ctr = envs->x_ctr[0];
        FINT j_ctr = envs->x_ctr[1];
        FINT k_ctr = envs->x_ctr[2];
        FINT di = _len_spinor(i_kp, i_l);
        FINT dj = _len_spinor(j_kp, j_l);
        FINT nfj = envs->nfj;
        FINT nfk = envs->nfk;
        FINT nf2j = nfj + nfj;
        FINT nf = envs->nf;
        FINT ni = dims[0];
        FINT nj = dims[1];
        FINT nk = dims[2];
        FINT ofj = ni * dj;
        FINT ofk = ni * nj * nfk;
        FINT ic, jc, kc;
        FINT len1 = di*nfk*nf2j;
        FINT len2 = di*nfk*dj;
        double *tmp1R, *tmp1I, *tmp2R, *tmp2I;
        MALLOC_INSTACK(tmp1R, len1);
        MALLOC_INSTACK(tmp1I, len1);
        MALLOC_INSTACK(tmp2R, len2);
        MALLOC_INSTACK(tmp2I, len2);
        double complex *pijk;

        for (kc = 0; kc < k_ctr; kc++) {
        for (jc = 0; jc < j_ctr; jc++) {
        for (ic = 0; ic < i_ctr; ic++) {
                a_bra_cart2spinor_sf(tmp1R, tmp1I, NULL, NULL, NULL, gctr, nfk*nfj, i_kp, i_l);
                a_ket_cart2spinor(tmp2R, tmp2I, tmp1R, tmp1I, di*nfk, j_kp, j_l);
                pijk = opijk + ofk * kc + ofj * jc + di * ic;
                zcopy_iklj(pijk, tmp2R, tmp2I, ni, nj, nk, 1, di, dj, nfk, 1);
                gctr += nf;
        } } }
}

void c2s_sf_3c2e1i_ssc(double complex *opijk, double *gctr, FINT *dims,
                       CINTEnvVars *envs, double *cache)
{
        FINT *shls = envs->shls;
        FINT *bas = envs->bas;
        FINT i_sh = shls[0];
        FINT j_sh = shls[1];
        FINT i_l = envs->i_l;
        FINT j_l = envs->j_l;
        FINT i_kp = bas(KAPPA_OF, i_sh);
        FINT j_kp = bas(KAPPA_OF, j_sh);
        FINT i_ctr = envs->x_ctr[0];
        FINT j_ctr = envs->x_ctr[1];
        FINT k_ctr = envs->x_ctr[2];
        FINT di = _len_spinor(i_kp, i_l);
        FINT dj = _len_spinor(j_kp, j_l);
        FINT nfj = envs->nfj;
        FINT nfk = envs->nfk;
        FINT nf2j = nfj + nfj;
        FINT nf = envs->nf;
        FINT ni = dims[0];
        FINT nj = dims[1];
        FINT nk = dims[2];
        FINT ofj = ni * dj;
        FINT ofk = ni * nj * nfk;
        FINT ic, jc, kc;
        FINT len1 = di*nfk*nf2j;
        FINT len2 = di*nfk*dj;
        double *tmp1R, *tmp1I, *tmp2R, *tmp2I;
        MALLOC_INSTACK(tmp1R, len1);
        MALLOC_INSTACK(tmp1I, len1);
        MALLOC_INSTACK(tmp2R, len2);
        MALLOC_INSTACK(tmp2I, len2);
        double complex *pijk;

        for (kc = 0; kc < k_ctr; kc++) {
        for (jc = 0; jc < j_ctr; jc++) {
        for (ic = 0; ic < i_ctr; ic++) {
                a_bra_cart2spinor_sf(tmp1R, tmp1I, NULL, NULL, NULL, gctr, nfk*nfj, i_kp, i_l);
                a_iket_cart2spinor(tmp2R, tmp2I, tmp1R, tmp1I, di*nfk, j_kp, j_l);
                pijk = opijk + ofk * kc + ofj * jc + di * ic;
                zcopy_iklj(pijk, tmp2R, tmp2I, ni, nj, nk, 1, di, dj, nfk, 1);
                gctr += nf;
        } } }
}
void c2s_si_3c2e1_ssc(double complex *opijk, double *gctr, FINT *dims,
                      CINTEnvVars *envs, double *cache)
{
        FINT *shls = envs->shls;
        FINT *bas = envs->bas;
        FINT i_sh = shls[0];
        FINT j_sh = shls[1];
        FINT i_l = envs->i_l;
        FINT j_l = envs->j_l;
        FINT i_kp = bas(KAPPA_OF, i_sh);
        FINT j_kp = bas(KAPPA_OF, j_sh);
        FINT i_ctr = envs->x_ctr[0];
        FINT j_ctr = envs->x_ctr[1];
        FINT k_ctr = envs->x_ctr[2];
        FINT di = _len_spinor(i_kp, i_l);
        FINT dj = _len_spinor(j_kp, j_l);
        FINT nfi = envs->nfi;
        FINT nfj = envs->nfj;
        FINT nfk = envs->nfk;
        FINT nf2i = nfi + nfi;
        FINT nf2j = nfj + nfj;
        FINT nf = envs->nf;
        FINT nfijdk = nfi * nfj * nfk;
        FINT ni = dims[0];
        FINT nj = dims[1];
        FINT nk = dims[2];
        FINT ofj = ni * dj;
        FINT ofk = ni * nj * nfk;
        FINT ic, jc, kc;
        double *gc_x = gctr;
        double *gc_y = gc_x + nf * i_ctr * j_ctr * k_ctr;
        double *gc_z = gc_y + nf * i_ctr * j_ctr * k_ctr;
        double *gc_1 = gc_z + nf * i_ctr * j_ctr * k_ctr;
        FINT len2 = di*nfk*dj;
        FINT len1 = di*nfk*nf2j;
        double *tmp1R, *tmp1I, *tmp2R, *tmp2I;
        MALLOC_INSTACK(tmp1R, len1);
        MALLOC_INSTACK(tmp1I, len1);
        MALLOC_INSTACK(tmp2R, len2);
        MALLOC_INSTACK(tmp2I, len2);
        double complex *pijk;

        for (kc = 0; kc < k_ctr; kc++) {
        for (jc = 0; jc < j_ctr; jc++) {
        for (ic = 0; ic < i_ctr; ic++) {
                a_bra_cart2spinor_si(tmp1R, tmp1I, gc_x, gc_y, gc_z, gc_1, nfk*nfj, i_kp, i_l);
                a_ket_cart2spinor(tmp2R, tmp2I, tmp1R, tmp1I, di*nfk, j_kp, j_l);
                pijk = opijk + ofk * kc + ofj * jc + di * ic;
                zcopy_iklj(pijk, tmp2R, tmp2I, ni, nj, nk, 1, di, dj, nfk, 1);
                gc_x += nf;
                gc_y += nf;
                gc_z += nf;
                gc_1 += nf;
        } } }
}
void c2s_si_3c2e1i_ssc(double complex *opijk, double *gctr, FINT *dims,
                       CINTEnvVars *envs, double *cache)
{
        FINT *shls = envs->shls;
        FINT *bas = envs->bas;
        FINT i_sh = shls[0];
        FINT j_sh = shls[1];
        FINT i_l = envs->i_l;
        FINT j_l = envs->j_l;
        FINT i_kp = bas(KAPPA_OF, i_sh);
        FINT j_kp = bas(KAPPA_OF, j_sh);
        FINT i_ctr = envs->x_ctr[0];
        FINT j_ctr = envs->x_ctr[1];
        FINT k_ctr = envs->x_ctr[2];
        FINT di = _len_spinor(i_kp, i_l);
        FINT dj = _len_spinor(j_kp, j_l);
        FINT nfi = envs->nfi;
        FINT nfj = envs->nfj;
        FINT nfk = envs->nfk;
        FINT nf2i = nfi + nfi;
        FINT nf2j = nfj + nfj;
        FINT nf = envs->nf;
        FINT nfijdk = nfi * nfj * nfk;
        FINT ni = dims[0];
        FINT nj = dims[1];
        FINT nk = dims[2];
        FINT ofj = ni * dj;
        FINT ofk = ni * nj * nfk;
        FINT ic, jc, kc;
        double *gc_x = gctr;
        double *gc_y = gc_x + nf * i_ctr * j_ctr * k_ctr;
        double *gc_z = gc_y + nf * i_ctr * j_ctr * k_ctr;
        double *gc_1 = gc_z + nf * i_ctr * j_ctr * k_ctr;
        FINT len2 = di*nfk*dj;
        FINT len1 = di*nfk*nf2j;
        double *tmp1R, *tmp1I, *tmp2R, *tmp2I;
        MALLOC_INSTACK(tmp1R, len1);
        MALLOC_INSTACK(tmp1I, len1);
        MALLOC_INSTACK(tmp2R, len2);
        MALLOC_INSTACK(tmp2I, len2);
        double complex *pijk;

        for (kc = 0; kc < k_ctr; kc++) {
        for (jc = 0; jc < j_ctr; jc++) {
        for (ic = 0; ic < i_ctr; ic++) {
                a_bra_cart2spinor_si(tmp1R, tmp1I, gc_x, gc_y, gc_z, gc_1, nfk*nfj, i_kp, i_l);
                a_iket_cart2spinor(tmp2R, tmp2I, tmp1R, tmp1I, di*nfk, j_kp, j_l);
                pijk = opijk + ofk * kc + ofj * jc + di * ic;
                zcopy_iklj(pijk, tmp2R, tmp2I, ni, nj, nk, 1, di, dj, nfk, 1);
                gc_x += nf;
                gc_y += nf;
                gc_z += nf;
                gc_1 += nf;
        } } }
}


/*************************************************
 *
 * 3-center 1-electron integral transformation
 *
 *************************************************/
void c2s_sph_3c1e(double *out, double *gctr, FINT *dims,
                  CINTEnvVars *envs, double *cache)
{
        c2s_sph_3c2e1(out, gctr, dims, envs, cache);
}

void c2s_cart_3c1e(double *out, double *gctr, FINT *dims,
                   CINTEnvVars *envs, double *cache)
{
        c2s_cart_3c2e1(out, gctr, dims, envs, cache);
}


/*************************************************
 *
 * transform vectors
 *
 *************************************************/
double *CINTc2s_bra_sph(double *gsph, FINT nket, double *gcart, FINT l)
{
        return (c2s_bra_sph[l])(gsph, nket, gcart, l);
}
double *CINTc2s_ket_sph(double *gsph, FINT nbra, double *gcart, FINT l)
{
        return (c2s_ket_sph[l])(gsph, gcart, nbra, nbra, l);
}
double *CINTc2s_ket_sph1(double *sph, double *cart, FINT lds, FINT ldc, FINT l)
{
        return (c2s_ket_sph1[l])(sph, cart, lds, ldc, l);
}

void CINTc2s_bra_spinor_e1sf(double complex *gsp, FINT nket,
                             double *gcart, FINT kappa, FINT l)
{
        FINT nf = _len_cart[l];
        FINT nd = _len_spinor(kappa, l);
        double *gspa = (double *)gsp;
        double *gspb = gspa + nket * nd * OF_CMPLX;
        double *coeffR, *coeffI;
        if (kappa < 0) { // j = l + 1/2
                coeffR = g_c2s[l].cart2j_gt_lR;
                coeffI = g_c2s[l].cart2j_gt_lI;
        } else {
                coeffR = g_c2s[l].cart2j_lt_lR;
                coeffI = g_c2s[l].cart2j_lt_lI;
        }

        FINT i, j, n;
        double saR, saI, sbR, sbI, caR, caI, cbR, cbI, v1;

        for (j = 0; j < nket; j++) {
        for (i = 0; i < nd; i++) {
                saR = 0;
                saI = 0;
                sbR = 0;
                sbI = 0;
#pragma GCC ivdep
                for (n = 0; n < nf; n++) {
                        v1 = gcart[j*nf+n];
                        caR = coeffR[i*nf*2   +n];
                        caI = coeffI[i*nf*2   +n];
                        cbR = coeffR[i*nf*2+nf+n];
                        cbI = coeffI[i*nf*2+nf+n];
                        saR += caR * v1;
                        saI +=-caI * v1;
                        sbR += cbR * v1;
                        sbI +=-cbI * v1;
                }
                gspa[(j*nd+i)*OF_CMPLX  ] = saR;
                gspa[(j*nd+i)*OF_CMPLX+1] = saI;
                gspb[(j*nd+i)*OF_CMPLX  ] = sbR;
                gspb[(j*nd+i)*OF_CMPLX+1] = sbI;
        } }
}

void CINTc2s_bra_spinor_sf(double complex *gsp, FINT nket,
                           double complex *gcart, FINT kappa, FINT l)
{
        FINT nf = _len_cart[l];
        FINT nd = _len_spinor(kappa, l);
        double *gspa = (double *)gsp;
        double *gspb = gspa + nket * nd * OF_CMPLX;
        double *coeffR, *coeffI;
        if (kappa < 0) { // j = l + 1/2
                coeffR = g_c2s[l].cart2j_gt_lR;
                coeffI = g_c2s[l].cart2j_gt_lI;
        } else {
                coeffR = g_c2s[l].cart2j_lt_lR;
                coeffI = g_c2s[l].cart2j_lt_lI;
        }

        FINT i, j, n;
        double saR, saI, sbR, sbI, caR, caI, cbR, cbI, v1R, v1I;

        for (j = 0; j < nket; j++) {
        for (i = 0; i < nd; i++) {
                saR = 0;
                saI = 0;
                sbR = 0;
                sbI = 0;
#pragma GCC ivdep
                for (n = 0; n < nf; n++) {
                        v1R = creal(gcart[j*nf+n]);
                        v1I = cimag(gcart[j*nf+n]);
                        caR = coeffR[i*nf*2   +n];
                        caI = coeffI[i*nf*2   +n];
                        cbR = coeffR[i*nf*2+nf+n];
                        cbI = coeffI[i*nf*2+nf+n];
                        saR += caR * v1R + caI * v1I;
                        saI +=-caI * v1R + caR * v1I;
                        sbR += cbR * v1R + cbI * v1I;
                        sbI +=-cbI * v1R + cbR * v1I;
                }
                gspa[(j*nd+i)*OF_CMPLX  ] = saR;
                gspa[(j*nd+i)*OF_CMPLX+1] = saI;
                gspb[(j*nd+i)*OF_CMPLX  ] = sbR;
                gspb[(j*nd+i)*OF_CMPLX+1] = sbI;
        } }
}

void CINTc2s_ket_spinor(double complex *gsp, FINT nbra,
                        double complex *gcart, FINT kappa, FINT l)
{
        FINT nf = _len_cart[l];
        FINT nf2 = nf * 2;
        FINT nd = _len_spinor(kappa, l);
        double *coeffR, *coeffI;
        double *gspz = (double *)gsp;
        if (kappa < 0) { // j = l + 1/2
                coeffR = g_c2s[l].cart2j_gt_lR;
                coeffI = g_c2s[l].cart2j_gt_lI;
        } else {
                coeffR = g_c2s[l].cart2j_lt_lR;
                coeffI = g_c2s[l].cart2j_lt_lI;
        }

        FINT i, j, n;
        double cR, cI, gR, gI;

        for (i = 0; i < nd; i++) {
#pragma GCC ivdep
                for (j = 0; j < nbra; j++) {
                        gspz[(j+i*nbra)*OF_CMPLX  ] = 0;
                        gspz[(j+i*nbra)*OF_CMPLX+1] = 0;
                }
                for (n = 0; n < nf2; n++) {
                        cR = coeffR[i*nf2+n];
                        cI = coeffI[i*nf2+n];
#pragma GCC ivdep
                        for (j = 0; j < nbra; j++) {
                                gR = creal(gcart[j+n*nbra]);
                                gI = cimag(gcart[j+n*nbra]);
                                gspz[(j+i*nbra)*OF_CMPLX  ] += cR * gR - cI * gI;
                                gspz[(j+i*nbra)*OF_CMPLX+1] += cI * gR + cR * gI;
                        }
                }
        }
}

void CINTc2s_iket_spinor(double complex *gsp, FINT nbra,
                         double complex *gcart, FINT kappa, FINT l)
{
        FINT nf = _len_cart[l];
        FINT nf2 = nf * 2;
        FINT nd = _len_spinor(kappa, l);
        double *coeffR, *coeffI;
        double *gspz = (double *)gsp;
        if (kappa < 0) { // j = l + 1/2
                coeffR = g_c2s[l].cart2j_gt_lR;
                coeffI = g_c2s[l].cart2j_gt_lI;
        } else {
                coeffR = g_c2s[l].cart2j_lt_lR;
                coeffI = g_c2s[l].cart2j_lt_lI;
        }

        FINT i, j, n;
        double cR, cI, gR, gI;

        for (i = 0; i < nd; i++) {
#pragma GCC ivdep
                for (j = 0; j < nbra; j++) {
                        gspz[(j+i*nbra)*OF_CMPLX  ] = 0;
                        gspz[(j+i*nbra)*OF_CMPLX+1] = 0;
                }
                for (n = 0; n < nf2; n++) {
                        cR = coeffR[i*nf2+n];
                        cI = coeffI[i*nf2+n];
#pragma GCC ivdep
                        for (j = 0; j < nbra; j++) {
                                gR = creal(gcart[j+n*nbra]);
                                gI = cimag(gcart[j+n*nbra]);
                                gspz[(j+i*nbra)*OF_CMPLX  ] -= cI * gR + cR * gI;
                                gspz[(j+i*nbra)*OF_CMPLX+1] += cR * gR - cI * gI;
                        }
                }
        }
}

void CINTc2s_bra_spinor(double complex *gsp, FINT nket,
                        double complex *gcart, FINT kappa, FINT l)
{
        FINT nf = _len_cart[l];
        FINT nf2 = nf * 2;
        FINT nd = _len_spinor(kappa, l);
        double *gspz = (double *)gsp;
        double *coeffR, *coeffI;
        if (kappa < 0) { // j = l + 1/2
                coeffR = g_c2s[l].cart2j_gt_lR;
                coeffI = g_c2s[l].cart2j_gt_lI;
        } else {
                coeffR = g_c2s[l].cart2j_lt_lR;
                coeffI = g_c2s[l].cart2j_lt_lI;
        }

        FINT i, j, n;
        double sR, sI, cR, cI, gR, gI;

        for (j = 0; j < nket; j++) {
        for (i = 0; i < nd; i++) {
                sR = 0;
                sI = 0;
#pragma GCC ivdep
                for (n = 0; n < nf2; n++) {
                        gR = creal(gcart[j*nf2+n]);
                        gI = cimag(gcart[j*nf2+n]);
                        cR = coeffR[i*nf2+n];
                        cI = coeffI[i*nf2+n];
                        sR += cR * gR + cI * gI;
                        sI += cR * gI - cI * gR;
                }
                gspz[(j*nd+i)*OF_CMPLX  ] = sR;
                gspz[(j*nd+i)*OF_CMPLX+1] = sI;
        } }
}

void CINTc2s_bra_spinor_si(double complex *gsp, FINT nket,
                           double complex *gcart, FINT kappa, FINT l)
{
        FINT nf = _len_cart[l];
        FINT nf2 = nf * 2;
        FINT nd = _len_spinor(kappa, l);
        double *gspz = (double *)gsp;
        double complex *gcarta = gcart;
        double complex *gcartb = gcarta + nf * nket;
        double *coeffR, *coeffI;
        if (kappa < 0) { // j = l + 1/2
                coeffR = g_c2s[l].cart2j_gt_lR;
                coeffI = g_c2s[l].cart2j_gt_lI;
        } else {
                coeffR = g_c2s[l].cart2j_lt_lR;
                coeffI = g_c2s[l].cart2j_lt_lI;
        }

        FINT i, j, n;
        double sR, sI, caR, caI, cbR, cbI, gaR, gaI, gbR, gbI;

        for (j = 0; j < nket; j++) {
        for (i = 0; i < nd; i++) {
                sR = 0;
                sI = 0;
#pragma GCC ivdep
                for (n = 0; n < nf; n++) {
                        caR = coeffR[i*nf*2   +n];
                        caI = coeffI[i*nf*2   +n];
                        cbR = coeffR[i*nf*2+nf+n];
                        cbI = coeffI[i*nf*2+nf+n];
                        gaR = creal(gcarta[j*nf+n]);
                        gaI = cimag(gcarta[j*nf+n]);
                        gbR = creal(gcartb[j*nf+n]);
                        gbI = cimag(gcartb[j*nf+n]);
                        sR += caR * gaR + caI * gaI + cbR * gbR + cbI * gbI;
                        sI += caR * gaI - caI * gaR + cbR * gbI - cbI * gbR;
                }
                gspz[(j*nd+i)*OF_CMPLX  ] = sR;
                gspz[(j*nd+i)*OF_CMPLX+1] = sI;
        } }
}


/*
 * vectors gspa and gspb are the upper and lower components of the
 * two-component vector
 */
void CINTc2s_ket_spinor_sf1(double complex *gspa, double complex *gspb, double *gcart,
                            FINT lds, FINT ldc, FINT nctr, FINT kappa, FINT l)
{
        FINT nf = _len_cart[l];
        FINT nd = _len_spinor(kappa, l);
        double *gspaz = (double *)gspa;
        double *gspbz = (double *)gspb;
        double *coeffR, *coeffI;
        if (kappa < 0) { // j = l + 1/2
                coeffR = g_c2s[l].cart2j_gt_lR;
                coeffI = g_c2s[l].cart2j_gt_lI;
        } else {
                coeffR = g_c2s[l].cart2j_lt_lR;
                coeffI = g_c2s[l].cart2j_lt_lI;
        }

        FINT i, j, k, n;
        double caR, caI, cbR, cbI, v1;

        for (k = 0; k < nctr; k++) {
                for (i = 0; i < nd; i++) {
#pragma GCC ivdep
                        for (j = 0; j < ldc; j++) {
                                gspaz[(j+i*lds)*OF_CMPLX  ] = 0;
                                gspaz[(j+i*lds)*OF_CMPLX+1] = 0;
                                gspbz[(j+i*lds)*OF_CMPLX  ] = 0;
                                gspbz[(j+i*lds)*OF_CMPLX+1] = 0;
                        }
                        for (n = 0; n < nf; n++) {
                                caR = coeffR[i*nf*2   +n];
                                caI = coeffI[i*nf*2   +n];
                                cbR = coeffR[i*nf*2+nf+n];
                                cbI = coeffI[i*nf*2+nf+n];
#pragma GCC ivdep
                                for (j = 0; j < ldc; j++) {
                                        v1 = gcart[j+n*ldc];
                                        gspaz[(j+i*lds)*OF_CMPLX  ] += caR * v1;
                                        gspaz[(j+i*lds)*OF_CMPLX+1] += caI * v1;
                                        gspbz[(j+i*lds)*OF_CMPLX  ] += cbR * v1;
                                        gspbz[(j+i*lds)*OF_CMPLX+1] += cbI * v1;
                                }
                        }
                }
                gspaz += nd * lds * OF_CMPLX;
                gspbz += nd * lds * OF_CMPLX;
                gcart += nf * ldc;
        }
}

void CINTc2s_iket_spinor_sf1(double complex *gspa, double complex *gspb, double *gcart,
                             FINT lds, FINT ldc, FINT nctr, FINT kappa, FINT l)
{
        FINT nf = _len_cart[l];
        FINT nd = _len_spinor(kappa, l);
        double *gspaz = (double *)gspa;
        double *gspbz = (double *)gspb;
        double *coeffR, *coeffI;
        if (kappa < 0) { // j = l + 1/2
                coeffR = g_c2s[l].cart2j_gt_lR;
                coeffI = g_c2s[l].cart2j_gt_lI;
        } else {
                coeffR = g_c2s[l].cart2j_lt_lR;
                coeffI = g_c2s[l].cart2j_lt_lI;
        }

        FINT i, j, k, n;
        double caR, caI, cbR, cbI, v1;

        for (k = 0; k < nctr; k++) {
                for (i = 0; i < nd; i++) {
#pragma GCC ivdep
                        for (j = 0; j < ldc; j++) {
                                gspaz[(j+i*lds)*OF_CMPLX  ] = 0;
                                gspaz[(j+i*lds)*OF_CMPLX+1] = 0;
                                gspbz[(j+i*lds)*OF_CMPLX  ] = 0;
                                gspbz[(j+i*lds)*OF_CMPLX+1] = 0;
                        }
                        for (n = 0; n < nf; n++) {
                                caR = coeffR[i*nf*2   +n];
                                caI = coeffI[i*nf*2   +n];
                                cbR = coeffR[i*nf*2+nf+n];
                                cbI = coeffI[i*nf*2+nf+n];
#pragma GCC ivdep
                                for (j = 0; j < ldc; j++) {
                                        v1 = gcart[j+n*ldc];
                                        gspaz[(j+i*lds)*OF_CMPLX  ] -= caI * v1;
                                        gspaz[(j+i*lds)*OF_CMPLX+1] += caR * v1;
                                        gspbz[(j+i*lds)*OF_CMPLX  ] -= cbI * v1;
                                        gspbz[(j+i*lds)*OF_CMPLX+1] += cbR * v1;
                                }
                        }
                }
                gspaz += nd * lds * OF_CMPLX;
                gspbz += nd * lds * OF_CMPLX;
                gcart += nf * ldc;
        }
}

void CINTc2s_ket_spinor_si1(double complex *gspa, double complex *gspb, double *gcart,
                            FINT lds, FINT ldc, FINT nctr, FINT kappa, FINT l)
{
        FINT nf = _len_cart[l];
        FINT nd = _len_spinor(kappa, l);
        FINT ngc = nf * ldc;
        double *gc_x = gcart;
        double *gc_y = gc_x + nctr*ngc;
        double *gc_z = gc_y + nctr*ngc;
        double *gc_1 = gc_z + nctr*ngc;
        double *gspaz = (double *)gspa;
        double *gspbz = (double *)gspb;
        double *coeffR, *coeffI;
        if (kappa < 0) { // j = l + 1/2
                coeffR = g_c2s[l].cart2j_gt_lR;
                coeffI = g_c2s[l].cart2j_gt_lI;
        } else {
                coeffR = g_c2s[l].cart2j_lt_lR;
                coeffI = g_c2s[l].cart2j_lt_lI;
        }

        FINT i, j, k, n;
        double caR, caI, cbR, cbI, v1, vx, vy, vz;

        for (k = 0; k < nctr; k++) {
                for (i = 0; i < nd; i++) {
#pragma GCC ivdep
                        for (j = 0; j < ldc; j++) {
                                gspaz[(j+i*lds)*OF_CMPLX  ] = 0;
                                gspaz[(j+i*lds)*OF_CMPLX+1] = 0;
                                gspbz[(j+i*lds)*OF_CMPLX  ] = 0;
                                gspbz[(j+i*lds)*OF_CMPLX+1] = 0;
                        }
                        for (n = 0; n < nf; n++) {
                                caR = coeffR[i*nf*2   +n];
                                caI = coeffI[i*nf*2   +n];
                                cbR = coeffR[i*nf*2+nf+n];
                                cbI = coeffI[i*nf*2+nf+n];
#pragma GCC ivdep
                                for (j = 0; j < ldc; j++) {
                                        v1 = gc_1[j+n*ldc];
                                        vx = gc_x[j+n*ldc];
                                        vy = gc_y[j+n*ldc];
                                        vz = gc_z[j+n*ldc];
                                        gspaz[(j+i*lds)*OF_CMPLX  ] += caR * v1 - caI * vz + cbR * vy - cbI * vx;
                                        gspaz[(j+i*lds)*OF_CMPLX+1] += caI * v1 + caR * vz + cbI * vy + cbR * vx;
                                        gspbz[(j+i*lds)*OF_CMPLX  ] += cbR * v1 + cbI * vz - caR * vy - caI * vx;
                                        gspbz[(j+i*lds)*OF_CMPLX+1] += cbI * v1 - cbR * vz - caI * vy + caR * vx;
                                }
                        }
                }
                gspaz += nd * lds * OF_CMPLX;
                gspbz += nd * lds * OF_CMPLX;
                gc_x += ngc;
                gc_y += ngc;
                gc_z += ngc;
                gc_1 += ngc;
        }
}

void CINTc2s_iket_spinor_si1(double complex *gspa, double complex *gspb, double *gcart,
                             FINT lds, FINT ldc, FINT nctr, FINT kappa, FINT l)
{
        FINT nf = _len_cart[l];
        FINT nd = _len_spinor(kappa, l);
        FINT ngc = nf * ldc;
        double *gc_x = gcart;
        double *gc_y = gc_x + nctr*ngc;
        double *gc_z = gc_y + nctr*ngc;
        double *gc_1 = gc_z + nctr*ngc;
        double *gspaz = (double *)gspa;
        double *gspbz = (double *)gspb;
        double *coeffR, *coeffI;
        if (kappa < 0) { // j = l + 1/2
                coeffR = g_c2s[l].cart2j_gt_lR;
                coeffI = g_c2s[l].cart2j_gt_lI;
        } else {
                coeffR = g_c2s[l].cart2j_lt_lR;
                coeffI = g_c2s[l].cart2j_lt_lI;
        }

        FINT i, j, k, n;
        double caR, caI, cbR, cbI, v1, vx, vy, vz;

        for (k = 0; k < nctr; k++) {
                for (i = 0; i < nd; i++) {
#pragma GCC ivdep
                        for (j = 0; j < ldc; j++) {
                                gspaz[(j+i*lds)*OF_CMPLX  ] = 0;
                                gspaz[(j+i*lds)*OF_CMPLX+1] = 0;
                                gspbz[(j+i*lds)*OF_CMPLX  ] = 0;
                                gspbz[(j+i*lds)*OF_CMPLX+1] = 0;
                        }
                        for (n = 0; n < nf; n++) {
                                caR = coeffR[i*nf*2   +n];
                                caI = coeffI[i*nf*2   +n];
                                cbR = coeffR[i*nf*2+nf+n];
                                cbI = coeffI[i*nf*2+nf+n];
#pragma GCC ivdep
                                for (j = 0; j < ldc; j++) {
                                        v1 = gc_1[j+n*ldc];
                                        vx = gc_x[j+n*ldc];
                                        vy = gc_y[j+n*ldc];
                                        vz = gc_z[j+n*ldc];

                                        gspaz[(j+i*lds)*OF_CMPLX  ] -= caI * v1 + caR * vz + cbI * vy + cbR * vx;
                                        gspaz[(j+i*lds)*OF_CMPLX+1] += caR * v1 - caI * vz + cbR * vy - cbI * vx;
                                        gspbz[(j+i*lds)*OF_CMPLX  ] -= cbI * v1 - cbR * vz - caI * vy + caR * vx;
                                        gspbz[(j+i*lds)*OF_CMPLX+1] += cbR * v1 + cbI * vz - caR * vy - caI * vx;
                                }
                        }
                }
                gspaz += nd * lds * OF_CMPLX;
                gspbz += nd * lds * OF_CMPLX;
                gc_x += ngc;
                gc_y += ngc;
                gc_z += ngc;
                gc_1 += ngc;
        }
}

/*
 * Spherical to Cartesian back transformation
 * The input gsph (Fortran contiguous) has l*2+1 rows
 * The output gcart (Fortran contiguous) has (l+1)*(l+2)/2 rows
 */
double *CINTs2c_bra_sph(double *gsph, FINT nket, double *gcart, FINT l)
{
        FINT nf = (l+1)*(l+2)/2;
        FINT nd = l * 2 + 1;
        CINTdgemm_NN1(nf, nket, nd, g_c2s[l].cart2sph, gsph, gcart, nf);
        return gcart;
}
double *CINTs2c_ket_sph(double *gsph, FINT nbra, double *gcart, FINT l)
{
        FINT nf = (l+1)*(l+2)/2;
        FINT nd = l * 2 + 1;
        CINTdgemm_NT(nbra, nf, nd, gsph, g_c2s[l].cart2sph, gcart);
        return gcart;
}
